﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;

#endregion

namespace Microsoft.Ceres.Optimizer
{
    /// <summary>
    /// Contains functionality for printing diagnostics information during optimization.
    /// Use the Diag methods for things that should be removed in non-diagnostics builds.
    /// </summary>
    public class DiagPrinter
    {
        private static readonly DiagPrinter instance = new DiagPrinter();
        private readonly Dictionary<string, CategoryInfo> categories = new Dictionary<string, CategoryInfo>();
        private int fileCount = 1;
        private StringBuilder totalBuffer = new StringBuilder();

        private DiagPrinter()
        {
            registerCategory("stats", true);
        }

        public static DiagPrinter Instance
        {
            get { return instance; }
        }

        private void registerCategory(string category, bool writeToConsole)
        {
            categories[category] = new CategoryInfo {WriteToConsole = writeToConsole};
        }


        [Conditional("DIAGNOSTICS")]
        public void DiagWriteLine(string category)
        {
            InfoWriteLine(category);
        }

        [Conditional("DIAGNOSTICS")]
        public void DiagWriteLine(string category, object value)
        {
            InfoWriteLine(category, value);
        }

        [Conditional("DIAGNOSTICS")]
        public void DiagWriteLine(string category, string format, params object[] args)
        {
            InfoWriteLine(category, format, args);
        }

        [Conditional("DIAGNOSTICS")]
        public void DiagIndent(string category)
        {
            Indent(category);
        }

        [Conditional("DIAGNOSTICS")]
        public void DiagUnindent(string category)
        {
            Unindent(category);
        }

        public void Indent(string category)
        {
            getCategoryInfo(category).Indent++;
        }

        public void Unindent(string category)
        {
            CategoryInfo categoryInfo = getCategoryInfo(category);
            if (categoryInfo.Indent == 0)
                throw new InvalidOperationException(string.Format("Indent for category {0} is already 0.", category));
            categoryInfo.Indent--;
        }

        public void InfoWriteLine(string category)
        {
            InfoWriteLine(category, string.Empty);
        }

        public void InfoWriteLine(string category, object value)
        {
            InfoWriteLine(category, value != null ? value.ToString() : "(null)");
        }

        public void InfoWriteLine(string category, string format, params object[] args)
        {
            CategoryInfo categoryInfo = getCategoryInfo(category);

            string line = string.Format("{0}{1}", new string('\t', categoryInfo.Indent),
                                        args.Length > 0 ? Utils.NullFormat(format, args) : format);
            categoryInfo.TextBuffer.AppendLine(line);
            totalBuffer.AppendLine(line);
            if (categoryInfo.WriteToConsole)
                Console.WriteLine(line);
        }

        private CategoryInfo getCategoryInfo(string category)
        {
            CategoryInfo categoryInfo;
            if (!categories.TryGetValue(category, out categoryInfo))
            {
                categoryInfo = new CategoryInfo();
                categories[category] = categoryInfo;
            }
            return categoryInfo;
        }

        public void WriteToFiles()
        {
            foreach (KeyValuePair<string, CategoryInfo> category in categories)
            {
                File.WriteAllText(string.Format("{0}-{1}.txt", category.Key, fileCount),
                                  category.Value.TextBuffer.ToString());
                category.Value.TextBuffer = new StringBuilder();
            }
            File.WriteAllText(string.Format("Total-{0}.txt", fileCount), totalBuffer.ToString());
            totalBuffer = new StringBuilder();
            fileCount++;
        }

        ~DiagPrinter()
        {
            WriteToFiles();
        }

        #region Nested type: CategoryInfo

        private class CategoryInfo
        {
            public int Indent;
            public StringBuilder TextBuffer = new StringBuilder();
            public bool WriteToConsole;
        }

        #endregion
    }
}