﻿#region

using System.Collections.Generic;
using System;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{
    /// <summary>
    /// Our simple cost model. Contains constants and cost primitives and also serves as the unit of cost.
    /// </summary>
    public struct BasicCost : ICost
    {
        /// <summary>CPU instruction cost/IO cost factor </summary>
        private const double CPU_FACTOR = 0.0005;

        /// <summary>Page size in the cost model.</summary>
        private const uint PAGE_SIZE = 8192;

        /// <summary>Cost for reading a page in 1/10ms.</summary>
        private const uint PAGE_READ_COSTS = 17;

        /// <summary>Cost for writing a page in 1/10ms.</summary>
        private const uint PAGE_WRITE_COSTS = 18;

        /// <summary>Cost to seek in 1/10ms.</summary>
        private const uint SEEK_COSTS = 95;

        public BasicCost(double firstFurtherReads)
            : this(firstFurtherReads, firstFurtherReads)
        {
        }

        public BasicCost(double firstRead, double furtherReads) : this()
        {
            FirstRead = firstRead;
            FurtherReads = furtherReads;
            Passes = 0;
        }

        /// <summary>Cost to read the plan the first time.</summary>
        public double FirstRead { get; set; }
        /// <summary>Cost for further reads of the plan after the first read.</summary>
        public double FurtherReads { get; set; }
        /// <summary>The number of plan executions calculated for so far. Used
        /// by the DAG cost algorithms.</summary>
        public int Passes { get; set; }

        /// <summary>Zero costs.</summary>
        public static BasicCost Zero
        {
            get { return new BasicCost(0d, 0d); }
        }

        /// <summary>Maximum costs.</summary>
        public static BasicCost Max
        {
            get { return new BasicCost(double.MaxValue, double.MaxValue); }
        }

        #region ICost Members

        public CostRelation Compare(ICost other)
        {
            BasicCost otherCost = (BasicCost) other;

            if (FirstRead < otherCost.FirstRead && FurtherReads <= otherCost.FurtherReads)
                return CostRelation.Better;
            if (FirstRead > otherCost.FirstRead && FurtherReads >= otherCost.FurtherReads)
                return CostRelation.Worse;
            if (FirstRead == otherCost.FirstRead)
            {
                if (FurtherReads < otherCost.FurtherReads)
                    return CostRelation.Better;
                if (FurtherReads > otherCost.FurtherReads)
                    return CostRelation.Worse;
                return CostRelation.Equal;
            }
            return CostRelation.Unknown;
        }

        public CostRelation CompareTotal(ICost other)
        {
            CostRelation result = Compare(other);
            if (result == CostRelation.Unknown)
            {
                if (FirstRead < ((BasicCost) other).FirstRead)
                    return CostRelation.Better;
                else
                    return CostRelation.Worse;
            }
            return result;
        }

        #endregion

        /// <summary>Sums the costs a and b.</summary>
        public static BasicCost operator +(BasicCost a, BasicCost b)
        {
            return new BasicCost(a.FirstRead + b.FirstRead, a.FurtherReads + b.FurtherReads);
        }

        /// <summary>Returns the cost of reading the specified input plan times number of times.
        /// Takes into account that the first read might be more expensive.</summary>
        public static BasicCost operator *(BasicCost input, double times)
        {
            return new BasicCost(input.FirstRead + (times - 1)*input.FurtherReads, times*input.FurtherReads);
        }

        public override string ToString()
        {
            return string.Format("BasicCost(FirstRead: {0:F3}, FurtherReads: {1:F3})", FirstRead, FurtherReads);
        }

        public override int GetHashCode()
        {
            return FirstRead.GetHashCode() ^ FurtherReads.GetHashCode();
        }

        public override bool Equals(object obj)
        {
            if (!(obj is BasicCost))
                return false;
            BasicCost other = (BasicCost) obj;
            return FirstRead.AccuracyEquals(other.FirstRead, 0.001) &&
                   FurtherReads.AccuracyEquals(other.FurtherReads, 0.001);
        }

        /// <summary>
        /// Calcualates the cost of reading a each plan in plans the number of times
        /// specified in the corresponding item in reads.
        /// DAG calculation is uded if the plans given forms as DAG.
        /// </summary>
        /// <param name="plans">Plans to be executed.</param>
        /// <param name="reads">The number of times to execute each of the plans.
        /// Supply null to read each plan once.</param>
        /// <returns>The cost of executing the plans.</returns>
        public static BasicCost InputCosts(IList<Plan> plans, IList<int> reads)
        {
            Validation.NoNulls("plans", plans);
            Validation.Assert("plans, reads", reads == null || plans.Count == reads.Count, "Must have the same length.");

            if (plans.Count == 0)
                return Zero;

            // Check if the Sharing property on any of the plans overlap by unioning
            // them along the way.
            bool overlap = false;
            BitSet temp = (BitSet)plans[0].Sharing.Clone();
            for (int i = 1; i < plans.Count; i++)
            {
                if (temp.Overlaps(plans[i].Sharing))
                {
                    overlap = true;
                    break;
                }
                temp.Or(plans[i].Sharing);
            }

            // If they overlap, we have a DAG and must use the more expensive DAG algorithm. Otherwise, calculate for tree.
            if (!overlap)
                return InputCostsTree(plans, reads);
            else
                return InputCostsDag(plans, reads);
        }

        // Cost for trees. Simply summarize the cost of reading each plan the specified number of times.
        private static BasicCost InputCostsTree(IList<Plan> plans, IList<int> reads)
        {
            BasicCost result = Zero;
            for (int i = 0; i < plans.Count; i++)
                result += (BasicCost) plans[i].Costs*(reads != null ? reads[i] : 1);

            return result;
        }

        // Cost for DAGs. Uses the DagCosts function on the rules to recursively
        // calculate costs for DAGs.
        private static BasicCost InputCostsDag(IList<Plan> plans, IList<int> reads)
        {
            BasicCost result = Zero;
            for (int i = 0; i < plans.Count; i++)
                result += (BasicCost) plans[i].Rule.DagCosts(plans[i], reads != null ? reads[i] : 1);

            foreach (Plan plan in plans)
                resetPasses(plan);

            return result;
        }

        // Helper function for DAG costing, reseting the state after each calculation.
        private static void resetPasses(Plan plan)
        {
            BasicCost cost = (BasicCost) plan.Costs;

            if (cost.Passes != 0)
            {
                cost.Passes = 0;
                plan.Costs = cost;
                foreach (Plan subPlan in plan.Children)
                    resetPasses(subPlan);
            }
        }

        /// <summary>Returns the number of page accesses for bytes amount of data in whole pages</summary>
        public static double PageAccesses(double bytes)
        {
            return Math.Ceiling(bytes/PAGE_SIZE);
        }

        /// <summary>Returns the cost of reading the specified number of pages sequentially.</summary>
        public static double ReadSequential(double pages)
        {
            return SEEK_COSTS + pages * PAGE_READ_COSTS;
        }

        /// <summary>Returns the cost of writing the specified number of pages sequentially.</summary>
        public static double WriteSequential(double pages)
        {
            return SEEK_COSTS + pages * PAGE_WRITE_COSTS;
        }

        /// <summary>Returns the cost of reading the specified number of pages randomly.</summary>
        public static double ReadRandom(double pages)
        {
            return pages* (SEEK_COSTS + PAGE_READ_COSTS);
        }

        /// <summary>Returns the cost of writing the specified number of pages randomly.</summary>
        public static double WriteRandom(double pages)
        {
            return pages * (SEEK_COSTS + PAGE_WRITE_COSTS);
        }

        /// <summary>
        /// Returns CPU costs.
        /// </summary>
        /// <param name="cardinality">Number of typles to execute for.</param>
        /// <param name="instructions">The number of instructions to execute.</param>
        public static double CpuCosts(double cardinality, double instructions)
        {
            return CPU_FACTOR*cardinality*instructions;
        }
    }
}