﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{

    /// <summary>Represents a bit set instance. Support state data is stored in BitSetManager for compactness.</summary>
    public struct BitSet : ICloneable
    {
        /// <summary>The bits are internally stored as ints. 1 int up to 32, 2 up to 64 etc.</summary>
        private int[] data;

        /// <summary>Current number of bits stored.</summary>
        private int length;

        private BitSetManager manager;

        private BitSet(BitSetManager manager, int[] data, int length)
        {
            this.manager = manager;
            this.data = data;
            this.length = length;
        }

        private BitSet(BitSetManager manager, int length)
        {
#if VALIDATE
            Validation.NotNull("manager", manager);
            Validation.Assert("length", length >= 0, "Must be 0 or more.");
#endif
            this.manager = manager;
            data = new int[(length + 31)/32]; // We have room for 32 bits in one int.
            this.length = length;
        }

        public BitSet(BitSetManager manager) : this(manager, false)
        {
        }

        public BitSet(BitSetManager manager, bool setAll)
            : this(manager, manager.Count)
        {
            if (setAll)
            {
                for (int i = 0, index = 0, mask = 1; i < length; ++i)
                {
                    if (mask == ((int) 1) << 31)
                    {
                        ++index;
                        mask = 1;
                    }
                    data[index] |= mask;
                    mask <<= 1;
                }
            }
        }

        /// <summary>Gets or sets the i-th bit in this bit set.</summary>
        public bool this[int index]
        {
            get
            {
                if (index < 0 || index > length)
                {
                    throw new ArgumentOutOfRangeException("index");
                }
                // Get the correct array item, AND in 1 bitshifted to the correct position and return if it is not 0.
                return ((data[index/32] & (((int) 1) << (index%32))) != 0);
            }
            set
            {
                if (index < 0 || index > length)
                {
                    throw new ArgumentOutOfRangeException("index");
                }
                if (value)
                {
                    // OR 1 bitshifted to the correct position.
                    data[index/32] |= ((int) 1) << (index%32);
                }
                else
                {
                    // AND, for example 1110111 (where 0 is the correct position).
                    data[index/32] &= ~(((int) 1) << (index%32));
                }
            }
        }

        #region ICloneable Members

        public object Clone()
        {
            return new BitSet(manager, (int[]) data.Clone(), length);
        }

        #endregion

        /// <summary>Adds the specified property to the set.</summary>
        public void Add(string property)
        {
            this[manager[property]] = true;
        }

        /// <summary>Checks if the specified property is in the set.</summary>
        public bool Contains(string property)
        {
            return this[manager[property]];
        }

        /// <summary>Removes the specified property to the set.</summary>
        public void Remove(string property)
        {
            this[manager[property]] = false;
        }

        /// <summary>
        /// Returns if this bit set overlaps with the given bit set.
        /// </summary>
        /// <param name="other">The bit set to compare with.</param>
        public bool Overlaps(BitSet other)
        {
            validate(this, other);

            // Intersect all the data items (&) and check if the result is > 0..
            int length = (this.length + 31)/32;
            for (int i = 0; i < length; i++)
            {
                if ((data[i] & other.data[i]) > 0)
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Return if all bit properties are equally set.
        /// </summary>
        public static bool operator ==(BitSet a, BitSet b)
        {
            validate(a, b);

            // Bit-operations on BitSets modifies it in-place, so we iterate over the bits instead of bit-mangling a clone.
            // Verify that all data items are equal.
            int length = (a.length + 31)/32;
            for (int i = 0; i < length; i++)
            {
                if (a.data[i] != b.data[i])
                {
                    return false;
                }
            }
            return true;
        }

        /// <summary>
        /// Returns if any bit property is not equally set.
        /// </summary>
        public static bool operator !=(BitSet a, BitSet b)
        {
            return !(a == b);
        }

        /// <summary>
        /// Const union operator, returning a new bit set.
        /// </summary>
        public static BitSet operator |(BitSet a, BitSet b)
        {
            validate(a, b);

            // Union all the data items (|) and construct a new BitSet.
            int length = (a.length + 31)/32;
            int[] data = new int[length];
            for (int i = 0; i < length; i++)
            {
                data[i] = a.data[i] | b.data[i];
            }
            return new BitSet(a.manager, data, a.length);
        }

        /// <summary>
        /// In-place union, i.e. adds all bits set in other to this bit set.
        /// </summary>
        /// <param name="other">The bit set to union into this bit set.</param>
        public BitSet Or(BitSet other)
        {
            validate(this, other);

            // Union all the data items (|).
            int length = (this.length + 31)/32;
            for (int i = 0; i < length; i++)
            {
                data[i] |= other.data[i];
            }
            return this;
        }

        /// <summary>
        /// Const intersection operator, returning a new bit set.
        /// </summary>
        public static BitSet operator &(BitSet a, BitSet b)
        {
            validate(a, b);

            // Intersect all the data items (&) and construct a new BitSet.
            int length = (a.length + 31)/32;
            int[] data = new int[length];
            for (int i = 0; i < length; i++)
            {
                data[i] = a.data[i] & b.data[i];
            }
            return new BitSet(a.manager, data, a.length);
        }

        /// <summary>
        /// Const difference operator, removing b's properties from a, returning a new bit set.
        /// </summary>
        public static BitSet operator -(BitSet a, BitSet b)
        {
            validate(a, b);

            // Negation (A - B) is implemented as A & !B.
            int length = (a.length + 31)/32;
            int[] data = new int[length];
            for (int i = 0; i < length; i++)
            {
                data[i] = a.data[i] & ~b.data[i];
            }
            return new BitSet(a.manager, data, a.length);
        }

        /// <summary>
        /// In-place difference operator, removing other's set bit from this instance.
        /// </summary>
        /// <param name="other">The bit set to remove from this bit set.</param>
        public BitSet Subtract(BitSet other)
        {
            validate(this, other);

            int length = (this.length + 31)/32;
            for (int i = 0; i < length; i++)
            {
                data[i] = data[i] & ~other.data[i];
            }
            return this;
        }

        /// <summary>
        /// If a is subset of b.
        /// </summary>
        public static bool operator <=(BitSet a, BitSet b)
        {
            validate(a, b);

            // Subset (A <= B) is implemented as (A & !B) == Ø.
            int length = (a.length + 31)/32;
            for (int i = 0; i < length; i++)
            {
                if ((a.data[i] & ~b.data[i]) != 0)
                {
                    return false;
                }
            }
            return true;
        }

        /// <summary>
        /// If b is subset of a.
        /// </summary>
        public static bool operator >=(BitSet a, BitSet b)
        {
            return b <= a;
        }

        /// <summary>
        /// Returns a BitSetWalker, which is an IEnumerable that generates all permutations
        /// of the set bits in this instance.
        /// </summary>
        public IEnumerable<BitSet> Walk()
        {
            return new BitSetWalker(this);
        }

        // Debug verification
        [Conditional("VERIFY")]
        private static void validate(BitSet a, BitSet b)
        {
            if (a.manager != b.manager)
            {
                throw new ArgumentException("The BitSets needs to have the same manager.");
            }
        }

        public override bool Equals(object obj)
        {
            if (!(obj is BitSet))
            {
                return false;
            }
            BitSet other = (BitSet) obj;
            validate(this, other);

            // Verify that all data items are equal.
            int length = (this.length + 31)/32;
            for (int i = 0; i < length; i++)
            {
                if (data[i] != other.data[i])
                {
                    return false;
                }
            }
            return true;
        }

        public override int GetHashCode()
        {
            int hash = data[0];
            int length = (this.length + 31)/32;
            for (int i = 1; i < length; i++)
                hash ^= data[i];
            return hash;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder("[");
            for (int i = 0; i < length; i++)
            {
                if (this[i])
                {
                    sb.AppendFormat("{{{0}}},", Utils.JoinEnumerable(manager[i]));
                }
            }
            if (sb.Length > 1)
            {
                sb.Remove(sb.Length - 1, 1);
            }
            sb.Append("]");
            return sb.ToString();
        }

        #region Nested type: BitSetWalker

        /// <summary>
        /// Helper class which efficiently generates all permutations of
        /// the bits set in the supplied bit set.
        /// </summary>
        public class BitSetWalker : IEnumerable<BitSet>, IEnumerator<BitSet>
        {
            private BitSet current; // Position in the walk.
            private bool first = true; // At first permutation?
            private BitSet source; // Original bit mask to walk

            public BitSetWalker(BitSet bitSet)
            {
                source = bitSet;
                current = source.manager.Empty;
            }

            #region IEnumerable<BitSet> Members

            public IEnumerator<BitSet> GetEnumerator()
            {
                return this;
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }

            #endregion

            #region IEnumerator<BitSet> Members

            public bool MoveNext()
            {
                // If this is the first time, do not produce a permutation
                // (leave it as 00000...)
                if (first)
                {
                    first = false;
                    return true;
                }

                // Iterate over the full bit mask length
                for (int i = 0, index = 0, mask = 1; i < source.length; ++i)
                {
                    if (mask == ((int) 1) << 31)
                    {
                        ++index;
                        mask = 1;
                    }

                    // if source[i] == true, we've found a bit to permutate.
                    if ((source.data[index] & mask) != 0)
                    {
                        // if current[i] == false, flip it and return true (we have more permutations).
                        if ((current.data[index] & mask) == 0)
                        {
                            // current[i] = true;
                            current.data[index] |= mask;
                            return true;
                        }
                        else // else, flip it back
                        {
                            // current[i] = false;
                            current.data[index] &= ~mask;
                        }
                    }
                    mask <<= 1;
                }
                // No more permutations
                return false;
            }

            public BitSet Current
            {
                get { return current; }
            }

            public void Reset()
            {
                current = source.manager.Empty;
                first = true;
            }

            public void Dispose()
            {
            }

            object IEnumerator.Current
            {
                get { return Current; }
            }

            #endregion
        }

        #endregion
    }
}