﻿#region

using System;
using System.Collections.Generic;
using System.Linq;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{
    /// <summary>
    /// How the BitSet is going to be used by rules.
    /// </summary>
    [Flags]
    public enum PropertyUsage
    {
        None = 0,
        Produced = 1,
        Required = 2,
        Goal = 4
    }

    /// <summary>
    /// Main class for bit sets. Manages creation and minimization.
    /// </summary>
    public class BitSetManager
    {
        private readonly List<HashSet<string>> indexToName = new List<HashSet<string>>();
        private readonly Dictionary<string, int> nameToIndex = new Dictionary<string, int>();

        private readonly Dictionary<NestableHashSet<string>, BitSetProxy> proxies =
            new Dictionary<NestableHashSet<string>, BitSetProxy>();

        private bool prepared;

        /// <summary>Number of bits in use.</summary>
        public int Count
        {
            get
            {
                ValidatePrepared(true);
                return indexToName.Count;
            }
        }

        /// <summary>Gets the bit index of the property name.</summary>
        public int this[string property]
        {
            get
            {
                ValidatePrepared(true);
                int index;
                if (nameToIndex.TryGetValue(property, out index))
                {
                    return index;
                }
                throw new ArgumentException("No property with this name registered: " + property);
            }
        }

        /// <summary>Gets the property names (multiple due to minimization) of the set index..</summary>
        public HashSet<string> this[int index]
        {
            get
            {
                ValidatePrepared(true);
                return indexToName[index];
            }
        }

        /// <summary>Gets an empty (no bits set) bit set.</summary>
        public BitSet Empty
        {
            get
            {
                ValidatePrepared(true);
                return new BitSet(this);
            }
        }

        /// <summary>Gets an full (all bits set) bit set.</summary>
        public BitSet All
        {
            get
            {
                ValidatePrepared(true);
                return new BitSet(this, true);
            }
        }

        /// <summary>Gets a bitset (in a proxy) with the specified properties and property usage.</summary>
        public BitSetProxy GetBitSet(PropertyUsage propertyUsage, params string[] properties)
        {
            ValidatePrepared(false);
            return GetBitSet(propertyUsage, new NestableHashSet<string>(properties));
        }

        /// <summary>Gets a bitset (in a proxy) with the specified properties and property usage.</summary>
        public BitSetProxy GetBitSet(PropertyUsage propertyUsage, NestableHashSet<string> properties)
        {
            Validation.NoNullsOrEmpty("properties", properties);
            BitSetProxy result;
            if (!proxies.TryGetValue(properties, out result))
            {
                result = new BitSetProxy {PropertyUsage = propertyUsage, Properties = properties};
                proxies.Add(properties, result);
            }
            return result;
        }

        /// <summary>Prepares the bit sets. This will minimze bit sets, assign bit indexes to properties
        /// and fill proxies.</summary>
        public void Prepare()
        {
            HashSet<NestableHashSet<string>> producedPropertysets = new HashSet<NestableHashSet<string>>();
            HashSet<NestableHashSet<string>> requiredPropertySets = new HashSet<NestableHashSet<string>>();

            // Analyze usage.
            foreach (BitSetProxy proxy in proxies.Values)
            {
                if ((proxy.PropertyUsage & PropertyUsage.Produced) == PropertyUsage.Produced)
                    producedPropertysets.Add(proxy.Properties);

                if ((proxy.PropertyUsage & PropertyUsage.Required) == PropertyUsage.Required)
                    requiredPropertySets.Add(proxy.Properties);
            }

            Prepare(producedPropertysets, requiredPropertySets);
        }

        /// <summary>
        /// Prepares the BitSetManager for use. This must be called before any bitsets can be used.
        /// It will minimize the bit set properties and register all the mappings.
        /// </summary>
        public void Prepare(HashSet<NestableHashSet<string>> producedPropertysets,
                            HashSet<NestableHashSet<string>> requiredPropertySets)
        {
            // Check that we haven't been prepared already
            ValidatePrepared(false);
            prepared = true;
            // Minize properties.
            Dictionary<string, string> mappings = MinimizeProperties(producedPropertysets, requiredPropertySets);
            // Registers mappings, assigning names to indexes.
            RegisterMappings(mappings);
            // Fill the registered proxies.
            FillProxies();
        }

        // Fills the proxies with bit sets.
        private void FillProxies()
        {
            foreach (var proxy in proxies)
            {
                proxy.Value.BitSet = GetWithValues(proxy.Key.ToArray());
            }
        }

        /// <summary>
        /// Minimize the properties. Currently, this includes merging all properties
        /// that are always produced together.
        /// </summary>
        /// <returns>A dictionary that maps property name -> property name.
        /// If for instance A and B are always produced together, it will contain {A->A, B->A}.</returns>
        private Dictionary<string, string> MinimizeProperties(HashSet<NestableHashSet<string>> producedPropertysets,
                                                              HashSet<NestableHashSet<string>> requiredPropertySets)
        {
            Dictionary<string, string> mappings = new Dictionary<string, string>();
            HashSet<string> allProducedProperties = new HashSet<string>();
            // Register all direct mappings in all produced sets by default.
            foreach (NestableHashSet<string> propertySet in producedPropertysets)
            {
                foreach (string property in propertySet)
                    mappings[property] = property;

                // Get all the produced properties unioned together.
                allProducedProperties.UnionWith(propertySet);
            }

            // Validate that no proxy is using properties that are not produced
            foreach (KeyValuePair<NestableHashSet<string>, BitSetProxy> proxy in proxies)
            {
                foreach (string property in proxy.Value.Properties)
                {
                    if (!allProducedProperties.Contains(property))
                        throw new OptimizerException("The requested property '{0}' in the proxy {1} is never produced.",
                                                     property, proxy.Value);
                }
            }

            // Now, foreach over all properties, visiting each combination {propA, propB} once,
            // where propA < propB.
            foreach (string propA in allProducedProperties)
            {
                foreach (string propB in allProducedProperties)
                {
                    if (propA.CompareTo(propB) < 0)
                    {
                        // Foreach over all produced sets. If a set contains propA or propB, add it
                        // to our set of sets.
                        HashSet<NestableHashSet<string>> setsContainingA = new HashSet<NestableHashSet<string>>();
                        HashSet<NestableHashSet<string>> setsContainingB = new HashSet<NestableHashSet<string>>();
                        foreach (var proxy in proxies)
                        {
                            if ((proxy.Value.PropertyUsage & PropertyUsage.Produced) == PropertyUsage.Produced)
                            {
                                if (proxy.Key.Contains(propA))
                                {
                                    setsContainingA.Add(proxy.Key);
                                }
                                if (proxy.Key.Contains(propB))
                                {
                                    setsContainingB.Add(proxy.Key);
                                }
                            }
                        }

                        // Now setsContainingA contains all sets containing A and setsContainingB
                        // contains all sets containing B.
                        // If these two sets are set equals (contain the same elements), it means
                        // that propA and propB are always produced together.
                        if (setsContainingA.SetEquals(setsContainingB))
                        {
                            // If so, merge the properties by making propB map to whatever propA
                            // maps to. The reason we're mapping propB to mappings[propA] and
                            // not to propA, is that propA could have been remapped earlier itself.
                            mappings[propB] = mappings[propA];
                        }
                    }
                }
            }
            return mappings;
        }

        /// <summary>
        /// Registers the minimized mappings. This means adding entries
        /// to the nameToIndex and indexToName dictionaries.
        /// </summary>
        /// <param name="mappings">The mappings to register.</param>
        private void RegisterMappings(Dictionary<string, string> mappings)
        {
            // mappings.Values contains all possible bit property variations.
            // Register a physical property for all these.
            foreach (string property in mappings.Values)
            {
                // Only add each possible variation once.
                if (!nameToIndex.ContainsKey(property))
                {
                    nameToIndex[property] = indexToName.Count;
                    indexToName.Add(new HashSet<string> {property});
                }
            }

            // Then, add all the remappings (ie. {B->A}).
            foreach (KeyValuePair<string, string> pair in mappings)
            {
                // Get the target index
                int index = nameToIndex[pair.Value];
                // Make the from-property point to the correct index
                nameToIndex[pair.Key] = index;
                // Add the from-property to the reverse lookup table
                indexToName[index].Add(pair.Key);
            }
        }

        /// <summary>Internal validation.</summary>
        private void ValidatePrepared(bool shouldBePrepared)
        {
            if (shouldBePrepared && !prepared)
            {
                throw new InvalidOperationException("The BitSetManager must be prepared before it can be used.");
            }
            if (!shouldBePrepared && prepared)
            {
                throw new InvalidOperationException("The BitSetManager is already prepared.");
            }
        }

        /// <summary>Gets a bit set with the specified values. Can only be called after preparation.</summary>
        public BitSet GetWithValues(params string[] properties)
        {
            ValidatePrepared(true);
            BitSet bitSet = new BitSet(this);
            foreach (string property in properties)
            {
                bitSet.Add(property);
            }
            return bitSet;
        }
    }
}