﻿#region

using System.Collections.Generic;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{
    /// <summary>
    /// Works like a regular HashSet, but supports being nested within other HashSet
    /// with the correct behaviour. This is done by having it implement GetHashCode
    /// and Equals, both driven by the contents of the set.
    /// Maintains the hash code on item insertion and removal.
    /// </summary>
    public class NestableHashSet<T> : HashSet<T>
    {
        private int hashCode = 0;

        public NestableHashSet()
        {
        }

        public NestableHashSet(IEnumerable<T> collection)
            : base(collection)
        {
        }

        public new bool Add(T item)
        {
            if (base.Add(item))
            {
                hashCode ^= item.GetHashCode();
                return true;
            }
            return false;
        }

        public new bool Remove(T item)
        {
            if (base.Remove(item))
            {
                hashCode ^= item.GetHashCode();
                return true;
            }
            return false;
        }

        public new void Clear()
        {
            hashCode = 0;
            base.Clear();
        }

        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is HashSet<T>))
            {
                return false;
            }
            return SetEquals((HashSet<T>) obj);
        }

        public override int GetHashCode()
        {
            return hashCode;
        }

        public override string ToString()
        {
            return string.Format("{{{0}}}", Utils.JoinEnumerable(this));
        }
    }
}