﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine.Orders
{
    /// <summary>
    /// Represents a DFSM.
    /// </summary>
    internal class DFA : IEnumerable<DFANode>
    {
        private readonly Dictionary<NestableHashSet<NFANode>, DFANode> nodes =
            new Dictionary<NestableHashSet<NFANode>, DFANode>();

        private List<DFANode> pendingIteration;

        #region IEnumerable<DFANode> Members
        /// <summary>Enumerates the nodes in the DFA.</summary>
        public IEnumerator<DFANode> GetEnumerator()
        {
            return nodes.Values.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        /// <summary>Gets or creates the node with the specified set of NFA nodes.</summary>
        public DFANode GetOrCreateNode(NestableHashSet<NFANode> description)
        {
            DFANode node;
            if (!nodes.TryGetValue(description, out node))
            {
                node = new DFANode {Description = description};
                nodes.Add(description, node);

                // If we are in the middle of a node iteration, add any nodes added during the
                // iteration to a queue of nodes pending to be iterated.
                if (pendingIteration != null)
                {
                    pendingIteration.Add(node);
                }
            }
            return node;
        }

        /// <summary>Converts all referenced NFA nodes from the specified DFA node to DFA nodes.</summary>
        private void convertNode(DFANode node)
        {
            List<LinkedListNode<Edge<NFANode>>> edges = new List<LinkedListNode<Edge<NFANode>>>();
            foreach (NFANode nfaNode in node.Description)
            {
                if (nfaNode.Edges.Count > 0)
                {
                    edges.Add(new LinkedList<Edge<NFANode>>(nfaNode.Edges.OrderBy(edge => edge)).First);
                }
            }

            while (edges.Count > 0)
            {
                // Finds smallest label of outgoing edges
                int label = edges[0].Value.Label;
                string description = edges[0].Value.Description;
                foreach (LinkedListNode<Edge<NFANode>> edge in edges)
                {
                    if (edge.Value.Label < label)
                    {
                        label = edge.Value.Label;
                        description = edge.Value.Description;
                    }
                }
                // Epsilon is the larges possible label - we're done.
                if (label == NFANode.EPSILON)
                {
                    break;
                }

                // Add the current available orderings or groupings.
                NestableHashSet<NFANode> nfaNodes = new NestableHashSet<NFANode>();
                foreach (NFANode nfaNode in node.Description)
                {
                    // Grouping or Order
                    if (nfaNode.OrderDescription == null || nfaNode.OrderDescription.Size > 0)
                    {
                        recursiveAdd(nfaNodes, nfaNode, false);
                    }
                }

                // Now gather all nodes reachable by following any edges with the found smallest label.
                // Also follow any epsilon edges recursively. Finally, remove the edge(s) followed from the queue.
                for (int i = 0; i < edges.Count; i++)
                {
                    while (edges[i].Value.Label == label)
                    {
                        recursiveAdd(nfaNodes, edges[i].Value.Target, true);
                        edges[i] = edges[i].Next;
                        if (edges[i] == null)
                        {
                            edges.RemoveAt(i);
                            i--;
                            break;
                        }
                    }
                }

                // Create a new node with the set of orderings and groupings reachable
                // and add an edge to with the found smallest label.
                node.AddEdge(new Edge<DFANode>(GetOrCreateNode(nfaNodes), label, description));
            }
        }

        /// <summary>
        /// Add the specified node to the node set, recursively following all
        /// epsilon edges if specified.
        /// </summary>
        /// <param name="nfaNodes">Set to add nodes to.</param>
        /// <param name="node">Node to start at.</param>
        /// <param name="withEpsilon">If all nodes reachable by epsilon edges should be added.</param>
        private void recursiveAdd(NestableHashSet<NFANode> nfaNodes, NFANode node, bool withEpsilon)
        {
            if (nfaNodes.Contains(node))
            {
                return;
            }
            nfaNodes.Add(node);

            if (withEpsilon)
            {
                foreach (Edge<NFANode> edge in node.Edges)
                {
                    if (edge.Label == NFANode.EPSILON)
                    {
                        recursiveAdd(nfaNodes, edge.Target, true);
                    }
                }
            }
        }

        /// <summary>
        /// Converts the specified NFA to a DFA using power set construction.
        /// </summary>
        /// <param name="nfa">NFA to convert.</param>
        /// <param name="nfaRoot">Root node of the NFA.</param>
        /// <returns>Root note of the DFA.</returns>
        internal DFANode Convert(NFA nfa, NFANode nfaRoot)
        {
            // Create the root
            DFANode root = GetOrCreateNode(new NestableHashSet<NFANode> {nfaRoot});
            // Make sure we keep references to nodes added in the process.
            pendingIteration = new List<DFANode>();
            // Convert all NFA nodes it references.
            convertNode(root);

            // Continue calling convertNode on all DFA nodes added until
            // no more DFA nodes are added.
            while (pendingIteration.Count > 0)
            {
                int startCount = pendingIteration.Count;
                for (int i = startCount - 1; i >= 0; i--)
                {
                    convertNode(pendingIteration[i]);
                }
                pendingIteration.RemoveRange(0, startCount);
            }
            pendingIteration = null;
            return root;
        }

        /// <summary>Insers HTML entities for some common symbols and terms.</summary>
        private static string formatAndInsertHtmlEntities(string input)
        {
            input = input ?? string.Empty;
            input = input.Replace("->", "&rarr;");
            input = input.Replace("ASC", "&darr;");
            input = input.Replace("DESC", "&uarr;");
            return input;
        }

        /// <summary>Generates and returns the dot graph description of this DFA instance.</summary>
        public string GetDot()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine("digraph test {");
            sb.AppendLine("node [fontsize=12];");
            sb.AppendLine("edge [fontsize=12];");

            foreach (DFANode node in nodes.Values)
            {
                // Output the node. Hash code is used as node id.
                string shape = "shape=box,style=rounded";
                sb.AppendFormat("{0} [label=\"{1}\",{2}];{3}", node.GetHashCode(),
                                (node.Order != null ? node.Order.Id + " " : "") +
                                formatAndInsertHtmlEntities("{" + Utils.JoinEnumerable(node.Description,
                                    n =>n.OrderDescription != null ? n.OrderDescription.ToString() : n.GroupingDescription.ToString(),
                                    string.Empty, ",\\n", true) + "}"),
                                shape, Environment.NewLine);

                // Output all outgoing edges.
                foreach (Edge<DFANode> edge in node.Edges)
                {
                    sb.AppendFormat("{0} -> {1} [label = \"{2}\"];{3}", node.GetHashCode(), edge.Target.GetHashCode(),
                                    formatAndInsertHtmlEntities(edge.Description), Environment.NewLine);
                }
            }

            sb.AppendLine("}");
            return sb.ToString();
        }
    }
}