﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine.Orders
{
    internal delegate void ForAllNodesCallback(NFA nfa, NFANode node, object data);

    /// <summary>
    /// Represent a non-deterministic finite state machine.
    /// </summary>
    internal class NFA
    {
        private readonly Dictionary<object, NFANode> nodes = new Dictionary<object, NFANode>();
        private List<NFANode> pendingIteration;
        private int sequenceCounter;

        /// <summary>Number of nodes.</summary>
        public int Size
        {
            get { return nodes.Count; }
        }

        /// <summary>
        /// Gets (if exists) or creates a node (state) for the specified order description.
        /// </summary>
        /// <param name="orderDescription">OrderDescription to get node for.</param>
        /// <returns>An existing or newly created node.</returns>
        public NFANode GetOrCreateNode(OrderDescription orderDescription)
        {
            NFANode node;
            // Check if it exists already
            if (! nodes.TryGetValue(orderDescription, out node))
            {
                // No, construct a new one -- sequenceCounter NOT thread safe.
                node = new NFANode {OrderDescription = orderDescription, SequenceNumber = sequenceCounter++};
                // Add it with the order description as key
                nodes.Add(orderDescription, node);

                // If we are in the middle of a node iteration, add any nodes added during the
                // iteration to a queue of nodes pending to be iterated.
                if (pendingIteration != null)
                {
                    pendingIteration.Add(node);
                }

                // Also add nodes for all prefixes of this ordeing and add epsilon edges to them.
                if (orderDescription.Size > 1)
                {
                    node.AddEdge(new Edge<NFANode>(
                                     GetOrCreateNode(orderDescription.RemoveAt(orderDescription.Size - 1)),
                                     NFANode.EPSILON, "ε"));
                }
            }
            return node;
        }

        /// <summary>
        /// Gets (if exists) or creates a node (state) for the specified grouping description.
        /// </summary>
        /// <param name="groupingDescription">Grouping description to get node for.</param>
        /// <returns>An existing or newly created node.</returns>
        public NFANode GetOrCreateNode(NestableHashSet<string> groupingDescription)
        {
            NFANode node;
            // Check if it exists
            if (!nodes.TryGetValue(groupingDescription, out node))
            {
                // No, construct a new one -- sequenceCounter NOT thread safe.
                node = new NFANode {GroupingDescription = groupingDescription, SequenceNumber = sequenceCounter++};

                // Add it with the grouping description as key.
                nodes.Add(groupingDescription, node);

                // If we are in the middle of a node iteration, add any nodes added during the
                // iteration to a queue of nodes pending to be iterated.
                if (pendingIteration != null)
                {
                    pendingIteration.Add(node);
                }
            }
            return node;
        }

        /// <summary>Try to map an attribute using the map. If there is no mapping for it,
        /// return the original value.</summary>
        private static string mapAttribute(Dictionary<string, string> shareEquivalentMappings, string attribute)
        {
            string result;
            if (shareEquivalentMappings.TryGetValue(attribute, out result))
                return result;
            return attribute;
        }

        /// <summary>
        /// Add edges for share equivalency to the NFA. This means adding epsilon edges
        /// from ordering states with equivalence class representatives to all share
        /// equivalent orderings.
        /// </summary>
        /// <param name="shareEquivalentMappings">Mappings for share equivalent attributes,
        /// mapping the class members to the representative.</param>
        public void AddShareEquivalencyEdges(Dictionary<string, string> shareEquivalentMappings)
        {
            // Iterate over all possible target for share equivalency edges.
            foreach (NFANode toNode in nodes.Values)
            {
                if (toNode.OrderDescription != null)
                {
                    // Rewrite the order description to use equivalence class representatives.
                    OrderDescription temp = new OrderDescription();
                    foreach (OrderDescriptionItem item in toNode.OrderDescription)
                        temp.Append(new OrderDescriptionItem(mapAttribute(shareEquivalentMappings, item.Property),
                                                             item.Ascending));

                    // If there is such an ordering present, add an edge from it to this node.
                    NFANode fromNode;
                    if (nodes.TryGetValue(temp, out fromNode))
                        fromNode.AddEdge(new Edge<NFANode>(toNode, NFANode.EPSILON, "ε"));
                }
                else
                {
                    // Rewrite the grouping description to use equivalence class representatives.
                    NestableHashSet<string> temp = new NestableHashSet<string>();
                    foreach (string property in toNode.GroupingDescription)
                        temp.Add(mapAttribute(shareEquivalentMappings, property));

                    // If there is such a grouping present, add an edge from it to this node.
                    NFANode fromNode;
                    if (nodes.TryGetValue(temp, out fromNode))
                        fromNode.AddEdge(new Edge<NFANode>(toNode, NFANode.EPSILON, "ε"));
                }
            }
        }

        /// <summary>
        /// Call the callback delegate for each node in the NFA, optionally passing the
        /// data argument to it.
        /// Any nodes added by the delegate will also be iterated.
        /// </summary>
        /// <param name="callback">Callback delegate.</param>
        /// <param name="data">Optiondal data argument to pass.</param>
        public void ForAllNodes(ForAllNodesCallback callback, object data)
        {
            // Keep added nodes during iteration.
            pendingIteration = new List<NFANode>();

            // Call delegate for each node.
            foreach (NFANode node in new List<NFANode>(nodes.Values))
            {
                callback(this, node, data);
            }

            // While nodes are being added, call the delegate again for these new nodes.
            while (pendingIteration.Count > 0)
            {
                int startCount = pendingIteration.Count;
                for (int i = startCount - 1; i >= 0; i--)
                {
                    callback(this, pendingIteration[i], data);
                }
                pendingIteration.RemoveRange(0, startCount);
            }

            pendingIteration = null;
        }


        /// <summary>
        /// Adds epsilon and other edges to the NFA. This involves two steps:
        ///  * Add transitive edges, ie. if we have a->b and b->c, then add a->c.
        ///  * Add edges from ordering to grouping, ie. (a,b) -> {a,b}, (a,b) -> {a}
        /// </summary>
        public void AddEpsilonEdges()
        {
            // Add transitive edges
            bool changed = true;
            // Continue as long as we are adding edges.
            while (changed)
            {
                changed = false;
                foreach (NFANode node in nodes.Values)
                {
                    // We cannot modify a collection we are iterating over so create a list.
                    foreach (Edge<NFANode> edge1 in new List<Edge<NFANode>>(node.Edges))
                    {
                        foreach (Edge<NFANode> edge2 in edge1.Target.Edges)
                        {
                            // If the labels of edge1 and edge 2 in  node1 -(edge1)-> node2 -(edge2)-> node3 are equal,
                            // ad an ede node1 --> node3.
                            if (edge1.Label == edge2.Label)
                            {
                                if (node.AddEdge(new Edge<NFANode>(edge2.Target, edge1.Label, edge1.Description)))
                                {
                                    changed = true;
                                }
                            }
                        }
                    }
                }
            }

            // Add edges from Ordering --> Grouping
            foreach (NFANode orderNode in nodes.Values)
            {
                if (orderNode.OrderDescription != null)
                {
                    for (int i = 1; i <= orderNode.OrderDescription.Size; i++)
                    {
                        // Create the corresponding grouping.
                        NestableHashSet<string> groupingDescription = new NestableHashSet<string>();
                        for (int j = 0; j < i; j++)
                            groupingDescription.Add(orderNode.OrderDescription[j].Property);

                        // If the grouping does not exist, add it.
                        foreach (NFANode groupingNode in nodes.Values)
                            if (groupingNode.GroupingDescription != null &&
                                groupingNode.GroupingDescription.SetEquals(groupingDescription))
                                orderNode.AddEdge(new Edge<NFANode>(groupingNode, NFANode.EPSILON, "ε"));
                    }
                }
            }
        }

        /// <summary>
        /// Removes artificial (not needed) nodes from NFA to minimize it before DFA conversion.
        /// </summary>
        public void PruneArtificialNodes()
        {
            bool changed = true;
            // Create a list since we cannot iterate over the collection while modifying it.
            List<NFANode> nodeList = new List<NFANode>(nodes.Values);

            while (changed)
            {
                changed = false;

                // Merge all artificial nodes that have equivalent outgoing edges.
                for (int i = 0; i < nodeList.Count; i++)
                {
                    if (nodeList[i].Id == 0)
                    {
                        for (int j = i + 1; j < nodeList.Count; j++)
                        {
                            if (nodeList[j].Id == 0 && nodeList[i].BehavesLike(nodeList[j]))
                            {
                                // We have found two node which behaves alike. Replace the latter with the first.
                                for (int k = 0; k < nodeList.Count; k++)
                                {
                                    nodeList[k].ReplaceTarget(nodeList[j], nodeList[i]);
                                }
                                // Remove the latter node and decrement the counter.
                                nodes.Remove((object)nodeList[j].OrderDescription ?? (object)nodeList[j].GroupingDescription);
                                nodeList.RemoveAt(j);
                                j--;
                                changed = true;
                            }
                        }
                    }
                }

                // Remove all artificial nodes that have no outgoing edges or only one outgoing epsilon-edge.
                for (int i = 0; i < nodeList.Count; i++)
                {
                    if (nodeList[i].Id == 0)
                    {
                        bool remove = false;
                        if (nodeList[i].Edges.Count == 0)
                        {
                            // No outgoing edges, remove all references to it and mark it for removal.
                            for (int k = 0; k < nodeList.Count; k++)
                            {
                                nodeList[k].RemoveTarget(nodeList[i]);
                            }
                            remove = true;
                        }
                        else if (nodeList[i].Edges.Count == 1 && nodeList[i].Edges.Single().Label == NFANode.EPSILON)
                        {
                            // Only one outgoing epsilon edge - remove it and redirect any incoming edges.
                            NFANode target = nodeList[i].Edges.Single().Target;
                            for (int k = 0; k < nodeList.Count; k++)
                            {
                                nodeList[k].ReplaceTarget(nodeList[i], target);
                            }
                            remove = true;
                        }
                        // Remove the node.
                        if (remove)
                        {
                            nodes.Remove(nodeList[i].OrderDescription ?? (object) nodeList[i].GroupingDescription);
                            nodeList.RemoveAt(i);
                            i--;
                            changed = true;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets the maximum label for any edge in the NFA.
        /// </summary>
        public int GetMaxLabel()
        {
            int max = 0;
            foreach (NFANode node in nodes.Values)
            {
                foreach (Edge<NFANode> edge in node.Edges)
                {
                    if (edge.Label != NFANode.EPSILON && edge.Label > max)
                    {
                        max = edge.Label;
                    }
                }
            }
            return max;
        }

        public override string ToString()
        {
            return string.Format("[NFA{0}{1}{0}]{0}", Environment.NewLine,
                                 Utils.JoinEnumerable(nodes.Values, node => node.ToString(), "\t",
                                                      "," + Environment.NewLine, true));
        }

        /// <summary>Insers HTML entities for some common symbols and terms.</summary>
        private static string formatAndInsertHtmlEntities(string input)
        {
            input = input ?? string.Empty;
            input = input.Replace("->", "&rarr;");
            input = input.Replace("ASC", "&darr;");
            input = input.Replace("DESC", "&uarr;");
            return input;
        }

        /// <summary>Generates and returns the dot graph description of this NFA instance.</summary>
        public string GetDot()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine("digraph test {");
            sb.AppendLine("node [fontsize=12];");
            sb.AppendLine("edge [fontsize=12];");

            // Output all the nodes.
            foreach (NFANode node in nodes.Values)
            {
                // Use a rounded rectangle for orders and a rectangle for groupings.
                if (node.OrderDescription != null)
                {
                    if (node.OrderDescription.Size > 0)
                        sb.AppendFormat("{0} [label=\"{1}\",shape=box,style=rounded];{2}", node.GetHashCode(),
                                        node.SequenceNumber + "," + node.Id + " " +
                                        formatAndInsertHtmlEntities(node.OrderDescription.ToString().Replace(",", ",\\n")),
                                        Environment.NewLine);
                    else // The root node is a circle
                        sb.AppendFormat("{0} [label=\"{1}\",shape=circle];{2}", node.GetHashCode(),
                                        node.SequenceNumber + "," + node.Id + " ", Environment.NewLine);
                }
                else
                {
                    sb.AppendFormat("{0} [label=\"{{{1}}}\",shape=box];{2}", node.GetHashCode(),
                                    node.SequenceNumber + "," + node.Id + " " +
                                    formatAndInsertHtmlEntities(node.GroupingDescription.ToString()),
                                    Environment.NewLine);
                }

                // Then output all the outgoing edges for this node.
                foreach (Edge<NFANode> edge in node.Edges)
                {
                    sb.AppendFormat("{0} -> {1} [label = \"{2}\"];{3}", node.GetHashCode(), edge.Target.GetHashCode(),
                                    formatAndInsertHtmlEntities(edge.Description.Replace("->", "->\\n")),
                                    Environment.NewLine);
                }
            }

            sb.AppendLine("}");
            return sb.ToString();
        }
    }
}