﻿#region

using System;
using System.Collections.Generic;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine.Orders
{
    /// <summary>
    /// Represents a node in the NFA.
    /// </summary>
    internal class NFANode : IComparable<NFANode>
    {
        /// <summary>Label for epsilon edges.</summary>
        public const int EPSILON = int.MaxValue;

        private HashSet<Edge<NFANode>> edges = new HashSet<Edge<NFANode>>();

        /// <summary>The outgoing edges from this node.</summary>
        public HashSet<Edge<NFANode>> Edges
        {
            get { return edges; }
        }

        /// <summary>Logical order description for this node, if any.</summary>
        public OrderDescription OrderDescription { get; set; }
        /// <summary>Logical grouping description for this node, if any.</summary>
        public NestableHashSet<string> GroupingDescription { get; set; }
        /// <summary>Id of this node.</summary>
        public int Id { get; set; }
        /// <summary>Ever-increasing sequence number.</summary>
        public int SequenceNumber { get; set; }

        #region IComparable<NFANode> Members

        public int CompareTo(NFANode other)
        {
            return SequenceNumber.CompareTo(other.SequenceNumber);
        }

        #endregion

        /// <summary>
        /// Adds the specified outgoing edge to this node. Will se origin.
        /// </summary>
        /// <param name="edge">The edge to add.</param>
        /// <returns>true if the edge was added, false if not (already there or self-edge).</returns>
        public bool AddEdge(Edge<NFANode> edge)
        {
            if (edge.Target == this)
                return false;
            if (edges.Contains(edge))
                return false;

            edge.Origin = this;
            edges.Add(edge);
            return true;
        }

        /// <summary>
        /// Determines if this node behaves like the specified node (has the same set of outgoing edges).
        /// </summary>
        /// <param name="other">The node to compare with.</param>
        /// <returns>true if it does, false if not.</returns>
        public bool BehavesLike(NFANode other)
        {
            return edges.SetEquals(other.edges);
        }

        /// <summary>
        /// For all edges to oldTarget, replace the target of the edge with newTarget.
        /// </summary>
        /// <param name="oldTarget">Look for edges targeting this node.</param>
        /// <param name="newTarget">New target for the edges.</param>
        public void ReplaceTarget(NFANode oldTarget, NFANode newTarget)
        {
            if (oldTarget == newTarget)
                return;

            foreach (Edge<NFANode> edge in new HashSet<Edge<NFANode>>(edges))
                if (edge.Target == oldTarget)
                    AddEdge(new Edge<NFANode>(newTarget, edge.Label, edge.Description));
            edges.RemoveWhere(edge => edge.Target == oldTarget);
        }

        /// <summary>
        /// Removes all outgoing edges to the specified target.
        /// </summary>
        /// <param name="target">Target for edges to remove.</param>
        public void RemoveTarget(NFANode target)
        {
            edges.RemoveWhere(edge => edge.Target == target);
        }

        public override string ToString()
        {
            return string.Format("{0} {1}", Id, OrderDescription);
        }
    }
}