﻿#region

using System.Collections;
using System.Collections.Generic;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine.Orders
{
    /// <summary>
    /// Represents an ordering state with a set of logical orderings and groupings
    /// in the API. Maps 1-to-1 with states in the final DFA.
    /// </summary>
    public class Order
    {
        public Order(int id, HashSet<OrderDescription> description)
        {
            Id = id;
            Description = description;
        }

        public Order(int id, HashSet<NestableHashSet<string>> groupingDescription)
        {
            Id = id;
            GroupingDescription = groupingDescription;
        }

        /// <summary>Stores which logical ordering and grouping this state satisfies.</summary>
        public BitArray Compatibility { get; set; }
        /// <summary>Bit map having '1' for all outgoing FDs from this state, indexed by FD Id.</summary>
        public int PossibleTransitions { get; set; }
        /// <summary>Array of pointers storing the transitions, indexed by FD Id.</summary>
        public Order[] Transitions { get; set; }
        /// <summary>Id of this state</summary>
        public int Id { get; set; }
        /// <summary>Set of logical orderings satisfied in this state. For display purposes only.</summary>
        public HashSet<OrderDescription> Description { get; set; }
        /// <summary>Set og logical groupings satisfie in this state. For display purposes only.</summary>
        public HashSet<NestableHashSet<string>> GroupingDescription { get; set; }

        /// <summary>Is this the null-order (no order)</summary>
        public bool IsNull
        {
            get { return Id == 0; }
        }

        /// <summary>Does order state b satisfy order a?</summary>
        public static bool operator <=(Order a, Order b)
        {
            return b.Compatibility[a.Id];
        }

        /// <summary>Does order state a satisfy order b?</summary>
        public static bool operator >=(Order a, Order b)
        {
            return b <= a;
        }

        /// <summary>Does order state b satisfy order a and is not the same order?</summary>
        public static bool operator <(Order a, Order b)
        {
            return a <= b && a != b;
        }

        /// <summary>Does order state a satisfy order b and is not the same order?</summary>
        public static bool operator >(Order a, Order b)
        {
            return b < a;
        }

        /// <summary>
        /// Apply the FD with the specified Id.
        /// </summary>
        /// <param name="dependencyId">Id of FD to apply.</param>
        /// <returns>The new ordering state.</returns>
        public Order Apply(int dependencyId)
        {
            return Transitions[dependencyId];
        }

        /// <summary>
        /// Apply the specified FD.
        /// </summary>
        /// <param name="dependency">FD to apply</param>
        /// <returns>The new ordering state.</returns>
        public Order Apply(Dependency dependency)
        {
            return Apply(dependency.Id);
        }

        public override string ToString()
        {
            return string.Format("Order(Id:{0})", Id,
                                 Description != null
                                     ? Utils.JoinEnumerable<OrderDescription>(Description)
                                     : GroupingDescription.ToString());
        }
    }
}