﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine.Orders
{
    /// <summary>
    /// An item in an order description, for example "a ASC".
    /// </summary>
    public struct OrderDescriptionItem
    {
        /// <summary>Ordering ascending (true) or descending (false) on this item.</summary>
        public bool Ascending;
        /// <summary>Property name of order item.</summary>
        public string Property;

        public OrderDescriptionItem(string property, bool ascending)
        {
            Property = property;
            Ascending = ascending;
        }
    }

    /// <summary>
    /// Description of an order, for example [a ASC, b DESC] and so on.
    /// </summary>
    public class OrderDescription : IEnumerable<OrderDescriptionItem>, ICloneable
    {
        private List<OrderDescriptionItem> data;

        /// <summary>
        /// Constructs a new order description by parsing a string description.
        /// </summary>
        /// <param name="description">String description on the format "prop1 ASC,prop2 DESC" and so on.</param>
        public OrderDescription(string description)
            : this(parseOrderDesc(description))
        {
        }

        public OrderDescription(params OrderDescriptionItem[] items)
            : this((IEnumerable<OrderDescriptionItem>) items)
        {
        }

        public OrderDescription(IEnumerable<OrderDescriptionItem> items)
            : this(new List<OrderDescriptionItem>(items))
        {
        }

        private OrderDescription(List<OrderDescriptionItem> items)
        {
            data = items;
        }

        /// <summary>Number of decription items.</summary>
        public int Size
        {
            get { return data.Count; }
        }

        /// <summary>Get the description item with the specified.</summary>
        public OrderDescriptionItem this[int index]
        {
            get { return data[index]; }
            set { data[index] = value; }
        }

        #region ICloneable Members

        public object Clone()
        {
            return new OrderDescription((IEnumerable<OrderDescriptionItem>) data);
        }

        #endregion

        #region IEnumerable<OrderDescriptionItem> Members

        public IEnumerator<OrderDescriptionItem> GetEnumerator()
        {
            return data.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        private static List<OrderDescriptionItem> parseOrderDesc(string description)
        {
            List<OrderDescriptionItem> items = new List<OrderDescriptionItem>();
            foreach (string item in description.Split(','))
            {
                string[] split = item.Trim().Split(' ');
                items.Add(new OrderDescriptionItem(split[0], split[1] == "ASC"));
            }
            return items;
        }

        /// <summary>Append a description item to the end.</summary>
        public void Append(OrderDescriptionItem item)
        {
            data.Add(item);
        }

        /// <summary>Limit (truncate) the order description to the specified length.</summary>
        public void LimitTo(int length)
        {
            if (data.Count > length)
            {
                data.RemoveRange(length, data.Count - length);
            }
        }

        /// <summary>Append the specified item to the end, returning a new order description
        /// and not modifying the original.</summary>
        public OrderDescription ConstAppend(OrderDescriptionItem item)
        {
            OrderDescription result = new OrderDescription((IEnumerable<OrderDescriptionItem>) data);
            result.data.Add(item);
            return result;
        }

        /// <summary>Remove the description item at the specified index.</summary>
        public OrderDescription RemoveAt(int index)
        {
            List<OrderDescriptionItem> items = new List<OrderDescriptionItem>(data.Count - 1);
            for (int i = 0; i < index; i++)
            {
                items.Add(data[i]);
            }
            for (int i = index + 1; i < data.Count; i++)
            {
                items.Add(data[i]);
            }
            return new OrderDescription(items);
        }

        /// <summary>
        /// Copies the OrderDescription and inserts <paramref name="item"/>
        /// at position index.
        /// </summary>
        public OrderDescription Insert(int index, OrderDescriptionItem item)
        {
            List<OrderDescriptionItem> items = new List<OrderDescriptionItem>(data.Count + 1);
            for (int i = 0; i < index; i++)
            {
                items.Add(data[i]);
            }
            items.Add(item);
            for (int i = index; i < data.Count; i++)
            {
                items.Add(data[i]);
            }
            return new OrderDescription(items);
        }

        /// <summary>
        /// Copies the OrderDescription and replaces the item at position <paramref name="index"/>
        /// with property <paramref name="newProperty"/> and the same order. If <paramref name="newProperty"/>
        /// is found before position <paramref name="index"/>, the previous item is just removed and 
        /// <paramref name="newProperty"/> not inserted. If replaced, any occurences of <paramref name="newProperty"/> after
        /// <paramref name="index"/> is removed.
        /// </summary>
        public OrderDescription Replace(int index, string newProperty)
        {
            OrderDescription orderDescription = new OrderDescription();
            int cut = index;
            for (int iter = 0; iter < data.Count; iter++)
            {
                if (iter == index)
                {
                    if (index == cut)
                    {
                        orderDescription.data.Add(new OrderDescriptionItem(newProperty, data[index].Ascending));
                    }
                }
                else if (data[iter].Property == newProperty)
                {
                    if (iter < cut)
                    {
                        orderDescription.data.Add(data[iter]);
                        cut = iter;
                    }
                    else
                    {
                        continue;
                    }
                }
                else
                {
                    orderDescription.data.Add(data[iter]);
                }
            }
            return orderDescription;
        }

        /// <summary>If a is a prefix of b</summary>
        public static bool operator <=(OrderDescription a, OrderDescription b)
        {
            if (a.data.Count > b.data.Count)
                return false;

            return Compare(a, b);
        }

        /// <summary>If b is a prefix of a.</summary>
        public static bool operator >=(OrderDescription a, OrderDescription b)
        {
            return b <= a;
        }

        /// <summary>If a is a proper prefix (not equal) of b.</summary>
        public static bool operator <(OrderDescription a, OrderDescription b)
        {
            if (a.data.Count >= b.data.Count)
                return false;

            return Compare(a, b);
        }

        /// <summary>If b is a proper prefix (not equal) of a.</summary>
        public static bool operator >(OrderDescription a, OrderDescription b)
        {
            return b < a;
        }

        /// <summary>If descriptions a and b are equal.</summary>
        public static bool operator ==(OrderDescription a, OrderDescription b)
        {
            return ReferenceEquals(a, b) || (!ReferenceEquals(null, a) && a.Equals(b));
        }

        /// <summary>If descriptions a and b are not equal.</summary>
        public static bool operator !=(OrderDescription a, OrderDescription b)
        {
            return !(a == b);
        }

        public override bool Equals(object obj)
        {
            OrderDescription other = obj as OrderDescription;
            if (ReferenceEquals(null, other))
                return false;

            // Verify that all data items are equal.
            if (data.Count != other.data.Count)
                return false;

            return Compare(this, other);
        }

        // Compare a and b for equality.
        private static bool Compare(OrderDescription a, OrderDescription b)
        {
            for (int i = 0; i < a.data.Count; i++)
                if (a.data[i].Property != b.data[i].Property || a.data[i].Ascending != b.data[i].Ascending)
                    return false;

            return true;
        }

        public override int GetHashCode()
        {
            int hash = 0;
            for (int i = 0; i < data.Count; i++)
                hash ^= data[i].Property.GetHashCode() ^ (i*data[i].Ascending.GetHashCode());

            return hash;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder("[");
            for (int i = 0; i < data.Count; i++)
                sb.AppendFormat("{0}{1},", data[i].Property, data[i].Ascending ? " ASC" : " DESC");

            if (sb.Length > 1)
                sb.Remove(sb.Length - 1, 1);

            sb.Append("]");
            return sb.ToString();
        }
    }
}