﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine.Orders
{
    /// <summary>
    /// Main class for the ordering component. Provides an API to get orderings, groupings and functional
    /// dependencies and orchestrates the construction of the finite state machines.
    /// What referred to as "Properties" in here is really attributes or "fields".
    /// </summary>
    public class OrderManager
    {
        /// <summary>TODO: Remove this. Due to a optimization bug, this is used to get realistic timings.</summary>
        public static bool LIMIT1 = false;
        // Maps string-desc to FDs to quicly look them up
        private readonly Dictionary<string, Dependency> dependencies = new Dictionary<string, Dependency>();
        private readonly List<Dependency> dependencyList = new List<Dependency>(); // List of FDs.

        // Proxies for requested groupings.
        private readonly Dictionary<NestableHashSet<string>, GroupingProxy> groupingProxies =
            new Dictionary<NestableHashSet<string>, GroupingProxy>();

        // Proxies for requested orderings.
        private readonly Dictionary<OrderDescription, OrderProxy> orderProxies =
            new Dictionary<OrderDescription, OrderProxy>();

        private int dependencyCount = 1; // Sequence counter for FDs.
        private List<Dependency> filteredDependencies; // List of filtered FDs.
        private int graphCount = 1; // Sequence counter for graph generation
        private Dictionary<string, HashSet<string>> groupingInfo;
        private HashSet<NestableHashSet<string>> groupingTemplates; // Equivalence (=, not share)-mapped interesting groupings.
        private List<OrderTemplate> orderTemplates; // Equivalence-mapped interesting orderings.

        private HashSet<string> propertyDomain; // Domain of all used properties.
        private Dictionary<string, PropertyFilter> propertyFilter; // Usage for each property
        private Dictionary<string, string> propertyMap; // Map for equivalent (=, not share equivalent) properties.

        /// <summary>
        /// Visualizer to be set if the order manager should visualize the NFSMs generated.
        /// </summary>
        public OperatorGraphVisualizer Visualizer { get; set; }
        /// <summary>
        /// Mapping of share equivalent attributes so the ordering manager can recogninze
        /// share equivalent orderings.
        /// </summary>
        public Dictionary<string, string> ShareEquivalentMappings { get; set; }

        /// <summary>Number of DFs in total.</summary>
        public int DependencyCount
        {
            get { return dependencyCount - 1; }
        }

        /// <summary>The null order - represents no ordering or grouping.</summary>
        public Order NullOrder { get; private set; }

        /// <summary>Ordering state for the null order.</summary>
        public OrderingState NullOrderingState
        {
            get { return GetOrderingState(new OrderDescription(), false); }
        }
        
        /// <summary>Filters the FDs, removing not used elements of them or entire FDs.</summary>
        private void filterDependencies()
        {
            // Initialize result dictionary to None for each element
            propertyFilter = new Dictionary<string, PropertyFilter>();
            foreach (string property in propertyDomain)
                propertyFilter[property] = PropertyFilter.None;

            // Flag all used properties as interesting 
            foreach (OrderProxy orderProxy in orderProxies.Values)
                foreach (OrderDescriptionItem item in orderProxy.Description)
                    propertyFilter[item.Property] |= item.Ascending ? PropertyFilter.Asc : PropertyFilter.Desc;

            // Propagate interesting properties transitively (given a -> b, if b is interesting, then a is also interesting).
            bool found = true;
            while (found)
            {
                found = false;
                foreach (Dependency dependency in dependencies.Values)
                {
                    foreach (string fromProperty in dependency.From)
                    {
                        PropertyFilter old = propertyFilter[fromProperty];
                        foreach (string toProperty in dependency.To)
                        {   // Propagate to -> from.
                            propertyFilter[fromProperty] |= propertyFilter[toProperty];
                            // If there's a change, set found to continue iterating.
                            if (propertyFilter[fromProperty] != old)
                                found = true;
                        }
                    }
                }
            }

            // Build a new list of dependencies by using the filter on the old list of FDs.
            filteredDependencies = new List<Dependency>();
            foreach (Dependency dependency in dependencies.Values)
            {
                Dependency newDependency = new Dependency(dependency.Id, dependency.Equal);

                foreach (string property in dependency.From)
                    if (propertyFilter[property] > 0)
                        newDependency.From.Add(property);

                foreach (string property in dependency.To)
                    if (propertyFilter[property] > 0)
                        newDependency.To.Add(property);

                if (dependency.To.Count > 0) // Zero props on left means constant.)
                    filteredDependencies.Add(newDependency);
            }
        }

        /// <summary>Builds equivalene classes of attributes/properties.</summary>
        private void mapEquivalentProperties()
        {
            // Create a map, mapping properties to their representative.
            propertyMap = new Dictionary<string, string>();
            // Initially they point to themselves.
            foreach (string property in propertyDomain)
                propertyMap[property] = property;

            // Continue until nothing changes to catch transitive equivalences.
            while (true)
            {
                bool changed = false;
                foreach (Dependency dependency in dependencies.Values)
                {
                    if (dependency.Equal)
                    {
                        // If we have an equivalency dependency (a=b), map the larger to the smaller one.
                        if (propertyMap[dependency.From.Single()].CompareTo(propertyMap[dependency.To.Single()]) < 0)
                        {
                            propertyMap[dependency.To.Single()] = propertyMap[dependency.From.Single()];
                            changed = true;
                        }
                        else if (propertyMap[dependency.From.Single()].CompareTo(propertyMap[dependency.To.Single()]) > 0)
                        {
                            propertyMap[dependency.From.Single()] = propertyMap[dependency.To.Single()];
                            changed = true;
                        }
                    }
                }
                if (!changed)
                    break;
            }
        }

        /// <summary>
        /// Create equivalence mappings for orderings with the 
        /// representatives in orderTemplates. orderTemplates will
        /// contain representatives for all interesting orderings.</summary>
        private void mapEquivalentOrderings()
        {
            orderTemplates = new List<OrderTemplate>();

            foreach (OrderProxy orderProxy in orderProxies.Values)
            {
                // Build an order description by sending the items through propertyMap.
                OrderTemplate current = new OrderTemplate();
                current.OrderDescription = new OrderDescription();
                foreach (OrderDescriptionItem item in orderProxy.Description)
                    current.OrderDescription.Append(new OrderDescriptionItem(propertyMap[item.Property], item.Ascending));

                // Add it to orderTemplates if it is unique.
                bool found = false;
                foreach (OrderTemplate orderTemplate in orderTemplates)
                {
                    if (orderTemplate.OrderDescription == current.OrderDescription)
                    {
                        found = true;
                        break;
                    }
                }
                if (!found)
                    orderTemplates.Add(current);
            }
        }

        /// <summary>
        /// Create equivalence mappings for groupings with the 
        /// representatives in groupingTemplates. groupingTemplates will
        /// contain representatives for all interesting groupings.</summary>
        private void mapEquivalentGroupings()
        {
            groupingTemplates = new HashSet<NestableHashSet<string>>();

            foreach (GroupingProxy groupingProxy in groupingProxies.Values)
            {
                // Build a grouping by sending all members through propertymap.
                NestableHashSet<string> current = new NestableHashSet<string>();
                foreach (string property in groupingProxy.Description)
                    current.Add(propertyMap[property]);
                // Add it to the set, which automatically rejects duplicates.
                groupingTemplates.Add(current);
            }
        }

        /// <summary>Create equivalence classes for groupings.</summary>
        private void calculateGroupingInfo()
        {
            groupingInfo = new Dictionary<string, HashSet<string>>();

            foreach (GroupingProxy groupingProxy in groupingProxies.Values)
                foreach (string property in groupingProxy.Description)
                    groupingInfo[property] = new HashSet<string> {property};

            HashSet<string> properties = new HashSet<string>();
            foreach (Dependency dependency in dependencies.Values)
            {
                if (dependency.From.Count == 0) // Constant
                    properties.UnionWith(dependency.To);
                else
                    properties.UnionWith(dependency.From);
            }
            foreach (string property in properties)
            {
                HashSet<string> value;
                if (!groupingInfo.TryGetValue(property, out value))
                    groupingInfo[property] = value = new HashSet<string>();
                value.Add(property);
            }

            // Transitivity
            while (true)
            {
                bool changed = false;
                foreach (KeyValuePair<string, HashSet<string>> pair in groupingInfo)
                {
                    foreach (Dependency dependency in dependencies.Values)
                    {
                        if (dependency.From.Count == 0) // Constant
                            continue;

                        bool mismatch = false;
                        foreach (string fromProperty in dependency.From)
                        {
                            if (!pair.Value.Contains(fromProperty))
                            {
                                mismatch = true;
                                break;
                            }
                        }
                        if (mismatch)
                            continue;

                        foreach (string toProperty in dependency.To)
                            if (pair.Value.Add(toProperty))
                                changed = true;
                    }
                }
                if (!changed)
                    break;
            }

            // Filter
            foreach (KeyValuePair<string, HashSet<string>> pair in groupingInfo)
                pair.Value.IntersectWith(groupingInfo.Keys);
        }

        /// <summary>Find all the used properties in orderings, groupings or FDs. </summary>
        private void determinePropertyDomain()
        {
            // Find all used properties in orderings.
            propertyDomain = new HashSet<string>();
            foreach (OrderProxy orderProxy in orderProxies.Values)
                foreach (OrderDescriptionItem item in orderProxy.Description)
                    propertyDomain.Add(item.Property);

            // ... and in groupings
            foreach (GroupingProxy groupingProxy in groupingProxies.Values)
                propertyDomain.UnionWith(groupingProxy.Description);

            // ... and on both sides of FDs.
            foreach (Dependency dependency in dependencies.Values)
            {
                propertyDomain.UnionWith(dependency.From);
                propertyDomain.UnionWith(dependency.To);
            }
        }


        /// <summary>
        /// Callback from the NFA, called for each node in the NFA.
        /// Will consider all FDs for appliance to this node.
        /// </summary>
        private void considerTransitions(NFA nfa, NFANode node, object data)
        {
            // For each dependency, consider if it leads to new interesting orderings from this node.
            foreach (Dependency dependency in filteredDependencies)
                // For all RHS elements, since a->bc is the same as a->b and a->c.
                foreach (string toProperty in dependency.To)
                    considerTransition(nfa, node, dependency, toProperty);
        }

        /// <summary>
        /// Consideres the specified toProperty for the specified
        /// dependency for appliance to the given node.
        /// </summary>
        private void considerTransition(NFA nfa, NFANode node, Dependency dependency, string toProperty)
        {
            if (node.OrderDescription == null)
                return;

            // Determine the relevant order templates. It is relevant if it contains
            // a RHS property of this dependency.
            List<OrderTemplate> relevantTemplates = new List<OrderTemplate>();
            foreach (OrderTemplate orderTemplate in orderTemplates)
            {
                for (int i = orderTemplate.OrderDescription.Size - 1; i >= 0; i--)
                {
                    if (orderTemplate.OrderDescription[i].Property == propertyMap[toProperty])
                    {
                        orderTemplate.MaxRelevant = i + 1;
                        relevantTemplates.Add(orderTemplate);
                        break;
                    }
                }
            }

            if (dependency.From.Count == 0) //  Ø -> ... (constant)
            {
                // If the toProperty is already in the ordering, remove it.
                // We don't want it twice, and this can be an opportunity to get it
                // earlier on (to the left) in the ordering.
                OrderDescription baseDesc = node.OrderDescription;
                for (int i = 0; i < node.OrderDescription.Size; i++)
                {
                    if (node.OrderDescription[i].Property == toProperty)
                    {
                        baseDesc = node.OrderDescription.RemoveAt(i);
                        break;
                    }
                }

                // Try inserting at all locations the the right.
                int limit = baseDesc.Size;
                int startAt = 0;
                for (int i = 0; i <= limit; i++)
                {
                    // Determine maximum interesting length of the ordering.
                    int max = 0;
                    for (int j = startAt; j < relevantTemplates.Count; j++)
                    {
                        int r = relevantTemplates[j].RelevantLength(baseDesc, i, toProperty, propertyMap);
                        if (r <= i)
                        {
                            startAt = j + 1;
                        }
                        else if (r > max)
                        {
                            max = r;
                        }
                    }

                    // Interesting length = 0, so abort.
                    if (max == 0)
                        break;

                    // Insert both ASC and DESC, if interesting.
                    for (int direction = 1; direction <= 2; direction++)
                    {
                        if ((int) propertyFilter[toProperty] == direction)
                        {
                            // Insert 
                            OrderDescription current = baseDesc.Insert(i,
                                                                       new OrderDescriptionItem(toProperty,
                                                                                                direction ==
                                                                                                (int) PropertyFilter.Asc));
                            // Limit to max interesting length
                            current.LimitTo(max);
                            
                            // Add a new node if not exists and add an edge to it.
                            NFANode target = nfa.GetOrCreateNode(current);
                            node.AddEdge(new Edge<NFANode>(target, dependency.Id, dependency.ToString()));
                        }
                    }
                }
            }
            else
            {
                // Iterate over this node's OrderDescription.
                //   1. If not all LHS properties are found, abort.
                //   2. If all LHS properties are found, pos will point at first one.
                int pos = node.OrderDescription.Size, remaining = dependency.From.Count;
                for (int i = 0; i < node.OrderDescription.Size; i++)
                {
                    foreach (string fromProperty in dependency.From)
                    {
                        if (node.OrderDescription[i].Property == fromProperty)
                        {
                            remaining--;
                            if (remaining == 0)
                            {
                                pos = i;
                            }
                            break;
                        }
                    }
                }

                // Abort if not all LHS are found.
                if (pos < node.OrderDescription.Size)
                {
                    // Check if the RHS property is interesting
                    if ((propertyFilter[toProperty] &
                         (node.OrderDescription[pos].Ascending ? PropertyFilter.Asc : PropertyFilter.Desc)) != 0)
                    {
                        // If the toProperty is already in the ordering, remove it.
                        // We don't want it twice, and this can be an opportunity to get it
                        // earlier on (to the left) in the ordering.
                        OrderDescription baseDesc = node.OrderDescription;
                        for (int i = 0; i < node.OrderDescription.Size; i++)
                        {
                            if (node.OrderDescription[i].Property == toProperty)
                            {
                                baseDesc = node.OrderDescription.RemoveAt(i);
                                break;
                            }
                        }
                        // If it's an equality, replace as well as insert.
                        if (dependency.Equal)
                        {
                            OrderDescription current = baseDesc.Replace(pos, toProperty);
                            NFANode target = nfa.GetOrCreateNode(current);

                            node.AddEdge(new Edge<NFANode>(target, dependency.Id, dependency.ToString()));
                        }

                        // Try inserting at all locations the the right.
                        int startAt = 0;
                        for (int i = pos + 1; i <= baseDesc.Size; i++)
                        {
                            // Determine maximum interesting length of the ordering.
                            int max = 0;
                            for (int j = startAt; j < relevantTemplates.Count; j++)
                            {
                                int r = relevantTemplates[j].RelevantLength(baseDesc, i, toProperty, propertyMap);
                                if (r <= i)
                                {
                                    startAt = j + 1;
                                }
                                else if (r > max)
                                {
                                    max = r;
                                }
                            }

                            if (max == 0)
                                break;
                            // TODO: Due to a bug not fixed, we allow certain tests to limit manually.
                            if (LIMIT1 && max > 1)
                                max = 1;

                            // Insert to the right.
                            OrderDescription current = baseDesc.Insert(i,
                                                                       new OrderDescriptionItem(toProperty,
                                                                                                node.OrderDescription[
                                                                                                    pos].Ascending));
                            // Limit to the maximum interesting length.
                            current.LimitTo(max);

                            // Add a new node if not exists and add an edge to id.
                            NFANode target = nfa.GetOrCreateNode(current);
                            node.AddEdge(new Edge<NFANode>(target, dependency.Id, dependency.ToString()));
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Callback from the NFA, called for each node in the NFA.
        /// Will consider all FDs for appliance to this node for groupings.
        /// </summary>
        private void considerGroupingTransitions(NFA nfa, NFANode node, object data)
        {
            // For each dependency, consider if it leads to new interesting groupings from this node.
            foreach (Dependency dependency in filteredDependencies)
                // For all RHS elements, since a->bc is the same as a->b and a->c.
                foreach (string toProperty in dependency.To)
                    considerGroupingTransition(nfa, node, dependency, toProperty);
        }

        /// <summary>
        /// Consideres the specified toProperty for the specified
        /// dependency for appliance to the given node.
        /// </summary>
        private void considerGroupingTransition(NFA nfa, NFANode node, Dependency dependency, string toProperty)
        {
            // Check if it is a grouping node (it could be an ordering node)
            if (node.GroupingDescription == null)
                return;

            // Check if we have a constant FD.
            if (dependency.From.Count == 0)
            {
                if (groupingInfo[toProperty].IsSubsetOf(node.GroupingDescription))
                    return;

                // Create a new grouping by adding the constant property.
                NestableHashSet<string> other = new NestableHashSet<string>(node.GroupingDescription);
                other.Add(toProperty);
                // Create a new grouping node if not exists and add an edge to it.
                NFANode target = nfa.GetOrCreateNode(other);
                if (target != node)
                    node.AddEdge(new Edge<NFANode>(target, dependency.Id, dependency.ToString()));
            }
            else
            {
                // Non-constant FD. All LHS properties must be in the grouping we're looking at to continue.
                if (!dependency.From.IsSubsetOf(node.GroupingDescription))
                    return;

                // Check if the new target grouping that can be achieved is interesting
                // by finding a matching grouping template
                bool found = false;
                string mappedTo = propertyMap[toProperty];
                foreach (NestableHashSet<string> groupingTemplate in groupingTemplates)
                {
                    // Check that the property we're about to insert is in the template.
                    if (groupingTemplate.Contains(mappedTo))
                    {
                        // Check that all the properties in the current node is in the template.
                        bool ok = true;
                        foreach (string property in node.GroupingDescription)
                        {
                            if (!groupingTemplate.Contains(propertyMap[property]))
                            {
                                ok = false;
                                break;
                            }
                        }

                        if (ok)
                        {
                            found = true;
                            break;
                        }
                    }
                }
                // No matching template found, abort
                if (!found)
                    return;

                // Create the new grouping by adding to toProperty to the old.
                NestableHashSet<string> other = new NestableHashSet<string>(node.GroupingDescription);
                other.Add(toProperty);
                // Create a new grouping node if not exists and add an edge to it.
                NFANode target = nfa.GetOrCreateNode(other);
                if (target != node)
                    node.AddEdge(new Edge<NFANode>(target, dependency.Id, dependency.ToString()));
                
                // If it's an equal FD (a=b), also create a grouping with a replaced with b.
                if (dependency.Equal)
                {
                    other = new NestableHashSet<string>(other);
                    other.Remove(dependency.From.Single());
                    target = nfa.GetOrCreateNode(other);
                    if (target != node)
                        node.AddEdge(new Edge<NFANode>(target, dependency.Id, dependency.ToString()));
                }
            }
        }

        private void GraphCurrentState(NFA nfa, string filename)
        {
            if (Visualizer.Visualize)
                Visualizer.GenerateGraph(nfa.GetDot(), filename);
        }

        private void GraphCurrentState(NFA nfa)
        {
            GraphCurrentState(nfa, string.Format("nfa{0}", graphCount++));
        }


        /// <summary>
        /// Creates the finite automation and fills orderings and grouping proxies.
        /// Call this after all interesting orders and groupings have been established
        /// and before using any of them.
        /// </summary>
        public void PrepareOrders()
        {
            OrderingState dummy = NullOrderingState;

            // Find all mentioned properties
            determinePropertyDomain();
            // Filter unused FDs.
            filterDependencies();

            // Set ids of order and grouping proxies.
            int orderCount = 0;
            foreach (OrderProxy orderProxy in orderProxies.Values)
                if (orderProxy.Description.Size > 0)
                    orderProxy.Id = ++orderCount;
            foreach (GroupingProxy groupingProxy in groupingProxies.Values)
                groupingProxy.Id = ++orderCount;

            // Construct the NFA
            NFA nfa = new NFA();
            NFANode nfaRoot = nfa.GetOrCreateNode(new OrderDescription());
            // Add a node with and edge from the root for each interesting order.
            foreach (OrderProxy orderProxy in orderProxies.Values)
            {
                if (orderProxy.Description.Size > 0)
                {
                    NFANode current = nfa.GetOrCreateNode(orderProxy.Description);
                    current.Id = orderProxy.Id;
                    nfaRoot.AddEdge(new Edge<NFANode>(current, dependencyCount + current.Id,
                                                      orderProxy.Description.ToString()));
                }
            }
            // Add a node with and edge from the root for each interesting grouping.
            foreach (GroupingProxy groupingProxy in groupingProxies.Values)
            {
                NFANode current = nfa.GetOrCreateNode(groupingProxy.Description);
                current.Id = groupingProxy.Id;
                nfaRoot.AddEdge(new Edge<NFANode>(current, dependencyCount + current.Id,
                                                  groupingProxy.Description.ToString()));
            }

            // Graph nfa1
            GraphCurrentState(nfa);
            // Create equivalence classes for attributes.
            mapEquivalentProperties();
            // Create equivalence classes for orderings and groupings
            mapEquivalentOrderings();
            mapEquivalentGroupings();
            calculateGroupingInfo();

            // Add FD transitions for orderings
            nfa.ForAllNodes(considerTransitions, null);
            // Add FD transitions for groupings
            nfa.ForAllNodes(considerGroupingTransitions, null);
            // Graph nfa2
            GraphCurrentState(nfa);

            // Add share equivalent edges and graph nfa3.
            if (ShareEquivalentMappings != null)
                nfa.AddShareEquivalencyEdges(ShareEquivalentMappings);
            GraphCurrentState(nfa);

            // Add transitive and grouping epsilon edges to the NFA and graph nfa4.
            nfa.AddEpsilonEdges();
            GraphCurrentState(nfa);

            // Optimize the NFA and graph nfa5.
            nfaRoot.Id = orderCount + 1; // Just to make sure it is not removed by the next step
            nfa.PruneArtificialNodes();
            nfaRoot.Id = 0;
            GraphCurrentState(nfa);

            // Convert to a DFA and visualize
            DFA dfa = new DFA();
            DFANode dfaRoot = dfa.Convert(nfa, nfaRoot);
            if (Visualizer.Visualize)
                Visualizer.GenerateGraph(dfa.GetDot(), "dfa1");

            // Fill proxies
            dfaRoot.Order = new Order(0, new HashSet<OrderDescription> {new OrderDescription()});
            NullOrder = dfaRoot.Order;
            foreach (OrderProxy orderProxy in orderProxies.Values)
            {
                if (orderProxy.Id == 0)
                {
                    orderProxy.Order = NullOrder;
                }
                else
                {
                    // Fill by using the edge from the root.
                    foreach (Edge<DFANode> edge in dfaRoot.Edges)
                    {
                        if (edge.Label == (dependencyCount + orderProxy.Id))
                        {
                            edge.Target.Order = new Order(orderProxy.Id,
                                                          new HashSet<OrderDescription> {orderProxy.Description});
                            orderProxy.Order = edge.Target.Order;
                        }
                    }
                }
            }

            // Fill grouping proxies using the edges from the root.
            NullOrder = dfaRoot.Order;
            foreach (GroupingProxy groupingProxy in groupingProxies.Values)
            {
                foreach (Edge<DFANode> edge in dfaRoot.Edges)
                {
                    if (edge.Label == (dependencyCount + groupingProxy.Id))
                    {
                        edge.Target.Order = new Order(groupingProxy.Id,
                                                      new HashSet<NestableHashSet<string>> {groupingProxy.Description});
                        groupingProxy.Order = edge.Target.Order;
                    }
                }
            }

            // Create orders
            foreach (DFANode node in dfa)
            {
                if (node.Order == null)
                {
                    node.Order = new Order(++orderCount,
                                           new HashSet<OrderDescription>(
                                               node.Description.Select(nfanode => nfanode.OrderDescription)));
                }
            }

            // Create compatibility matrices
            foreach (DFANode dfaNode in dfa)
            {
                dfaNode.Order.Compatibility = new BitArray(orderCount + 1, false);
                foreach (NFANode nfaNode in dfaNode.Description)
                {
                    dfaNode.Order.Compatibility[nfaNode.Id] = true;
                }
                dfaNode.Order.Compatibility[dfaNode.Order.Id] = true;
                dfaNode.Order.Compatibility[NullOrder.Id] = true;
            }

            // Compare artificial orderings for subsets to increase compatibility.
            foreach (DFANode dfaNode1 in dfa)
                foreach (DFANode dfaNode2 in dfa)
                    if (dfaNode2.IsSubsetOf(dfaNode1))
                        dfaNode1.Order.Compatibility[dfaNode2.Order.Id] = true;

            // Create transition functions
            foreach (DFANode node in dfa)
            {
                node.Order.Transitions = new Order[dependencyCount];
                foreach (Dependency dependency in dependencies.Values)
                {
                    foreach (Edge<DFANode> edge in node.Edges)
                    {
                        if (dependency.Id == edge.Label)
                        {
                            node.Order.Transitions[dependency.Id] = edge.Target.Order;
                            node.Order.PossibleTransitions |= (1 << dependency.Id - 1);
                        }
                    }
                }
            }
            // Finally visualize.
            if (Visualizer != null)
                Visualizer.GenerateGraph(dfa.GetDot(), "dfa2");
        }

        /// <summary>Gets the specified ordering.</summary>
        public OrderProxy GetOrder(OrderDescription desc, bool produced)
        {
            OrderProxy proxy;
            if (orderProxies.TryGetValue(desc, out proxy))
            {
                if (produced && !proxy.Produced)
                    proxy.Produced = true;
                return proxy;
            }
            return orderProxies[desc] = new OrderProxy {Description = desc, Produced = produced};
        }

        /// <summary>Gets the specified ordering wrapped in an orderingstate.</summary>
        public OrderingState GetOrderingState(OrderDescription desc, bool produced)
        {
            return new OrderingState(this, GetOrder(desc, produced));
        }

        /// <summary>Gets the specified grouping.</summary>
        public GroupingProxy GetGrouping(NestableHashSet<string> description, bool produced)
        {
            GroupingProxy proxy;
            if (groupingProxies.TryGetValue(description, out proxy))
            {
                if (produced && !proxy.Produced)
                    proxy.Produced = true;
                return proxy;
            }

            return groupingProxies[description] = new GroupingProxy {Description = description, Produced = produced};
        }

        /// <summary>Creates and returns a new dependency from a description (a,b->c,d) or (a=b).</summary>
        public Dependency GetDependency(string description)
        {
            string[] split = description.Split(new[] {"->", "="}, StringSplitOptions.None);

            HashSet<string>[] sets = new[] {new HashSet<string>(), new HashSet<string>()};
            for (int i = 0; i < 2; i++)
                foreach (string part in split[i].Trim().Split(new[] {','}, StringSplitOptions.RemoveEmptyEntries))
                    sets[i].Add(part.Trim());
            return GetDependency(sets[0], sets[1], description.Contains("="));
        }

        /// <summary>Creates and returns a new dependency with one or zero (null) elements LHS and one RHS.</summary>
        public Dependency GetDependency(string from, string to, bool equal)
        {
            Validation.NotEquals("from", from, string.Empty);
            Validation.NotNullOrEmtpy("to", to);
            return GetDependency(from == null ? new HashSet<string>() : new HashSet<string> {from},
                                 new HashSet<string> {to}, equal);
        }

        /// <summary>Creates and returns a new dependency.</summary>
        public Dependency GetDependency(HashSet<string> from, HashSet<string> to, bool equal)
        {
            Validation.NoNulls("from", from);
            Validation.NoNullsOrEmpty("to", to);
            Validation.Assert("from, equal", !(from.Count == 0 && equal),
                              "Constant functional dependencies cannot be equalities.");

            Dependency dependency;
            string key = Utils.JoinEnumerable(from) + "#" + Utils.JoinEnumerable(to) + "#" + equal;

            if (!dependencies.TryGetValue(key, out dependency))
            {
                dependency = new Dependency(dependencyCount, from, to, equal);
                dependencies[key] = dependency;
                dependencyList.Add(dependency);
                if (equal)
                {
                    string inverseKey = Utils.JoinEnumerable(to) + "#" + Utils.JoinEnumerable(from) + "#" + equal;
                    dependencies[inverseKey] = new Dependency(dependencyCount, to, from, equal);
                }
                dependencyCount++;
            }
            return dependency;
        }

        public Dependency GetDependency(int dependencyId)
        {
            return dependencyList[dependencyId - 1];
        }

        #region Nested type: PropertyFilter

        [Flags]
        private enum PropertyFilter
        {
            None = 0,
            Asc = 1,
            Desc = 2,
            Group = 4
        }

        #endregion
    }
}