﻿#region

using System.Collections.Generic;
using System.Text;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine.Orders
{
    /// <summary>
    /// Ordering state for a plan. Wraps the Order class
    /// and additionally stores the FDs applied so far. Contains
    /// functionality to apply FDs and merge OrderingStates.
    /// </summary>
    public class OrderingState
    {
        // Maps the numeric value of (2^n) to (n + 1) for n < 32,
        // i.e. map the bit mask value to the Id of the dependency.
        private static readonly Dictionary<int, int> LOG2MAP;

        private int appliedDependencies; // Limited to 32 dependencies for now.
        private Order order;
        private OrderManager orderManager;
        private Proxy proxy;

        // Fill the log 2 map.
        static OrderingState()
        {
            LOG2MAP = new Dictionary<int, int>();
            for (byte i = 0; i < 32; i++)
                LOG2MAP[1 << i] = i + 1;
        }

        public OrderingState(OrderManager orderManager, Proxy proxy)
            : this(orderManager, 0, proxy, null)
        {
        }

        public OrderingState(OrderManager orderManager, Order order)
            : this(orderManager, 0, null, order)
        {
        }

        private OrderingState(OrderManager orderManager, int appliedDependencies, Proxy proxy, Order order)
        {
            this.orderManager = orderManager;
            this.appliedDependencies = appliedDependencies;
            this.proxy = proxy;
            this.order = order;
        }

        /// <summary>
        /// Returns the order for this instance, either if set directly (for a plan, for instance),
        /// or from the proxy (for a rule, for instance).
        /// </summary>
        public Order Order
        {
            get { return order ?? proxy.Order; }
        }

        /// <summary>
        /// Apply the dependency to this ordering instance, returning a new ordering instance.
        /// </summary>
        public OrderingState Apply(Dependency dependency)
        {
            // Union already applied FDs bitmask and 1 bitshifted to represent the newly applied FD.
            int newAppliedDependencies = appliedDependencies | 1 << (dependency.Id - 1);

            // Iteratively apply and return new instance.
            Order newOrder = IterativelyApplyOrder(Order, newAppliedDependencies);
            return new OrderingState(orderManager, newAppliedDependencies, null, newOrder);
        }

        /// <summary>
        /// Apply a set of dependencies to this ordering instance, returning a new ordering instance.
        /// </summary>
        public OrderingState Apply(IEnumerable<Dependency> dependencies)
        {
            // Union already applied FDs with each of the newly applied FDs,
            // 1 bitshifted to represent the applied FD.
            int newAppliedDependencies = appliedDependencies;
            foreach (Dependency dependency in dependencies)
                newAppliedDependencies |= 1 << (dependency.Id - 1);

            // Iteratively apply and return new instance.
            Order newOrder = IterativelyApplyOrder(Order, newAppliedDependencies);
            return new OrderingState(orderManager, newAppliedDependencies, null, newOrder);
        }

        /// <summary>
        /// Merge this and the specified ordering instance. This means unioning the set
        /// of applied dependecies and using this ordering instance as a starting point,
        /// iteratively applying the dependencies.
        /// </summary>
        public OrderingState Apply(OrderingState other)
        {
            // Union this and other's applied FDs.
            int newAppliedDependencies = appliedDependencies | other.appliedDependencies;

            // Iteratively apply and return new instance.
            Order newOrder = IterativelyApplyOrder(Order, newAppliedDependencies);
            return new OrderingState(orderManager, newAppliedDependencies, null, newOrder);
        }

        /// <summary>
        /// Iteratively apply the specified dependencies (bit mask) to the specified order
        /// instance, returning the resulting order instance.
        /// </summary>
        private Order IterativelyApplyOrder(Order input, int dependencies)
        {
            Order oldOrder = null;
            // Continue while the ordering state is changing.
            while (oldOrder != input)
            {
                // Reset stop criteria
                oldOrder = input;
                // Intersect the FDs to apply with the possible transitions at this node
                // to get the possible transitions to make.
                int applicableDependencies = dependencies & input.PossibleTransitions;
                // Intersect with the negated self to get the lowest bit set for the bit mask.
                int dependencyToApplyMask = applicableDependencies & -applicableDependencies;

                // If there is a possible transition...
                if (dependencyToApplyMask > 0)
                    // Find the Id of the FD using the precomputed log 2 map.
                    input = input.Apply(LOG2MAP[dependencyToApplyMask]);
            }
            return input;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder("Applied FDs: {{");

            for (int i = 0; i < 32; i++)
                if ((appliedDependencies & (1 << i)) > 0)
                    sb.AppendFormat("{0},", orderManager.GetDependency(i + 1).Id);

            if (sb.Length > 14) // Length of leading string.
                sb.Remove(sb.Length - 1, 1);

            sb.AppendFormat("}} Order: {0}", Order);
            return sb.ToString();
        }
    }
}