﻿#region

using System;
using System.Collections.Generic;
using System.Threading;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;
using System.Diagnostics;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{
    /// <summary>
    /// Lightweight structure for representing a sub-plan for the optimized query.
    /// </summary>
    public class Plan
    {
        private static int planCounter;

        public Plan(params Plan[] children)
        {
            Children = new List<Plan>(children);
            Id = Interlocked.Increment(ref planCounter); // Thread safe increment.
        }

        /// <summary>The containing plan set.</summary>
        public PlanSet PlanSet { get; set; }
        /// <summary>The rule that created us.</summary>
        public IRule Rule { get; set; }
        /// <summary>Ordering state for this plan.</summary>
        public OrderingState OrderingState { get; set; }
        /// <summary>Costs for this plan.</summary>
        public ICost Costs { get; set; }
        /// <summary>Unique Id for this plan, used for debugging.</summary>
        public int Id { get; set; }
        /// <summary>How much of this plan that can be potentially shared.</summary>
        public BitSet Sharing { get; set; }
        /// <summary>If this exact plan is shared.</summary>
        public bool Shared { get; set; }

        // TODO: Plan should be a Node.
        /// <summary>List of children for this plan, left to right.</summary>
        public List<Plan> Children { get; set; }

        /// <summary>Returns the child with the specified index.</summary>
        public Plan this[int index]
        {
            get { return Children[index]; }
            set { Children[index] = value; }
        }

        /// <summary>Returns the only child of this plan (throws exception if |Childen| != 1).</summary>
        public virtual Plan OnlyChild
        {
            get { return getChildAndVerifyCount(0, 1); }
        }

        /// <summary>Returns the left child of this plan (throws exception if |Childen| != 2).</summary>
        public virtual Plan LeftChild
        {
            get { return getChildAndVerifyCount(0, 2); }
        }

        /// <summary>Returns the right child of this plan (throws exception if |Childen| != 2).</summary>
        public virtual Plan RightChild
        {
            get { return getChildAndVerifyCount(1, 2); }
        }

        private Plan getChildAndVerifyCount(int index, int count)
        {
            Debug.Assert(index < count);
            if (Children.Count != count)
                throw new InvalidOperationException(string.Format("Plan does not have exactly {0} child(ren).", count));
            return Children[index];
        }

        public override string ToString()
        {
            return Utils.NullFormat("Plan-{0}({1}, Cost: {2}, Sharing: {3}, Shared: {4}, Ordering: {5}, Children: {6})",
                                    Id, Rule, Costs, Sharing, Shared, OrderingState.ToString().LimitTo(50),
                                    Children.Count);
        }

        /// <summary>
        /// Compares this plan to the specified plan in terms of cost, sharing and ordering.
        /// </summary>
        /// <param name="other">The plan to compare with.</param>
        /// <returns>Any PlanRelation value describing the relation.</returns>
        public PlanRelation Compare(Plan other)
        {
            return internalCompare(other, Costs.Compare(other.Costs));
        }

        /// <summary>
        /// Compares this plan to the specified plan in terms of cost, sharing and ordering, but
        /// Unknown is not allowed to be returned.
        /// </summary>
        /// <param name="other">The plan to compare with.</param>
        /// <returns>Any PlanRelation value EXCEPT Unknown describing the relation.</returns>
        public PlanRelation CompareTotal(Plan other)
        {
            CostRelation costRelation = Costs.CompareTotal(other.Costs);
            PlanRelation result = internalCompare(other, Costs.CompareTotal(other.Costs));

            if (result != PlanRelation.Unknown)
                return result;

            switch (costRelation)
            {
                case CostRelation.Better:
                    return PlanRelation.Dominates;
                case CostRelation.Worse:
                    return PlanRelation.Dominated;
            }

            return PlanRelation.Equal;
        }

        private PlanRelation internalCompare(Plan other, CostRelation costRelation)
        {
            // If everything is equal, the plans are equal.
            if (Sharing == other.Sharing && OrderingState.Order == other.OrderingState.Order &&
                costRelation == CostRelation.Equal)
                return PlanRelation.Equal;

            // If everything is equal or better, this plan dominates the other plan.
            if (Sharing >= other.Sharing && OrderingState.Order >= other.OrderingState.Order &&
                (costRelation == CostRelation.Better || costRelation == CostRelation.Equal))
                return PlanRelation.Dominates;

            // If everything is equal or worse, this plan is dominated by the other plan
            if (Sharing <= other.Sharing && OrderingState.Order <= other.OrderingState.Order &&
                (costRelation == CostRelation.Worse || costRelation == CostRelation.Equal))
                return PlanRelation.Dominated;

            // If none of the above is the case, we can't say if any of the plans dominates the other
            return PlanRelation.Unknown;
        }
    }

    /// <summary>How two plan instances relate to each other.</summary>
    public enum PlanRelation
    {
        /// <summary>None of the plans dominate each other and they are not equal.</summary>
        Unknown,
        Equal,
        Dominates,
        Dominated
    }
}