﻿#region

using System.Collections;
using System.Collections.Generic;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using System.Diagnostics;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{
    /// <summary>
    /// A PlanSet is a set of plans and can be viewed as a container for the solutions to a sub-graph of the
    /// entire solution operator graph (the equivalent for trees would be a
    /// \emph{branch}).
    /// PlanSet prunes dominated plans whenever they are added. Thereby, the
    /// set will never contain any plan dominated by any other plan (e.g.\ in
    /// terms of cost, sharing and ordering), and hence the optimizer will never store
    /// such plans.
    /// </summary>
    public class PlanSet : IEnumerable<Plan>
    {
        // Internal list of plans. It is a list to ease removal of plans during iteration of it.
        private readonly List<Plan> plans = new List<Plan>();

        /// <summary>The properties satisfied by all the plans in the set.</summary>
        public BitSet Properties { get; set; }
        /// <summary>Common state for all the plans in the set (i.e. cardinality, tuplesize etc).</summary>
        public IPlanSetState State { get; set; }


        /// <summary>Returns the number of plans in the set.</summary>
        public int Count
        {
            get { return plans.Count; }
        }

        /// <summary>Returns the plan with the specified index.</summary>
        public Plan this[int index]
        {
            get { return plans[index]; }
        }

        #region IEnumerable<Plan> Members

        public IEnumerator<Plan> GetEnumerator()
        {
            return plans.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        // TODO: Maybe make this one smarter, ie. take
        // the cost of a sort into consideration
        /// <summary>
        /// Adds the specified plan to the plan set. Will compare it
        /// to the other plans in the set and prune dominated plans, either it is
        /// existing plans or the added plan.
        /// </summary>
        /// <param name="planToAdd">The plan to add.</param>
        public void AddPlan(Plan planToAdd)
        {
            Debug.Assert(State != null, "Cannot add plans to a PlanSet without its state set.");

            for (int i = 0; i < plans.Count; ++i)
            {
                Plan existingPlan = plans[i];

                // Compare and make a descision.
                switch (planToAdd.Compare(existingPlan))
                {
                    case PlanRelation.Equal:
                    case PlanRelation.Dominated:
                        DiagPrinter.Instance.DiagWriteLine("search", "Not adding 1, keeping 2: {0}, {1},", planToAdd,
                                                           existingPlan);
                        return;
                    case PlanRelation.Dominates:
                        DiagPrinter.Instance.DiagWriteLine("search", "Adding 1, throwing out 2: {0}, {1},", planToAdd,
                                                           existingPlan);
                        RemovePlan(existingPlan);
                        --i;
                        break;
                    case PlanRelation.Unknown:
                        DiagPrinter.Instance.DiagWriteLine("search", "Adding 1, keeping 2: {0}, {1},", planToAdd,
                                                           existingPlan);
                        break;
                }
            }
            // At this point, the plan is worth keeping.

            if (plans.Count == 0)
                DiagPrinter.Instance.DiagWriteLine("search", "Initial add: {0}", planToAdd);

            plans.Add(planToAdd);
            if (planToAdd.PlanSet == null)
                planToAdd.PlanSet = this;
        }

        /// <summary>Removes the specified plan.</summary>
        public void RemovePlan(Plan planToRemove)
        {
            plans.Remove(planToRemove);
            planToRemove.PlanSet = null;
        }

        /// <summary>If the plan set contains the specified plan.</summary>
        public bool Contains(Plan plan)
        {
            return plans.Contains(plan);
        }

        /// <summary>Gets the cheapest plan in terms of cost.</summary>
        internal Plan GetCheapest()
        {
            if (plans.Count == 0)
                return null;

            Plan cheapest = plans[0];
            for (int i = 1; i < plans.Count; i++)
                if (plans[i].Costs.CompareTotal(cheapest.Costs) == CostRelation.Better)
                    cheapest = plans[i];

            return cheapest;
        }

        /// <summary>Gets the cheapest plan in terms of cost that satisfies the given order.</summary>
        internal Plan GetCheapest(Order requiredOrder)
        {
            Plan cheapest = null;
            for (int i = 0; i < plans.Count; i++)
                if (plans[i].OrderingState.Order >= requiredOrder
                    && (cheapest == null || plans[i].CompareTotal(cheapest) == PlanRelation.Dominates))
                    cheapest = plans[i];

            return cheapest;
        }

        public override string ToString()
        {
            return string.Format("PlanSet(Props: {0}, State: {1}, Plans: {2})", Properties, State, plans.Count);
        }
    }
}