﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Infrastructure;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Postprocessors;
using Microsoft.Ceres.Optimizer.Preprocessors;
using Microsoft.Ceres.Optimizer.Rules;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{
    /// <summary>Join ordering enumeration schemes.</summary>
    public enum JoinEnumeration
    {
        /// <summary>All layouts allowed.</summary>
        Bushy,
        /// <summary>Joins can have a join input as one ONE of the inputs.</summary>
        ZigZag,
        /// <summary>Joins can have a join input only as the left input.</summary>
        LeftDeep,
        /// <summary>Joins can have a join input only as the right input.</summary>
        RightDeep
    }

    /// <summary>
    /// Main class for a query optimization. It keeps all the state associated with a
    /// query and pulls all treads during optimization from end to end.
    /// NOT THREAD SAFE.
    /// </summary>
    public class QueryOptimization
    {
        #region Delegates

        public delegate void OptimizationHookHandler(object sender, OptimizationEventArgs e);

        #endregion
        private static readonly Type iBaseRuleType = typeof (IBaseRule); // Cache type for performance
        private static readonly Type iSearchRuleType = typeof(ISearchRule); // Cache type for performance
        private List<IProducerRule> rules; // All instantiated rules.
        private List<IBaseRule> baseRules; // Instantiated base rules.
        private List<ISearchRule> searchRules; // Instantiated search rules.
        private List<QueryOutput> queryOutputs; // Outputs of a query (>1 if multi-query)

        private Dictionary<BitSet, PlanSet> plansCache; // Memoization table
        private int plansFromCache = 0; // For statistics
        private int plansGenerated = 0; // For statistics
        private int rulesInvoked = 0; // For statistics
        private int unreachablePlans = 0; // For statistics
        private StopWatchTimer timer; // For statistics
        
        private List<Pair<BitSet, BitSet>> shareEquivalenceMap; // Map of share equivalent rules.
        private BitSet shareEquivalenceRepresentatives; // Union of all class representatives.
        
        static QueryOptimization()
        {
            // Find available rule binders. Only do this once for
            // each system startup, thereby in the static constructor.
            InitRuleBinders();
        }

        public QueryOptimization()
        {
            Visualizer = new OperatorGraphVisualizer();
            JoinEnumeration = JoinEnumeration.Bushy;
            EnableDAG = true;
            InitializeProcessors();
        }

        /// <summary>List if available, reflected rule binders.</summary>
        public static List<IRuleBinder> RuleBinders { get; private set; }
        /// <summary>BitSetManager for the query.</summary>
        public BitSetManager BitSetManager { get; private set; }
        /// <summary>OrderManager for the query.</summary>
        public OrderManager OrderManager { get; private set; }
        /// <summary>Reconstruction table used during the last phase of plan generation.</summary>
        public Dictionary<Plan, OperatorNode> ReconstructionTable { get; private set; }
        /// <summary>Join enumaration scheme to use during plan generation.</summary>
        public JoinEnumeration JoinEnumeration { get; set; }
        /// <summary>Query to dot-visualizer to use for the query.</summary>
        public OperatorGraphVisualizer Visualizer { get; set; }
        /// <summary>If statistics should be printed to standard out during optimization.</summary>
        public bool PrintStats { get; set; }
        /// <summary>The pre-processor for the query.</summary>
        public TransformationProcessor<PreprocessorAttribute> Preprocessor { get; set; }
        /// <summary>The post-processing for the query.</summary>
        public TransformationProcessor<PostprocessorAttribute> Postprocessor { get; set; }
        /// <summary>System catalog used by the rules.</summary>
        public ISystemMetadata SystemMetadata { get; set; }
        /// <summary>Mappings for share equivalent attibutes.</summary>
        public Dictionary<NodeAttribute, NodeAttribute> SEQAttributeMappings { get; private set; }
        /// <summary>Map of equivalence class representative to the equivalence class.</summary>
        public Dictionary<IProducerRule, HashSet<IProducerRule>> ShareEquivalenceClasses { get; private set; }
        /// <summary>If DAG construction is enabled.</summary>
        public bool EnableDAG { get; set; }

        /// <summary>Initialized pre- and post-processors.</summary>
        private void InitializeProcessors()
        {
            Preprocessor = new TransformationProcessor<PreprocessorAttribute>(this);
            Postprocessor = new TransformationProcessor<PostprocessorAttribute>(this);

            Preprocessor.Initialize();
            Postprocessor.Initialize();
        }

        /// <summary>Resets the state before optimizing a query.</summary>
        private void ResetState()
        {
            Visualizer.BumpCounter();
            BitSetManager = new BitSetManager();
            OrderManager = new OrderManager() {Visualizer = Visualizer};
            plansGenerated = plansFromCache = unreachablePlans = rulesInvoked = 0;
            timer = new StopWatchTimer() {Enabled = PrintStats};
        }

        /// <summary>
        /// Optimizes the given query. The optimizer WILL modify this input query.
        /// </summary>
        /// <param name="query">The query to optimize.</param>
        /// <returns>The optimizer query.</returns>
        public OperatorNode Optimize(OperatorNode query)
        {
            Validation.NotNull("query", query);
            Validation.Assert("query", query.OperatorType.Name == "OutputOperator",
                              string.Format("Root node must be OutputOperator, not {0}", query.OperatorType.Name));
            // Reset state before optimization.
            ResetState();

            // Visualize before-query.
            Visualizer.GenerateGraph(query, "Before Optimization", "before");

            timer.Start();

            // Preprocess the graph.
            Preprocess(query);
            timer.PrintIntermediatePoint("Time spent preprocessing: {0}");
            Visualizer.GenerateGraph(query, "After Preprocessing", "preproc");

            // Find and instantiate rules.
            InitializeRules(query);
            timer.PrintIntermediatePoint("Time spent loading rules: {0}");

            // Prepare bit sets.
            DeterminePropertySetsAndPrepare();
            timer.PrintIntermediatePoint("Time spent minimizing bit sets: {0}");

            // Validate rules.
            ValidateRules();
            timer.PrintIntermediatePoint("Time spent validating rules: {0}");

            // Determine share equivalence classes.
            constructShareEquivalenceClasses();
            timer.PrintIntermediatePoint("Time spent constructing share equivalence classes: {0}");

            // Find out the goals of the sub-plans.
            DetermineGoals(query);
            timer.PrintIntermediatePoint("Time spent determining goals: {0}");

            // All requested and produced orders have now been set, so prepare orders.
            // Supply the order manager with information about what attributes
            // are share equivalent so it can recognize share equivalent orderings.
            OrderManager.ShareEquivalentMappings = getShareEquivalentOrderings();
            OrderManager.PrepareOrders();
            timer.PrintIntermediatePoint("Time spent preparing orders: {0}");

            // Instatiate the memorization table, using a rough estimate for entries.
            plansCache = new Dictionary<BitSet, PlanSet>(BitSetManager.Count*10000);

            // Make plans for leaf-nodes, i.e. scans.
            InitializeBasePlans();
            timer.PrintIntermediatePoint("Time spent initializing base plans: {0}");

            // Go plan!
            GeneratePlans(query);
            OnOptimized(new OptimizationEventArgs(query));

            timer.PrintIntermediatePoint("Time spent generating plans: {0}");

            // Convert plans back to nodes.
            query = MakePhysicalPlan(query);
            timer.PrintIntermediatePoint("Time spent making physical plans: {0}");
            Visualizer.GenerateGraph(query, "After Planning", "planning");

            // Do any post processing.
            Postprocess(query);
            timer.PrintIntermediatePoint("Time spent postprocessing: {0}");
            Visualizer.GenerateGraph(query, "After Optimization", "after");

            // Output summary information, such as plan count, total time spent, etc.
            if (PrintStats)
                PrintSummary();

            // Dump debug logs.
            if (PrintStats)
            {
                DumpPlans();
                DiagPrinter.Instance.WriteToFiles();
            }
            
            // Return the optimized query.
            return query;
        }

        /// <summary>Perform some sanity checks on the instantiated rules to uncover 
        /// bugs in the rule implementations. Only include in debug builds.</summary>
        [Conditional("DIAGNOSTICS")]
        private void ValidateRules()
        {
            // Check that the rules' properties aren't insane.
            foreach (ISearchRule rule in searchRules)
            {
                BitSet required = BitSetManager.Empty;
                foreach (BitSetProxy requiredProxy in rule.Required)
                    required |= requiredProxy.BitSet;

                if (required.Overlaps(rule.Produced.BitSet))
                    // Clearly a bug when the rule produces stuff it requires.
                    throw new OptimizerException("Rule {0}'s required and produced overlaps.", rule);
            }
        }

        /// <summary>Extracts a map of share equivalent attributes,
        /// mapping all attributes to the class representative except the representative itself.</summary>
        private Dictionary<string, string> getShareEquivalentOrderings()
        {
            // Extract all share equivalence attribute mappings, but exclude the mappings
            // from the class representative to itself, as this it not interesting to the
            // order manager.
            Dictionary<string, string> result = new Dictionary<string, string>();
            foreach (KeyValuePair<NodeAttribute, NodeAttribute> mapping in SEQAttributeMappings)
                if (mapping.Key != mapping.Value)
                    result.Add(mapping.Key.PropertyName, mapping.Value.PropertyName);
            return result;
        }

        /// <summary>Determine the produced and required property sets for all the rules
        /// and prepare the bit set manager. It is done this way and not through the proxies
        /// to enable optimal support for compund rules in the future.</summary>
        private void DeterminePropertySetsAndPrepare()
        {
            HashSet<NestableHashSet<string>> produced = new HashSet<NestableHashSet<string>>();
            HashSet<NestableHashSet<string>> required = new HashSet<NestableHashSet<string>>();

            // Gather both kinds of bitsets.
            foreach (IProducerRule rule in rules)
            {
                produced.Add(rule.Produced.Properties);
                ISearchRule searchRule = rule as ISearchRule;
                if (searchRule == null)
                    continue;

                foreach (BitSetProxy proxy in searchRule.Required)
                    required.Add(proxy.Properties);
            }

            // All gathered, now prepare.
            // At this point, the properties must no longer change -- which is only by convention at this point.
            BitSetManager.Prepare(produced, required);

            // Dump debug output.
            foreach (IProducerRule rule in rules)
                DiagPrinter.Instance.InfoWriteLine("rules", rule);
        }

        /// <summary>Dump all the plans generated to debug output, but only for debug builds.</summary>
        [Conditional("DIAGNOSTICS")]
        private void DumpPlans()
        {
            // Write the determined goals and orderings for every query.
            foreach (QueryOutput query in queryOutputs)
                DiagPrinter.Instance.DiagWriteLine("plans", "GOAL: {0}, ORDER: {1}", query.Goal, query.Order);
            DiagPrinter.Instance.DiagWriteLine("plans");

            // Dump all plans.
            foreach (KeyValuePair<BitSet, PlanSet> planSet in plansCache)
            {
                if (planSet.Value == null)
                    // Cached unavailability-ness, so no plan to dump.
                    continue;

                // Dump the planset's state.
                DiagPrinter.Instance.DiagWriteLine("plans", planSet.Key);
                DiagPrinter.Instance.DiagWriteLine("plans", planSet.Value.State);
                DiagPrinter.Instance.DiagWriteLine("plans", "-----------------------------------------");

                // Now dump all plans belonging to the planset.
                foreach (Plan plan in planSet.Value)
                    DumpPlan(plan);
                DiagPrinter.Instance.DiagWriteLine("plans");
            }

            // Dump all the keys of the unavailable plans we skipped above.
            DiagPrinter.Instance.DiagWriteLine("plans", "Unreachable plans:");
            DiagPrinter.Instance.DiagWriteLine("plans", "-----------------------------------------");
            foreach (KeyValuePair<BitSet, PlanSet> planSet in plansCache)
                if (planSet.Value == null)
                    DiagPrinter.Instance.DiagWriteLine("plans", planSet.Key);
        }

        /// <summary>Dump a plan recursively.</summary>
        private static void DumpPlan(Plan plan)
        {
            DiagPrinter.Instance.DiagIndent("plans");
            DiagPrinter.Instance.DiagWriteLine("plans", plan);
            foreach (Plan subPlan in plan.Children)
                DumpPlan(subPlan);
            DiagPrinter.Instance.DiagUnindent("plans");
        }

        /// <summary>
        /// Reconstructs the node structure from the plan hierarchy.
        /// </summary>
        /// <param name="outputOperator">Root of the original query.</param>
        private OperatorNode MakePhysicalPlan(OperatorNode outputOperator)
        {
            Validation.Assert("query", outputOperator.OperatorType.Name == "OutputOperator",
                              string.Format("Root node must be OutputOperator, not {0}", outputOperator.OperatorType.Name));

            // Since multiple operators may refer to the same plan, we need to have a
            // dictionary that maps the plans to the already constructed operator node
            // to ensure a DAG is created.
            ReconstructionTable = new Dictionary<Plan, OperatorNode>();

            // Construct a new output operator.
            OperatorNode finalOutputOperator = new OperatorNode(typeof (OutputOperator))
                                        {
                                            Properties = outputOperator.Properties,
                                            Id = outputOperator.Id
                                        };

            // Construct the algebra for all queries, and add them to the output operator.
            foreach (QueryOutput query in queryOutputs)
                finalOutputOperator.AddChild(query.FinalPlan.Rule.BuildAlgebra(query.FinalPlan));

            return finalOutputOperator;
        }

        /// <summary>Determines the goal properties for query or queries.</summary>
        private void DetermineGoals(OperatorNode multiQuery)
        {
            Validation.NotNull("query", multiQuery);
            Validation.Assert("query", multiQuery.OperatorType.Name == "OutputOperator",
                              string.Format("Root node must be OutputOperator, not {0}", multiQuery.OperatorType.Name));

            queryOutputs = new List<QueryOutput>();

            // Get required order from the root, set by OrderingExtractor.
            IList<OrderDescription> orders = null;
            if (multiQuery.Properties.ContainsKey("Orders"))
                orders = (IList<OrderDescription>) multiQuery["Orders"];

            for (int i = 0; i < multiQuery.Children.Count; i++)
            {
                // Set the goal of each query.
                BitSet goal = BitSetManager.Empty;
                SetGoalsRecursively((OperatorNode) multiQuery.Children[i], ref goal);

                // Also set the required orders, if any.
                OrderProxy order = null;
                if (orders != null && orders[i] != null)
                    order = OrderManager.GetOrder(orders[i], true);

                // Add this output to the list.
                queryOutputs.Add(new QueryOutput() {Goal = goal, Order = order});
            }
        }

        /// <summary>Sets goals recursively for the query with the specified root.</summary>
        private static void SetGoalsRecursively(OperatorNode node, ref BitSet goal)
        {
            // Accumulate produced properties for all logical rules.
            foreach(OperatorNode descendant in node.IterateDepthFirst(true))
            {
                Validation.Assert("rulesExist", descendant.HasRules, 
                    string.Format(
                        "Operator node '{0}' not represented in the query (do you lack a rule implementation for it?).",
                        node.OperatorType));

                foreach (IProducerRule rule in descendant.Rules)
                    goal |= rule.Produced.BitSet;
            }
        }

        // Adds the base plans to the memozation table.
        private void InitializeBasePlans()
        {
            // First add base plans for rules that are share equivalence class representatives.
            // Add them to the memoization table using their own produced bit sets.
            foreach (IBaseRule baseRule in baseRules)
            {
                BitSet shared;
                if ( ShareEquivalentGoal(baseRule.Produced.BitSet, out shared) )
                    continue;

                DiagPrinter.Instance.DiagWriteLine("search", "Adding base plan for rule {0}.", baseRule);

                PlanSet plans = new PlanSet() { Properties = baseRule.Produced.BitSet };
                baseRule.Initialize(plans);
                Plan plan = new Plan() { Rule = baseRule };
                baseRule.UpdatePlan(plan);

                plans.AddPlan(plan);
                plansCache[baseRule.Produced.BitSet] = plans;
            }

            // Then add plans for all rules that are not share equivalence class representatives.
            // Add them to the memoization table using their own produced bit sets, but ADDITIONALLY,
            // add plans for their share equivalence representative.
            foreach (IBaseRule baseRule in baseRules)
            {
                BitSet shared;
                if (! ShareEquivalentGoal(baseRule.Produced.BitSet, out shared))
                    continue;

                DiagPrinter.Instance.DiagWriteLine("search", "Adding base plan for rule {0}.", baseRule);

                PlanSet plans = new PlanSet() { Properties = baseRule.Produced.BitSet };
                baseRule.Initialize(plans);

                if (!EnableDAG)
                {
                    // Add plan for this rule.
                    Plan plan = new Plan() { Rule = baseRule };
                    baseRule.UpdatePlan(plan);
                    plans.AddPlan(plan);
                }
                else
                {
                    // Add plans for their share equivalence representative.
                    PlanSet equivalentPlans = plansCache[shared];
                    foreach (Plan plan in equivalentPlans)
                        plans.AddPlan(plan);
                }
                plansCache[baseRule.Produced.BitSet] = plans;
            }
        }

        /// <summary>Instantiates the rules for the specified query.</summary>
        private void InitializeRules(OperatorNode query)
        {
            rules = new List<IProducerRule>();
            baseRules = new List<IBaseRule>();
            searchRules = new List<ISearchRule>();

            // Give the rule binders a go at the query.
            foreach (IRuleBinder ruleBinder in RuleBinders)
            {
                // If the rule binders pattern matches anywhere in the graph,
                // invoke the rule binder with the matching nodes and have it
                // instantiate rules. Register those rules.
                GraphSearcher graphSearcher = ruleBinder.Pattern.GetSearcher();
                if (graphSearcher.Search(query))
                    RegisterRules(ruleBinder.GetRules(this, graphSearcher.MatchedNodes.Cast<OperatorNode>()));
            }

            if (PrintStats)
                DiagPrinter.Instance.InfoWriteLine("stats", "{0} base rules and {1} search rules loaded.",
                                                   baseRules.Count, searchRules.Count);
        }

        /// <summary>Add rules to the correct lists.</summary>
        private void RegisterRules(IEnumerable<IProducerRule> rulesToRegister)
        {
            rules.AddRange(rulesToRegister);
            foreach (IProducerRule rule in rulesToRegister)
            {
                if (iBaseRuleType.IsInstanceOfType(rule))
                    baseRules.Add((IBaseRule) rule);
                if (iSearchRuleType.IsInstanceOfType(rule))
                    searchRules.Add((ISearchRule) rule);
            }
        }

        /// <summary>Returns a set of rules the searchRule depends on.</summary>
        private HashSet<IProducerRule> ruleDependsOn(ISearchRule searchRule)
        {
            HashSet<IProducerRule> result = new HashSet<IProducerRule>();

            // Gather requirements.
            BitSet required = BitSetManager.Empty;
            foreach (BitSetProxy bitsetProxy in searchRule.Required)
                required |= bitsetProxy.BitSet;

            // Which rules produce something we require?
            foreach (IProducerRule rule in rules)
                if (rule.Produced.BitSet.Overlaps(required))
                    result.Add(rule);

            return result;
        }

        /// <summary>Returns whether the two rules are share equivalent.</summary>
        private bool shareEquivalentRules(IProducerRule rule1, IProducerRule rule2)
        {
            // Rules that aren't structurally identical cannot possibly be share equivalent.
            if (!rule1.StructurallyIdentical(rule2))
                return false;

            // If they are structually identical and both base rules, they are trivially share equivalent.
            if (rule1 is IBaseRule && rule2 is IBaseRule)
                return true;

            // Comparing a baserule and a search rules is clearly a bug.
            Validation.Assert("rule-equivalency", !(rule1 is IBaseRule || rule2 is IBaseRule),
                              string.Format(
                                  "Bad rule equivalency test: {0} and {1} are structurally identical but not both IBaseRules or ISearchRules.",
                                  rule1, rule2));

            // Need to reason about the share equivalency of the inputs.
            HashSet<IProducerRule> requiredRules1 = ruleDependsOn((ISearchRule) rule1);
            HashSet<IProducerRule> requiredRules2 = ruleDependsOn((ISearchRule) rule2);

            // Each required rule for rule1 needes a share equivalent in required rules for rule2.
            foreach (IProducerRule i in requiredRules1)
                if (!requiredRules2.Any(j => shareEquivalentRules(i, j)))
                    return false;
            // Each required rule for rule2 needes a share equivalent in required rules for rule1.
            foreach (IProducerRule i in requiredRules2)
                if (!requiredRules1.Any(j => shareEquivalentRules(i, j)))
                    return false;

            return true;
        }

        //TODO: Rule should really be a Node.
        /// <summary>Get rules topologically ordered on produced and required properties.</summary>
        private IEnumerable<IProducerRule> getRulesTopologicallyOrdered()
        {
            Dictionary<IProducerRule, RuleNode> nodeForRule = new Dictionary<IProducerRule, RuleNode>();
            HashSet<IProducerRule> added = new HashSet<IProducerRule>();

            // Create a node for each rule
            RuleNode root = new RuleNode();
            foreach (IProducerRule rule in rules)
                nodeForRule[rule] = new RuleNode() {Rule = rule};

            // Connect parent-children
            foreach (ISearchRule rule in searchRules)
            {
                foreach (IProducerRule dependsOn in ruleDependsOn(rule))
                {
                    nodeForRule[rule].AddChild(nodeForRule[dependsOn]);
                    added.Add(dependsOn);
                }
            }

            foreach (IProducerRule rule in rules)
                if (!added.Contains(rule))
                    root.AddChild(nodeForRule[rule]);

            // Get rules for descendants of the root, not including the root itself, since Output has no rules.
            return root.GetTopologicalOrdering().Where(node => node != root).Select(node => ((RuleNode) node).Rule);
        }

        /// <summary>Construct equivalence classes for the instantiated rules..</summary>
        private void constructShareEquivalenceClasses()
        {
            Dictionary<IProducerRule, HashSet<IProducerRule>> eqClasses =
                new Dictionary<IProducerRule, HashSet<IProducerRule>>();
            SEQAttributeMappings = new Dictionary<NodeAttribute, NodeAttribute>();

            // Iterate through all rules topologically. We need to do so because of
            // share equivalent attribute mapping dependencies
            foreach (IProducerRule rule in getRulesTopologicallyOrdered())
            {
                // Iterate over all equivalence classes found so far
                bool foundEquivalentRulePair = false;
                foreach (KeyValuePair<IProducerRule, HashSet<IProducerRule>> eqClass in eqClasses)
                {
                    // Compare the representative with the rule we're working on
                    if (shareEquivalentRules(eqClass.Key, rule))
                    {
                        // If they are share equivalent, add the rule to the class and map attributes.
                        eqClass.Key.MapShareEquivalentAttributes(rule, SEQAttributeMappings);
                        eqClass.Value.Add(rule);
                        foundEquivalentRulePair = true;
                        break;
                    }
                }
                if (!foundEquivalentRulePair)
                {
                    // It was not share equivalent to any class - create a new class
                    // and let it map attributes to itself.
                    eqClasses.Add(rule, new HashSet<IProducerRule>() {rule});
                    rule.MapShareEquivalentAttributes(rule, SEQAttributeMappings);
                }
            }

            // Construct the final result dictionary
            ShareEquivalenceClasses = new Dictionary<IProducerRule, HashSet<IProducerRule>>();
            // Only add equivalence classes with more than one member
            foreach (KeyValuePair<IProducerRule, HashSet<IProducerRule>> eqClass in eqClasses)
                if (eqClass.Value.Count > 1)
                    ShareEquivalenceClasses.Add(eqClass.Key, eqClass.Value);

            // Now construct the share equivalent bit set map, mapping produced properties for
            // all rules in the class to the produces properties of the representative.
            shareEquivalenceMap = new List<Pair<BitSet, BitSet>>();
            // And the bitmask of all representatives
            shareEquivalenceRepresentatives = BitSetManager.Empty;
            foreach (KeyValuePair<IProducerRule, HashSet<IProducerRule>> eqClass in ShareEquivalenceClasses)
            {
                shareEquivalenceRepresentatives.Or(eqClass.Key.Produced.BitSet);
                // Add a mapping for all rules except the representative itself.
                foreach (IProducerRule rule in eqClass.Value)
                    if (rule != eqClass.Key)
                        shareEquivalenceMap.Add(new Pair<BitSet, BitSet>(rule.Produced.BitSet,
                                                                         eqClass.Key.Produced.BitSet));
            }

            // Dump to debug output
            DiagPrinter.Instance.InfoWriteLine("shareequivalence", "Share Equivalence Classes");
            DiagPrinter.Instance.InfoWriteLine("shareequivalence", "-------------------------");
            foreach (KeyValuePair<IProducerRule, HashSet<IProducerRule>> eqClass in ShareEquivalenceClasses)
            {
                DiagPrinter.Instance.InfoWriteLine("shareequivalence", eqClass.Key.ToString());
                DiagPrinter.Instance.Indent("shareequivalence");
                foreach (IProducerRule rule in eqClass.Value)
                    DiagPrinter.Instance.InfoWriteLine("shareequivalence", rule.ToString());
                DiagPrinter.Instance.Unindent("shareequivalence");
            }

            DiagPrinter.Instance.InfoWriteLine("shareequivalence", "Share Equivalence Map");
            DiagPrinter.Instance.InfoWriteLine("shareequivalence", "---------------------");
            foreach (Pair<BitSet, BitSet> mapping in shareEquivalenceMap)
                DiagPrinter.Instance.InfoWriteLine("shareequivalence", mapping.ToString());
        }


        /// <summary>Reflect, find and instantiate rule binders.</summary>
        private static void InitRuleBinders()
        {
            RuleBinders = new List<IRuleBinder>();
            // Add all rulebinders marked with [RuleBinder].
            foreach (Type type in ReflectionPluginHelper.GetTypesWithAttribute(typeof (RuleBinderAttribute)))
                AddRuleBinder(type);
        }

        /// <summary>Instantiate and add the rule binder with the given type.</summary>
        /// <param name="type">Type of the rule binder to add</param>
        private static void AddRuleBinder(Type type)
        {
            RuleBinders.Add((IRuleBinder) Activator.CreateInstance(type));
        }

        /// <summary>
        /// Generate plans for the given query.
        /// </summary>
        /// <param name="multiQuery">Ouput operator node of the query.</param>
        private void GeneratePlans(OperatorNode multiQuery)
        {
            // Iterate over all sub query in the possible multi-query.
            foreach (QueryOutput queryOutput in queryOutputs)
            {
                DiagPrinter.Instance.DiagWriteLine("search", "Searching for GOAL {0}", queryOutput.Goal);

                if (queryOutput.Order != null)
                {
                    // The query requires a specific ordering, so make sure it's provided.
                    SortRule enforcer = new SortRule(this, queryOutput.Order);
                    queryOutput.Plans = GeneratePlans(queryOutput.Goal, BasicCost.Max, queryOutput.Order.Order, enforcer);
                }
                else
                {
                    // No specific ordering required.
                    queryOutput.Plans = GeneratePlans(queryOutput.Goal, BasicCost.Max);
                }

                if (queryOutput.Plans == null)
                    throw new OptimizerException("Optimizer failure, could not produce plan for query {0}.", queryOutput.Goal);
            }

            // Now we want to find the best combination of plans.
            ICost best = null;
            Plan[] finalPlans = null;

            // Try all combinations of the plans of all queries to find the best combination.
            // Plan[] is an array with one plan per query.
            List<Plan[]> permutations = new List<Plan[]>();
            permute(permutations, new int[queryOutputs.Count], 0);
            foreach (Plan[] plans in permutations)
            {
                ICost current = BasicCost.InputCosts(plans, null);

                // Consider whether this is the best plan so far.
                if (best == null || current.CompareTotal(best) == CostRelation.Better)
                {
                    best = current;
                    finalPlans = plans;
                }
            }

            // The cost of the query is that of the best plan combination.
            multiQuery["Cost"] = best;

            // Set the plans found for every query.
            for (int i = 0; i < queryOutputs.Count; i++)
                queryOutputs[i].FinalPlan = finalPlans[i];
        }

        /// <summary>Generates all permutations for plans.</summary>
        private void permute(List<Plan[]> result, int[] permutation, int branchIndex)
        {
            // Check if we're reached an permutation.
            if (branchIndex == queryOutputs.Count)
            {
                // Verify that this permutation satisfies the requested output orderings.
                for (int i = 0; i < queryOutputs.Count; i++)
                    if (queryOutputs[i].Order != null &&
                        !(queryOutputs[i].Plans[permutation[i]].OrderingState.Order >= queryOutputs[i].Order.Order))
                        return;

                // Insert the permutation
                Plan[] temp = new Plan[queryOutputs.Count];
                for (int i = 0; i < queryOutputs.Count; i++)
                    temp[i] = queryOutputs[i].Plans[permutation[i]];
                result.Add(temp);
            }
            else
            {
                // Call recursively with each alternative at this index.
                for (int i = 0; i < queryOutputs[branchIndex].Plans.Count; i++)
                {
                    permutation[branchIndex] = i;
                    permute(result, permutation, branchIndex + 1);
                }
            }
        }

        /// <summary>
        /// Generates plans satisfying the given goal and order. Multiple plans can be returned
        /// if they are not comparable in cost or have different orderings. Will add the sort rule
        /// to the cheapest plan.
        /// </summary>
        /// <param name="goal">The goal to fulfill.</param>
        /// <param name="limit">Current cost limit for pruning (abort search if passing this).</param>
        /// <param name="order">Order at least one of the plans should satisfy.</param>
        /// <param name="enforcer">Enforer rule (sort) that will produce the requested ordering.</param>
        /// <returns>A PlanSet containing the best plans found.</returns>
        public PlanSet GeneratePlans(BitSet goal, ICost limit, Order order, IRule enforcer)
        {
            PlanSet plans = GeneratePlans(goal, limit);
            if (plans == null)
                return null;

            Plan planWithSort = new Plan(plans.GetCheapest());
            enforcer.UpdatePlan(planWithSort);
            plans.AddPlan(planWithSort);

            return plans;
        }

        /// <summary>
        /// Generates plans satisfying the given goal. Multiple plans can be returned
        /// if they are not comparable in cost or have different orderings.
        /// </summary>
        /// <param name="goal">The goal to fulfill.</param>
        /// <param name="limit">Current cost limit for pruning (abort search if passing this).</param>
        /// <returns>A PlanSet containing the best plans found.</returns>
        public PlanSet GeneratePlans(BitSet goal, ICost limit)
        {
            DiagPrinter.Instance.DiagIndent("search");
            PlanSet plans;
            // If we already have a plan that satisfies the goal, return it.
            if (plansCache.TryGetValue(goal, out plans))
            {
                plansFromCache++;
                if (plans == null)
                    DiagPrinter.Instance.DiagWriteLine("search", "UNREACHABLE from cache: {0}", goal);
                else
                    DiagPrinter.Instance.DiagWriteLine("search", "CACHE return: {0}", goal);

                DiagPrinter.Instance.DiagUnindent("search");
                return plans;
            }

            // Check if we can reach this goal with the current rule set.
            if (GoalIsUnreachable(goal))
            {
                DiagPrinter.Instance.DiagWriteLine("search", "UNREACHABLE: {0}", goal);
                DiagPrinter.Instance.DiagUnindent("search");

                // Memoize the fact that it is unreachable, to avoid unnecessary
                // GoalIsUnreachable()-calls that are O(n) for n rules
                plansCache[goal] = null;
                unreachablePlans++;
                return null;
            }

            // We have a plan we have not made before, and we are able to create it!
            plans = new PlanSet() {Properties = goal};

            // If we've got a disjunct goal, try to get plans for this one also.
            BitSet shared;
            if (EnableDAG && ShareEquivalentGoal(goal, out shared))
            {
                DiagPrinter.Instance.DiagWriteLine("search", "Rewrote {0} to {1}, firing off share equivalent search.",
                                                   goal, shared);

                PlanSet sharingPlans = GeneratePlans(shared, BasicCost.Max);
                if (sharingPlans != null)
                {
                    // Consider the plans where we share sub-plans.
                    plans.State = sharingPlans.State;
                    foreach (Plan plan in sharingPlans)
                        plans.AddPlan(plan);
                }
            }

            // Use the search rule to search for plans.
            foreach (ISearchRule searchRule in searchRules)
            {
                if (! searchRule.IsRelevantTo(goal))
                    // The rule cannot satisfy the goal.
                    continue;

                rulesInvoked++;
                searchRule.Search(plans, limit);
            }

            if (plans.Count == 0)
                plans = null;

            plansCache[goal] = plans;
            plansGenerated++;
            DiagPrinter.Instance.DiagUnindent("search");
            return plans;
        }

        // Rewrite the goal using share equivalent class representatives
        public BitSet ShareEquivalentGoal(BitSet goal)
        {
            // Copy the goal to get a working copy
            BitSet sharedGoal = (BitSet)goal.Clone();
            // shareEquivalenceMap contains all mappings BitSet -> Share Equivalent Representative BitSet
            foreach (Pair<BitSet, BitSet> mapper in shareEquivalenceMap)
                // If all of entire mapper.First is in sharedGoal...
                if ((sharedGoal & mapper.First) == mapper.First)
                    // Then remove the "from" bits and add in the "to" bits, i.e. the share equivalent representatives
                    sharedGoal.Subtract(mapper.First).Or(mapper.Second);
            return sharedGoal;
        }

        // Try to rewrite the goal using share equivalent class representatives and return true
        // and the rewritten goal if successful, otherwise false.
        public bool ShareEquivalentGoal(BitSet goal, out BitSet sharedGoal)
        {
            sharedGoal = ShareEquivalentGoal(goal);

            // If the new goal is different from the old and does not overlap the old except for any already
            // existing representatives in it, we have a successful rewrite.
            return sharedGoal != goal && !sharedGoal.Overlaps(goal - shareEquivalenceRepresentatives);
        }

        private bool GoalIsUnreachable(BitSet goal)
        {
            BitSet mask = BitSetManager.Empty;
            foreach (IProducerRule producerRule in rules)
                if (producerRule.Filter <= goal)
                    mask.Or(producerRule.Filter);
            return mask != goal;
        }

        /// <summary>Prints statistics summary to the console.</summary>
        private void PrintSummary()
        {
            timer.PrintTotal("Total time spent optimizing: {0}");

            int planSetCount = 0, planCount = 0;
            foreach (PlanSet plans in plansCache.Values)
            {
                if (plans != null)
                {
                    planSetCount++;
                    planCount += plans.Count;
                }
            }
            DiagPrinter.Instance.InfoWriteLine("stats",
                                               "Total memorization entries: {0}. Actual plan sets: {1}, actual plans {2}.",
                                               plansCache.Count, planSetCount, planCount);
            DiagPrinter.Instance.InfoWriteLine("stats",
                                               "Plans generated: {0}, Unreachable plans: {1}, plans returned from memorization table: {2}, rule appliances: {3}",
                                               plansGenerated, unreachablePlans, plansFromCache, rulesInvoked);
        }

        #region Events
        public event OptimizationHookHandler Preprocessed , Optimized , Postprocessed;

        protected virtual void OnPreprocessed(OptimizationEventArgs e)
        {
            if (Preprocessed != null)
                Preprocessed(this, e);
        }

        protected virtual void OnOptimized(OptimizationEventArgs e)
        {
            if (Optimized != null)
                Optimized(this, e);
        }

        protected virtual void OnPostprocessed(OptimizationEventArgs e)
        {
            if (Postprocessed != null)
                Postprocessed(this, e);
        }
        #endregion

        /// <summary>Preprocesses the specified query.</summary>
        public void Preprocess(OperatorNode optimizableGraph)
        {
            if (Preprocessor == null)
                return;

            Preprocessor.Transform(optimizableGraph);
            OnPreprocessed(new OptimizationEventArgs(optimizableGraph));
        }

        /// <summary>Postprocesses the specified query.</summary>
        public void Postprocess(OperatorNode optimizableGraph)
        {
            if (Postprocessor == null)
                return;

            Postprocessor.Transform(optimizableGraph);
            OnPostprocessed(new OptimizationEventArgs(optimizableGraph));
        }

        #region Nested type: QueryOutput

        private class QueryOutput
        {
            public BitSet Goal { get; set; }
            public OrderProxy Order { get; set; }
            public PlanSet Plans { get; set; }
            public Plan FinalPlan { get; set; }
        }

        #endregion

        #region Nested type: RuleNode

        private class RuleNode : Node
        {
            public IProducerRule Rule { get; set; }

            public override string ToString()
            {
                return "RuleNode: " + Rule ?? "";
            }
        }

        #endregion
    }

    /// <summary>
    /// Event args for optimization events.
    /// </summary>
    public class OptimizationEventArgs : EventArgs
    {
        /// <summary>Query being optimized.</summary>
        public OperatorNode node;

        public OptimizationEventArgs(OperatorNode node)
        {
            this.node = node;
        }
    }
}