﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using Microsoft.Ceres.Optimizer.GraphMatching;
using System.Collections;

#endregion

namespace Microsoft.Ceres.Optimizer.Engine
{
    /// <summary>
    /// A transformation processor applies a chain of ITransformationRules on the nodes on which Process() is invoked.
    /// </summary>
    /// <typeparam name="T">Attribute-type of processors to reflect.</typeparam>
    public class TransformationProcessor<T>
    {
        private QueryOptimization optimizationInstance;

        /// <summary>
        /// The constructor takes a QueryOptimization instance as some processors need access to the available rule binders configured.
        /// </summary>
        /// <param name="optimizationInstance"></param>
        public TransformationProcessor(QueryOptimization optimizationInstance)
        {
            this.optimizationInstance = optimizationInstance;
        }

        public List<ITransformationRule> TransformationChain { get; set; }

        public void Initialize()
        {
            LoadProcessors();
            ResolveDependencies();
            ConfigureProcessors();
        }

        public void LoadProcessors()
        {
            // Find all preprocessors annotated with the provided attribute.
            TransformationChain = new List<ITransformationRule>();
            foreach (Type type in ReflectionPluginHelper.GetTypesWithAttribute(typeof (T)))
            {
                Debug.Assert(typeof (ITransformationRule).IsAssignableFrom(type), type.ToString(),
                    string.Format("The class marked with {0} is not a transformation rule.", type.Name));

                // Instantiate the rule and add it to the chain.
                TransformationChain.Add((ITransformationRule) Activator.CreateInstance(type));
            }
        }

        public void ResolveDependencies()
        {
            Dictionary<Type, TransformationNode> nodeForType = new Dictionary<Type, TransformationNode>();

            // We need to find the topological ordering of the processors, to invoke them in the right order.
            // To do so we make an auxiliary graph of the processors, where we add edges from the
            // depended-on transformers to the ones that depend on them.

            // First make a node for every transformer.
            foreach (ITransformationRule transformer in TransformationChain)
            {
                nodeForType[transformer.GetType()] = new TransformationNode() {Transformer = transformer};
            }

            TransformationNode root = new TransformationNode();

            foreach (ITransformationRule transformer in TransformationChain)
            {
                // The root of our dummy-graph has an edge to every processor, for the edge case
                // where no processors in the chain depend on each other.
                root.AddChild(nodeForType[transformer.GetType()]);

                foreach (Type dependedOnTransformer in transformer.DependsOn)
                {
                    if (!nodeForType.ContainsKey(dependedOnTransformer))
                        throw new InvalidOperationException(
                            "Processor {0} depends on {1}, which is NOT present in the transformation chain.");
                    nodeForType[dependedOnTransformer].AddChild(nodeForType[transformer.GetType()]);
                }
            }

            List<ITransformationRule> orderedTransformers = new List<ITransformationRule>();
            foreach (TransformationNode node in root.GetTopologicalOrdering())
            {
                if (node == root)
                    continue; // Just a dummy.

                orderedTransformers.Add(node.Transformer);
            }

            orderedTransformers.Reverse();
            TransformationChain = orderedTransformers;
        }

        public void ConfigureProcessors()
        {
            // Inject the QueryOptimization-instance for those who need it.
            foreach (ITransformationRule transformer in TransformationChain)
            {
                transformer.Configure(optimizationInstance);
            }
        }

        public void Transform(Node root)
        {
            bool graphHasChanged;

            List<ITransformationRule> transformationRules = new List<ITransformationRule>(TransformationChain);
            HashSet<ITransformationRule> completedTransformationRules = new HashSet<ITransformationRule>();

            do
            {
                graphHasChanged = false;
                foreach (ITransformationRule transformer in transformationRules)
                {
                    if (completedTransformationRules.Contains(transformer))
                        continue;

                    GraphSearcher matcher = transformer.Pattern.GetSearcher();
                    if (matcher.Search(root))
                    {
                        graphHasChanged |= transformer.Fire(matcher);

                        // Iterative rules want to be invoked whenever the graph changes.
                        if (! transformer.Iterative)
                            completedTransformationRules.Add(transformer);
                    }
                }
            } while (graphHasChanged);
        }

        #region Nested type: TransformationNode

        private class TransformationNode : Node
        {
            public ITransformationRule Transformer { get; set; }
        }

        #endregion
    }
}