﻿#region

using System.Collections.Generic;
using System.Diagnostics;

#endregion

namespace Microsoft.Ceres.Optimizer.GraphMatching
{
    /// <summary>
    /// Base clase for all node matchers.
    /// 
    /// A node matcher is a stateful object that has a HasMatched-property and a Search-method.
    /// 
    /// The HasMatched-property obviously returns whether the pattern has matched.
    /// 
    /// The Search-method is used for actual searching. If the pattern figures that the input node does not move the pattern into an invalid state, Search() returns true.
    /// </summary>
    public abstract class AbstractNodeMatcher : Node, IGraphPattern
    {
        public AbstractNodeMatcher()
        {
            MatchedNodes = new List<Node>();
            HasMatched = false;
        }

        public List<Node> MatchedNodes { get; set; }

        public Node OnlyMatch
        {
            get
            {
                Debug.Assert(MatchedNodes.Count == 1);
                return MatchedNodes[0];
            }
        }

        #region IGraphPattern Members

        public virtual bool HasMatched { get; set; }

        public abstract bool Search(Node node);

        #endregion

        // Sugar for binary operators.
        public static AbstractNodeMatcher operator |(AbstractNodeMatcher a, AbstractNodeMatcher b)
        {
            return new OrMatcher(a, b);
        }

        public static AbstractNodeMatcher operator &(AbstractNodeMatcher a, AbstractNodeMatcher b)
        {
            return new AndMatcher(a, b);
        }

        public virtual void Reset()
        {
            HasMatched = false;
            MatchedNodes.Clear();
        }

        public virtual AbstractNodeMatcher WithChild(AbstractNodeMatcher stopPattern)
        {
            return new ChildMatcher {RootMatcher = this, StopPattern = stopPattern};
        }

        public virtual AbstractNodeMatcher WithSingleParent()
        {
            return new ParentMatcher {RootMatcher = this, TakeWhile = new RangeQuantifier(new AnythingMatcher(), 1, 1)};
        }

        /// <summary>
        /// A branching node.
        /// </summary>
        /// <returns></returns>
        public virtual AbstractNodeMatcher WithMultipleParents()
        {
            return new ParentMatcher
                       {RootMatcher = this, TakeWhile = new RangeQuantifier(new AnythingMatcher(), 2, int.MaxValue)};
        }

        public virtual AbstractNodeMatcher WithAllChildren(AbstractNodeMatcher takeWhile)
        {
            return new ChildMatcher {RootMatcher = this, TakeWhile = new AllMatcher(takeWhile)};
        }

        public virtual AbstractNodeMatcher WithDescendant(AbstractNodeMatcher stopPattern)
        {
            return new DescendantTraverser {RootMatcher = this, TakeWhile = new AnythingMatcher(), StopPattern = stopPattern};
        }

        public virtual AbstractNodeMatcher WithDescendant(AbstractNodeMatcher takeWhile, AbstractNodeMatcher stopPattern)
        {
            return new DescendantTraverser {RootMatcher = this, TakeWhile = takeWhile, StopPattern = stopPattern};
        }

        public virtual AbstractNodeMatcher GroupAs(string groupName)
        {
            return new MatchGrouper(groupName, this);
        }

        /// <summary>
        /// Returns a GraphSearcher with this pattern as its sub-pattern.
        /// 
        /// Invoke the GraphSearcher to do find all occurrences of the pattern in an entire graph.
        /// </summary>
        /// <returns></returns>
        public virtual GraphSearcher GetSearcher()
        {
            return new GraphSearcher(this);
        }
    }
}