﻿#region

using System.Collections.Generic;
using System.Diagnostics;

#endregion

namespace Microsoft.Ceres.Optimizer.GraphMatching
{
    /// <summary>
    /// Sub-classes of AbstractTraverser are expected to just return different iterators.
    /// 
    /// For every node matched by the RootMatcher, as long as TakeWhile (if any) matches, match until StopPattern (if any) matches.
    /// 
    /// These simple patterns combined provide a flexible, yet simple, means of applying pattern matching of more than a single node.
    /// 
    /// Note that traversers without a StopPattern are useful with stateful TakeWhile-patterns. However, both TakeWhile and StopPattern cannot both be null.
    /// </summary>
    public abstract class AbstractTraverser : AbstractCompoundMatcher
    {
        private AbstractNodeMatcher rootMatcher;
        private AbstractNodeMatcher stopPattern;

        private AbstractNodeMatcher takeWhile;

        /// <summary>
        /// Optional matcher that may return multiple nodes that are subsequently traversed.
        /// 
        /// If no RootMatcher is provided, the node given to the Search()-method is used as the only root.
        /// </summary>
        public AbstractNodeMatcher RootMatcher
        {
            get { return rootMatcher; }
            set
            {
                if (rootMatcher != null)
                    RemoveChild(rootMatcher);

                rootMatcher = value;
                if (rootMatcher != null)
                    AddChild(rootMatcher);
            }
        }

        /// <summary>
        /// Optional pattern that determines whether the matching should continue.
        /// </summary>
        public AbstractNodeMatcher TakeWhile
        {
            get { return takeWhile; }
            set
            {
                if (takeWhile != null)
                    RemoveChild(takeWhile);

                takeWhile = value;
                if (takeWhile != null)
                    AddChild(takeWhile);
            }
        }

        /// <summary>
        /// Optional pattern that determines whether matching should stop.
        /// </summary>
        public AbstractNodeMatcher StopPattern
        {
            get { return stopPattern; }
            set
            {
                if (stopPattern != null)
                    RemoveChild(stopPattern);

                stopPattern = value;
                if (stopPattern != null)
                    AddChild(stopPattern);
            }
        }

        /// <summary>
        /// Invoked for every root found. Steers the direction of the traversal.
        /// </summary>
        public abstract IEnumerable<Node> IteratorFor(Node node);

        public override bool Search(Node node)
        {
            Debug.Assert(!(TakeWhile == null && StopPattern == null), "TakeWhile and StopPattern cannot both be null");

            // First, determine roots we'll find iterators for.
            HashSet<Node> roots;
            if (RootMatcher != null)
            {
                if (!RootMatcher.Search(node))
                    return false;
                roots = new HashSet<Node>(RootMatcher.MatchedNodes);
            }
            else
            {
                roots = new HashSet<Node>() {node};
            }

            // Iterate every root.
            foreach (Node root in roots)
            {
                foreach (Node descendant in IteratorFor(root))
                {
                    // If we have a stop pattern and it matches, we're done.
                    if (StopPattern != null && StopPattern.Search(descendant))
                    {
                        return HasMatched = true;
                    }
                    // If not, continue if we can.
                    if (! (TakeWhile == null || TakeWhile.Search(descendant)))
                        break;
                }
            }

            return HasMatched = (StopPattern == null || StopPattern.HasMatched) && (TakeWhile == null || TakeWhile.HasMatched);
        }

        public override void Reset()
        {
            base.Reset();
            foreach (AbstractNodeMatcher child in Children)
            {
                child.Reset();
            }
        }

        public override string ToString()
        {
            return string.Format("{3}({0}, {1}, {2})", RootMatcher, TakeWhile, StopPattern, GetType().Name);
        }
    }
}