﻿#region

using System.Collections.Generic;

#endregion

namespace Microsoft.Ceres.Optimizer.GraphMatching
{
    /// <summary>
    /// Traverses the input graph breadth first and checks the wrapped pattern.
    /// Particularly useful is the availability of named matches through the Groups property.
    /// 
    /// Caveat: If there are multiple GroupMatchers with the same name, only the top-most is bound.
    /// </summary>
    public class GraphSearcher : MatcherWrapper
    {
        private List<Pair<string, List<Node>>> groupMatches;

        public GraphSearcher(AbstractNodeMatcher pattern)
        {
            AddChild(pattern);
            MapGroups();
        }

        public Dictionary<string, List<Node>> Groups { get; set; }

        /// <summary>
        /// If FirstHitOnly is true, then only the top-most hit is found. The search is aborted after the first hit.
        /// </summary>
        public bool FirstHitOnly { get; set; }
        public override bool HasMatched { get; set; }

        public void MapGroups()
        {
            Groups = new Dictionary<string, List<Node>>();
            groupMatches = new List<Pair<string, List<Node>>>();


            // Map all the group names to their respective MatchedNodes-lists.
            foreach (AbstractNodeMatcher pattern in Pattern.IterateBreadthFirst())
            {
                if (! (pattern is MatchGrouper))
                    continue;

                MatchGrouper grouper = (MatchGrouper)pattern;
                groupMatches.Add(new Pair<string, List<Node>>(grouper.GroupName, grouper.MatchedNodes));

                // We do not overwrite nor throw an exception here, because it is occasionally useful
                // to have the same GroupMatcher-object occur multiple times in the pattern.
                if (!Groups.ContainsKey(grouper.GroupName))
                    Groups.Add(grouper.GroupName, new List<Node>());
            }
        }

        public override bool Search(Node root)
        {
            foreach (Node node in root.IterateBreadthFirst())
            {
                Pattern.Reset();

                if (!Pattern.Search(node))
                    continue;

                MatchedNodes.Add(node);
                copyGroupMatches();
                HasMatched = true;
                if (FirstHitOnly)
                    return true;
            }
            return HasMatched;
        }

        public override void Reset()
        {
            base.Reset();
            Pattern.Reset();
        }

        public override string ToString()
        {
            return string.Format("GraphSearcher({0})", Pattern);
        }

        private void copyGroupMatches()
        {
            foreach (Pair<string, List<Node>> pair in groupMatches)
            {
                if (pair.Second.Count == 0)
                    continue; // No matches.

                string groupName = pair.First;
                Groups[groupName].AddRange(pair.Second);
            }
        }
    }
}