﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.Ceres.Optimizer.Engine;

#endregion

namespace Microsoft.Ceres.Optimizer
{
    /// <summary>
    /// Allows for build graph structures in memory. Contains functionality for iteration etc.
    /// </summary>
    public class Node : ICloneable
    {
        #region Delegates

        public delegate void NodeChangeEventHandler(object sender, NodeChangeEventArgs e);

        public delegate bool StoppingVisitorDelegate(Node node);

        public delegate void VisitorDelegate(Node node);

        #endregion

        public string Name = "Node";

        public Node()
        {
            Children = new List<Node>();
            Parents = new List<Node>();
            Properties = new Dictionary<string, object>();
        }

        public Node(string name)
            : this()
        {
            Name = name;
        }

        public Dictionary<string, object> Properties { get; set; }

        public virtual List<Node> Parents { get; protected set; }

        public virtual Node FirstParent
        {
            get { return Parents[0]; }
        }

        public virtual Node OnlyParent
        {
            get
            {
                if (Parents.Count > 1)
                    throw new InvalidOperationException("Assumed tree is a DAG");
                return FirstParent;
            }
            set
            {
                if (Parents.Count > 1)
                    throw new InvalidOperationException("Assumed tree is a DAG");

                if (Parents.Count == 0)
                    Parents.Add(value);
                else
                    Parents[0] = value;
            }
        }

        public virtual List<Node> Children { get; protected set; }

        public virtual Node FirstChild
        {
            get { return Children[0]; }
        }

        public virtual Node OnlyChild
        {
            get
            {
                if (Children.Count != 1)
                    throw new InvalidOperationException("Node does not have exactly one child.");

                return FirstChild;
            }
        }

        public virtual bool IsRoot
        {
            get { return Parents.Count == 0; }
        }

        public virtual bool IsLeaf
        {
            get { return Children.Count == 0; }
        }

        public virtual bool HasParents
        {
            get { return Parents.Count > 0; }
        }

        public virtual bool HasChildren
        {
            get { return Children.Count > 0; }
        }

        public virtual object this[string propertyName]
        {
            get
            {
                object o;
                if (!Properties.TryGetValue(propertyName, out o))
                    throw new OptimizerException(string.Format(@"Expected property ""{0}"" not present.", propertyName));

                return o;
            }
            set { Properties[propertyName] = value; }
        }

        #region ICloneable Members

        public virtual object Clone()
        {
            Node clone = new Node();

            SetVariablesForClone(clone);

            return clone;
        }

        #endregion

        public override string ToString()
        {
            return string.Format("{0}<{1}>", Name, GetHashCode());
        }

        public event NodeChangeEventHandler ChildAdded , ChildRemoved , ParentAdded , ParentRemoved;

        protected void OnChildAdded(NodeChangeEventArgs e)
        {
            if (ChildAdded != null)
                ChildAdded(this, e);
        }

        protected void OnChildRemoved(NodeChangeEventArgs e)
        {
            if (ChildRemoved != null)
                ChildRemoved(this, e);
        }

        protected void OnParentAdded(NodeChangeEventArgs e)
        {
            if (ParentAdded != null)
                ParentAdded(this, e);
        }

        protected void OnParentRemoved(NodeChangeEventArgs e)
        {
            if (ParentRemoved != null)
                ParentRemoved(this, e);
        }

        public virtual void AddChild(Node child)
        {
            if (Children.Contains(child) || child.Parents.Contains(this))
            {
                /* Corner case, when B is extracted from
                 *  A - B - C
                 *   \-----/
                 */
                return;
            }

            child.Parents.Add(this);
            child.OnParentAdded(new NodeChangeEventArgs(this));

            Children.Add(child);
            OnChildAdded(new NodeChangeEventArgs(child));
        }

        public virtual void AddChildren(IEnumerable<Node> children)
        {
            foreach (var child in children)
                AddChild(child);
        }

        public virtual void RemoveChild(Node child)
        {
            Debug.Assert(Children.Contains(child), "Child not contained in children!");
            Debug.Assert(child.Parents.Contains(this), "I'm not this child's parent!");

            child.Parents.Remove(this);
            child.OnParentRemoved(new NodeChangeEventArgs(this));

            Children.Remove(child);
            OnChildRemoved(new NodeChangeEventArgs(child));
        }

        public virtual void ReplaceChild(Node oldChild, Node newChild)
        {
            Debug.Assert(Children.Contains(oldChild), "Child not contained in children!");
            Debug.Assert(oldChild.Parents.Contains(this), "I'm not this child's parent!");

            // We do not oldChild.RemoveChild() as we do not want Children changed yet.
            oldChild.Parents.Remove(this);
            oldChild.OnParentRemoved(new NodeChangeEventArgs(this));

            // We want to keep the orderings of the children.
            int oldIndex = Children.IndexOf(oldChild);

            newChild.Parents.Add(this);
            newChild.OnParentAdded(new NodeChangeEventArgs(this));

            Children.RemoveAt(oldIndex);
            OnChildRemoved(new NodeChangeEventArgs(oldChild));
            Children.Insert(oldIndex, newChild);
            OnChildAdded(new NodeChangeEventArgs(newChild));
        }

        /// <summary>
        /// Makes newParent the only parent of this node, but make our current outputs the
        /// outputs of newParent.
        ///  
        ///      B   C
        ///       \ /
        /// Given  A , A.InsertBetweenAllParents(D) results in:
        /// 
        ///  B   C
        ///   \ /
        ///    D
        ///    |
        ///    A
        /// </summary>
        /// <param name="newParent"></param>
        public virtual void InsertBetweenAllParents(Node newParent)
        {
            foreach (Node parent in new List<Node>(Parents))
                parent.ReplaceChild(this, newParent);
            newParent.AddChild(this);
        }

        /// <summary>
        /// Extracts the node from the graph, adding the extracted node's children to the parent.
        /// </summary>
        public virtual void Extract()
        {
            // Copy lists, as we cannot modify a collection we're iterating over.
            var existingParents = new List<Node>(Parents);
            var existingChildren = new List<Node>(Children);

            foreach (var parent in existingParents)
            {
                parent.RemoveChild(this);
                foreach (var child in existingChildren)
                {
                    RemoveChild(child);
                    parent.AddChild(child);
                }
            }
        }

        public virtual void ReplaceWith(Node replacementNode)
        {
            var existingParents = new List<Node>(Parents);
            var existingChildren = new List<Node>(Children);

            foreach (var parent in existingParents)
            {
                if (parent == replacementNode)
                    parent.RemoveChild(this);
                else
                    parent.ReplaceChild(this, replacementNode);
            }

            foreach (var child in existingChildren)
            {
                RemoveChild(child);
                if (child == replacementNode)
                    continue;
                replacementNode.AddChild(child);
            }
        }

        public virtual IEnumerable<Node> IterateDepthFirst(bool inclusive)
        {
            if (inclusive)
            {
                yield return this;
            }
            if (Children.Count == 0)
            {
                yield break;
            }

            var visitedNodes = new HashSet<Node>() {this};
            var stack = new Stack<Node>(Children.Reverse<Node>());
            while (stack.Count > 0)
            {
                var node = stack.Pop();
                if (visitedNodes.Contains(node))
                    continue;

                visitedNodes.Add(node);
                yield return node;
                foreach (Node child in node.Children.Reverse<Node>())
                {
                    if (!visitedNodes.Contains(child))
                        stack.Push(child);
                }
            }
        }

        public virtual IEnumerable<Node> IterateDepthFirst()
        {
            return IterateDepthFirst(true);
        }


        public virtual IEnumerable<Node> IterateBreadthFirst(bool inclusive)
        {
            if (inclusive)
            {
                yield return this;
            }
            if (Children.Count == 0)
            {
                yield break;
            }

            var visitedNodes = new HashSet<Node>() {this};
            var queue = new Queue<Node>(Children);
            while (queue.Count > 0)
            {
                var node = queue.Dequeue();
                if (visitedNodes.Contains(node))
                    continue;

                visitedNodes.Add(node);
                yield return node;
                foreach (var child in node.Children)
                {
                    if (!visitedNodes.Contains(child))
                        queue.Enqueue(child);
                }
            }
        }

        public virtual IEnumerable<Node> IterateBreadthFirst()
        {
            return IterateBreadthFirst(true);
        }

        public virtual IEnumerable<Node> GetTopologicalOrdering()
        {
            HashSet<Node> discovered = new HashSet<Node>();
            HashSet<Node> explored = new HashSet<Node>();

            Stack<Node> stack = new Stack<Node>();
            stack.Push(this);
            while (stack.Count > 0)
            {
                Node node = stack.Peek();
                if (explored.Contains(node))
                {
                    stack.Pop();
                    continue;
                }
                discovered.Add(node);

                List<Node> newNodes = new List<Node>();
                foreach (var child in node.Children)
                {
                    if (!explored.Contains(child))
                    {
                        if (discovered.Contains(child))
                        {
                            throw new InvalidOperationException("Graph contains cycle.");
                        }
                        newNodes.Add(child);
                    }
                }

                if (newNodes.Count > 0)
                {
                    foreach (var newNode in newNodes)
                    {
                        stack.Push(newNode);
                    }
                }
                else
                {
                    explored.Add(node);
                    yield return node;
                }
            }
        }

        public virtual void Traverse(IEnumerable<Node> iterator, VisitorDelegate visitor)
        {
            foreach (Node node in iterator)
            {
                visitor(node);
            }
        }

        public virtual void Traverse(IEnumerable<Node> iterator, StoppingVisitorDelegate visitor)
        {
            foreach (Node node in iterator)
            {
                bool shouldContinue = visitor(node);
                if (!shouldContinue)
                {
                    break;
                }
            }
        }

        public virtual void TraverseDepthFirst(VisitorDelegate visitor)
        {
            Traverse(IterateDepthFirst(), visitor);
        }

        public virtual bool IsAcyclic()
        {
            try
            {
                // We must go through the iteration, but we do not need the values.
                foreach (var node in GetTopologicalOrdering())
                {
                }
                return true;
            }
            catch (InvalidOperationException)
            {
                return false;
            }
        }

        public virtual IEnumerable<Node> GetAscendants()
        {
            Stack<Node> stack = new Stack<Node>(Parents);
            var visitedNodes = new HashSet<Node>();
            while (stack.Count > 0)
            {
                var node = stack.Pop();
                if (node == this)
                {
                    throw new InvalidOperationException("Graph has a cycle");
                }
                if (visitedNodes.Contains(node))
                {
                    continue;
                }
                visitedNodes.Add(node);
                yield return node;
                foreach (var parent in node.Parents)
                {
                    stack.Push(parent);
                }
            }
        }

        public virtual void Bubble(VisitorDelegate visitor)
        {
            visitor(this);
            foreach (var ancestor in GetAscendants())
            {
                visitor(ancestor);
            }
        }

        public virtual void SetVariablesForClone(Node clone)
        {
            clone.Parents = new List<Node>(Parents);
            clone.Children = new List<Node>(Children);
            clone.Properties = new Dictionary<string, object>(Properties);
            clone.Name = Name;
        }

        public virtual Node CloneNodeAndSubgraph()
        {
            Dictionary<Node, Node> cloneMap = new Dictionary<Node, Node>();

            foreach (Node node in IterateBreadthFirst())
            {
                Node clonedNode = (Node) node.Clone();
                cloneMap.Add(node, clonedNode);

                // We're adding cloned children from below.
                clonedNode.Children = new List<Node>();

                List<Node> parents = new List<Node>(clonedNode.Parents);
                foreach (Node parent in parents)
                {
                    Node newParent;
                    if (cloneMap.TryGetValue(parent, out newParent))
                    {
                        newParent.AddChild(clonedNode);
                    }
                    clonedNode.Parents.Remove(parent);
                }
            }

            return cloneMap[this];
        }
    }

    public class NodeChangeEventArgs : EventArgs
    {
        public Node Node { get; set; }
        public NodeChangeEventArgs(Node node)
        {
            Node = node;
        }
    }
}