﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Web;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer
{
    public class OperatorGraphVisualizer
    {
        private static readonly string[] DOT_PATHS = new[] {@"..\..\..\bin\Graphviz\bin", @".\bin"};
        private int graphsMade;

        public OperatorGraphVisualizer()
        {
            Visualize = true;
            InvokeDot = false;
            AutoOpen = false;
            PreserveDotFiles = true;
            OutputFormat = "png";
            TempFileSuffix = ".temp.dot";
            EnumerateFiles = true;
            graphsMade = 0;
            ShowAttributeOrigin = false;
            ShowDependencies = false;
        }

        public bool ListAttributes { get; set; }
        public bool ShowPlanStats { get; set; }

        public bool Visualize { get; set; }
        public bool AutoOpen { get; set; }
        public bool PreserveDotFiles { get; set; }
        public string OutputFormat { get; set; }
        public string TempFileSuffix { get; set; }
        public bool InvokeDot { get; set; }
        public bool EnumerateFiles { get; set; }
        public bool ShowAttributeOrigin { get; set; }
        public bool ShowDependencies { get; set; }

        private string TempFileName(string fileName)
        {
            return fileName + "-" + (EnumerateFiles ? graphsMade.ToString() : string.Empty) + TempFileSuffix;
        }

        public void GenerateGraph(string dot, string outputFile)
        {
            if (!Visualize)
                return;
            string dotFileName = TempFileName(outputFile);
            File.WriteAllText(dotFileName, CleanupLabel(dot));

            if (InvokeDot)
            {
                RunDot(dotFileName, outputFile);

                if (AutoOpen)
                    Process.Start(outputFile);
            }
        }

        public void GenerateGraph(Node root, string graphLabel, string outputFile)
        {
            if (!Visualize)
                return;
            String dot = DotForGraph(root, graphLabel);
            GenerateGraph(dot, outputFile);
        }

        private static string CleanupLabel(string label)
        {
            // Switch to regexes when a lot more types are used.
            return label.Replace("AT_", "").Replace("AP_", "");
        }

        private static string SmallText(string text)
        {
            return string.Format("<font point-size=\"10\">{0}</font>", text);
        }

        private void RunDot(string inputFile, string outputFile)
        {
            string dotPath = null;
            foreach (string path in DOT_PATHS)
            {
                if (File.Exists(Path.Combine(path, "dot.exe")))
                {
                    dotPath = Path.Combine(path, "dot.exe");
                    break;
                }
            }
            if (dotPath == null)
            {
                throw new Exception("Could not find dot.exe.");
            }

            outputFile = string.Format("{0}.{1}", outputFile, OutputFormat);

            ProcessStartInfo startInfo = new ProcessStartInfo
                                             {
                                                 FileName = dotPath,
                                                 Arguments =
                                                     string.Format("-T{0} -o{1} {2}", OutputFormat, outputFile,
                                                                   inputFile),
                                                 WindowStyle = ProcessWindowStyle.Hidden,
                                                 RedirectStandardError = true,
                                                 UseShellExecute = false
                                             };
            Process p = Process.Start(startInfo);


            if (!p.WaitForExit(10000))
            {
                p.Kill();
                throw new Exception("Dot did not finish in 10 seconds.");
            }

            if (p.ExitCode != 0)
            {
                throw new Exception(string.Format("Dot returned exit code {0}, standard out is: {1}",
                                                  p.ExitCode, p.StandardError.ReadToEnd()));
            }

            if (!PreserveDotFiles)
            {
                // Delete temporary file.
                File.Delete(inputFile);
            }
        }

        private static string HtmlEncode(string text)
        {
            char[] chars = HttpUtility.HtmlEncode(text).ToCharArray();
            StringBuilder result = new StringBuilder(text.Length + (int) (text.Length*0.1));

            foreach (char c in chars)
            {
                int value = Convert.ToInt32(c);
                if (value > 127)
                    result.AppendFormat("&#{0};", value);
                else
                    result.Append(c);
            }

            return result.ToString();
        }

        // Enumerate IDictionaries and ILists recursively, turning them into readable strings.
        private static string PropertyToString(object value)
        {
            if (value is IDictionary)
            {
                // Somehow I didn't get LINQ's IEnumerable.Cast<T> to work and did it this way instead.
                IList<DictionaryEntry> list = new List<DictionaryEntry>();
                foreach (DictionaryEntry entry in (IDictionary) value)
                    list.Add(entry);
                return Utils.JoinEnumerable(list,
                                            entry => string.Format("{{{0},{1}}}",
                                                                   PropertyToString(entry.Key),
                                                                   PropertyToString(entry.Value)), string.Empty, ",",
                                            true);
            }
            else if (value is IList)
                return string.Format("{{{0}}}",
                                     Utils.JoinEnumerable(((IList) value).Cast<object>(), item => PropertyToString(item),
                                                          string.Empty, ",", true));
            else
                return value != null ? value.ToString() : string.Empty;
        }

        private string DotForNode(Node node)
        {
            StringBuilder dot = new StringBuilder();
            dot.AppendLine(node.GetHashCode() + " [");

            // TODO - This knowing about OperatorNode is fugly.
            OperatorNode operatorNode = node as OperatorNode;

            // Shape
            if (operatorNode == null || operatorNode.Rules == null || !operatorNode.Rules.Any(r => r is IBaseRule))
                dot.AppendLine("style=\"rounded\"");

            // Label
            dot.AppendFormat("label=<<font point-size=\"14\">");

            if (node is OperatorNode)
            {
                if (operatorNode.OperatorType != null)
                    dot.Append(operatorNode.OperatorType.Name);
                else
                    dot.Append((node.Name != null) ? node.Name : "N/A");
                dot.AppendFormat("({0})", operatorNode.Id);
            }
            else if (node is AbstractNodeMatcher)
                dot.Append(node.ToString().Replace("<", "&lt;").Replace(">", "&gt;"));
            else
                dot.Append(node.Name);
            dot.AppendLine("</font> <br />");

            // Add the Produces-string, if any.
            if (operatorNode != null && operatorNode.ProducesString != null)
                dot.AppendFormat("Produces: {0}<br />{1}", HtmlEncode(operatorNode.ProducesString), Environment.NewLine);
            // Requirements, if any.
            if (operatorNode != null && operatorNode.RequiresString != null)
                dot.AppendFormat("Requires: {0}<br />{1}", HtmlEncode(operatorNode.RequiresString), Environment.NewLine);

            if (operatorNode.Plan != null)
            {
                if (operatorNode.Plan.Costs != null)
                    dot.AppendFormat("{0}<br />{1}", HtmlEncode(operatorNode.Plan.Costs.ToString()), Environment.NewLine);
                if (operatorNode.Plan.OrderingState != null)
                    dot.AppendFormat("{0}<br />{1}", HtmlEncode(operatorNode.Plan.OrderingState.ToString().LimitTo(150)),
                                     Environment.NewLine);
                if (operatorNode.Plan.PlanSet != null && operatorNode.Plan.PlanSet.State != null)
                    dot.AppendFormat("{0}<br />{1}", HtmlEncode(operatorNode.Plan.PlanSet.State.ToString()),
                                     Environment.NewLine);
            }

            foreach (string property in node.Properties.Keys)
                dot.AppendLine(string.Format("{0} = {1}<br />", property,
                                             HtmlEncode(PropertyToString(node.Properties[property]).LimitTo(100))));

            dot.AppendLine(">]");
            return dot.ToString();
        }

        private string DotForEdge(Node from, Node to)
        {
            // Surely we'll add more attributes here.
            string label = "";

            if (from is OperatorNode && ((OperatorNode) from).OutputTypeDescriptor != null)
            {
                label = ((OperatorNode) from).OutputTypeDescriptor.ToString();
                label = CleanupLabel(label);
                label = label.Replace("],[", "], <br/> [");
            }

            string edgeToParent = string.Format("{0} -> {1} [label=<{2}>]{3}", from.GetHashCode(), to.GetHashCode(),
                                                label, Environment.NewLine);

            if (ShowAttributeOrigin && from is OperatorNode)
            {
                StringBuilder edges = new StringBuilder(edgeToParent);
                foreach (Field field in ((OperatorNode) from).RecordDescriptor)
                {
                    edges.AppendFormat(@"{0} -> {1} [label=<{2}>,style=""dashed""]{3}", from.GetHashCode(),
                                       field.NodeAttribute.Node.GetHashCode(), field.Name, Environment.NewLine);
                }
                edgeToParent = edges.ToString();
            }

            if (ShowDependencies && from is OperatorNode)
            {
                StringBuilder edges = new StringBuilder(edgeToParent);
                foreach (HashSet<NodeAttribute> dependencySet in ((OperatorNode) from).Dependencies)
                    foreach (NodeAttribute dependency in dependencySet)
                        edges.AppendFormat(@"{0} -> {1} [label=<{2}>,style=""bold""]{3}", from.GetHashCode(),
                                           dependency.Node.GetHashCode(), dependency.AttributeName, Environment.NewLine);

                edgeToParent = edges.ToString();
            }

            return edgeToParent;
        }

        private string DotForGraph(Node root, string graphLabel)
        {
            StringBuilder graph = new StringBuilder();

            // The header. Add any graph options here.
            graph.AppendLine("digraph test {");
            graph.AppendLine("node [shape=\"box\" fontsize=\"12\"];");
            graph.AppendLine("edge [fontsize=\"12\"];");
            graph.AppendLine("rankdir=BT;");
            graph.AppendLine("labelloc=t;");
            graph.AppendLine("ordering=in;");
            graph.AppendLine(string.Format("label=<{0}>;", graphLabel));

            StringBuilder nodeLabels = new StringBuilder();
            StringBuilder edgeLabels = new StringBuilder();

            // Track leaves to make them the same rank.
            List<Node> leaveNodes = new List<Node>();

            root.TraverseDepthFirst(delegate(Node node)
                                        {
                                            nodeLabels.AppendLine(DotForNode(node));
                                            foreach (var parent in node.Parents)
                                            {
                                                edgeLabels.AppendLine(DotForEdge(node, parent));
                                            }

                                            if (node.IsLeaf)
                                            {
                                                leaveNodes.Add(node);
                                            }
                                        });

            graph.AppendLine();
            graph.AppendLine("// Node labels");
            graph.Append(nodeLabels.ToString());
            if (leaveNodes.Count >= 2)
            {
                graph.AppendLine("{ rank=same; ");
                graph.AppendLine(Utils.JoinEnumerable(from leaf in leaveNodes select leaf.GetHashCode(), ";"));
                graph.AppendLine("}");
            }
            graph.AppendLine("// Edges");
            graph.Append(edgeLabels.ToString());

            graph.AppendLine("}");
            return graph.ToString();
        }


        public void BumpCounter()
        {
            graphsMade++;
        }
    }
}