﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using Microsoft.Ceres.Evaluation.DataModel.Types;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer
{
    public class OperatorNode : Node, ICloneable
    {
        private static int nodeCounter = -1;
        public Type ByRule;

        public OperatorNode()
        {
            Id = Interlocked.Increment(ref nodeCounter);
            RecordDescriptor = new RecordDescriptor();
            Behaviour = OperatorBehaviour.None;
            Dependencies = new List<HashSet<NodeAttribute>>();
            Dependers = new HashSet<Node>();
        }

        public OperatorNode(Type operatorType) : this()
        {
            OperatorType = operatorType;
        }

        public int Id { get; set; }
        public RecordDescriptor RecordDescriptor { get; set; }
        public OperatorBehaviour Behaviour { get; set; }
        public List<HashSet<NodeAttribute>> Dependencies { get; set; }
        public HashSet<Node> Dependers { get; set; }
        public Type OperatorType { get; set; }
        public List<IRule> Rules { get; set; }
        public IRecordSetTypeDescriptor OutputTypeDescriptor { get; set; }
        public Plan Plan { get; set; }

        public IEnumerable<OperatorNode> Outputs
        {
            get { return Parents.Cast<OperatorNode>(); }
        }

        public OperatorNode OnlyOutput
        {
            get { return (OperatorNode) OnlyParent; }
        }

        public IEnumerable<OperatorNode> Inputs
        {
            get { return Children.Cast<OperatorNode>(); }
        }

        public OperatorNode OnlyInput
        {
            get { return (OperatorNode) OnlyChild; }
        }

        public bool HasRules
        {
            get { return Rules != null && Rules.Count > 0; }
        }

        public string RequiresString
        {
            get
            {
                // If it's not an ISearchRule, it has no requirements.
                if (!HasRules || !(Rules[0] is ISearchRule))
                {
                    return null;
                }
                ISearchRule firstRule = (ISearchRule) Rules[0];
                return "[" + Utils.JoinEnumerable(from bitSet in firstRule.Required select bitSet.BitSet, "; ") + "]";
            }
        }

        public string ProducesString
        {
            get
            {
                // If it's not an IProdocerRule, it produces nothing.
                if (!HasRules || !(Rules[0] is IProducerRule))
                {
                    return null;
                }
                IProducerRule firstRule = (IProducerRule) Rules[0];
                return firstRule.Produced.BitSet.ToString();
            }
        }

        #region ICloneable Members

        public override object Clone()
        {
            OperatorNode clone = new OperatorNode();
            SetVariablesForClone(clone);
            return clone;
        }

        #endregion

        public NestableHashSet<string> GetDepencyProperties(int index)
        {
            return new NestableHashSet<string>(Dependencies[index].Select(na => na.PropertyName));
        }

        public HashSet<OperatorNode> GetAllNodeDependencies()
        {
            HashSet<OperatorNode> nodes = new HashSet<OperatorNode>();
            foreach (HashSet<NodeAttribute> dependencies in Dependencies)
                foreach (NodeAttribute dependency in dependencies)
                    nodes.Add(dependency.Node);
            return nodes;
        }

        public void PassthroughInputAttributeOrigins()
        {
            foreach (OperatorNode input in Inputs)
                foreach (Field field in input.RecordDescriptor)
                    RecordDescriptor.AddFieldForName(field.CloneWithoutPosition());
        }

        public void PassthroughFieldNames()
        {
            PassthroughFieldNames(int.MaxValue);
        }

        public void PassthroughFieldNames(int childPrefix)
        {
            for (int i = 0; i < Math.Min(childPrefix, Children.Count); i++)
                foreach (Field field in ((OperatorNode) Children[i]).RecordDescriptor)
                    RecordDescriptor.AddFieldForAttr(field.CloneWithoutPosition());
        }

        public override string ToString()
        {
            if (OperatorType != null)
                return string.Format("OperatorNode({0})<{1}>", OperatorType.Name, Id);
            return base.ToString();
        }

        public void SetVariablesForClone(OperatorNode clone)
        {
            base.SetVariablesForClone((Node) clone);
            clone.Rules = Rules;
            clone.OperatorType = OperatorType;
            clone.OutputTypeDescriptor = OutputTypeDescriptor;
        }

        public override int GetHashCode()
        {
            return Id.GetHashCode();
        }

        public override bool Equals(object obj)
        {
            OperatorNode other = obj as OperatorNode;
            return !ReferenceEquals(other, null) && Id == other.Id;
        }
    }
}