﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Preprocessors
{
    /// <summary>
    /// An annotator is a transformation rule that only adds information to the graph in some way.
    /// Annotator should not change the graph structure in any way that iterative rules need to consider.
    /// </summary>
    /// <typeparam name="MapperType"></typeparam>
    public abstract class AbstractAnnotator<MapperType> : AbstractTransformationRule where MapperType : IMapper
    {
        protected Dictionary<Type, MapperType> mapperForType = new Dictionary<Type, MapperType>();
        protected QueryOptimization queryOptimization;

        public override AbstractNodeMatcher Pattern
        {
            // Annotation should happen from Output and below. Going through graph matching and all
            // just to pass the root-operator on to Annotate() may seem clunky, but it allows us
            // to use the same infrastructure for annotators and processors, and they need to depend
            // on each other.
            get { return Match.NodeOfType("OutputOperator").GroupAs("root"); }
        }

        public override void Configure(QueryOptimization queryOptimization)
        {
            // Annotators typically consult the rule binders.
            this.queryOptimization = queryOptimization;
            foreach (IRuleBinder ruleBinder in QueryOptimization.RuleBinders)
                if (ruleBinder is MapperType)
                    foreach (Type type in ((MapperType) ruleBinder).MapperFor)
                        mapperForType.Add(type, (MapperType) ruleBinder);
        }

        public override bool Fire(GraphSearcher graphSearcher)
        {
            Annotate((OperatorNode) graphSearcher.Groups["root"].Single());

            // Do NOT change the graph structure in annotators such that other rules need re-firing!
            return false;
        }

        public abstract void Annotate(OperatorNode root);
    }
}