﻿#region

using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Optimizer.GraphMatching;

#endregion

namespace Microsoft.Ceres.Optimizer.Preprocessors
{
    /// <summary>
    /// CopyRemover removes all Copy-operators from the query, effectively converting a DAG into an equivalent tree.
    /// </summary>
    [Preprocessor]
    public class CopyRemover : AbstractTransformationRule
    {
        public override AbstractNodeMatcher Pattern
        {
            // We'll traverse the root. We could have Fire()d with a CopyOperator-matcher, but
            // going in topological ordering is easier.
            get { return Match.NodeOfType("OutputOperator").GroupAs("root"); }
        }

        public override bool Fire(GraphSearcher graphSearcher)
        {
            OperatorNode outputNode = (OperatorNode) graphSearcher.Groups["root"].Single();

            bool changed = false;
            foreach (OperatorNode node in outputNode.GetTopologicalOrdering())
            {
                if (node.OperatorType != typeof (CopyOperator))
                    continue;

                // We've found a Copy-operator and will for sure change the graph.
                changed = true;

                // If the Copy-operator doesn't actually copy, simply remove it.
                if (node.Parents.Count == 1)
                {
                    node.Extract();
                    continue;
                }

                List<OperatorNode> outputs = new List<OperatorNode>(node.Outputs);

                // The only input of the copy operator needs to be copied --- one for each output of Copy.
                OperatorNode inputToClone = (OperatorNode) node.OnlyChild;
                // We'll make new ones, so just remove the existing one.
                node.RemoveChild(inputToClone);

                foreach (OperatorNode output in outputs)
                {
                    OperatorNode clonedInput = (OperatorNode) inputToClone.CloneNodeAndSubgraph();
                    output.AddChild(clonedInput);
                    output.RemoveChild(node);
                }
            }

            return changed;
        }
    }
}