﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;

#endregion

namespace Microsoft.Ceres.Optimizer.Preprocessors
{
    /// <summary>
    ///  Analyzes whether the output of a query (or
    ///  queries, in a multi-query) should have a specific ordering --- i.e.
    ///  that the ordering of the output of a Sort-operator is
    ///  guaranteed to be preserved until it reaches the
    ///  Output -operator.
    ///  When an ordering for a result set is desired, the relevant
    ///  Output-inputs (i.e. queries) are marked as such.
    /// </summary>
    [Preprocessor]
    public class OrderingExtracter : AbstractTransformationRule
    {
        public override IEnumerable<Type> DependsOn
        {
            get { yield return typeof (ProducesAnnotator); }
        }

        public override bool Iterative
        {
            get { return false; }
        }

        public override AbstractNodeMatcher Pattern
        {
            get
            {
                // Match SortOperators below Output and Trim.
                return (Match.NodeOfType("OutputOperator") | Match.NodeOfType("TrimOperator")).GroupAs("root")
                    .WithDescendant(Match.NodeOfType("SortOperator").GroupAs("sort"));
            }
        }

        public override bool Fire(GraphSearcher graphSearcher)
        {
            Node root = graphSearcher.Groups["root"].Single();

            // Match order preserving operators with a sort descendant, as long as all operators in between
            // also are order preserving.
            AbstractNodeMatcher subPattern = Match.NodeWithBehaviour(OperatorBehaviour.OrderPreserving)
                .WithDescendant(Match.NodeWithBehaviour(OperatorBehaviour.OrderPreserving),
                                Match.NodeOfType("SortOperator").GroupAs("sort"));

            // The n-th OrderDescription is for the n-th input. n can only be > 1 for Output-operator.
            IList<OrderDescription> orders = new List<OrderDescription>();
            foreach (Node child in root.Children)
            {
                GraphSearcher subSearcher = subPattern.GetSearcher();
                subSearcher.FirstHitOnly = true; // Taking one at the time is simpler.

                OperatorNode sort = null;
                if (subSearcher.Search(child))
                    // The sort operator is a descendant.
                    sort = (OperatorNode) subSearcher.Groups["sort"].Single();
                else if (((OperatorNode) child).OperatorType == typeof (SortOperator))
                    // The sort operator is a direct child. The pattern is simpler by keeping the trivial case out of it.
                    sort = (OperatorNode) child;

                if (sort != null)
                {
                    // We've found a sort. Extract its ordering requirements, to be set as a property.
                    orders.Add(getOrderDescription(sort));
                }
                else
                {
                    orders.Add(null);
                }
            }
            root["Orders"] = orders;
            return false;
        }

        private static OrderDescription getOrderDescription(OperatorNode sort)
        {
            OrderDescription orderDesc = new OrderDescription();
            IList<SortDefinition> sortFields = (IList<SortDefinition>) sort["SortFields"];
            foreach (SortDefinition sortDef in sortFields)
                orderDesc.Append(
                    new OrderDescriptionItem(
                        sort.OnlyInput.RecordDescriptor[sortDef.Field].NodeAttribute.PropertyName,
                        sortDef.NaturalOrder));
            return orderDesc;
        }
    }

    [Preprocessor]
    public class SortRemover : AbstractTransformationRule
    {
        public override bool Iterative
        {
            get { return false; }
        }

        public override IEnumerable<Type> DependsOn
        {
            // Clearly we must have extracted the useful information before removing the operators.
            get { return new List<Type> {typeof (OrderingExtracter)}; }
        }

        public override AbstractNodeMatcher Pattern
        {
            get { return Match.NodeOfType("SortOperator"); }
        }

        public override bool Fire(GraphSearcher graphSearcher)
        {
            foreach (Node node in graphSearcher.MatchedNodes)
            {
                if ((long) node["TrimHitCount"] > 0 || (long) node["TrimOffset"] > 0)
                {
                    // If the Sort-operator also does trimming, we need to replace it with a
                    // Trim-operator, with the ordering requirements.
                    OperatorNode trim = new OperatorNode(typeof (TrimOperator));
                    trim["Hits"] = node["TrimHitCount"];
                    trim["Offset"] = node["TrimOffset"];
                    trim["Order"] = node["SortFields"];
                    trim.OutputTypeDescriptor = ((OperatorNode) node).OutputTypeDescriptor;
                    node.ReplaceWith(trim);
                }
                else
                {
                    // Remove the Sort operator.
                    node.Extract();
                }
            }

            // If we've fired at all, we've surely changed the graph.
            return true;
        }
    }
}