﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.DataModel.Types;
using Microsoft.Ceres.Optimizer.Engine;

#endregion

namespace Microsoft.Ceres.Optimizer
{
    public class Field
    {
        public Field(NodeAttribute nodeAttribute, Type type)
            : this(nodeAttribute, nodeAttribute.AttributeName, type)
        {
        }

        public Field(NodeAttribute nodeAttribute, string name, Type type)
        {
            NodeAttribute = nodeAttribute;
            Name = name;
            Position = -1;
            Type = type;
        }

        public NodeAttribute NodeAttribute { get; set; }
        public string Name { get; set; }
        public int Position { get; set; }
        public Type Type { get; set; }

        public Field CloneWithoutPosition()
        {
            return new Field(NodeAttribute, Name, Type);
        }

        public override string ToString()
        {
            return Utils.NullFormat("Field(NA: {0}, Name: {1}, Pos: {2}, Type: {3})", NodeAttribute, Name, Position,
                                    Type);
        }
    }

    public class RecordDescriptor : IEnumerable<Field>
    {
        private HashSet<Field> fields = new HashSet<Field>();
        private Dictionary<string, Field> fieldsByName = new Dictionary<string, Field>();
        private Dictionary<NodeAttribute, Field> fieldsByNodeAttribute = new Dictionary<NodeAttribute, Field>();

        public RecordDescriptor(params Field[] fields)
        {
            foreach (Field field in fields)
                AddFieldForName(field);
        }

        public Field this[int position]
        {
            get { return fields.Single(f => f.Position == position); }
        }

        public Field this[string name]
        {
            get { return fieldsByName[name]; }
        }

        public Field this[QueryOptimization queryOptimization, NodeAttribute nodeAttribute]
        {
            get
            {
                Field result;
                if (fieldsByNodeAttribute.TryGetValue(nodeAttribute, out result))
                    return result;
                if (fieldsByNodeAttribute.TryGetValue(queryOptimization.SEQAttributeMappings[nodeAttribute], out result))
                    return result;

                throw new OptimizerException("Field {0} not available, available: {1}.", nodeAttribute,
                                             Utils.JoinEnumerable(fieldsByNodeAttribute.Keys));
            }
        }

        #region IEnumerable<Field> Members

        public IEnumerator<Field> GetEnumerator()
        {
            return fields.OrderBy(f => f.Position).GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        public void AddFieldForName(Field field)
        {
            if (field.Position == -1)
                field.Position = fields.Count;
            fields.Add(field);
            fieldsByName.Add(field.Name, field);
        }

        public void AddFieldForAttr(Field field)
        {
            if (field.Position == -1)
                field.Position = fields.Count;
            fields.Add(field);
            fieldsByNodeAttribute.Add(field.NodeAttribute, field);
        }

        public bool Contains(string name)
        {
            return fieldsByName.ContainsKey(name);
        }

        public void Clear()
        {
            fields.Clear();
            fieldsByName.Clear();
            fieldsByNodeAttribute.Clear();
        }

        public IEnumerable<Field> GetFieldsWithMappings(QueryOptimization queryOptimization)
        {
            foreach (Field field in this)
            {
                yield return field;
                foreach (KeyValuePair<NodeAttribute, NodeAttribute> mapping in queryOptimization.SEQAttributeMappings)
                    if (mapping.Value == field.NodeAttribute && mapping.Key != mapping.Value)
                        yield return new Field(mapping.Key, field.Name, field.Type);
            }
        }

        public IRecordTypeDescriptor AsRecordTypeDescriptor()
        {
            return new RecordTypeDescriptorProxy(this);
        }

        #region Nested type: RecordTypeDescriptorProxy

        public class RecordTypeDescriptorProxy : IRecordTypeDescriptor
        {
            private RecordDescriptor recordDescriptor;

            public RecordTypeDescriptorProxy(RecordDescriptor recordDescriptor)
            {
                this.recordDescriptor = recordDescriptor;
            }

            #region IRecordTypeDescriptor Members

            public bool Contains(string name)
            {
                return recordDescriptor.fieldsByName.ContainsKey(name);
            }

            public int Count
            {
                get { return recordDescriptor.fieldsByName.Count; }
            }

            public string Name(int pos)
            {
                return recordDescriptor[pos].NodeAttribute.PropertyName;
            }

            public int Position(string name)
            {
                return recordDescriptor[name].Position;
            }

            public FieldOccurrence this[int pos]
            {
                get { return ToFieldOccurrence(recordDescriptor[pos]); }
            }

            public FieldOccurrence this[string name]
            {
                get { return ToFieldOccurrence(recordDescriptor[name]); }
            }

            public IEnumerator<FieldOccurrence> GetEnumerator()
            {
                return recordDescriptor.Select(f => ToFieldOccurrence(f)).GetEnumerator();
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }

            #endregion

            private FieldOccurrence ToFieldOccurrence(Field field)
            {
                return new FieldOccurrence(field.NodeAttribute.PropertyName,
                                           TypeRegistry.Singleton.GetType(field.Type).Type, field.Position);
            }
        }

        #endregion
    }
}