﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>Abstract base class for all rules.</summary>
    public abstract class AbstractRule : IRule
    {
        protected QueryOptimization queryOptimization;

        public AbstractRule(QueryOptimization queryOptimization)
        {
            this.queryOptimization = queryOptimization;
        }

        /// <summary>Nodes this rule was instantiated from.</summary>
        public IList<OperatorNode> Nodes { get; set; }

        // IRule

        #region IRule Members

        /// <summary>
        /// Updates the properties of the plan (costs, ordering, sharing).
        /// Will calculate costs for both DAGs and trees.
        /// </summary>
        /// <param name="plan">The plan to update.</param>
        public virtual void UpdatePlan(Plan plan)
        {
            plan.Rule = this;

            UpdateCosts(plan);
        }

        /// <summary>
        /// Updates the costs for the given plan, automatically choosing between DAG and tree calculation
        /// algorithms.
        /// </summary>
        /// <param name="plan">The plan to calculate costs for.</param>
        public virtual void UpdateCosts(Plan plan)
        {
            // Get the cost for this operator.
            BasicCost cost = Cost(plan);
            // If only one child, just sum them.
            if (plan.Children.Count == 1)
                cost += (BasicCost) plan.OnlyChild.Costs;
            // If more than one child, we may need to do DAG costing.
            else if (plan.Children.Count > 1)
                cost += BasicCost.InputCosts(plan.Children, null);
            plan.Costs = cost;
        }

        /// <summary>
        /// Builds the Node in the physical algebra graph for this rule and calls
        /// its children recursively.
        /// </summary>
        /// <param name="plan">Plan to build algebra for.</param>
        public abstract OperatorNode BuildAlgebra(Plan plan);

        /// <summary>
        /// Return the cost for the subtree with this rule on the top,
        /// using the cost algorithms for DAGs.
        /// </summary>
        /// <param name="plan">The plan (constructed by this rule).</param>
        /// <param name="reads">How many reads of the plan.</param>
        /// <returns>The costs for the subtree.</returns>
        public virtual ICost DagCosts(Plan plan, int reads)
        {
            BasicCost planCosts = (BasicCost) plan.Costs;
            // If we've already been executed this many times, the reads are free.
            if (reads < planCosts.Passes)
                return BasicCost.Zero;

            // Our cost:
            int additional = reads - planCosts.Passes;
            planCosts.Passes = reads;
            plan.Costs = planCosts; // BasicCost is a struct, so set the entire thing.
            BasicCost result = Cost(plan)*additional;

            // Cost of inputs:
            foreach (Plan subPlan in plan.Children)
                result += (BasicCost) subPlan.Rule.DagCosts(subPlan, reads);

            return result;
        }

        #endregion

        /// <summary>
        /// Sets the node out-parameter to the operator root-node for the plan.
        /// Returns true if the node was created and false if it was found in the reconstruction table.
        /// </summary>
        protected bool GetOrCreatePhysicalNode(Plan plan, Type operatorType, out OperatorNode node)
        {
            if (queryOptimization.ReconstructionTable.TryGetValue(plan, out node))
                return false;

            node = createPhysicalNode(plan, operatorType);
            return true;
        }

        // Creates the physical operator node for the given plan and operator type.
        private OperatorNode createPhysicalNode(Plan plan, Type operatorType)
        {
            // Root node for the plan:
            OperatorNode node = new OperatorNode(operatorType);
            node.ByRule = GetType();
            if (Nodes != null && Nodes.Count > 0)
                node.Id = Nodes.First().Id;
            node.Plan = plan;

            // Also create children.
            foreach (Plan subPlan in plan.Children)
                node.AddChild(subPlan.Rule.BuildAlgebra(subPlan));

            // Memoize it so new requests for the same plan gets the same node,
            // important for creating DAGs.
            queryOptimization.ReconstructionTable[plan] = node;

            return node;
        }

        /// <summary>
        /// The cost for only the operator represented by this rule, applied
        /// on the top of the given plan.
        /// </summary>
        /// <param name="plan">Plan with this rule on the top.</param>
        /// <returns>The cost of this exact rule.</returns>
        protected abstract BasicCost Cost(Plan plan);
    }
}