﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Base class for rule binders, offering common functionality.
    /// </summary>
    /// <typeparam name="RuleType">The type of rule to bind.</typeparam>
    public abstract class AbstractRuleBinder<RuleType> : IRuleBinder where RuleType : IProducerRule
    {
        #region IRuleBinder Members

        /// <summary>Node pattern to match in the operator graph during initialization.</summary>
        public abstract AbstractNodeMatcher Pattern { get; }

        /// <summary>
        /// This method is called by the optimizer to get the rules this rule binder produces.
        /// The default implementation instantiates a rule for each pattern match
        /// and sets its produced properties to RULE_APPLIED.
        /// </summary>
        /// <param name="queryOptimization">The currrent QueryOptimization instance.</param>
        /// <param name="matches">IEnumerable of matches to the pattern declared by the rule binder.</param>
        /// <returns>IEnumerable of rules that should be added to the search by the optimizer.</returns>
        public virtual IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                           IEnumerable<OperatorNode> nodes)
        {
            List<IProducerRule> rules = new List<IProducerRule>();
            // For each match, instantiate a rule and set properties.
            foreach (OperatorNode node in nodes)
            {
                RuleType rule = CreateRule(queryOptimization, node);
                rule.Produced = queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Produced,
                                                                          new NodeAttribute(rule.Nodes.Single(), null).
                                                                              PropertyName);
                rules.Add(rule);
            }
            return rules;
        }

        #endregion

        /// <summary>
        /// Creates a rule instance and sets common properties.
        /// </summary>
        /// <param name="queryOptimization">The query optimization instance.</param>
        /// <param name="node">Node to get commoin properties from.</param>
        /// <returns>The created rule instance.</returns>
        protected RuleType CreateRule(QueryOptimization queryOptimization, OperatorNode node)
        {
            RuleType rule = (RuleType) Activator.CreateInstance(typeof (RuleType), queryOptimization);
            rule.Id = node.Id;
            rule.Nodes = new List<OperatorNode> {node};
            node.Rules = new List<IRule> {rule};
            return rule;
        }
    }
}