﻿#region

using System.Collections.Generic;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Base class for search rules, implementing required members and
    /// Filter caching.
    /// </summary>
    public abstract class AbstractSearchRule : AbstractRule, ISearchRule
    {
        // We're using a bool on the side instead of BitSet? (nullable bitset) becase it is actually less overhead.
        private bool cached = false;
        private BitSet cachedFilter;

        public AbstractSearchRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
            // Default selectivity as this may also be a non-limiting rule.
            Selectivity = 1;
        }

        /// <summary>Selectivity for this rule. 1 for non-limiting rules.</summary>
        public double Selectivity { get; set; }

        #region ISearchRule Members

        /// <summary>Properties produced by this rule.</summary>
        public BitSetProxy Produced { get; set; }

        /// <summary>List of properties required for this rule.</summary>
        public IList<BitSetProxy> Required { get; set; }

        /// <summary>
        /// Automatically generates and caches the Filter property based
        /// on the Produced and Required properties.
        /// </summary>
        public virtual BitSet Filter
        {
            get
            {
                if (cached)
                    return cachedFilter;

                cached = true;
                cachedFilter = (BitSet)Produced.BitSet.Clone();
                foreach (BitSetProxy bitsetProxy in Required)
                    cachedFilter.Or(bitsetProxy.BitSet);

                return cachedFilter;
            }
        }

        /// <summary>
        /// Determine if this rule is relevant to reach the given goal.
        /// </summary>
        public bool IsRelevantTo(BitSet goal)
        {
            return Filter <= goal;
        }

        /// <summary>Guides the search for this rule instance.</summary>
        /// <param name="plans">PlanSet to add plans to (also defines goal properties).</param>
        /// <param name="limit">Abort the search if passing this cost limit (pruning).</param>
        public abstract void Search(PlanSet planSet, ICost limit);

        /// <summary>Id for this rule as given by the optimizer.</summary>
        public int Id { get; set; }

        /// <summary>
        /// Determines if this and the other rule is structurally identical, i.e. if they have
        /// the same configured properties, the same join keys etc.
        /// </summary>
        /// <param name="other">Rule to compare with.</param>
        /// <returns>If the rules are structurally identical.</returns>
        public abstract bool StructurallyIdentical(IProducerRule other);

        /// <summary>
        /// This method will be called if this rule is determined to be share equivalent with another rule.
        /// The default implementaion is empty as many rules do not produce attributes.
        /// </summary>
        /// <param name="from">The rule to map attributes from.</param>
        /// <param name="equivalenceClassAttributeMappings">The dictionary the mappings should be added to.</param>
        public virtual void MapShareEquivalentAttributes(IProducerRule from,
                                                         Dictionary<NodeAttribute, NodeAttribute>
                                                             equivalenceClassAttributeMappings)
        {
        }

        #endregion
    }
}