﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Aggregation;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Represents a group operator (both hash and streaming) in a query.
    /// </summary>
    public class GroupOperatorRule : UnaryRule
    {
        private const double FIELD_SIZE = 10;
        private const double HASH_INSTRUCTION_COUNT = 50;
        private const double MAX_MEMORY = 1024*1024*10;
        private const double STREAM_INSTRUCTION_COUNT = 20;

        public GroupOperatorRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
        }

        public GroupingProxy Grouping { get; set; }
        public IList<NodeAttribute> GroupingFields { get; set; }
        public IDictionary<string, Aggregator> Aggregators { get; set; }

        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);

            // Update ordering stuff.
            Plan input = plan.Children[0];
            if (input.OrderingState.Order < Grouping.Order)
                plan.OrderingState =
                    new OrderingState(queryOptimization.OrderManager, Grouping).Apply(input.OrderingState);
            else
                plan.OrderingState = input.OrderingState;
        }

        public bool CanStreamGrouping(Plan plan) {
            // If the input has ordering that satisfies the requested grouping, we need not group first.
            return plan.OnlyChild.OrderingState.Order >= Grouping.Order;
        }

        public override OperatorNode BuildAlgebra(Plan plan)
        {
            OperatorNode node;
            if (GetOrCreatePhysicalNode(plan, typeof (GroupOperator), out node))
            {
                node.Properties = Nodes.First().Properties;
                node.Rules = Nodes.First().Rules;

                if (CanStreamGrouping(plan))
                    node["Type"] = "Stream";
                else
                    node["Type"] = "Hash";
            }
            return node;
        }

        public override bool StructurallyIdentical(IProducerRule other)
        {
            // TODO: Implement this. None of our example queries have SHEQ groups, so this is fine for now.
            return false;
        }

        public override string ToString()
        {
            return string.Format("GroupOperator(Produced: {0}, Required: {1}", Produced.BitSet, OnlyRequired);
        }

        protected override BasicPlanSetState CalcPlanSetState(BasicPlanSetState input)
        {
            return base.CalcPlanSetState(input).Expand((GroupingFields.Count + Aggregators.Count)*FIELD_SIZE);
        }

        protected override BasicCost Cost(Plan plan)
        {
            BasicPlanSetState childState = (BasicPlanSetState) plan.OnlyChild.PlanSet.State;
            double cost = CanStreamGrouping(plan)
                              ?
                                costForStreamingGroup(childState)
                              : costForHashingGroup(childState);
  
            return new BasicCost(cost, cost);
        }

        private double costForStreamingGroup(BasicPlanSetState childState)
        {
            return BasicCost.CpuCosts(childState.Cardinality,
                                      (GroupingFields.Count + Aggregators.Count)*STREAM_INSTRUCTION_COUNT);
        }

        private double costForHashingGroup(BasicPlanSetState childState)
        {
            double cost = BasicCost.CpuCosts(childState.Cardinality,
                                             (GroupingFields.Count + Aggregators.Count)*HASH_INSTRUCTION_COUNT);

            if (childState.ResultSetSize > MAX_MEMORY)
            {
                // We need to spool to disk.
                cost += (BasicCost.WriteRandom(BasicCost.PageAccesses(childState.ResultSetSize))
                         + BasicCost.ReadSequential(BasicCost.PageAccesses(childState.ResultSetSize))
                         +
                         BasicCost.CpuCosts(childState.Cardinality*Selectivity,
                                            (GroupingFields.Count + Aggregators.Count)*STREAM_INSTRUCTION_COUNT));
            }
            return cost;
        }
    }

    [RuleBinder]
    public class GroupOperatorRuleBinder : AbstractRuleBinder<GroupOperatorRule>, IProducesMapper, IDependencyMapper,
                                           IFieldMapper
    {
        public override AbstractNodeMatcher Pattern
        {
            get { return Match.NodeOfType("GroupOperator"); }
        }

        #region IDependencyMapper Members

        public void SetDependencies(OperatorNode groupNode)
        {
            Debug.Assert(groupNode.OperatorType == typeof (GroupOperator));

            groupNode.Dependencies.Add(new HashSet<NodeAttribute>());
            // Add grouping fields.
            foreach (NodeAttribute groupingField in (IEnumerable<NodeAttribute>) groupNode["GroupingFields"])
                groupNode.Dependencies[0].Add(groupingField);

            // Add all operators below it
            // TODO: This heuristic is too simple, but okay for a prototype.
            foreach (OperatorNode node in groupNode.IterateBreadthFirst(false))
                groupNode.Dependencies[0].Add(new NodeAttribute(node, null));
            // TODO: Add depency on this for all operators above it.
        }

        #endregion

        #region IFieldMapper Members

        public void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            // Grouping fields.
            IList<string> groupingFields = new List<string>();
            foreach (NodeAttribute groupingField in (IList<NodeAttribute>) node.Properties["GroupingFields"])
            {
                groupingFields.Add(node.OnlyInput.RecordDescriptor[queryOptimization, groupingField].Name);
                node.RecordDescriptor.AddFieldForAttr(
                    node.OnlyInput.RecordDescriptor[queryOptimization, groupingField].CloneWithoutPosition());
            }
            node.Properties["GroupingFields"] = groupingFields;

            // Finally, add any aggregates.
            // Todo: Transform refs to string.
            foreach (
                string aggregateAttribute in ((IDictionary<string, Aggregator>) node.Properties["Aggregators"]).Keys)
                node.RecordDescriptor.AddFieldForAttr(new Field(new NodeAttribute(node, aggregateAttribute), null));
        }

        #endregion

        #region IProducesMapper Members

        public IEnumerable<Type> MapperFor
        {
            get { yield return typeof (GroupOperator); }
        }

        public void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            // Grouping fields.
            IList<NodeAttribute> groupingFields = new List<NodeAttribute>();
            foreach (string groupingField in (IList<string>) node.Properties["GroupingFields"])
            {
                groupingFields.Add(node.OnlyInput.RecordDescriptor[groupingField].NodeAttribute);
                node.RecordDescriptor.AddFieldForName(
                    node.OnlyInput.RecordDescriptor[groupingField].CloneWithoutPosition());
            }
            node.Properties["GroupingFields"] = groupingFields;

            // Finally, add any aggregates.
            // Todo: Transform refs to NodeAttribute.
            foreach (
                string aggregateAttribute in ((IDictionary<string, Aggregator>) node.Properties["Aggregators"]).Keys)
                node.RecordDescriptor.AddFieldForName(new Field(new NodeAttribute(node, aggregateAttribute),
                                                                node.OutputTypeDescriptor.RecordType[aggregateAttribute]
                                                                    .Type.FieldType));
        }

        #endregion

        public override IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                            IEnumerable<OperatorNode> nodes)
        {
            IEnumerable<IProducerRule> rules = base.GetRules(queryOptimization, nodes);
            foreach (GroupOperatorRule rule in rules)
            {
                rule.Selectivity = 0.01; //TODO: Hard coded.
                rule.GroupingFields = (IList<NodeAttribute>) rule.Nodes.Single()["GroupingFields"];
                rule.Aggregators = (IDictionary<string, Aggregator>) rule.Nodes.Single()["Aggregators"];
                rule.Required = new List<BitSetProxy>
                                    {
                                        queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Required,
                                                                                  rule.Nodes.Single().
                                                                                      GetDepencyProperties(0))
                                    };
                rule.Grouping = queryOptimization.OrderManager.GetGrouping(
                    new NestableHashSet<string>(rule.GroupingFields.Select(attr => attr.PropertyName)), true);
            }
            return rules;
        }
    }
}