﻿#region

using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Join;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Helper rule for JoinRule, offering cost calculation and
    /// physical algebra building.
    /// </summary>
    public class HybridHashJoinRule : AbstractRule, IHelperRule
    {
        private const double INSTRUCTION_COUNT = 35;
        private const double MAX_MEMORY = 1024*1024*300;
        private JoinRule parent;

        public HybridHashJoinRule(QueryOptimization queryOptimization, JoinRule parent)
            : base(queryOptimization)
        {
            this.parent = parent;
        }

        public HashSet<Dependency> InducedDependencies { get; set; }

        #region IHelperRule Members

        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);
            plan.OrderingState = queryOptimization.OrderManager.NullOrderingState
                .Apply(InducedDependencies).Apply(plan.LeftChild.OrderingState).Apply(plan.RightChild.OrderingState);
        }

        public override OperatorNode BuildAlgebra(Plan plan)
        {
            OperatorNode node;
            if (GetOrCreatePhysicalNode(plan, typeof (HybridHashJoinOperator), out node))
            {
                node.Properties = parent.Nodes.Single().Properties;
                parent.SwapJoinKeyIfNecessary((IList<NodeAttribute>)node["JoinKey"], plan);
                node["JoinKey"] = new List<IList<NodeAttribute>>() {(IList<NodeAttribute>) node["JoinKey"]};
                node.Rules = parent.Nodes.First().Rules;
            }
            return node;
        }

        #endregion

        public override string ToString()
        {
            return "HybridHashJoin";
        }

        protected override BasicCost Cost(Plan plan)
        {
            BasicPlanSetState leftState = (BasicPlanSetState) plan.LeftChild.PlanSet.State;
            BasicPlanSetState rightState = (BasicPlanSetState) plan.RightChild.PlanSet.State;

            double cost = BasicCost.CpuCosts(parent.GetCardinality(leftState, rightState), INSTRUCTION_COUNT);
            if (leftState.ResultSetSize > MAX_MEMORY && rightState.ResultSetSize > MAX_MEMORY)
            {
                // We need to spool to disk.
                cost += (BasicCost.WriteRandom(BasicCost.PageAccesses(leftState.ResultSetSize + rightState.ResultSetSize))
                         +
                         BasicCost.ReadSequential(
                             BasicCost.PageAccesses(leftState.ResultSetSize + rightState.ResultSetSize))
                         + BasicCost.CpuCosts(leftState.Cardinality + rightState.Cardinality, INSTRUCTION_COUNT));
            }
            return new BasicCost(cost, cost);
        }
    }
}