﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Join;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Operators;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Represents binary joins. 
    /// </summary>
    public class JoinRule : BinaryLogicalRule
    {
        public JoinRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
            HybridHashJoin = new HybridHashJoinRule(queryOptimization, this);
            MergeJoin = new MergeJoinRule(queryOptimization, this);
        }

        public HybridHashJoinRule HybridHashJoin { get; set; }
        public MergeJoinRule MergeJoin { get; set; }

        public IList<NodeAttribute> JoinKey { get; set; }

        public OrderProxy OrderLeft { get; set; }
        public OrderProxy OrderRight { get; set; }
        public IRule SortLeft { get; set; }
        public IRule SortRight { get; set; }

        public BitSetProxy OtherJoinRules { get; set; }
        public JoinEnumeration JoinEnumeration { get; set; }

        public override void Search(PlanSet planSet, ICost limit)
        {
            // TODO: Make sure we do not perform subsearches for unreachable plans? Reckon map-c-discussion.
            // We want plans with properties not determined by the join requirements and what we produce ourselves.
            BitSet wantedProperties, remainingJoins;
            if (JoinEnumeration == JoinEnumeration.Bushy)
            {
                wantedProperties = planSet.Properties - Filter;
                InternalSearch(planSet, limit, ref wantedProperties, RequiredLeft, RequiredRight,
                    OrderLeft.Order, OrderRight.Order, SortLeft, SortRight);
            }
            else
            {
                wantedProperties = planSet.Properties - (Filter | OtherJoinRules.BitSet);
                remainingJoins = (OtherJoinRules.BitSet & planSet.Properties) - Filter;

                if (JoinEnumeration == JoinEnumeration.LeftDeep)
                {
                    // Note that this tries to swap input plans for the join. This gives rise to the
                    // SwapJoinKeyIfNecessary method, as the join keys for the resulting operator must be swapped as well.
                    InternalSearch(planSet, limit, ref wantedProperties, RequiredLeft | remainingJoins, RequiredRight,
                        OrderLeft.Order, OrderRight.Order, SortLeft, SortRight);
                    InternalSearch(planSet, limit, ref wantedProperties, RequiredRight | remainingJoins, RequiredLeft,
                        OrderRight.Order, OrderLeft.Order, SortRight, SortLeft);
                }
                else if (JoinEnumeration == JoinEnumeration.RightDeep)
                {
                    InternalSearch(planSet, limit, ref wantedProperties, RequiredLeft, RequiredRight | remainingJoins,
                        OrderLeft.Order, OrderRight.Order, SortLeft, SortRight);
                    InternalSearch(planSet, limit, ref wantedProperties, RequiredRight, RequiredLeft | remainingJoins,
                        OrderRight.Order, OrderLeft.Order, SortRight, SortLeft);
                }
                else if (JoinEnumeration == JoinEnumeration.ZigZag)
                {
                    InternalSearch(planSet, limit, ref wantedProperties, RequiredLeft, RequiredRight | remainingJoins,
                        OrderLeft.Order, OrderRight.Order, SortLeft, SortRight);
                    InternalSearch(planSet, limit, ref wantedProperties, RequiredRight, RequiredLeft | remainingJoins,
                        OrderRight.Order, OrderLeft.Order, SortRight, SortLeft);

                    InternalSearch(planSet, limit, ref wantedProperties, RequiredLeft | remainingJoins, RequiredRight,
                        OrderLeft.Order, OrderRight.Order, SortLeft, SortRight);
                    InternalSearch(planSet, limit, ref wantedProperties, RequiredRight | remainingJoins, RequiredLeft,
                        OrderRight.Order, OrderLeft.Order, SortRight, SortLeft);
                }
            }
        }

        public double GetCardinality(BasicPlanSetState leftState, BasicPlanSetState rightState)
        {
            if (Selectivity == -1)
                return leftState.Cardinality;

            return leftState.Cardinality * rightState.Cardinality * Selectivity;
        }

        /// <summary>
        /// Swap join keys if the input plans have been swapped during left deep join enumeration.
        /// </summary>
        public void SwapJoinKeyIfNecessary(IList<NodeAttribute> joinKeys, Plan plan)
        {
            if (!(RequiredLeft <= plan.LeftChild.PlanSet.Properties
                || queryOptimization.ShareEquivalentGoal(RequiredLeft) <= plan.LeftChild.PlanSet.Properties))
            {
                NodeAttribute temp = joinKeys[0];
                joinKeys[0] = joinKeys[1];
                joinKeys[1] = temp;
            }
        }

        private void InternalSearch(PlanSet planSet, ICost limit, ref BitSet wantedProperties, BitSet left, BitSet right,
            Order orderLeft, Order orderRight, IRule sortLeft, IRule sortRight)
        {
            foreach (BitSet leftProperties in wantedProperties.Walk())
            {
                DiagPrinter.Instance.DiagWriteLine("search", "JoinRule producing {0} searching for {1} LEFT.",
                                                   Produced.BitSet, left | leftProperties);

                PlanSet leftPlans = queryOptimization.GeneratePlans(left | leftProperties, limit, orderLeft,
                                                                    sortLeft);
                if (leftPlans == null || leftPlans.GetCheapest().Costs.Compare(limit) == CostRelation.Worse)
                {
                    DiagPrinter.Instance.DiagWriteLine("search", "Found 0 plans.");
                    continue;
                }

                BitSet rightProperties = wantedProperties - leftProperties;
                DiagPrinter.Instance.DiagWriteLine("search", "Found {0} plans.", leftPlans.Count);
                DiagPrinter.Instance.DiagWriteLine("search", "JoinRule producing {0} searching for {1} RIGHT.",
                                                   Produced.BitSet, right | rightProperties);

                PlanSet rightPlans = queryOptimization.GeneratePlans(right | rightProperties, limit, orderRight,
                                                                     sortRight);
                if (rightPlans == null)
                {
                    DiagPrinter.Instance.DiagWriteLine("search", "Found 0 plans.");
                    continue;
                }

                DiagPrinter.Instance.DiagWriteLine("search", "Found {0} plans.", rightPlans.Count);
                DiagPrinter.Instance.DiagWriteLine("search", "JOINPLAN.", rightPlans.Count);

                if (planSet.Count == 0)
                {
                    // First plan, so set some state.
                    planSet.State =
                        new BasicPlanSetState(
                            GetCardinality((BasicPlanSetState) leftPlans.State, (BasicPlanSetState) rightPlans.State),
                            ((BasicPlanSetState) leftPlans.State).TupleSize +
                            ((BasicPlanSetState) rightPlans.State).TupleSize
                            );
                }

                foreach (Plan leftPlan in leftPlans)
                {
                    foreach (Plan rightPlan in rightPlans)
                    {
                        Plan plan;
                        // Merge join
                        if (leftPlan.OrderingState.Order >= orderLeft &&
                            rightPlan.OrderingState.Order >= orderRight)
                        {
                            plan = new Plan(leftPlan, rightPlan);
                            MergeJoin.UpdatePlan(plan);
                            UpdatePlan(plan);
                            planSet.AddPlan(plan);
                        }

                        // Hybrid Hash join
                        plan = new Plan(leftPlan, rightPlan);
                        HybridHashJoin.UpdatePlan(plan);
                        UpdatePlan(plan);
                        planSet.AddPlan(plan);
                    }
                }
            }
        }

        public override OperatorNode BuildAlgebra(Plan plan)
        {
            throw new OptimizerException("This should not happen, as this is delegated to helper rules.");
        }

        protected override BasicCost Cost(Plan plan)
        {
            throw new OptimizerException("This should not happen, as this is delegated to helper rules.");
        }

        public override bool StructurallyIdentical(IProducerRule other)
        {
            JoinRule otherJoin = other as JoinRule;

            // If it's not a join or the number of join keys differ, they are trivially not identical.
            if (otherJoin == null || JoinKey.Count != otherJoin.JoinKey.Count)
                return false;

            // Ensure that all the keys are equal. Note that we need to consider equivalence classes of keys.
            for (int i = 0; i < JoinKey.Count; i++)
                if (queryOptimization.SEQAttributeMappings[JoinKey[i]] !=
                    queryOptimization.SEQAttributeMappings[otherJoin.JoinKey[i]])
                    return false;

            // All keys are equal
            return true;
        }

        public override string ToString()
        {
            return string.Format("JoinRule(Produced: {0}, RL: {1}, RR: {2}", Produced.BitSet, RequiredLeft,
                                 RequiredRight);
        }
    }

    [RuleBinder]
    public class JoinRuleBinder : AbstractRuleBinder<JoinRule>, IProducesMapper, IDependencyMapper, IFieldMapper
    {
        public override AbstractNodeMatcher Pattern
        {
            get { return Match.NodeOfType("JoinOperator"); }
        }

        #region IDependencyMapper Members

        public void SetDependencies(OperatorNode node)
        {
            // We depend on the operators producing the join keys.t
            IList<NodeAttribute> joinKey = (IList<NodeAttribute>) node["JoinKey"];
            for (int i = 0; i < joinKey.Count; i++)
                node.Dependencies.Add(new HashSet<NodeAttribute> {joinKey[i]});
        }

        #endregion

        #region IFieldMapper Members

        public void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            IList<NodeAttribute> joinKey;
            if (node.OperatorType == typeof (MergeJoinOperator))
                joinKey = (IList<NodeAttribute>) node["JoinKey"];
            else if (node.OperatorType == typeof (HybridHashJoinOperator))
                joinKey = ((IList<IList<NodeAttribute>>) node["JoinKey"]).Single();
            else
                throw new OptimizerException("Unkown physical join operator.");

            IList<string> newJoinKey = new List<string>();

            // First deal with all keys.
            for (int i = 0; i < node.Children.Count; i++)
            {
                newJoinKey.Add(((OperatorNode) node.Children[i]).RecordDescriptor[queryOptimization, joinKey[i]].Name);
                foreach (Field field in ((OperatorNode) node.Children[i]).RecordDescriptor)
                    if ((joinKey[i] == field.NodeAttribute ||
                         queryOptimization.SEQAttributeMappings[joinKey[i]] == field.NodeAttribute))
                        node.RecordDescriptor.AddFieldForAttr(new Field(field.NodeAttribute, field.Name, field.Type));
            }

            // Rest of the attributes.
            // TODO: We assume non-composite keys.
            for (int i = 0; i < node.Children.Count; i++)
            {
                foreach (Field field in ((OperatorNode) node.Children[i]).RecordDescriptor)
                    if (joinKey[i] != field.NodeAttribute &&
                        queryOptimization.SEQAttributeMappings[joinKey[i]] != field.NodeAttribute)
                        node.RecordDescriptor.AddFieldForAttr(new Field(field.NodeAttribute,
                                                                        string.Format("Input{0}.{1}", i, field.Name),
                                                                        field.Type));
            }
            if (node.OperatorType == typeof (MergeJoinOperator))
                node["JoinKey"] = newJoinKey;
            else if (node.OperatorType == typeof (HybridHashJoinOperator))
                node["JoinKey"] = new List<IList<string>> {newJoinKey};
        }

        #endregion

        #region IProducesMapper Members

        public IEnumerable<Type> MapperFor
        {
            get
            {
                return new List<Type>
                           {typeof (JoinOperator), typeof (MergeJoinOperator), typeof (HybridHashJoinOperator)};
            }
        }

        public void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            IList<string> joinKey = (IList<string>) node["JoinKey"];
            IList<NodeAttribute> newJoinKey = new List<NodeAttribute>();

            // Key attributes
            for (int i = 0; i < node.Children.Count; i++)
            {
                newJoinKey.Add(((OperatorNode) node.Children[i]).RecordDescriptor[joinKey[i]].NodeAttribute);
                foreach (Field field in ((OperatorNode) node.Children[i]).RecordDescriptor)
                    if (newJoinKey[i] == field.NodeAttribute && !node.RecordDescriptor.Contains(field.Name))
                        node.RecordDescriptor.AddFieldForName(field.CloneWithoutPosition());
            }

            // Non-key attributes.
            for (int i = 0; i < node.Children.Count; i++)
            {
                foreach (Field field in ((OperatorNode) node.Children[i]).RecordDescriptor)
                    if (newJoinKey[i] != field.NodeAttribute)
                        node.RecordDescriptor.AddFieldForName(new Field(field.NodeAttribute,
                                                                        string.Format("Input{0}.{1}", i, field.Name),
                                                                        field.Type));
            }
            node["JoinKey"] = newJoinKey;
        }

        #endregion

        public override IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                            IEnumerable<OperatorNode> nodes)
        {
            List<IProducerRule> rules = (List<IProducerRule>) base.GetRules(queryOptimization, nodes);
            foreach (JoinRule rule in rules)
            {
                setRuleProperties(queryOptimization, rule);
            }

            // Initialize the OtherJoinRules property.
            for (int i = 0; i < rules.Count; i++)
            {
                NestableHashSet<string> other = new NestableHashSet<string>();
                for (int j = 0; j < rules.Count; j++)
                {
                    if (i == j)
                        continue;

                    other.UnionWith(rules[j].Produced.Properties);
                    foreach (BitSetProxy bitSetProxy in ((JoinRule) rules[j]).Required)
                        other.UnionWith(bitSetProxy.Properties);
                }
                ((JoinRule) rules[i]).OtherJoinRules = queryOptimization.BitSetManager.GetBitSet(PropertyUsage.None,
                                                                                                 other);
            }
            return rules;
        }

        private static void setRuleProperties(QueryOptimization queryOptimization, JoinRule rule)
        {
            rule.MergeJoin.Nodes = rule.Nodes;
            rule.HybridHashJoin.Nodes = rule.Nodes;

            if (rule.Nodes.Single().Properties.ContainsKey("Selectivity"))
                rule.Selectivity = (double) rule.Nodes.Single()["Selectivity"];
            else
                rule.Selectivity = -1d;

            rule.JoinKey = (IList<NodeAttribute>) rule.Nodes.Single()["JoinKey"];

            rule.Required = new List<BitSetProxy>
                                {
                                    queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Required,
                                                                              rule.Nodes.Single().
                                                                                  GetDepencyProperties(0)),
                                    queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Required,
                                                                              rule.Nodes.Single().
                                                                                  GetDepencyProperties(1))
                                };

            rule.OrderLeft = queryOptimization.OrderManager.GetOrder(new OrderDescription(
                                                                         new OrderDescriptionItem(
                                                                             rule.JoinKey[0].PropertyName, true)),
                                                                     true);
            rule.OrderRight = queryOptimization.OrderManager.GetOrder(new OrderDescription(
                                                                          new OrderDescriptionItem(
                                                                              rule.JoinKey[1].PropertyName, true)),
                                                                      true);
            rule.SortLeft = new SortRule(queryOptimization, rule.OrderLeft);
            rule.SortRight = new SortRule(queryOptimization, rule.OrderRight);

            rule.HybridHashJoin.InducedDependencies = rule.MergeJoin.InducedDependencies
                                                      = new HashSet<Dependency>
                                                            {
                                                                queryOptimization.OrderManager.GetDependency(
                                                                    new HashSet<string>
                                                                        {rule.JoinKey[0].PropertyName},
                                                                    new HashSet<string>
                                                                        {rule.JoinKey[1].PropertyName},
                                                                    true)
                                                            };

            rule.JoinEnumeration = queryOptimization.JoinEnumeration;
        }

        public IEnumerable<HashSet<NodeAttribute>> GetEquivalences(OperatorNode node)
        {
            // Not done.
            yield break;
        }
    }
}