﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;
using Microsoft.Ceres.SearchCore.Search.Operators;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Base rule representing MARS' Lookup operator.
    /// </summary>
    public class LookupRule : AbstractBaseRule
    {
        public LookupRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
        }

        /// <summary>Statistics for this index lookup.</summary>
        public BasicPlanSetState Stats { get; set; }
        /// <summary>Name of the index to look up into.</summary>
        public string Index { get; set; }
        /// <summary>Word to look up.</summary>
        public string Word { get; set; }
        /// <summary>The order produced by this lookup (if it us clustered).</summary>
        public OrderProxy Order { get; set; }

        /// <summary>
        /// Initialize a plan set with properties from the rule, like tuple size, cardinality etc.
        /// </summary>
        /// <param name="plans">The plan set to initialize.</param>
        public override void Initialize(PlanSet plans)
        {
            plans.State = Stats;
        }

        /// <summary>
        /// Updates the given plan with information from this rule, like
        /// costs, order and so on.
        /// </summary>
        /// <param name="plan">The plan (constructed by this rule).</param>
        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);
            
            // Initialize to the ordering provided by the index.
            plan.OrderingState = new OrderingState(queryOptimization.OrderManager, Order.Order);

            // Set sharing for the plan if this rule is the equivalence class representative.
            plan.Sharing = queryOptimization.BitSetManager.Empty;
            if (queryOptimization.ShareEquivalenceClasses.ContainsKey(this))
            {
                plan.Sharing.Or(
                    queryOptimization.BitSetManager.GetWithValues(new NodeAttribute(Nodes.Single(), null).PropertyName));
                plan.Shared = true;
            }
        }

        /// <summary>Builds the physical algebra node for the given plan.</summary>
        /// <param name="plan"></param>
        /// <returns>The physical algebra node</returns>
        public override OperatorNode BuildAlgebra(Plan plan)
        {
            OperatorNode node;
            if (GetOrCreatePhysicalNode(plan, typeof (LookupOperator), out node))
            {
                node.Properties = Nodes.Single().Properties;
                node.Rules = Nodes.Single().Rules;
            }
            return node;
        }

        /// <summary>
        /// Determines if this and the other rule is structurally identical, i.e. if they have
        /// the same configured properties, the same join keys etc.
        /// </summary>
        /// <param name="other">Rule to compare with.</param>
        /// <returns>If the rules are structurally identical.</returns>
        public override bool StructurallyIdentical(IProducerRule other)
        {
            // Lookups are structurally identical if they look up the same word in the same index.
            LookupRule otherLookup = other as LookupRule;
            return otherLookup != null && Index.Equals(otherLookup.Index) && Word.Equals(otherLookup.Word);
        }

        /// <summary>
        /// This method will be called if this rule is determined to be share equivalent with another rule.
        /// It should map any attributes the other rule produces to be share equivalent to the corresponding attributes
        /// produced by this rule.
        /// </summary>
        /// <param name="from">The rule to map attributes from.</param>
        /// <param name="equivalenceClassAttributeMappings">The dictionary the mappings should be added to.</param>
        public override void MapShareEquivalentAttributes(IProducerRule from,
                                                          Dictionary<NodeAttribute, NodeAttribute>
                                                              equivalenceClassAttributeMappings)
        {
            // Map all output attributes.
            foreach (Field field in Nodes.Single().RecordDescriptor)
                equivalenceClassAttributeMappings[from.Nodes.Single().RecordDescriptor[field.Name].NodeAttribute]
                                                      = field.NodeAttribute;
        }

        public override string ToString()
        {
            return string.Format("LookupRule(Index:{0}, Word:{1}, Produces: {2}", Index, Word, Produced.BitSet);
        }

        /// <summary>
        /// The cost for only the operator represented by this rule, applied
        /// on the top of the given plan.
        /// </summary>
        /// <param name="plan">Plan with this rule on the top.</param>
        /// <returns>The cost of this exact rule.</returns>
        protected override BasicCost Cost(Plan plan)
        {
            double cost = BasicCost.ReadSequential(BasicCost.PageAccesses(Stats.ResultSetSize));
            return new BasicCost(cost, cost);
        }
    }

    /// <summary>Rule binder for Lookup operators.</summary>
    [RuleBinder]
    public class LookupRuleBinder : AbstractRuleBinder<LookupRule>, IProducesMapper, IFieldMapper
    {
        /// <summary>The pattern in the input query this rule binder will instantiate rules for.</summary>
        public override AbstractNodeMatcher Pattern
        {
            get { return Match.NodeOfType("LookupOperator"); }
        }

        public IEnumerable<Type> MapperFor
        {
            get { return new List<Type> { typeof(LookupOperator) }; }
        }

        #region IFieldMapper Members

        // Returns all the output fields for the given lookup node.
        private IEnumerable<Field> FieldsFor(QueryOptimization queryOptimization, OperatorNode node)
        {
            return queryOptimization.SystemMetadata.GetIndexSchema((string) node["IndexName"], (string) node["Word"]);
        }

        /// <summary>
        /// This method is called on the corresponding mapper for all nodes in the query in
        /// topological order and should set field names for all NodeAttributes in the nodes
        /// output using OperatorNode's RecordDescriptor property. It should also convert
        /// all NodeAttribute references to field names.
        /// </summary>
        /// <param name="queryOptimization">The current instance of QueryOptimization.</param>
        /// <param name="node">The node to set field names for.</param>
        public void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            //Add all output attributes.
            foreach (Field field in FieldsFor(queryOptimization, node))
                node.RecordDescriptor.AddFieldForAttr(
                    new Field(new NodeAttribute(node, field.NodeAttribute.AttributeName), field.Name, field.Type));
        }

        #endregion

        #region IProducesMapper Members

        /// <summary>
        /// This method is called on the corresponding mapper for all nodes in the query in
        /// topological order and should set the produced attributes
        /// using OperatorNode's RecordDescriptor property. It should also convert
        /// all field name references to NodeAttribute.
        /// </summary>
        /// <param name="queryOptimization">The current instance of QueryOptimization.</param>
        /// <param name="node">The node to set produced attributes for.</param>
        public virtual void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            //Add all output attributes.
            foreach (Field field in FieldsFor(queryOptimization, node))
                node.RecordDescriptor.AddFieldForName(
                    new Field(new NodeAttribute(node, field.NodeAttribute.AttributeName), field.Name, field.Type));
        }

        #endregion

        /// <summary>
        /// This method is called by the optimizer to get the rules this rule binder produces.
        /// </summary>
        /// <param name="queryOptimization">The currrent QueryOptimization instance.</param>
        /// <param name="matches">IEnumerable of matches to the pattern declared by the rule binder.</param>
        /// <returns>IEnumerable of rules that should be added to the search by the optimizer.</returns>
        public override IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                            IEnumerable<OperatorNode> nodes)
        {
            List<IProducerRule> rules = new List<IProducerRule>();

            // For each node, instantiate a rule and set properties.
            foreach (OperatorNode node in nodes)
            {
                // Instantiate rule
                LookupRule rule = CreateRule(queryOptimization, node);
                rules.Add(rule);

                // Index and lookup term.
                rule.Index = (string) rule.Nodes.Single()["IndexName"];
                rule.Word = (string) rule.Nodes.Single()["Word"];

                // Produced attributes.
                NestableHashSet<string> produced = new NestableHashSet<string>
                                                       {new NodeAttribute(rule.Nodes.Single(), null).PropertyName};
                foreach (Field field in node.RecordDescriptor)
                    produced.Add(field.NodeAttribute.PropertyName);
                rule.Produced = queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Produced, produced);

                // Orderings.
                OrderDescription orderDesc = new OrderDescription(
                    queryOptimization.SystemMetadata.GetIndexOrder(rule.Index)
                        .Select(
                        item =>
                        new OrderDescriptionItem(new NodeAttribute(node, item.Property).PropertyName, item.Ascending)));
                rule.Order = queryOptimization.OrderManager.GetOrder(orderDesc, true);
                rule.Stats = queryOptimization.SystemMetadata.GetIndexStats(rule.Index, rule.Word);
            }
            return rules;
        }
    }
}