﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Infrastructure;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;
using Regex = System.Text.RegularExpressions;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// This rule is reponsible for constructing map operators in query plans.
    /// </summary>
    public class MapRule : UnaryRule
    {
        private const double FIELD_SIZE = 20;

        public MapRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
        }

        /// <summary>Cost of evaluating the predicates for each record.</summary>
        public double PredicateCost { get; set; }

        public Dictionary<IdentifierProperty, ExpressionProperty> ParameterMap { get; set; }
        public HashSet<Dependency> InducedDependencies { get; set; }

        /// <summary>
        /// Updates the properties of the plan (costs, ordering, sharing).
        /// </summary>
        /// <param name="plan">The plan to update.</param>
        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);
            plan.OrderingState = plan.OnlyChild.OrderingState.Apply(InducedDependencies);
        }

        /// <summary>
        /// Builds the Node in the physical algebra graph for this rule and calls
        /// its children recursively.
        /// </summary>
        /// <param name="plan">Plan to build algebra for.</param>
        public override OperatorNode BuildAlgebra(Plan plan)
        {
            OperatorNode node;
            if (GetOrCreatePhysicalNode(plan, typeof (MapOperator), out node))
            {
                node["ParameterMap"] = ParameterMap;
                node["RetainInput"] = true;
                node.Rules = Nodes.First().Rules;
            }
            return node;
        }

        public override bool StructurallyIdentical(IProducerRule other)
        {
            MapRule otherMap = other as MapRule;
            return otherMap != null && Utils.DictionaryEquals(ParameterMap, otherMap.ParameterMap);
        }

        public override void MapShareEquivalentAttributes(IProducerRule from,
                                                          Dictionary<NodeAttribute, NodeAttribute>
                                                              equivalenceClassAttributeMappings)
        {
            foreach (KeyValuePair<IdentifierProperty, ExpressionProperty> parameter in ParameterMap)
                equivalenceClassAttributeMappings.Add(new NodeAttribute(from.Nodes.Single(), parameter.Key.Expression),
                                                      new NodeAttribute(Nodes.Single(), parameter.Key.Expression));
        }

        public override string ToString()
        {
            return string.Format("Map(Produced: {0}, Required: {1}", Produced.BitSet, OnlyRequired);
        }

        protected override BasicPlanSetState CalcPlanSetState(BasicPlanSetState input)
        {
            return input.Expand(ParameterMap.Count*FIELD_SIZE);
        }

        protected override BasicCost Cost(Plan plan)
        {
            BasicPlanSetState childState = (BasicPlanSetState) plan.OnlyChild.PlanSet.State;
            double cost = BasicCost.CpuCosts(childState.Cardinality, PredicateCost);
            return new BasicCost(cost, cost);
        }
    }

    /// <summary>
    /// Rule binder for MapRule.
    /// </summary>
    [RuleBinder]
    public class MapRuleBinder : AbstractRuleBinder<MapRule>, IDependencyMapper, IProducesMapper, IBehaviorMapper,
                                 IFieldMapper
    {
        private const double FUNCTION_COST = 30;

        /// <summary>Node pattern to match in the operator graph during initialization.</summary>
        public override AbstractNodeMatcher Pattern
        {
            get { return Match.NodeOfType("MapOperator"); }
        }

        #region IBehaviorMapper Members

        public void SetBehaviour(OperatorNode node)
        {
            node.Behaviour = OperatorBehaviour.OrderPreserving | OperatorBehaviour.SetPreserving;
        }

        #endregion

        #region IDependencyMapper Members

        public IEnumerable<Type> MapperFor
        {
            get { yield return typeof (MapOperator); }
        }

        public static Regex.Regex RegexForField(string propertyName)
        {
            return new Regex.Regex("\\b" + Regex.Regex.Escape(propertyName) + "\\b");
        }

        public void SetDependencies(OperatorNode mapNode)
        {
            Debug.Assert(mapNode.OperatorType == typeof (MapOperator));

            /* For every input attribute, if it's part of an expression, we depend on it.
               This should really have used a parser for the map expressions, but it's not easily
               accessible. */
            mapNode.Dependencies.Add(new HashSet<NodeAttribute>());
            foreach (Field field in mapNode.OnlyInput.RecordDescriptor)
            {
                // Wrap in word boundaries (\b) so "Score" doesn't match "ProxScore".
                Regex.Regex r = RegexForField(field.NodeAttribute.PropertyName);
                foreach (
                    ExpressionProperty e in
                        ((Dictionary<IdentifierProperty, ExpressionProperty>) mapNode.Properties["ParameterMap"]).Values
                    )
                {
                    if (r.IsMatch(e.Expression))
                    {
                        mapNode.Dependencies[0].Add(field.NodeAttribute);
                        break;
                    }
                }
            }
        }

        #endregion

        #region IFieldMapper Members

        public void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            Dictionary<IdentifierProperty, ExpressionProperty> expressions =
                new Dictionary<IdentifierProperty, ExpressionProperty>();
            foreach (
                KeyValuePair<IdentifierProperty, ExpressionProperty> pair in
                    ((Dictionary<IdentifierProperty, ExpressionProperty>) node["ParameterMap"]))
            {
                string expression = pair.Value.Expression;
                NodeAttribute nodeAttr = new NodeAttribute(node, pair.Key.Expression);
                node.RecordDescriptor.AddFieldForAttr(new Field(nodeAttr, nodeAttr.PropertyName, null));
                foreach (Field field in node.OnlyInput.RecordDescriptor.GetFieldsWithMappings(queryOptimization))
                {
                    expression = RegexForField(field.NodeAttribute.PropertyName).Replace(expression, field.Name);
                }
                expressions.Add(new IdentifierProperty(nodeAttr.PropertyName), new ExpressionProperty(expression));
            }
            node["ParameterMap"] = expressions;
            node.PassthroughFieldNames();
        }

        #endregion

        #region IProducesMapper Members

        public void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            if ((bool) node["RetainInput"])
                node.PassthroughInputAttributeOrigins();

            Dictionary<IdentifierProperty, ExpressionProperty> expressions =
                new Dictionary<IdentifierProperty, ExpressionProperty>();
            foreach (
                KeyValuePair<IdentifierProperty, ExpressionProperty> pair in
                    (Dictionary<IdentifierProperty, ExpressionProperty>) node["ParameterMap"])
            {
                string expression = pair.Value.Expression;
                if (node.OnlyInput.RecordDescriptor.Contains(expression))
                {
                    // We do not produce it.
                    Field input = node.OnlyInput.RecordDescriptor[expression];
                    node.RecordDescriptor.AddFieldForName(new Field(input.NodeAttribute, pair.Key.Expression, input.Type));
                }
                else
                {
                    /* If it's simply not a pass-through or a rename of an input attribute, it's produced
                    by the map operator, so replace field refs with NodeAttribute. */
                    foreach (Field field in node.OnlyInput.RecordDescriptor)
                    {
                        expression = RegexForField(field.Name).Replace(expression, field.NodeAttribute.PropertyName);
                    }
                    expressions.Add(new IdentifierProperty(pair.Key.Expression), new ExpressionProperty(expression));
                    node.RecordDescriptor.AddFieldForName(new Field(new NodeAttribute(node, pair.Key.Expression),
                                                                    node.OutputTypeDescriptor.RecordType[
                                                                        pair.Key.Expression].Type.FieldType));
                }
            }

            // Consider whether to remove the Map-operator:
            if (expressions.Count > 0)
            {
                // Complex expressions exists, keep the map rule.
                node["ParameterMap"] = expressions;
            }
            else
            {
                // There are no complex expressions. Replace AttributeOrigin for the child node with
                // what would be the result of the map and extract (remove) the map.
                node.OnlyInput.RecordDescriptor.Clear();
                foreach (Field field in node.RecordDescriptor)
                    node.OnlyInput.RecordDescriptor.AddFieldForName(field);
                node.Extract();
            }
        }

        #endregion

        public override IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                            IEnumerable<OperatorNode> nodes)
        {
            IList<IProducerRule> rules = new List<IProducerRule>();
            foreach (OperatorNode node in nodes)
            {
                Dictionary<IdentifierProperty, ExpressionProperty> parameterMap =
                    (Dictionary<IdentifierProperty, ExpressionProperty>) node["ParameterMap"];

                MapRule rule = CreateRule(queryOptimization, node);
                rules.Add(rule);

                rule.ParameterMap = parameterMap;
                rule.PredicateCost = FUNCTION_COST*parameterMap.Count;
                rule.InducedDependencies = new HashSet<Dependency>();

                setInducedDependenciesForRule(node, queryOptimization, rule, parameterMap);

                NestableHashSet<string> produced = new NestableHashSet<string>
                                                       {new NodeAttribute(node, null).PropertyName};
                foreach (KeyValuePair<IdentifierProperty, ExpressionProperty> pair in parameterMap)
                    produced.Add(new NodeAttribute(node, pair.Key.Expression).PropertyName);
                rule.Produced = queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Produced, produced);

                setDependenciesForRule(node, queryOptimization, rule);
            }
            return rules;
        }

        private static void setDependenciesForRule(OperatorNode node, QueryOptimization queryOptimization, MapRule rule)
        {
            NestableHashSet<string> required = node.GetDepencyProperties(0);
            // Add dependency on all nodes below it
            foreach (OperatorNode descendandt in node.IterateBreadthFirst(false))
                required.Add(new NodeAttribute(descendandt, null).PropertyName);
            // Add dependency on nodes above it that does not depend on the map itself (not transitively either).
            foreach (OperatorNode ascendant in node.GetAscendants())
                if (!node.Dependers.Contains(ascendant) && ascendant.OperatorType != typeof (OutputOperator))
                    required.Add(new NodeAttribute(ascendant, null).PropertyName);
            rule.Required = new List<BitSetProxy>
                                {queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Required, required)};
        }

        private static void setInducedDependenciesForRule(OperatorNode node, QueryOptimization queryOptimization, MapRule rule, Dictionary<IdentifierProperty, ExpressionProperty> parameterMap)
        {
            foreach (KeyValuePair<IdentifierProperty, ExpressionProperty> pair in parameterMap)
            {
                HashSet<string> dependencyLHS = new HashSet<string>();
                foreach (Field field in node.OnlyInput.RecordDescriptor)
                {
                    Regex.Regex r = RegexForField(field.NodeAttribute.PropertyName);
                    if (r.IsMatch(pair.Value.Expression))
                        dependencyLHS.Add(field.NodeAttribute.PropertyName);
                }
                rule.InducedDependencies.Add(queryOptimization.OrderManager.GetDependency(
                                                 dependencyLHS,
                                                 new HashSet<string>
                                                     {new NodeAttribute(node, pair.Key.Expression).PropertyName},
                                                 false));
            }
        }
    }
}