﻿#region

using System;
using System.Collections.Generic;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Infrastructure;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// This class is the produces mapper and field mapper for the output operator. It has two tasks:
    /// 1. Set the order and names of the output fields for each query in the output operator.
    /// 2. After plan generation, make sure the fields still have the same names and order. If
    ///     necessary, insert a map operator to fix wrong order og wrong naming.
    /// </summary>
    [RuleBinder]
    public class OutputOperatorMapper : IRuleBinder, IProducesMapper, IFieldMapper
    {
        #region IFieldMapper Members

        public void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            IList<IList<Pair<string, NodeAttribute>>> outputDescriptors
                = (IList<IList<Pair<string, NodeAttribute>>>) node["OutputDescriptors"];

            for (int q = 0; q < node.Children.Count; q++)
            {
                OperatorNode query = (OperatorNode) node.Children[q];
                IList<Pair<string, NodeAttribute>> outputDescriptor = outputDescriptors[q];

                // Verify output field order and names, and add a Map operator if necessary
                // TODO: Does not take column order into account yet.
                bool insertMap = false;
                IList<Pair<string, string>> fieldMapping = new List<Pair<string, string>>();

                // Find out whether we need to insert a Map-operator that realiases attribute names to the expected ones.
                for (int i = 0; i < outputDescriptor.Count; i++)
                {
                    Pair<string, NodeAttribute> requestedOutputField = outputDescriptor[i];
                    Field producedOutputField = query.RecordDescriptor[queryOptimization, requestedOutputField.Second];
                    fieldMapping.Add(new Pair<string, string>(requestedOutputField.First, producedOutputField.Name));

                    // When the expected and actual names differ, we need to remap.
                    if (requestedOutputField.First != producedOutputField.Name || i != producedOutputField.Position)
                        insertMap = true;
                }
                if (!insertMap)
                    continue;

                // We need to add a map
                OperatorNode mapOperator = new OperatorNode(typeof (MapOperator));
                mapOperator["RetainInput"] = false;

                // Make a parameter map that realiases.
                IDictionary<IdentifierProperty, ExpressionProperty> parameterMap
                    = new Dictionary<IdentifierProperty, ExpressionProperty>(fieldMapping.Count);
                foreach (Pair<string, string> field in fieldMapping)
                    parameterMap.Add(new IdentifierProperty(field.First), new ExpressionProperty(field.Second));
                mapOperator["ParameterMap"] = parameterMap;

                // Insert the Map between the Output-operator and the input that needs realiasing.
                mapOperator.AddChild(query);
                node.ReplaceChild(query, mapOperator);
            }
        }

        #endregion

        #region IProducesMapper Members

        public IEnumerable<Type> MapperFor
        {
            get { yield return typeof (OutputOperator); }
        }

        public void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            IList<IList<Pair<string, NodeAttribute>>> outputDescriptor = new List<IList<Pair<string, NodeAttribute>>>();

            foreach (OperatorNode child in node.Children)
            {
                IList<Pair<string, NodeAttribute>> fields = new List<Pair<string, NodeAttribute>>();
                foreach (Field field in child.RecordDescriptor)
                    fields.Add(new Pair<string, NodeAttribute>(field.Name, field.NodeAttribute));
                outputDescriptor.Add(fields);
            }
            node["OutputDescriptors"] = outputDescriptor;
        }

        #endregion

        #region IRuleBinder Members

        public AbstractNodeMatcher Pattern
        {
            get { return new NothingMatcher(); }
        }

        public IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                   IEnumerable<OperatorNode> matches)
        {
            // Should never be invoked!
            throw new NotImplementedException();
        }

        #endregion
    }
}