﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;
using Microsoft.Ceres.SearchCore.Search.Operators;
using Microsoft.Ceres.SearchCore.Search.Scorers;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Represents a score occurrences operator in the query.
    /// </summary>
    public class ScoreOccurrencesRule : AbstractSearchRule
    {
        private const double INSTRUCTION_COUNT = 30;

        public ScoreOccurrencesRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
        }

        public IDictionary<string, Scorer> Scorers { get; set; }
        public ScoreCombiner Combiner { get; set; }
        public int GroupPrefix { get; set; }
        public int RequiredProp { get; set; }
        public OrderProxy Order { get; set; }

        public override void Search(PlanSet planSet, ICost limit)
        {
            if (planSet.Properties != Filter)
                return;

            //TODO: THis is a hack, Permute instead
            BasicPlanSetState state = new BasicPlanSetState();
            Plan plan = new Plan();
            foreach (BitSetProxy requiredProxy in Required)
            {
                PlanSet plans = queryOptimization.GeneratePlans(requiredProxy.BitSet, limit);
                plan.Children.Add(plans[0]);
                BasicPlanSetState inputState = (BasicPlanSetState) plans.State;

                if (inputState.Cardinality > state.Cardinality)
                    state.Cardinality = inputState.Cardinality;
                if (inputState.TupleSize > state.TupleSize)
                    state.TupleSize = inputState.TupleSize;
            }

            if (planSet.Count == 0)
                planSet.State = state;

            UpdatePlan(plan);
            planSet.AddPlan(plan);
        }

        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);
            plan.OrderingState = new OrderingState(queryOptimization.OrderManager, Order);

            updateSharing(plan);
        }

        private void updateSharing(Plan plan)
        {
            bool allShared = true;
            plan.Sharing = queryOptimization.BitSetManager.Empty;
            foreach (Plan subPlan in plan.Children)
            {
                plan.Sharing.Or(subPlan.Sharing);
                allShared &= subPlan.Shared;
            }

            if (! (queryOptimization.ShareEquivalenceClasses.ContainsKey(this) && allShared))
                return;

            plan.Sharing.Or(
                queryOptimization.BitSetManager.GetWithValues(new NodeAttribute(Nodes.Single(), null).PropertyName));
            plan.Shared = true;
        }

        public override OperatorNode BuildAlgebra(Plan plan)
        {
            OperatorNode node;
            if (GetOrCreatePhysicalNode(plan, typeof (ScoreOccurrencesOperator), out node))
            {
                node["Scorers"] = Scorers;
                node["Combiner"] = Combiner;
                node["GroupPrefix"] = GroupPrefix;
                node["Required"] = RequiredProp;
                node.Rules = Nodes.First().Rules;
            }
            return node;
        }

        protected override BasicCost Cost(Plan plan)
        {
            double cardinalities = 0;
            foreach (Plan subplan in plan.Children)
                cardinalities += ((BasicPlanSetState) subplan.PlanSet.State).Cardinality;

            double cost = BasicCost.CpuCosts(cardinalities, INSTRUCTION_COUNT);
            return new BasicCost(cost, cost);
        }

        public override bool StructurallyIdentical(IProducerRule other)
        {
            return other is ScoreOccurrencesRule; //TODO: Obviously not done.
        }

        public override void MapShareEquivalentAttributes(IProducerRule from, Dictionary<NodeAttribute, NodeAttribute> equivalenceClassAttributeMappings)
        {
            foreach (KeyValuePair<string, Scorer> scorer in Scorers)
                equivalenceClassAttributeMappings.Add(new NodeAttribute(from.Nodes.Single(), scorer.Key),
                                                      new NodeAttribute(Nodes.Single(), scorer.Key));
        }

        public override string ToString()
        {
            return string.Format("ScoreOccurrence(Produced: {0}, Required: {1}", Produced.BitSet,
                                 Utils.JoinEnumerable(Required, p => p.BitSet.ToString(), "", ",", true));
        }
    }

    [RuleBinder]
    public class ScoreOccurrencesRuleBinder : AbstractRuleBinder<ScoreOccurrencesRule>, IProducesMapper, IBehaviorMapper,
                                              IDependencyMapper, IFieldMapper
    {
        public override AbstractNodeMatcher Pattern
        {
            get { return Match.NodeOfType("ScoreOccurrencesOperator"); }
        }

        #region IBehaviorMapper Members

        public void SetBehaviour(OperatorNode node)
        {
            // We add attributes for score and group by group prefix, so order is the only thing kept.
            // TODO: However, instead of being ordered by (Document ID, Pos), the order will be (Document ID).
            node.Behaviour = OperatorBehaviour.OrderPreserving;
        }

        #endregion

        #region IDependencyMapper Members

        public void SetDependencies(OperatorNode node)
        {
            Debug.Assert(node.OperatorType == typeof (ScoreOccurrencesOperator));

            // ScoreOccurrences always depend directly on its input.
            foreach (OperatorNode input in node.Inputs)
                node.Dependencies.Add(new HashSet<NodeAttribute> {new NodeAttribute(input, null)});
            // Everything above should depend on this
            //TODO
        }

        #endregion

        #region IFieldMapper Members

        public void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            node.PassthroughFieldNames(1);
            foreach (string scorerName in ((IDictionary<string, Scorer>) node.Properties["Scorers"]).Keys)
                node.RecordDescriptor.AddFieldForAttr(new Field(new NodeAttribute(node, scorerName), null));
        }

        #endregion

        #region IProducesMapper Members

        public IEnumerable<Type> MapperFor
        {
            get { yield return typeof (ScoreOccurrencesOperator); }
        }

        public virtual void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            IList<NodeAttribute> groupingFields = new List<NodeAttribute>();
            for (int i = 0; i < (int) node["GroupPrefix"]; i++)
            {
                Field source = ((OperatorNode) node.FirstChild).RecordDescriptor[i];
                groupingFields.Add(source.NodeAttribute);
                node.RecordDescriptor.AddFieldForName(source.CloneWithoutPosition());
            }
            node["_GroupingFields"] = groupingFields;

            foreach (string scorerName in ((IDictionary<string, Scorer>) node["Scorers"]).Keys)
                node.RecordDescriptor.AddFieldForName(new Field(new NodeAttribute(node, scorerName),
                                                                node.OutputTypeDescriptor.RecordType[scorerName].Type.
                                                                    FieldType));
        }

        #endregion

        public override IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                            IEnumerable<OperatorNode> nodes)
        {
            IList<IProducerRule> rules = new List<IProducerRule>();
            foreach (OperatorNode node in nodes)
            {
                ScoreOccurrencesRule rule = CreateRule(queryOptimization, node);
                rules.Add(rule);

                // Set the trivial properties.
                rule.Scorers = (Dictionary<string, Scorer>) rule.Nodes.Single()["Scorers"];
                rule.Combiner = (ScoreCombiner) rule.Nodes.Single()["Combiner"];
                rule.GroupPrefix = (int) rule.Nodes.Single()["GroupPrefix"];
                rule.RequiredProp = (int) rule.Nodes.Single()["Required"];

                // Set produced bits.
                NestableHashSet<string> produced = new NestableHashSet<string>();
                produced.Add(new NodeAttribute(node, null).PropertyName);
                foreach (string scorerName in ((IDictionary<string, Scorer>) node["Scorers"]).Keys)
                    produced.Add(new NodeAttribute(node, scorerName).PropertyName);
                rule.Produced = queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Produced, produced);

                // Set required bits.
                rule.Required = new List<BitSetProxy>();
                foreach (HashSet<NodeAttribute> dependencies in rule.Nodes.Single().Dependencies)
                    rule.Required.Add(queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Required,
                                                                                new NestableHashSet<string>(
                                                                                    dependencies.Select(
                                                                                        na => na.PropertyName))));

                // Set ordering state.
                rule.Order = queryOptimization.OrderManager.GetOrder(
                    new OrderDescription(((IList<NodeAttribute>) rule.Nodes.Single()["_GroupingFields"])
                                             .Select(attr => new OrderDescriptionItem(attr.PropertyName, true))), true);
            }
            return rules;
        }
    }
}