﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using Microsoft.Ceres.Evaluation.DataModel.Types;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Evaluation.Operators.Expressions;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;
using Regex = System.Text.RegularExpressions;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// This rule is reponsible for constructing selection operators in query plans.
    /// </summary>
    public class SelectionRule : UnaryRule
    {
        public SelectionRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
        }

        /// <summary>Cost of evaluating the predicates for each record.</summary>
        public double PredicateCost { get; set; }
        /// <summary>Set of FDs induced by the selection.</summary>
        public HashSet<Dependency> InducedDependencies { get; set; }
        /// <summary>The selection predicate.</summary>
        public ExpressionProperty SelectionFilter { get; set; }

        /// <summary>
        /// Updates the given plan with ordering, leaves the rest to UnaryRule.
        /// </summary>
        /// <param name="plan">The plan (constructed by this rule).</param>
        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);
            plan.OrderingState = plan.OnlyChild.OrderingState.Apply(InducedDependencies);
        }

        /// <summary>Builds the physical algebra node for the given plan recursively.</summary>
        /// <param name="plan">Plan to build node for</param>
        /// <returns>The physical algebra node</returns>
        public override OperatorNode BuildAlgebra(Plan plan)
        {
            OperatorNode node;
            if (GetOrCreatePhysicalNode(plan, typeof (SelectOperator), out node))
            {
                node.Properties = Nodes.First().Properties;
                node.Rules = Nodes.First().Rules;
            }
            return node;
        }

        /// <summary>
        /// The cost for the selection, which is predicate cost times cardinality.
        /// </summary>
        /// <param name="plan">Plan with this rule on the top.</param>
        /// <returns>The cost of this exact rule.</returns>
        protected override BasicCost Cost(Plan plan)
        {
            BasicPlanSetState childState = (BasicPlanSetState) plan.OnlyChild.PlanSet.State;
            double cost = BasicCost.CpuCosts(childState.Cardinality, PredicateCost);
            return new BasicCost(cost, cost);
        }

        /// <summary>
        /// Determines if this and the other rule is structurally identical,
        /// Selections are structurally identical if the filter predicates are equal.
        /// </summary>
        /// <param name="other">Rule to compare with.</param>
        /// <returns>If the rules are structurally identical.</returns>
        public override bool StructurallyIdentical(IProducerRule other)
        {
            SelectionRule otherSelection = other as SelectionRule;
            return otherSelection != null && SelectionFilter.Equals(otherSelection.SelectionFilter);
        }
    }

    /// <summary>
    /// Rule binder for SelectionRule.
    /// </summary>
    [RuleBinder]
    public class SelectionRuleBinder : AbstractFilterRuleBinder<SelectionRule>, IDependencyMapper, IProducesMapper
    {
        /// <summary>Node pattern to match in the operator graph during initialization, matching SelectOperator.</summary>
        public override AbstractNodeMatcher Pattern
        {
            get { return Match.NodeOfType("SelectOperator"); }
        }

        #region IDependencyMapper Members

        /// <summary>The types of operator the implementes is a mapper for: SelectOperator.</summary>
        public override IEnumerable<Type> MapperFor
        {
            get { return new List<Type> {typeof (SelectOperator)}; }
        }

        /// <summary>
        /// This method is called on the corresponding mapper for all nodes in the query in
        /// topological order and should set the node's dependencies.
        /// Selection depends on all attributes accessed by the predicate.
        /// </summary>
        /// <param name="node">The node to set dependencies for.</param>
        public void SetDependencies(OperatorNode selectNode)
        {
            // To properly set the dependencies, we'll invoke MARS' parser of the expression.

            // Prepare some stuff needed for parsing.
            IRecordTypeDescriptor inputFieldTypes = selectNode.RecordDescriptor.AsRecordTypeDescriptor();
            string expression = ((ExpressionProperty) selectNode["Filter"]).Expression;
            Dictionary<string, object> symbolsMap = MakeSymbolsMap(expression, inputFieldTypes);

            // Parse the expression, which gives us an easily traversable AST.
            Expression linqExpression = ParseExpression(expression, symbolsMap);

            // Visit the expression. The visitor keeps references to depended-on attributes.
            FilterVisitor visitor = new FilterVisitor();
            visitor.Initialize();
            visitor.Visit(linqExpression);

            // Get the dependencies from the visitor.
            selectNode.Dependencies.Add(new HashSet<NodeAttribute>());
            foreach (int visitedIndex in visitor.ReferencedIndexes)
                selectNode.Dependencies[0].Add(selectNode.OnlyInput.RecordDescriptor[visitedIndex].NodeAttribute);
        }

        #endregion

        #region IProducesMapper Members

        /// <summary>
        /// This method is called on the corresponding mapper for all nodes in the query in
        /// topological order and should set the produced attributes.
        /// Pass through the input attributes and replace references in the predicate with NodeAttributes.
        /// </summary>
        /// <param name="queryOptimization">The current instance of QueryOptimization.</param>
        /// <param name="node">The node to set produced attributes for.</param>
        public override void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            base.SetProduces(queryOptimization, node);

            string expression = ((ExpressionProperty) node["Filter"]).Expression;
            foreach (Field field in node.OnlyInput.RecordDescriptor)
            {
                System.Text.RegularExpressions.Regex r =
                    new System.Text.RegularExpressions.Regex("\\b" +
                                                             System.Text.RegularExpressions.Regex.Escape(
                                                                 field.NodeAttribute.AttributeName) + "\\b");
                expression = r.Replace(expression, field.NodeAttribute.PropertyName);
            }
            node["Filter"] = new ExpressionProperty(expression);
        }

        #endregion

        /// <summary>
        /// This method is called by the optimizer to get the rules this rule binder produces.
        /// </summary>
        /// <param name="queryOptimization">The currrent QueryOptimization instance.</param>
        /// <param name="matches">IEnumerable of matches to the pattern declared by the rule binder.</param>
        /// <returns>IEnumerable of rules that should be added to the search by the optimizer.</returns>
        public override IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                            IEnumerable<OperatorNode> nodes)
        {
            IEnumerable<IProducerRule> rules = base.GetRules(queryOptimization, nodes);
            foreach (SelectionRule rule in rules)
            {
                setSelectivity(rule);

                // This is to be able to set selectivity and predicate cost in tests.
                if (rule.Nodes.Single()["Selectivity"] != null)
                    rule.Selectivity = (double)rule.Nodes.Single()["Selectivity"];
                if (rule.Nodes.Single()["PredicateCost"] != null)
                    rule.PredicateCost = (double)rule.Nodes.Single()["PredicateCost"];

                setRuleDependencies(queryOptimization, rule);
            }
            return rules;
        }

        // Sets the selectivity for the rule.
        private static void setSelectivity(SelectionRule rule)
        {
            // TODO: We don't have anything to guesstimate selectivities from yet.
            rule.Selectivity = 0.1;
        }

        // Sets dependecies for the selection rule by parsing the predicate expression.
        private static void setRuleDependencies(QueryOptimization queryOptimization, SelectionRule rule)
        {
            // Get the LINQ Expression for this selection
            RecordDescriptor inputDescriptor = rule.Nodes.Single().RecordDescriptor;
            rule.SelectionFilter = (ExpressionProperty) rule.Nodes.Single()["Filter"];
            string expression = rule.SelectionFilter.Expression;
            Dictionary<string, object> symbolsMap = MakeSymbolsMap(expression,
                                                                   inputDescriptor.AsRecordTypeDescriptor());
            Expression linqExpression = ParseExpression(expression, symbolsMap);

            // Visit the expression
            FilterVisitor visitor = new FilterVisitor();
            visitor.Initialize();
            visitor.Visit(linqExpression);

            // Add required properties found by the visitor
            NestableHashSet<string> required = new NestableHashSet<string>();
            foreach (int index in visitor.ReferencedIndexes)
                required.Add(rule.Nodes.Single().RecordDescriptor[index].NodeAttribute.PropertyName);

            rule.Required = new List<BitSetProxy>
                                {queryOptimization.BitSetManager.GetBitSet(PropertyUsage.Required, required)};

            // Determine dependencies
            rule.InducedDependencies = new HashSet<Dependency>();
            foreach (Pair<object, object> equivalence in visitor.Equivalences)
            {
                Dependency dependency = null;

                // Both sides of the equivalence are columns in the data stream, add COL=COL dependency
                if (equivalence.First is FieldIndex && equivalence.Second is FieldIndex)
                    dependency = queryOptimization.OrderManager.GetDependency(
                        inputDescriptor[((FieldIndex) equivalence.First).Index].NodeAttribute.PropertyName,
                        inputDescriptor[((FieldIndex) equivalence.Second).Index].NodeAttribute.PropertyName, true);

                // The left operand is a constant, right one is a column. Add -->COL dependency
                if (equivalence.First is Constant && equivalence.Second is FieldIndex)
                    dependency = queryOptimization.OrderManager.GetDependency(
                        null,
                        inputDescriptor[((FieldIndex) equivalence.Second).Index].NodeAttribute.PropertyName, false);

                // The right operand is a constant, left one is a column. Add -->COL dependency
                if (equivalence.First is FieldIndex && equivalence.Second is Constant)
                    dependency = queryOptimization.OrderManager.GetDependency(
                        null,
                        inputDescriptor[((FieldIndex) equivalence.First).Index].NodeAttribute.PropertyName, false);

                if (dependency != null)
                    rule.InducedDependencies.Add(dependency);
            }
        }

        /// <summary>
        /// This method is called on the corresponding mapper for all nodes in the query in
        /// topological order and should set field names for all NodeAttributes in the nodes
        /// output using OperatorNode's RecordDescriptor property. It should also convert
        /// all NodeAttribute references to field names.
        /// Passes through input field names and convert NodeAttribute refs back to field names.
        /// </summary>
        /// <param name="queryOptimization">The current instance of QueryOptimization.</param>
        /// <param name="node">The node to set field names for.</param>
        public override void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            base.SetFields(queryOptimization, node);
            string expression = ((ExpressionProperty) node["Filter"]).Expression;
            foreach (Field field in node.OnlyInput.RecordDescriptor.GetFieldsWithMappings(queryOptimization))
            {
                Regex.Regex r =
                    new Regex.Regex("\\b" + Regex.Regex.Escape(field.NodeAttribute.PropertyName) + "\\b");
                expression = r.Replace(expression, field.Name);
            }
            node["Filter"] = new ExpressionProperty(expression);
        }

        #region MARS Reflection Code

        // Calls the MakeSymbolsMap function in MARS' Function Binder class using reflection,
        // as the function in question is private. It is not intended to be used externally, but we needed to anyway.
        // This should be considered a hack since we do not have access to MARS' source code.
        // It is needed to parse the selection predicate expression.
        private static Dictionary<string, object> MakeSymbolsMap(string expression, IRecordTypeDescriptor descriptor)
        {
            ParameterExpression expression2, expression3;
            return (Dictionary<string, object>)
                   typeof (FunctionBinder).InvokeMember("MakeSymbolsMap",
                                                        BindingFlags.InvokeMethod | BindingFlags.NonPublic |
                                                        BindingFlags.Static, null, null,
                                                        new object[]
                                                            {
                                                                descriptor,
                                                                (expression2 =
                                                                 Expression.Parameter(typeof (IFieldTuple),
                                                                                      "Argument/Record")),
                                                                (expression3 = Expression.Parameter(
                                                                                   Assembly.Load(
                                                                                       "Microsoft.Ceres.Evaluation.Operators")
                                                                                       .GetType(
                                                                                       "Microsoft.Ceres.Evaluation.Operators.Expressions.ExpressionContext"),
                                                                                   "Closure/Context"))
                                                            }
                       );
        }

        // Calls the ParseExpression function in MARS' DynamicExpression class using reflection,
        // as the class is private. It parses a selection predicate to a programmatic LINQ Expression
        // which we can work with.
        private static Expression ParseExpression(string expression, Dictionary<string, object> symbolsMap)
        {
            Type t =
                Assembly.Load("Microsoft.Ceres.Evaluation.Operators").GetType(
                    "Microsoft.Ceres.Evaluation.Operators.Expressions.DynamicExpression");
            return (Expression) t.InvokeMember("Parse",
                                               BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Static,
                                               null, null, new object[]
                                                               {
                                                                   typeof (bool?), expression, FunctionRegistry.Instance
                                                                   , new object[] {symbolsMap}
                                                               });
        }
        #endregion
    }

    // Represents an index to a field in a selection predicate.
    internal class FieldIndex
    {
        public int Index;
    }

    // Represents a constant in a selection predicate.
    internal class Constant
    {
        public object Value;
    }

    /// <summary>
    /// Visits all parts of a filter expression, getting the columns referenced
    /// and equivalences (... = ... expressions) applied.
    /// </summary>
    internal class FilterVisitor : ExpressionVisitor
    {
        private bool collectEquivalences;

        public FilterVisitor()
        {
            ReferencedIndexes = new HashSet<int>();
            Equivalences = new HashSet<Pair<object, object>>();
        }

        /// <summary>
        /// Set of all referenced field indexes.
        /// </summary>
        public HashSet<int> ReferencedIndexes { get; private set; }

        /// <summary>
        /// Set of equivalences at the root or under ONLY AND operators.
        /// </summary>
        public HashSet<Pair<object, object>> Equivalences { get; private set; }

        public void Initialize()
        {
            collectEquivalences = true;
        }

        protected override Expression VisitMethodCall(MethodCallExpression m)
        {
            bool oldCollectValue = collectEquivalences;

            // Check for fields.
            int index = fieldIndex(m);
            if (index >= 0)
            {
                ReferencedIndexes.Add(index);
            }
                // Check for equivalence
            else if (m.Method.Name == "EQ" && collectEquivalences)
            {
                object[] operands = new object[2];
                // Iterate over the operands.
                for (int i = 0; i < 2; i++)
                {
                    index = fieldIndex(m.Arguments[i]);
                    if (index >= 0)
                        operands[i] = new FieldIndex {Index = index};

                    object operand = constant(m.Arguments[i]);
                    if (operand != null)
                        operands[i] = new Constant {Value = operand};
                }
                if (operands[0] != null && operands[1] != null)
                    Equivalences.Add(new Pair<object, object>(operands[0], operands[1]));
            }
            else if (m.Method.Name != "And")
            {
                // We're passing an OR operator or something like that,
                // make sure we don't collect equivalences below here.
                collectEquivalences = false;
            }

            Expression returnValue = base.VisitMethodCall(m);

            // Restore the old value in the way up again.
            collectEquivalences = oldCollectValue;
            return returnValue;
        }

        /// <summary>
        /// Returns the index of the field referenced by the expression, -1 if none.
        /// </summary>
        private static int fieldIndex(Expression m)
        {
            MethodCallExpression call = m as MethodCallExpression;
            if (call == null)
                return -1;

            if (call.Method.Name == "get_Item")
            {
                ConstantExpression indexExpression = (ConstantExpression) call.Arguments[0];
                return (int) indexExpression.Value;
            }

            if (call.Object is UnaryExpression)
                return fieldIndex(((UnaryExpression) call.Object).Operand);

            return -1;
        }

        /// <summary>
        /// Returns the value of the constant referenced by the expression, null if none.
        /// </summary>
        private static object constant(Expression e)
        {
            ConstantExpression constant = e as ConstantExpression;
            if (constant == null)
                return null;

            return constant.Value;
        }
    }
}