﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Rules.Infrastructure;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// Represents a sort operator and is a helper rule which does
    /// not directly participate in the plan search.
    /// </summary>
    public class SortRule : AbstractRule, IHelperRule
    {
        private const double CPU_INSTRUCTIONS = 30;
        private const double MAX_MEMORY = 1024*1024*300;

        public SortRule(QueryOptimization queryOptimization, OrderProxy order)
            : base(queryOptimization)
        {
            Order = order;
        }

        /// <summary>Order for the sort.</summary>
        public OrderProxy Order { get; set; }

        #region IHelperRule Members

        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);
            plan.OrderingState = new OrderingState(queryOptimization.OrderManager, Order.Order)
                .Apply(plan.OnlyChild.OrderingState);
            plan.Sharing = plan.OnlyChild.Sharing;
        }

        public override OperatorNode BuildAlgebra(Plan plan)
        {
            OperatorNode node;
            if (GetOrCreatePhysicalNode(plan, typeof (SortOperator), out node))
            {
                // Make sure MARS can utilize the ordering later.
                node["Reorder"] = true;
                node["SortFields"] = OrderDescriptionToSortFields(Order.Description);
            }
            return node;
        }

        #endregion

        protected override BasicCost Cost(Plan plan)
        {
            BasicPlanSetState childState = (BasicPlanSetState) plan.OnlyChild.PlanSet.State;

            double cost = BasicCost.CpuCosts(childState.Cardinality*Math.Log(childState.Cardinality + 1, 2),
                                             CPU_INSTRUCTIONS);
            if (childState.ResultSetSize > MAX_MEMORY)
                // Too large for memory, assume it is enough to spill to disk once.
                cost += (BasicCost.ReadSequential(childState.ResultSetSize) +
                         BasicCost.ReadSequential(childState.ResultSetSize));

            return new BasicCost(cost, cost);
        }

        public override string ToString()
        {
            return "Sort";
        }

        private static IList<SortDefinition> OrderDescriptionToSortFields(OrderDescription orderDescription)
        {
            return new List<SortDefinition>(
                orderDescription.Select(item => new SortDefinition(item.Property, item.Ascending)));
        }
    }

    [RuleBinder]
    public class SortRuleBinder : IRuleBinder, IProducesMapper, IFieldMapper
    {
        #region IFieldMapper Members

        public void SetFields(QueryOptimization queryOptimization, OperatorNode node)
        {
            IList<SortDefinition> sortFields = (IList<SortDefinition>) node["SortFields"];

            foreach (Field field in node.OnlyInput.RecordDescriptor.GetFieldsWithMappings(queryOptimization))
                foreach (SortDefinition sortDef in sortFields)
                    if (field.NodeAttribute.PropertyName == sortDef.Field)
                        sortDef.Field = field.Name;

            node.PassthroughFieldNames();
        }

        #endregion

        #region IProducesMapper Members

        public IEnumerable<Type> MapperFor
        {
            get { yield return typeof (SortOperator); }
        }

        public void SetProduces(QueryOptimization queryOptimization, OperatorNode node)
        {
            node.PassthroughInputAttributeOrigins();
        }

        #endregion

        #region IRuleBinder Members

        public AbstractNodeMatcher Pattern
        {
            get { return new NothingMatcher(); }
        }

        public IEnumerable<IProducerRule> GetRules(QueryOptimization queryOptimization,
                                                   IEnumerable<OperatorNode> matches)
        {
            throw new OptimizerException("GetRules() should not be called for SortRule.");
        }

        #endregion
    }
}