﻿#region

using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Optimizer.Engine;

#endregion

namespace Microsoft.Ceres.Optimizer.Rules
{
    /// <summary>
    /// General base class for all unary rules (one input), offering
    /// basic search functionality.
    /// </summary>
    public abstract class UnaryRule : AbstractSearchRule
    {
        public UnaryRule(QueryOptimization queryOptimization)
            : base(queryOptimization)
        {
        }

        /// <summary>
        /// The rquired bitset for the only input.
        /// </summary>
        public BitSet OnlyRequired
        {
            get { return Required[0].BitSet; }
        }

        /// <summary>
        /// Guides the search for this rule instance. Offers the basic search functionality
        /// for unary rules: generating all possible plans with this rule on the top.
        /// </summary>
        /// <param name="plans">PlanSet to add plans to (also defines goal properties).</param>
        /// <param name="limit">Abort the search if passing this cost limit (pruning).</param>
        public override void Search(PlanSet plans, ICost limit)
        {
            DiagPrinter.Instance.DiagWriteLine("search", "UnaryRule producing {0} searching for {1}.", Produced.BitSet,
                                               (plans.Properties - Produced.BitSet) | Required[0].BitSet);
            // Get the possible input plans, i.e. plans with needed properties, 
            // except the ones we produce ourselves.
            PlanSet input = queryOptimization.GeneratePlans((plans.Properties - Produced.BitSet) | Required[0].BitSet,
                                                            limit);
            if (input == null)
            {
                DiagPrinter.Instance.DiagWriteLine("search", "0 plans.");
                return; // No plans
            }
            DiagPrinter.Instance.DiagWriteLine("search", "{0} plans.", input.Count);

            if (plans.Count == 0)
                // First plan, so set some state.
                plans.State = CalcPlanSetState((BasicPlanSetState) input.State);

            // Add each input plan to the PlanSet, updating their properties.
            foreach (Plan plan in input)
            {
                Plan newPlan = new Plan {Children = new List<Plan> {plan}};
                UpdatePlan(newPlan);
                plans.AddPlan(newPlan);
            }
        }

        /// <summary>
        /// Calculates the PlansSetState for the output based on the input.
        /// </summary>
        /// <param name="input">Input PlanSetState.</param>
        /// <returns>Output PlanSetState.</returns>
        protected virtual BasicPlanSetState CalcPlanSetState(BasicPlanSetState input)
        {
            return input.Filter(Selectivity);
        }

        public override void UpdatePlan(Plan plan)
        {
            base.UpdatePlan(plan);

            // Sharing is at least the sharing of the sub plan
            plan.Sharing = (BitSet) plan.OnlyChild.Sharing.Clone();

            // If this is not a representative or the child is not shared, no more sharing is possible.
            if (!queryOptimization.ShareEquivalenceClasses.ContainsKey(this) || !plan.OnlyChild.Shared)
                return;

            // Add this rule to the sharing bitset and set this plan as sharable.
            plan.Sharing.Or(
                queryOptimization.BitSetManager.GetWithValues(new NodeAttribute(Nodes.Single(), null).PropertyName));
            plan.Shared = true;
        }
    }
}