﻿#region

using System;
using System.Collections.Generic;
using System.Text;

#endregion

namespace Microsoft.Ceres.Optimizer
{
    /// <summary>
    /// Static class containing various utility functions.
    /// </summary>
    internal static class Utils
    {
        /// <summary>
        /// Limit a string to the specified length. If lengt if longer than
        /// the length of the string, the entire string is returned.
        /// </summary>
        /// <param name="value">String to limit.</param>
        /// <param name="length">Length to limit to.</param>
        /// <returns>The limited string.</returns>
        public static string LimitTo(this string value, int length)
        {
            if (value.Length <= length)
                return value;
            else
                return value.Substring(0, length);
        }

        /// <summary>
        /// Works just like <see cref="string.Format(string, object[])"/>, but
        /// handles null values for the <paramref name="args"/> parameter by
        /// inserting '(null)' instead of throwing an exception.
        /// </summary>
        /// <param name="format">Format string, see <see cref="string.Format(string, object[])"/>.</param>
        /// <param name="args">Arguments, see <see cref="string.Format(string, object[])"/>.</param>
        /// <returns>The formatted string.</returns>
        public static string NullFormat(string format, params object[] args)
        {
            object[] args2 = new object[args.Length];
            for (int i = 0; i < args.Length; ++i)
                args2[i] = args[i] ?? "(null)";
            return string.Format(format, args2);
        }

        public static string JoinEnumerable<T>(IEnumerable<T> enumerable)
        {
            return JoinEnumerable(enumerable, item => item.ToString(), string.Empty, ",", true);
        }

        public static string JoinEnumerable<T>(IEnumerable<T> enumerable, string separator)
        {
            return JoinEnumerable(enumerable, item => item.ToString(), string.Empty, separator, true);
        }

        /// <summary>
        /// Joins all elements in an IEnumerable to a string.
        /// </summary>
        /// <param name="enumerable">The sequence to enumerate.</param>
        /// <returns>elem1.Tostring(),elem2.ToString(),elem3.ToString()...</returns>
        public static string JoinEnumerable<T>(IEnumerable<T> enumerable, Func<T, string> projection, string before,
                                               string after, bool removeLast)
        {
            if (enumerable == null)
                return "";
            StringBuilder sb = new StringBuilder();
            foreach (T item in enumerable)
            {
                if (item != null)
                {
                    sb.AppendFormat("{0}{1}{2}", before, projection(item), after);
                }
            }
            if (sb.Length > 0 && removeLast)
            {
                sb.Remove(sb.Length - after.Length, after.Length);
            }
            return sb.ToString();
        }

        /// <summary>
        /// Approximate equals for doubles.
        /// </summary>
        /// <param name="first">The first double to compare.</param>
        /// <param name="other">The second double to compare.</param>
        /// <param name="accuracy">Accuracy limit.</param>
        /// <returns>If the doubles are within accurary of each other.</returns>
        public static bool AccuracyEquals(this double first, double other, double accuracy)
        {
            return Math.Abs(first - other) < accuracy;
        }

        /// <summary>
        /// Compares two dictionaries. Set equivalence for keys is required
        /// and values must be equal.
        /// </summary>
        /// <typeparam name="TKey">Key type for the dictionaries.</typeparam>
        /// <typeparam name="TValue">Value type for the dictionaries.</typeparam>
        /// <param name="first">The first dictionary to compare.</param>
        /// <param name="other">The second dictionary to compare.</param>
        /// <returns>If the dictionaries are equal.</returns>
        public static bool DictionaryEquals<TKey, TValue>(IDictionary<TKey, TValue> first,
                                                          IDictionary<TKey, TValue> other)
        {
            foreach (KeyValuePair<TKey, TValue> pair in first)
            {
                TValue value;
                if (!other.TryGetValue(pair.Key, out value) || !pair.Value.Equals(value))
                    return false;
            }

            foreach (TKey key in other.Keys)
                if (!first.ContainsKey(key))
                    return false;
            return true;
        }
    }
}