﻿#region

using System;
using System.Collections.Generic;
using Microsoft.Ceres.Optimizer.Engine;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests
{
    /// <summary>
    /// Tests of BitSet and BitSetManager.
    /// </summary>
    [TestFixture]
    public class BitSetTests : TestBase
    {
        #region Setup/Teardown

        [SetUp]
        public void SetUp()
        {
            manager = new BitSetManager();
            manager.GetBitSet(PropertyUsage.Produced, "TESTPROP1");
            manager.GetBitSet(PropertyUsage.Produced, "TESTPROP2");
            manager.GetBitSet(PropertyUsage.Produced, "TESTPROP3");
            manager.Prepare();
        }

        #endregion

        private BitSetManager manager;

        private void GetBitSets(out BitSet a, out BitSet aCheck, out BitSet b, out BitSet bCheck)
        {
            a = manager.Empty;
            aCheck = manager.Empty;
            a.Add("TESTPROP1");
            aCheck.Add("TESTPROP1");
            a.Add("TESTPROP3");
            aCheck.Add("TESTPROP3");

            b = manager.Empty;
            bCheck = manager.Empty;
            b.Add("TESTPROP2");
            bCheck.Add("TESTPROP2");
            b.Add("TESTPROP3");
            bCheck.Add("TESTPROP3");
        }

        [Test]
        public void ProxyTest()
        {
            manager = new BitSetManager();
            BitSetProxy p1 = manager.GetBitSet(PropertyUsage.Produced, "TESTPROP1.1", "TESTPROP1.2");
            BitSetProxy p2 = manager.GetBitSet(PropertyUsage.Produced, "TESTPROP2");
            manager.Prepare();

            Assert.IsTrue(p1.BitSet.Contains("TESTPROP1.1"));
            Assert.IsTrue(p1.BitSet.Contains("TESTPROP1.2"));
            Assert.IsFalse(p1.BitSet.Contains("TESTPROP2"));
            Assert.IsFalse(p2.BitSet.Contains("TESTPROP1.1"));
            Assert.IsFalse(p2.BitSet.Contains("TESTPROP1.1"));
            Assert.IsTrue(p2.BitSet.Contains("TESTPROP2"));
        }

        [Test]
        public void SimpleTest()
        {
            BitSet a = manager.Empty;
            a.Add("TESTPROP2");
            Assert.IsFalse(a.Contains("TESTPROP1"));
            Assert.IsTrue(a.Contains("TESTPROP2"));
            Assert.IsFalse(a.Contains("TESTPROP3"));

            a.Add("TESTPROP3");
            Assert.IsFalse(a.Contains("TESTPROP1"));
            Assert.IsTrue(a.Contains("TESTPROP2"));
            Assert.IsTrue(a.Contains("TESTPROP3"));

            a.Remove("TESTPROP2");
            Assert.IsFalse(a.Contains("TESTPROP1"));
            Assert.IsFalse(a.Contains("TESTPROP2"));
            Assert.IsTrue(a.Contains("TESTPROP3"));
        }

        [Test]
        public void TestDifferenceOperator()
        {
            BitSet a, aCheck, b, bCheck;
            GetBitSets(out a, out aCheck, out b, out bCheck);

            BitSet difference = a - b;
            BitSet differenceCheck = manager.Empty;
            differenceCheck.Add("TESTPROP1");

            Assert.That(difference.Equals(differenceCheck));
            Assert.That(a.Equals(aCheck));
            Assert.That(b.Equals(bCheck));
        }

        [Test]
        public void TestEqualsNotEqualsOperators()
        {
            BitSet a, aCheck, b, bCheck;
            GetBitSets(out a, out aCheck, out b, out bCheck);

            Assert.IsTrue(a == aCheck);
            Assert.IsFalse(a != aCheck);
            Assert.IsTrue(a.Equals(aCheck));
            Assert.IsFalse(a.Equals(new object()));
            Assert.IsFalse(a.Equals(null));

            aCheck.Add("TESTPROP2");
            Assert.IsFalse(a == aCheck);
            Assert.IsTrue(a != aCheck);
            Assert.IsFalse(a.Equals(aCheck));
        }

        [Test]
        public void TestIntersectionOperator()
        {
            BitSet a, aCheck, b, bCheck;
            GetBitSets(out a, out aCheck, out b, out bCheck);

            BitSet intersection = a & b;
            BitSet intersectionCheck = manager.Empty;
            intersectionCheck.Add("TESTPROP3");

            Assert.That(intersection.Equals(intersectionCheck));
            Assert.That(a.Equals(aCheck));
            Assert.That(b.Equals(bCheck));
        }

        [Test]
        public void TestMinimization()
        {
            BitSetManager manager = new BitSetManager();
            BitSetProxy p1 = manager.GetBitSet(PropertyUsage.Produced, "TESTPROP1.1", "TESTPROP1.2");
            BitSetProxy p2 = manager.GetBitSet(PropertyUsage.Produced, "TESTPROP2.1");
            BitSetProxy p3 = manager.GetBitSet(PropertyUsage.Produced, "TESTPROP3.1", "TESTPROP3.2", "TESTPROP3.3");
            manager.Prepare();

            Assert.AreEqual(3, manager.Count);
            BitSet a = manager.GetWithValues("TESTPROP1.1");
            Assert.IsTrue(a.Contains("TESTPROP1.1"));
            Assert.IsTrue(a.Contains("TESTPROP1.2"));
            Assert.IsFalse(a.Contains("TESTPROP2.1"));
            Assert.IsFalse(a.Contains("TESTPROP3.1"));
            Assert.IsFalse(a.Contains("TESTPROP3.2"));
            Assert.IsFalse(a.Contains("TESTPROP3.3"));
            a.Add("TESTPROP3.3");
            Assert.IsTrue(a.Contains("TESTPROP1.1"));
            Assert.IsTrue(a.Contains("TESTPROP1.2"));
            Assert.IsFalse(a.Contains("TESTPROP2.1"));
            Assert.IsTrue(a.Contains("TESTPROP3.1"));
            Assert.IsTrue(a.Contains("TESTPROP3.2"));
            Assert.IsTrue(a.Contains("TESTPROP3.3"));
        }

        [Test]
        public void TestNonExisting()
        {
            ArgumentException ex = XAssert.Throws<ArgumentException>(delegate
                                                                         {
                                                                             BitSet a = manager.Empty;
                                                                             a.Add("BING");
                                                                         });
            Assert.AreEqual("No property with this name registered: BING", ex.Message);
        }

        [Test]
        public void TestNotPrepared()
        {
            BitSetManager manager = new BitSetManager();
            InvalidOperationException ex = XAssert.Throws<InvalidOperationException>(
                delegate { BitSet dummy = manager.Empty; });
            Assert.AreEqual("The BitSetManager must be prepared before it can be used.", ex.Message);
        }

        [Test]
        public void TestOverlaps()
        {
            BitSet a, aCheck, b, bCheck;
            GetBitSets(out a, out aCheck, out b, out bCheck);

            Assert.IsTrue(a.Overlaps(b));
            b.Remove("TESTPROP3");
            Assert.IsFalse(a.Overlaps(b));
        }

        [Test]
        public void TestPrepared()
        {
            InvalidOperationException ex = XAssert.Throws<InvalidOperationException>(() => manager.Prepare());
            Assert.AreEqual("The BitSetManager is already prepared.", ex.Message);
        }

        [Test]
        public void TestPrepared2()
        {
            InvalidOperationException ex = XAssert.Throws<InvalidOperationException>(()
                                                                                     =>
                                                                                     manager.GetBitSet(
                                                                                         PropertyUsage.Produced,
                                                                                         "TESTPROP1.1", "TESTPROP1.2"));
            Assert.AreEqual("The BitSetManager is already prepared.", ex.Message);
        }

        // <= and >=
        [Test]
        public void TestSmallerEqualsLargerEqualsOperators()
        {
            BitSet a, aCheck, b, bCheck;
            GetBitSets(out a, out aCheck, out b, out bCheck);

            BitSet temp = manager.Empty;
            Assert.IsTrue(temp <= a);
            Assert.IsFalse(temp >= a);
            temp.Add("TESTPROP1");
            Assert.IsTrue(temp <= a);
            Assert.IsFalse(temp >= a);
            temp.Add("TESTPROP3");
            Assert.IsTrue(temp <= a);
            Assert.IsTrue(temp >= a);
            temp.Add("TESTPROP2");
            Assert.IsFalse(temp <= a);
            Assert.IsTrue(temp >= a);
        }

        [Test]
        public void TestUnionOperator()
        {
            BitSet a, aCheck, b, bCheck;
            GetBitSets(out a, out aCheck, out b, out bCheck);

            BitSet union = a | b;
            BitSet unionCheck = manager.Empty;
            unionCheck.Add("TESTPROP1");
            unionCheck.Add("TESTPROP2");
            unionCheck.Add("TESTPROP3");

            Assert.That(union.Equals(unionCheck));
            Assert.That(a.Equals(aCheck));
            Assert.That(b.Equals(bCheck));
        }

        [Test]
        public void TestWalk()
        {
            BitSetManager manager = new BitSetManager();
            manager.GetBitSet(PropertyUsage.Produced, "TESTPROP1");
            manager.GetBitSet(PropertyUsage.Produced, "TESTPROP2.1", "TESTPROP2.2");
            manager.GetBitSet(PropertyUsage.Produced, "TESTPROP3");
            manager.GetBitSet(PropertyUsage.Produced, "TESTPROP4");
            manager.Prepare();
            BitSet bitSet = manager.GetWithValues("TESTPROP1", "TESTPROP2.2", "TESTPROP4");

            List<BitSet> actual = new List<BitSet>();
            foreach (BitSet walk in bitSet.Walk())
            {
                actual.Add((BitSet) walk.Clone());
            }

            Assert.AreEqual(8, actual.Count);
            CollectionAssert.Contains(actual, manager.GetWithValues());
            CollectionAssert.Contains(actual, manager.GetWithValues("TESTPROP4"));
            CollectionAssert.Contains(actual, manager.GetWithValues("TESTPROP2.1")); //same as 2.2 (minimized)
            CollectionAssert.Contains(actual, manager.GetWithValues("TESTPROP2.1", "TESTPROP4"));
            CollectionAssert.Contains(actual, manager.GetWithValues("TESTPROP1"));
            CollectionAssert.Contains(actual, manager.GetWithValues("TESTPROP1", "TESTPROP4"));
            CollectionAssert.Contains(actual, manager.GetWithValues("TESTPROP1", "TESTPROP2.1"));
            CollectionAssert.Contains(actual, manager.GetWithValues("TESTPROP1", "TESTPROP2.1", "TESTPROP4"));
        }
    }
}