﻿#region

using System.Collections.Generic;
using Microsoft.Ceres.Optimizer.GraphMatching;
using NMock2;
using NUnit.Framework;

#endregion

//using Microsoft.Ceres.Optimizer.Tests.GraphMatching.Extensions;

namespace Microsoft.Ceres.Optimizer.Tests.GraphMatching
{
    /// <summary>
    /// Simple smoke tests for the graph matching component.
    /// </summary>
    [TestFixture]
    public class SimpleGraphMatchingTests
    {
        #region Setup/Teardown

        [SetUp]
        public void SetUp()
        {
            /* A <- B <- C <- D
             * ^ -------------/
             */

            a = new Node("A");
            b = new Node("B");
            c = new Node("C");
            d = new Node("D");

            any = new Node();

            a.AddChild(b);
            b.AddChild(c);
            c.AddChild(d);
            a.AddChild(d);

            aMatcher = new NamedNodeMatcher("A");
            bMatcher = new NamedNodeMatcher("B");
            cMatcher = new NamedNodeMatcher("C");
            dMatcher = new NamedNodeMatcher("D");

            mocks = new Mockery();

            visualizer = new OperatorGraphVisualizer();
        }

        #endregion

        private Node a, b, c, d, any;
        private NamedNodeMatcher aMatcher, bMatcher, cMatcher, dMatcher;
        private OperatorGraphVisualizer visualizer;

        private Mockery mocks;

        [Test]
        public void TestAllChildren()
        {
            AbstractNodeMatcher pattern = Match.AnyNode().WithAllChildren(
                Match.NodeWithName("B")
                );

            pattern.AssertNotMatches(a, b, c, d);

            pattern = Match.AnyNode().WithAllChildren(
                Match.NodeWithName("D")
                );

            pattern.AssertMatches(c);
            pattern.AssertNotMatches(a, b, d);
        }

        [Test]
        public void TestAllMatching()
        {
            // Match everything.
            AbstractNodeMatcher pattern = new AllMatcher(new AnythingMatcher());
            pattern.AssertMatches(any);

            pattern.Reset();

            // Check that it matches two nodes as well.
            pattern.Search(any);
            pattern.AssertMatches(any);

            // Match just A-nodes.
            pattern = new AllMatcher(aMatcher);
            pattern.AssertMatches(a);

            pattern.Reset();
            // First A-occurrance is OK.
            Assert.That(pattern.Search(a));
            // Does not match any.
            Assert.That(!pattern.Search(any));
            // All-pattern has already failed.
            Assert.That(!pattern.Search(a));
        }

        [Test]
        public void TestAnd()
        {
            // Match nodes that match both A and anything.
            AbstractNodeMatcher pattern = new AndMatcher(aMatcher, new AnythingMatcher());
            pattern.AssertMatches(a);
            // any does not match A:
            pattern.AssertNotMatches(any);

            pattern = new AndMatcher(aMatcher);
            // A matches A.
            pattern.AssertMatches(a);

            // AND-ing in a NothingMatcher makes the pattern a contradiction.
            pattern &= new NothingMatcher();
            // ... so these must not match.
            pattern.AssertNotMatches(a);
            pattern.AssertNotMatches(any);

            // Same, but the other way.
            pattern = new NothingMatcher();
            pattern &= aMatcher;
            pattern.AssertNotMatches(a);
        }

        [Test]
        public void TestCombiningChildAndParentTraverser()
        {
            /* Match a node A whose children all have one parent. This must not
               match, since A has B and D as children, and D has A and C as parents. */

            AbstractNodeMatcher pattern = Match.NodeWithName("A").WithAllChildren(
                Match.AnyNode().WithSingleParent()
                );

            pattern.AssertNotMatches(a, b, c, d);

            // Same as above, but starting on B instead.
            pattern = Match.NodeWithName("B").WithAllChildren(
                Match.AnyNode().WithSingleParent()
                );

            pattern.AssertMatches(b);
            pattern.AssertNotMatches(a, c, d);

            // Same as above, but starting on any node.
            pattern = Match.AnyNode().WithAllChildren(
                Match.AnyNode().WithSingleParent()
                );
            // B's child is C, whose only parent is B.
            pattern.AssertMatches(b);
            // A and C both have D as a child. D has no children.
            pattern.AssertNotMatches(a, c, d);
        }

        [Test]
        public void TestMatchCapturing()
        {
            // Pattern that matches anything, but with two operators.
            AbstractNodeMatcher inner = new AnythingMatcher();
            AbstractNodeMatcher outer = new OrMatcher(inner);

            outer.AssertMatches(any);

            // Check that both nodes have captured the node.
            Assert.AreEqual(new List<Node>() {any}, outer.MatchedNodes);
            Assert.AreEqual(inner.MatchedNodes, outer.MatchedNodes);
        }

        [Test]
        public void TestNodeWithChildWithSingleParent()
        {
            AbstractNodeMatcher pattern = Match.NodeWithName("A");
            pattern.AssertMatches(a);
            // Match A if it has a child B
            pattern = pattern.WithChild(bMatcher);
            pattern.AssertMatches(a);
            pattern.AssertNotMatches(b, c, d);

            // Match a node A if it has a child with a single parent.

            pattern = Match.NodeWithName("A").WithChild(
                Match.AnyNode().WithSingleParent()
                );

            // A has B as a child, whose only parent is A, and thus matches.
            pattern.AssertMatches(a);
            pattern.AssertNotMatches(b, c, d);
        }

        [Test]
        public void TestNodeWithSingleParent()
        {
            AbstractNodeMatcher nodesWithOneParent = Match.AnyNode().WithSingleParent();
            nodesWithOneParent.AssertMatches(b, c);
            nodesWithOneParent.AssertNotMatches(a, d);
        }

        [Test]
        public void TestOr()
        {
            // Nothing OR everything is a tautology.
            AbstractNodeMatcher pattern = new OrMatcher(new NothingMatcher(), new AnythingMatcher());
            pattern.AssertMatches(any);

            // Same, but the other way.
            pattern = (new NothingMatcher()) | (new AnythingMatcher());
            pattern.AssertMatches(any);

            // Nothing matches nothing.
            pattern = new NothingMatcher();
            pattern.AssertNotMatches(any);
            // ... nor two nothing matchers.
            pattern |= new NothingMatcher();
            pattern.AssertNotMatches(any);
            // Adding an anything matcher makes it a tautology.
            pattern |= new AnythingMatcher();
            pattern.AssertMatches(any);

            // Test operator overloading.
            pattern = aMatcher | bMatcher;
            pattern.AssertMatches(a);
            pattern.AssertMatches(b);
            pattern.AssertNotMatches(c);
        }

        [Test]
        public void TestParentBranching()
        {
            var singleParentPattern = new ParentMatcher() {TakeWhile = new RangeQuantifier(new AnythingMatcher(), 1, 1)};
            singleParentPattern.AssertMatches(b, c);
            // A has no parent, and D has two.
            singleParentPattern.AssertNotMatches(a, d);

            var branchingNodePattern = new ParentMatcher()
                                           {TakeWhile = new RangeQuantifier(new AnythingMatcher(), 2, -1)};
            branchingNodePattern.AssertMatches(d);
            branchingNodePattern.AssertNotMatches(a, b, c);
        }

        [Test]
        public void TestRangeQuantifier()
        {
            // Match nodes having at least two children that are either B or D.
            AbstractNodeMatcher pattern = new ChildMatcher()
                                              {TakeWhile = new RangeQuantifier(bMatcher | dMatcher, 2, -1)};
            pattern.AssertMatches(a);
            pattern.AssertNotMatches(b, c, d);

            // Match nodes with at most one descendant.
            pattern = new DescendantTraverser() {TakeWhile = new RangeQuantifier(new AnythingMatcher(), 0, 1)};
            pattern.AssertMatches(c, d);
            pattern.AssertNotMatches(a, b); // A's descendants are B, C, D. B's descendants are C and D.
        }

        [Test]
        public void TestSimpleChildMatcher()
        {
            // Match nodes with children that are just B or D or both.
            var pattern = new ChildMatcher()
                              {
                                  TakeWhile = new AllMatcher(bMatcher | dMatcher)
                              };

            pattern.AssertMatches(a, c);
            pattern.AssertNotMatches(b, d);
        }

        [Test]
        public void TestSimpleChildTraverser()
        {
            // Match a node "A" that has a descendant "C".
            AbstractTraverser pattern = new DescendantTraverser()
                                            {
                                                RootMatcher = aMatcher,
                                                TakeWhile = new AnythingMatcher(),
                                                StopPattern = cMatcher
                                            };

            pattern.AssertMatches(a);
            Assert.AreEqual(new List<Node>() {a}, pattern.RootMatcher.MatchedNodes);
            Assert.AreEqual(new List<Node>() {b, d}, pattern.TakeWhile.MatchedNodes);
            Assert.AreEqual(new List<Node>() {c}, pattern.StopPattern.MatchedNodes);

            pattern.AssertNotMatches(b, c, d);

            // Match nodes that have a descendant D.
            pattern = new DescendantTraverser()
                          {
                              TakeWhile = new AnythingMatcher(),
                              StopPattern = dMatcher
                          };

            pattern.AssertMatches(a, b, c);
            pattern.AssertNotMatches(d);
        }

        [Test]
        public void TestSimpleNamedNodeMatcher()
        {
            aMatcher.AssertMatches(a);
        }
    }
}