﻿#region

using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Threading;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests
{
    /// <summary>
    /// Contains automation code to start up a MARS node in console mode.
    /// It will redirect standard in and out for the node so users of
    /// this class can execute queries and get results back.
    /// </summary>
    public class NodeRunnerAutomation
    {
        private const string EVENTS_DIR = @"C:\Mars.Net\var\log\events";
        // Limit for one line in query feeder.
        private const int LINE_LIMIT = 250;
        private const int RUN_TIME = 400000;
        //NodeRunner process
        // If the node is ready (reported from the status command)
        private bool nodeIsReady;
        private Process nodeRunner;
        // Event to communicate that the query feeder is ready ("Query>") from the output reader thread.
        // Collects the output from NodeRunner.
        private StringBuilder outputCollector;
        private object outputCollectorLock = new object();
        private bool printToConsole = true;
        private ManualResetEvent queryFeederReadyEvent;
        private Thread readerThread;

        /// <summary>Start NodeRunner and await commands.</summary>
        public void Start()
        {
            // Hack. Remove events dir, otherwise NodeRunner fails.
            if (Directory.Exists(EVENTS_DIR))
                Directory.Delete(EVENTS_DIR, true);

            // Set up
            queryFeederReadyEvent = new ManualResetEvent(false);
            nodeIsReady = false;
            outputCollector = new StringBuilder();
            // Set up process, redirect stdin and stdout and start NodeRunner.
            nodeRunner = new Process();
            nodeRunner.StartInfo = new ProcessStartInfo(@"C:\Mars.NET\NodeRunner.exe", "--shell");
            nodeRunner.StartInfo.WorkingDirectory = @"C:\Mars.NET";
            nodeRunner.StartInfo.RedirectStandardInput = true;
            nodeRunner.StartInfo.RedirectStandardOutput = true;
            nodeRunner.StartInfo.UseShellExecute = false;
            nodeRunner.Start();

            // Start the thread reading NodeRunner's stdout
            readerThread = new Thread(new ThreadStart(OutputThread));
            readerThread.Start();

            // Wait for the node to become ready
            while ((DateTime.Now - nodeRunner.StartTime).Milliseconds < RUN_TIME && !nodeIsReady)
            {
                Thread.Sleep(1000);
                if (nodeRunner.HasExited)
                {
                    readerThread.Abort();
                    throw new Exception("MARS exited. Are you sure you started VS/NUnit from cmd?");
                }
                WriteLine("status");
            }

            // Interact with QueryFeeder.
            nodeRunner.StandardInput.WriteLine("i QueryFeeder1");
            // Set columns widths to avoid having NodeRunner call Console.WindowWidth, which
            // fails when we have redirected the output.
            nodeRunner.StandardInput.WriteLine("ColumnWidths={{30;all}}");

            // Wait for the node to start
            if (!nodeIsReady || !queryFeederReadyEvent.WaitOne(RUN_TIME))
            {
                readerThread.Abort();

                // Kill everything if it takes too long time.
                nodeRunner.Kill();
                throw new Exception(string.Format("MARS did not start within {0} seconds.", RUN_TIME/1000));
            }
            // We're past the startup garbage, enable output.
        }

        /// <summary>Shut down NodeRunner.</summary>
        public void Stop()
        {
            // Exit QueryFeeder and shut down the node.
            nodeRunner.StandardInput.WriteLine("exit");
            nodeRunner.StandardInput.WriteLine("sh");

            // Wait for the process to exit.
            if (!nodeRunner.WaitForExit(RUN_TIME))
            {
                // Kill everything if it does not within the time limit.
                readerThread.Abort();
                nodeRunner.Kill();
                throw new Exception(string.Format("Node runner did not stop within {0} seconds.", RUN_TIME/1000));
            }
            // Kill the reader thread anyways.
            readerThread.Abort();
        }

        /// <summary>
        /// Writes a line to the stdin of the QueryFeeder.
        /// </summary>
        /// <param name="line">Line to write.</param>
        public void WriteLine(string line)
        {
            if (printToConsole)
                Console.WriteLine("STDIN: {0}", line);
            nodeRunner.StandardInput.WriteLine(line);
        }

        /// <summary>
        /// Writes a line to the stdin of the QueryFeeder.
        /// </summary>
        /// <param name="format">String format.</param>
        /// <param name="arg">Arguments.</param>
        public void WriteLine(string format, params object[] arg)
        {
            if (printToConsole)
                Console.WriteLine("STDIN: {0}", string.Format(format, arg));
            nodeRunner.StandardInput.WriteLine(format, arg);
        }


        public string RunQuery(string query)
        {
            return RunQuery(query, true, true);
        }

        /// <summary>
        /// Runs the specified query and returns the standard output.
        /// Will throw exception if the word "exception" is encountered in stdout.
        /// </summary>
        /// <param name="query">Query text to run.</param>
        /// <returns>Standard output.</returns>
        public string RunQuery(string query, bool printToConsole, bool assertException)
        {
            this.printToConsole = printToConsole;
            // Setup
            queryFeederReadyEvent = new ManualResetEvent(false);
            lock (outputCollectorLock)
            {
                outputCollector = new StringBuilder();
            }

            WriteLine(query);

            // Wait for the query to finish.
            if (!queryFeederReadyEvent.WaitOne(RUN_TIME))
            {
                // If not, kill everything.
                nodeRunner.Kill();
                readerThread.Abort();
                throw new Exception(string.Format("Query did not finish in {0} seconds.", RUN_TIME/1000));
            }
            // Get the result
            string result;
            lock (outputCollectorLock)
            {
                result = outputCollector.ToString();
            }

            // Throw exception if we find 'Exception' in stdout.
            if (assertException && result.Contains("Exception"))
                throw new Exception("Word 'Exception' contained in query output.");

            return result;
        }

        /// <summary>Entry point for the output reader thread.</summary>
        private void OutputThread()
        {
            while (!nodeRunner.StandardOutput.EndOfStream)
            {
                // Read one character at a time (not a line, as we want to detect Query>, which is an unfinished line).
                char[] buffer = new char[1];
                nodeRunner.StandardOutput.ReadBlock(buffer, 0, 1);
                if (printToConsole)
                    Console.Write(buffer[0]);
                // Only flush the console at new lines to not slow it down.
                if (printToConsole && buffer[0] == '\n' || buffer[0] == '\r')
                    Console.Out.Flush();

                // Collect output
                lock (outputCollectorLock)
                {
                    outputCollector.Append(buffer[0]);
                }

                // Check if the last 6 characters form stdout is 'Query>'. In that
                // case, the QueryFeeder is ready, so set the event.
                lock (outputCollectorLock)
                {
                    if (outputCollector.Length >= 15)
                    {
                        char[] checkBuffer = new char[15];
                        outputCollector.CopyTo(outputCollector.Length - 15, checkBuffer, 0, 15);
                        if (new string(checkBuffer) == "Node is Running")
                            nodeIsReady = true;
                    }
                }

                // Check if the last 6 characters form stdout is 'Query>'. In that
                // case, the QueryFeeder is ready, so set the event.
                lock (outputCollectorLock)
                {
                    if (outputCollector.Length >= 6)
                    {
                        char[] checkBuffer = new char[6];
                        outputCollector.CopyTo(outputCollector.Length - 6, checkBuffer, 0, 6);
                        if (new string(checkBuffer) == "Query>")
                            queryFeederReadyEvent.Set();
                    }
                }
            }
        }
    }
}