﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests
{
    /// <summary>
    /// Tests automatically execution queries against a running MARS node with
    /// the optimizer injected.
    /// </summary>
    [TestFixture, Explicit]
    public class NodeRunnerTests
    {
        protected NodeRunnerAutomation automation = new NodeRunnerAutomation();

        [TestFixtureSetUp]
        public void FixtureSetUp()
        {
            automation.Start();
        }

        [TestFixtureTearDown]
        public void FixtureTearDown()
        {
            automation.Stop();
        }

        [Test]
        public void TestSortTiming()
        {
            TimeSpan sort = TimeSpan.Zero;
            TimeSpan nosort = TimeSpan.Zero;
            for (int i = 0; i < 10; i++)
            {
                Stopwatch watch = Stopwatch.StartNew();
                automation.RunQuery(@"Output(label=""NO_OPTIMIZATION"";Trim(hits=10;Sort(sortFields=[(""Position"", true)];Select(filter=""Position=30"";Lookup(index=""Occurrence1"",word=""is"")))))");
                watch.Stop();
                Console.Write("SORT: {0}", watch.Elapsed);
                sort += watch.Elapsed;
                watch.Reset();
                watch.Start();
                automation.RunQuery(@"Output(label=""NO_OPTIMIZATION"";Trim(hits=10;Select(filter=""Position=30"";Lookup(index=""Occurrence1"",word=""is""))))");
                watch.Stop();
                Console.Write("NO SORT: {0}", watch.Elapsed);
                nosort += watch.Elapsed;
            }

            Console.WriteLine(sort.TotalMilliseconds / 10);
            Console.WriteLine(nosort.TotalMilliseconds / 10);
        }


        private void doTimings(List<Pair<string, int>> words, Func<string, string, string> queryFunc,
                               Func<string, string, string> queryFunc2)
        {
            List<TimeSpan> timings = new List<TimeSpan>();

            int count = 0;
            for (int i = 0; i < words.Count - 1; i++)
            {
                Stopwatch watch = new Stopwatch();
                string word1 = words[i].First, word2 = words[i + 1].First;
                string result;

                do
                {
                    watch.Reset();
                    watch.Start();
                    result = automation.RunQuery(queryFunc(word1, word2), false, false);
                    if (queryFunc2 != null)
                        result += automation.RunQuery(queryFunc2(word1, word2), false, false);
                    watch.Stop();
                    if (result.Contains("errors"))
                        Console.WriteLine("{0} - {1}, {2}: Error recorded", i, word1, word2);
                } while (result.Contains("errors"));

                Console.WriteLine("{0};{1};{2};{3};{4:F6};{5}", word1, words[i].Second, word2, words[i + 1].Second,
                                  (words[i].Second/20000.0)*(words[i + 1].Second/20000.0)*250000,
                                  watch.ElapsedMilliseconds);
                timings.Add(watch.Elapsed);
                count++;
            }

            TimeSpan total = TimeSpan.Zero;
            foreach (TimeSpan t in timings)
                total += t;
            Console.WriteLine("TOTAL: {0}, {1}", count, total);
        }

        /// <summary>Returns the left part of the tree benchmarking query.</summary>
        private string getNonDagQuery1(string word1, string word2)
        {
            return
                @"Output(label=""NO_OPTIMIZATION"";\
    Sort(trimHitCount=10,sortFields=[(""Score"",False)];\
        Map(maps={Score=Input0.Score,DocumentName=Input1.DocumentName,DocumentType=Input2.DocumentType};\
            MergeJoin(joinPrefix=1;\
                Map(maps={DocumentId=DocumentId,Score=""(70 * Score + 30 * ProxScore ) / 100""};\
                    ScoreOccurrences(groupPrefix=1,required=2,scorers={""Score""=SimpleTerms(),""ProxScore""=SimpleProximity()};\
                        Lookup(index=""Occurrence1"",word=""" +
                word1 + @"""),\
                        Lookup(index=""Occurrence1"",word=""" + word2 +
                @"""))),\
                Lookup(index=""Occurrence2"",word=""DocumentName""),\
                Lookup(index=""Occurrence2"",word=""DocumentType"")))))";
        }

        /// <summary>Returns the right part of the tree benchmarking query.</summary>
        private string getNonDagQuery2(string word1, string word2)
        {
            return
                @"Output(label=""NO_OPTIMIZATION"";\
    Sort(trimOffset=0,trimHitCount=-1,reorder=True,sortFields=[(""Count"",False)];\
        Group(sorted=False,groupFields=[""DocumentType""],aggs={""Count""=Count(args=Null,filter=Null)};\
            Map(maps={Score=Input0.Score,DocumentName=Input1.DocumentName,DocumentType=Input2.DocumentType};\
                MergeJoin(joinPrefix=1;\
                    Map(maps={DocumentId=DocumentId,Score=""(70 * Score + 30 * ProxScore ) / 100""};\
                        ScoreOccurrences(groupPrefix=1,required=2,scorers={""Score""=SimpleTerms(),""ProxScore""=SimpleProximity()};\
                            Lookup(index=""Occurrence1"",word=""" +
                word1 + @"""),\
                            Lookup(index=""Occurrence1"",word=""" + word2 +
                @"""))),\
                    Lookup(index=""Occurrence2"",word=""DocumentName""),\
                    Lookup(index=""Occurrence2"",word=""DocumentType""))))))";
        }

        /// <summary>Returns the DAG benchmarking query.</summary>
        private string getDagQuery(string word1, string word2)
        {
            return
                @"Output(label=""NO_OPTIMIZATION"";\
    Sort(trimHitCount=10,sortFields=[(""Score"",False)];\
        Copy(additionalSuccessors=[""foo""]; \
            Map(maps={Score=Input0.Score,DocumentName=Input1.DocumentName,DocumentType=Input2.DocumentType};\
                MergeJoin(joinPrefix=1;\
                    Map(maps={DocumentId=DocumentId,Score=""(70 * Score + 30 * ProxScore ) / 100""};\
                        ScoreOccurrences(groupPrefix=1,required=2,scorers={""Score""=SimpleTerms(),""ProxScore""=SimpleProximity()};\
                            Lookup(index=""Occurrence1"",word=""" +
                word1 + @"""),\
                            Lookup(index=""Occurrence1"",word=""" + word2 +
                @"""))),\
                    Lookup(index=""Occurrence2"",word=""DocumentName""),\
                    Lookup(index=""Occurrence2"",word=""DocumentType""))))),\
   Sort(trimOffset=0,trimHitCount=-1,reorder=True,sortFields=[(""Count"",False)];\
      Group(sorted=False,groupFields=[""DocumentType""],aggs={""Count""=Count(args=Null,filter=Null)},label=""foo"")))";
        }

        /// <summary>Returns the optimized DAG benchmarking query.</summary>
        private string getOptimizedQuery(string word1, string word2)
        {
            return
                @"Output(label=""NO_OPTIMIZATION"";\
    Sort(sortFields=[(""Score"",False)];\
        Map(maps={Score=Input0.Input0.Score,DocumentName=Input1.DocumentName,DocumentType=Input0.Input1.DocumentType};\
            MergeJoin(joinPrefix=1;\
                Sort(sortFields=[(""DocumentId"",True)],reorder=True;\
                    Sort(trimHitCount=10,sortFields=[(""Input0.Score"",False)];\
                          Copy(additionalSuccessors=[""foo""];\
                               MergeJoin(joinPrefix=1;\
                                  Map(maps={DocumentId=DocumentId,Score=""(70 * Score + 30 * ProxScore ) / 100""};\
                                      ScoreOccurrences(groupPrefix=1,required=2,scorers={""Score""=SimpleTerms(),""ProxScore""=SimpleProximity()};\
                                          Lookup(index=""Occurrence1"",word=""" +
                word1 + @"""),\
                                          Lookup(index=""Occurrence1"",word=""" + word2 +
                @"""))),\
                              Lookup(index=""Occurrence2"",word=""DocumentType""))))),\
                    Lookup(index=""Occurrence2"",word=""DocumentName"")))),\
    Sort(sortFields=[(""Count"", False)];\
       Group(sorted=False,groupFields=[""Input1.DocumentType""],aggs={""Count""=Count(args=Null,filter=Null)},label=""foo"")))";
        }

        /// <summary>
        /// Benchmarking test.
        /// </summary>
        [Test, Explicit("LongRunning")]
        public void SpeedTest()
        {
            Dictionary<int, string> temp = new Dictionary<int, string>();
            int max = 0;
            foreach (string line in File.ReadAllLines("words.txt"))
            {
                string[] word = line.Split(',');
                int count = int.Parse(word[1]);
                temp[count] = word[0];
                if (count > max)
                    max = count;
            }
            List<Pair<string, int>> words = new List<Pair<string, int>>();

            for (int i = 1; i < max; i *= 2)
            {
                string word = null;
                int counter = i;
                while (i < max && !temp.TryGetValue(counter, out word))
                {
                    counter++;
                }
                if (i < max && word != null)
                    words.Add(new Pair<string, int>(word, counter));
            }

            //doTimings(words, getNonDagQuery);
            //automation.Stop(); automation.Start();
            //doTimings(words, getDagQuery);
            //automation.Stop(); automation.Start();
            doTimings(words, getNonDagQuery1, getNonDagQuery2);
            doTimings(words, getDagQuery, null);
            //doTimings(words, getNonDagQuery1, getNonDagQuery2);
        }

        /// <summary>
        /// Test a simple DAG without optimization.
        /// </summary>
        [Test]
        public void TestDAGWitNoOptimization()
        {
            automation.RunQuery(
                @"Output(label=""NO_OPTIMIZATION"";\
                                    Copy(additionalSuccessors=[""foobar""];\
                                        Make(start=1,end=10)),\
                                    Trim(label=""foobar"",hits=5,offset=2))");
        }

        /// <summary>
        /// Test running a larger query without optimization.
        /// </summary>
        [Test]
        public void TestLargeQuery()
        {
            automation.RunQuery(
                @"Output(label=""NO_OPTIMIZATION"";\
                    Sort(trimHitCount=10,sortFields=[(""Score"",False)];\
                            Map(maps={Score=Input0.Score,DocumentName=Input1.DocumentName,DocumentType=Input2.DocumentType};\
                                MergeJoin(joinPrefix=1;\
                                    Map(maps={DocumentId=DocumentId,Score=""(70 * Score + 30 * ProxScore ) / 100""};\
                                        ScoreOccurrences(groupPrefix=1,required=2,scorers={""Score""=SimpleTerms(),""ProxScore""=SimpleProximity()};\
                                            Lookup(index=""Occurrence1"",word=""connect""),\
                                            ONear(;\
                                                Lookup(index=""Occurrence1"",word=""in""),\
                                                Lookup(index=""Occurrence1"",word=""the"")))),\
                                    Lookup(index=""Occurrence2"",word=""DocumentName""),\
                                    Lookup(index=""Occurrence2"",word=""DocumentType"")))))");
        }

        /// <summary>
        /// Test running a query with sharing opportunities.
        /// </summary>
        [Test]
        public void TestSharing()
        {
            automation.RunQuery(
                @"Output(;
                    Group(sorted=False,groupFields=[""Input1.DocumentType""],aggs={""Count""=Count(args=Null,filter=Null)};\
                        MergeJoin(joinPrefix=1;\
                            Lookup(word=""olstad"",index=""Occurrence1""),\
                            Lookup(word=""DocumentType"",index=""Occurrence2"")\
                        )\
                    ),\
                    MergeJoin(joinPrefix=1;\
                        MergeJoin(joinPrefix=1;\
                           Lookup(word=""olstad"",index=""Occurrence1""),\
                           Lookup(word=""DocumentType"",index=""Occurrence2"")\
                        ),\
                        Lookup(word=""DocumentName"", index=""Occurrence2"")\
                    )\
                 )");
        }

        /// <summary>
        /// Test running a simple query with optimization.
        /// </summary>
        [Test]
        public void TestSimpleQuery()
        {
            automation.RunQuery("Output(;Lookup(index=\"Occurrence1\",word=\"foo\"))");
        }
    }
}