﻿#region

using System;
using System.Collections.Generic;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests
{
    /// <summary>
    /// Various tests for the Node class.
    /// </summary>
    [TestFixture]
    public class NodeTests
    {
        #region Setup/Teardown

        // A bunch of nodes we'll reuse in several tests.
        // They are grouped into graphs of four nodes with different structures.
        private Node a, b, c, d,
                     m, n, o, p,
                     w, x, y, z;

        [SetUp]
        public void SetUp()
        {
            /* A <- B <- C <- D
             * ^ -------------/
             */

            a = new Node("A");
            b = new Node("B");
            c = new Node("C");
            d = new Node("D");

            a.AddChild(b);
            b.AddChild(c);
            c.AddChild(d);
            a.AddChild(d);

            /*   /- N - \
             *  M        P
             *   \- O - /
             */

            m = new Node("M");
            n = new Node("N");
            o = new Node("O");
            p = new Node("P");

            m.AddChild(n);
            m.AddChild(o);
            n.AddChild(p);
            o.AddChild(p);

            /*  /  X
             * W - Y
             *  \  Z
             */
            w = new Node("W");
            x = new Node("X");
            y = new Node("Y");
            z = new Node("Z");

            w.AddChild(x);
            w.AddChild(y);
            w.AddChild(z);
        }

        #endregion

        [Test]
        public void TestAcyclicity()
        {
            // These graphs should all be acyclic.
            foreach (var node in new List<Node>() {a, m, w})
            {
                Assert.IsTrue(node.IsAcyclic());
            }

            // Make a cycle between Q and R.
            var q = new Node("Q");
            var r = new Node("R");

            q.AddChild(r);
            r.AddChild(q);

            Assert.IsFalse(q.IsAcyclic());
            Assert.IsFalse(r.IsAcyclic());
        }

        [Test]
        [ExpectedException(typeof (InvalidOperationException))]
        public void TestAscendantsWithCycle()
        {
            var r = new Node();
            var s = new Node();
            r.AddChild(s);
            s.AddChild(r);

            new List<Node>(s.GetAscendants());
        }

        [Test]
        public void TestBubble()
        {
            List<Node> visited = new List<Node>();
            d.Bubble(delegate(Node node) { visited.Add(node); });

            var expected = new List<Node>() {d, a, c, b};
            Assert.AreEqual(expected, visited);
        }

        [Test]
        public void TestDepthFirstIteratorOfTree()
        {
            var root = new Node("R");
            var child = new Node("C");
            root.AddChild(child);
            var grandChild = new Node("GC");
            child.AddChild(grandChild);

            // First inclusive
            var resultingTraversal = new List<Node>(root.IterateDepthFirst());
            var expectedTraversal = new List<Node>() {root, child, grandChild};
            Assert.AreEqual(expectedTraversal, resultingTraversal);

            // Excluding root.
            Assert.AreEqual(
                new List<Node>(root.IterateDepthFirst(false)),
                new List<Node>() {child, grandChild});
        }

        [Test]
        public void TestDepthFirstTraversal()
        {
            var expected = new List<Node>(a.IterateDepthFirst());
            var actual = new List<Node>();
            a.TraverseDepthFirst(delegate(Node node) { actual.Add(node); });
            Assert.AreEqual(expected, actual);

            // Test that the traversal stops when requested.
            expected = new List<Node>() {a};
            actual = new List<Node>();
            a.Traverse(a.IterateDepthFirst(true), delegate(Node node)
                                                      {
                                                          actual.Add(node);
                                                          if (node == a)
                                                          {
                                                              return false;
                                                          }
                                                          return true;
                                                      });
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void TestExtractNode()
        {
            // A <- B <- C
            var a = new Node();
            var b = new Node();
            var c = new Node();

            a.AddChild(b);
            b.AddChild(c);
            b.Extract();

            Assert.AreEqual(a.Children, new List<Node>() {c});
            Assert.AreEqual(c.Parents, new List<Node>() {a});

            var emptyList = new List<Node>();
            Assert.AreEqual(b.Parents, emptyList);
            Assert.AreEqual(b.Children, emptyList);
        }

        [Test]
        public void TestExtractNode2()
        {
            /* A <- B <- C <- D    =>   A <- D  
             * ^ -------------/  
             */

            b.Extract();
            c.Extract();

            Assert.AreEqual(a.Children, new List<Node>() {d});
            Assert.AreEqual(d.Parents, new List<Node>() {a});
        }

        [Test]
        public void TestNodeAscendants()
        {
            var actual = new List<Node>(d.GetAscendants());
            var expected = new List<Node>() {a, c, b};
            Assert.AreEqual(expected, actual);

            actual = new List<Node>(c.GetAscendants());
            expected = new List<Node>() {b, a};
            Assert.AreEqual(expected, actual);

            actual = new List<Node>(p.GetAscendants());
            expected = new List<Node>() {o, m, n};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void TestPartialTopologicalOrdering()
        {
            var actual = new List<Node>(b.GetTopologicalOrdering());
            var expected = new List<Node>() {d, c, b};
            Assert.AreEqual(expected, actual);

            actual = new List<Node>(n.GetTopologicalOrdering());
            expected = new List<Node>() {p, n};
            Assert.AreEqual(expected, actual);

            actual = new List<Node>(z.GetTopologicalOrdering());
            expected = new List<Node>() {z};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void TestReplace()
        {
            b.ReplaceWith(c);
            Assert.AreEqual(new List<Node>() {c, d}, a.Children);
            Assert.AreEqual(new List<Node>() {a}, c.Parents);
            Assert.AreEqual(new List<Node>(), b.Parents);
            Assert.AreEqual(new List<Node>(), b.Children);

            Assert.AreEqual(new List<Node>() {c, d}, new List<Node>(a.IterateDepthFirst(false)));
        }

        [Test]
        public void TestReplacementWithExtract()
        {
            // Achieving the same as above.
            b.ReplaceWith(c);
            c.Extract();
            Assert.AreEqual(a.Children, new List<Node>() {d});
            Assert.AreEqual(d.Parents, new List<Node>() {a});
        }

        [Test]
        public void TestSimpleNodeProperties()
        {
            Assert.IsTrue(a.IsRoot);
            Assert.IsFalse(b.IsRoot || c.IsRoot);

            Assert.IsTrue(d.IsLeaf);
            Assert.IsFalse(b.IsLeaf || c.IsLeaf);
        }

        [Test]
        public void TestTopologicalOrdering()
        {
            var actual = new List<Node>(a.GetTopologicalOrdering());
            var expected = new List<Node>() {d, c, b, a};
            Assert.AreEqual(expected, actual);

            actual = new List<Node>(m.GetTopologicalOrdering());
            expected = new List<Node>() {p, o, n, m};
            Assert.AreEqual(expected, actual);

            actual = new List<Node>(w.GetTopologicalOrdering());
            expected = new List<Node>() {z, y, x, w};
            Assert.AreEqual(expected, actual);
        }
    }
}