﻿#region

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.DataModel.Types;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Aggregation;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Aggregation.BasicAggregators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Infrastructure;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Join;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Search;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using Microsoft.Ceres.SearchCore.Search.Operators;
using Microsoft.Ceres.SearchCore.Search.Operators.Combiners;
using Microsoft.Ceres.SearchCore.Search.Scorers;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests
{
    public abstract class OptimizationTestBase
    {
        // Default cardinalities used for testing
        protected readonly double[] cardinalities = new double[]
                                                        {
                                                            1000000, 2000000, 142200, 1300, 10, 6, 34630, 463600, 34458, 38
                                                            , 150, 19000, 10, 5, 400, 500000
                                                        };
        // Defalt selectivities used for testing
        protected readonly double[] selectivities = new[]
                                                        {
                                                            0.0006, 0.001, 0.5, 0.2, 0.01, 0.0004, 0.6, 0.001, 0.2, 0.1,
                                                            0.01, 0.9, 1.0, 0.5, 0.01
                                                        };

        protected SystemMetadataMock metadata;
        public QueryOptimization queryOptimization;

        [SetUp]
        public virtual void SetUp()
        {
            OrderManager.LIMIT1 = false; // Do not turn on the hack by default.
            queryOptimization = new QueryOptimization();
            queryOptimization.PrintStats = true;
            InitializeIndexStats();
        }

        /// <summary>
        /// Initialize index statistics for the specified cardinalities for indexes Mock0...MockN.
        /// </summary>
        protected virtual void InitializeIndexStats(double[] cardinalities)
        {
            for (int i = 0; i < cardinalities.Length; i++)
            {
                metadata.SetIndexSchema("Mock" + i, null,
                                        new RecordDescriptor(
                                            new Field(new NodeAttribute(null, "DocumentId" + i), typeof (long)),
                                            new Field(new NodeAttribute(null, "Position"), typeof (long)),
                                            new Field(new NodeAttribute(null, "Scope"), typeof (string))));
                metadata.SetIndexOrder("Mock" + i, new OrderDescription(
                                                       new OrderDescriptionItem("DocumentId" + i, true)));//,
                                                       //new OrderDescriptionItem("Position", true)));

                metadata.SetIndexState("Mock" + i, "word", new BasicPlanSetState(cardinalities[i], 100));
            }
        }

        /// <summary>
        /// Initialize standard index statistics.
        /// </summary>
        protected virtual void InitializeIndexStats()
        {
            metadata = new SystemMetadataMock();
            metadata.SetIndexState("Occurrence1", "search", new BasicPlanSetState(10000, 100));
            metadata.SetIndexState("Occurrence1", "engine", new BasicPlanSetState(5000, 100));
            metadata.SetIndexState("Occurrence1", "olstad", new BasicPlanSetState(500, 100));
            metadata.SetIndexState("Occurrence1", "information", new BasicPlanSetState(2000, 100));
            metadata.SetIndexState("Occurrence1", null, new BasicPlanSetState(500, 100));
            metadata.SetIndexSchema("Occurrence1", null,
                                    new RecordDescriptor(
                                        new Field(new NodeAttribute(null, "DocumentId"), typeof (long)),
                                        new Field(new NodeAttribute(null, "Position"), typeof (long)),
                                        new Field(new NodeAttribute(null, "Scope"), typeof (string))));
            metadata.SetIndexOrder("Occurrence1", new OrderDescription(
                                                      new OrderDescriptionItem("DocumentId", true)));//,
                                                      //new OrderDescriptionItem("Position", true)));

            metadata.SetIndexState("Occurrence2", "DocumentName", new BasicPlanSetState(8000, 35));
            metadata.SetIndexSchema("Occurrence2", "DocumentName",
                                    new RecordDescriptor(
                                        new Field(new NodeAttribute(null, "DocumentId"), typeof (long)),
                                        new Field(new NodeAttribute(null, "DocumentName"), typeof (string))));
            metadata.SetIndexState("Occurrence2", "DocumentType", new BasicPlanSetState(8000, 10));
            metadata.SetIndexSchema("Occurrence2", "DocumentType",
                                    new RecordDescriptor(
                                        new Field(new NodeAttribute(null, "DocumentId"), typeof (long)),
                                        new Field(new NodeAttribute(null, "DocumentType"), typeof (string))));
            metadata.SetIndexOrder("Occurrence2", new OrderDescription(
                                                      new OrderDescriptionItem("DocumentId", true)));

            queryOptimization.SystemMetadata = metadata;
        }

        /// <summary>Return a query node representing MARS' Output operator.</summary>
        public static TestNode GetOutput(params TestNode[] children)
        {
            TestNode query = new TestNode(typeof (OutputOperator));
            query.AddChildren(children);
            return query;
        }

        /// <summary>Return a query node representing MARS' Group operator.</summary>
        public static TestNode GetGroup(IList<string> groupingFields, bool sorted,
                                        IDictionary<string, Aggregator> aggregators,
                                        Dictionary<string, Type> outputTypes, TestNode child)
        {
            TestNode group = new TestNode(typeof (GroupOperator));
            group["GroupingFields"] = groupingFields;
            group["Sorted"] = sorted;
            group["Aggregators"] = aggregators;
            group.AddChild(child);
            group.OutputTypeDescriptor = getOutputTypeDescriptor(outputTypes);
            return group;
        }

        /// <summary>Return a query node representing MARS' Sort operator.</summary>
        public static TestNode GetSort(long trimHitCount, long trimOffset, bool reorder, List<SortDefinition> sortFields,
                                       TestNode child)
        {
            TestNode sort = new TestNode(typeof (SortOperator));
            sort["TrimHitCount"] = trimHitCount;
            sort["TrimOffset"] = trimOffset;
            sort["Reorder"] = reorder;
            sort["SortFields"] = sortFields;
            sort.AddChild(child);
            return sort;
        }

        /// <summary>Return a query node representing MARS' Trim operator.</summary>
        public static TestNode GetTrim(int hits, int offset, TestNode child)
        {
            TestNode trim = new TestNode(typeof (TrimOperator));
            trim["Hits"] = hits;
            trim["Offset"] = offset;
            trim.AddChild(child);
            return trim;
        }

        /// <summary>Return a query node representing MARS' Map operator.</summary>
        public static TestNode GetMap(bool retainInput, Dictionary<IdentifierProperty, ExpressionProperty> parameterMap,
                                      Dictionary<string, Type> outputTypes, TestNode child)
        {
            TestNode map = new TestNode(typeof (MapOperator));
            map["RetainInput"] = retainInput;
            if (parameterMap != null)
                map["ParameterMap"] = parameterMap;
            map.AddChild(child);
            if (outputTypes != null)
                map.OutputTypeDescriptor = getOutputTypeDescriptor(outputTypes);
            return map;
        }

        private static IRecordSetTypeDescriptor getOutputTypeDescriptor(Dictionary<string, Type> outputTypes)
        {
            List<FieldOccurrence> fields = new List<FieldOccurrence>();
            foreach (KeyValuePair<string, Type> field in outputTypes)
                fields.Add(new FieldOccurrence(field.Key, TypeRegistry.Singleton.GetType(field.Value).Type, fields.Count));
            return RecordSetTypeDescriptors.Create(0, null, fields.ToArray());
        }

        /// <summary>Return a query node representing MARS' ScoreOccurrence operator with default config.</summary>
        public static TestNode GetScoreOccurences(params TestNode[] children)
        {
            return GetScoreOccurences(
                new Dictionary<string, Scorer>()
                    {
                        {"Score", new SimpleTermsScorer()},
                        {"ProxScore", new SimpleProximityScorer()}
                    },
                new AverageCombiner(),
                new Dictionary<string, Type>() {{"Score", typeof (double)}, {"ProxScore", typeof (double)}},
                children);
        }

        /// <summary>Return a query node representing MARS' ScoreOccurrence operator.</summary>
        public static TestNode GetScoreOccurences(Dictionary<string, Scorer> scorers, ScoreCombiner combiner,
                                                  Dictionary<string, Type> outputTypes, params TestNode[] children)
        {
            TestNode scoreOccurences = new TestNode(typeof (ScoreOccurrencesOperator));
            scoreOccurences["Scorers"] = scorers;
            scoreOccurences["Combiner"] = combiner;
            scoreOccurences["GroupPrefix"] = 1;
            scoreOccurences["Required"] = -1;
            scoreOccurences.AddChildren(children);
            scoreOccurences.OutputTypeDescriptor = getOutputTypeDescriptor(outputTypes);
            return scoreOccurences;
        }

        /// <summary>Return a query node representing MARS' ONear operator.</summary>
        public static TestNode GetONear(int maxDistance, params TestNode[] children)
        {
            TestNode oNear = new TestNode(typeof (ONearOperator));
            oNear["MaxDistance"] = maxDistance;
            oNear["GroupPrefix"] = 1;
            oNear.AddChildren(children);
            return oNear;
        }

        /// <summary>Return a query node representing MARS' Lookup operator.</summary>
        public static TestNode GetLookup(string word, string indexName)
        {
            TestNode lookup = new TestNode(typeof (LookupOperator));
            lookup["Word"] = word;
            lookup["IndexName"] = indexName;
            return lookup;
        }

        /// <summary>Return a query node representing MARS' Selection operator.</summary>
        public static TestNode GetSelection(string filter, TestNode child)
        {
            return GetSelection(filter, null, null, child);
        }

        /// <summary>Return a query node representing MARS' Selection operator.</summary>
        public static TestNode GetSelection(string filter, double? selectivity, double? predicateCost, TestNode child)
        {
            TestNode select = new TestNode(typeof (SelectOperator));
            select.AddChild(child);
            select["Filter"] = new ExpressionProperty(filter);
            if (selectivity != null)
                select["Selectivity"] = selectivity.Value;
            if (predicateCost != null)
                select["PredicateCost"] = predicateCost.Value;
            return select;
        }

        /// <summary>Return a query node representing MARS' HybridHashJoin operator.</summary>
        public static TestNode GetJoin(string leftJoinKey, string rightJoinKey, double? selectivity, TestNode leftChild,
                                       TestNode rightChild)
        {
            return GetJoin(new List<IList<string>>() {new List<string>() {leftJoinKey, rightJoinKey}},
                           selectivity, leftChild, rightChild);
        }

        /// <summary>Return a query node representing MARS' HybridHashJoin operator.</summary>
        public static TestNode GetJoin(IList<IList<string>> joinKeys, double? selectivity, TestNode leftChild,
                                       TestNode rightChild)
        {
            TestNode join = new TestNode(typeof (HybridHashJoinOperator));
            join.AddChild(leftChild);
            join.AddChild(rightChild);
            join["JoinKeys"] = joinKeys;

            if (selectivity.HasValue)
                join["Selectivity"] = selectivity.Value;
            return join;
        }

        /// <summary>Return a query node representing MARS' MergeJoin operator.</summary>
        public static TestNode GetMergeJoin(string key1, string key2, double? selectivity, params TestNode[] children)
        {
            return GetMergeJoin(new List<string>() {key1, key2}, selectivity, children);
        }

        /// <summary>Return a query node representing MARS' MergeJoin operator.</summary>
        public static TestNode GetMergeJoin(IList<string> joinKeys, double? selectivity, params TestNode[] children)
        {
            TestNode join = new TestNode(typeof (MergeJoinOperator));
            join["JoinPrefix"] = 1;
            join["RequiredPrefix"] = 1;
            if (selectivity.HasValue)
                join["Selectivity"] = selectivity.Value;
            join.AddChildren(children);
            if (joinKeys != null)
            {
                for (int i = 0; i < joinKeys.Count; i++)
                {
                    IRecordSetTypeDescriptor newDescriptor = RecordSetTypeDescriptors.Create(0, null,
                                                                                             new FieldOccurrence(
                                                                                                 joinKeys[i], null, 0));
                    if (children[i].OutputTypeDescriptor != null)
                        newDescriptor = newDescriptor.Append(children[i].OutputTypeDescriptor.RecordType.ToArray());
                    children[i].OutputTypeDescriptor = newDescriptor;
                }
            }
            return join;
        }

        /// <summary>Return a query node representing MARS' Copy operator.</summary>
        public static TestNode GetCopy(TestNode input)
        {
            TestNode copy = new TestNode(typeof (CopyOperator));
            copy.AddChild(input);
            return copy;
        }

        /// <summary>Return a simple query with three chain joins and four lookups.</summary>
        public static TestNode GetSimpleJoinsInput(double[] cardinalities, double[] selectivities)
        {
            // We input the worst plan possible with cost 222.
            TestNode[] lookups = new[]
                                     {
                                         GetLookup("word", "Mock0"),
                                         GetLookup("word", "Mock1"),
                                         GetLookup("word", "Mock2"),
                                         GetLookup("word", "Mock3")
                                     };
            TestNode[] joins = new TestNode[3];
            joins[0] = GetJoin("DocumentId1", "DocumentId2", selectivities[0], lookups[1], lookups[2]);
            joins[1] = GetJoin("DocumentId1", "DocumentId0", selectivities[1], joins[0], lookups[0]);
            joins[2] = GetJoin("Input0.DocumentId2", "DocumentId3", selectivities[2], joins[1], lookups[3]);
            return GetOutput(joins[2]);
        }

        /// <summary>
        /// Return the subtree upto the map operator for the large query used to demontrate
        /// sharing:
        /// MAP - JOIN -    JOIN -  MAP - ScoreOcc - Lookup
        ///         \- Lookup  \- Lookup    \- Lookup
        /// </summary>
        public static TestNode GetSampleQuerySubtree()
        {
            return
                GetMap(false,
                       new Dictionary<IdentifierProperty, ExpressionProperty>()
                           {
                               {
                                   new IdentifierProperty("DocumentType"),
                                   new ExpressionProperty("Input0.Input1.DocumentType")
                                   },
                               {new IdentifierProperty("DocumentName"), new ExpressionProperty("Input1.DocumentName")},
                               {new IdentifierProperty("Score"), new ExpressionProperty("Input0.Input0.Score")},
                           },
                       new Dictionary<string, Type>()
                           {
                               {"DocumentType", typeof (string)},
                               {"DocumentName", typeof (string)},
                               {"Score", typeof (double)}
                           },
                       GetJoin("DocumentId", "DocumentId", 1.0/8000,
                               GetJoin("DocumentId", "DocumentId", 1.0/8000,
                                       GetMap(false,
                                              new Dictionary<IdentifierProperty, ExpressionProperty>()
                                                  {
                                                      {
                                                          new IdentifierProperty("DocumentId"),
                                                          new ExpressionProperty("DocumentId")
                                                          },
                                                      {
                                                          new IdentifierProperty("Score"),
                                                          new ExpressionProperty("(70 * Score + 30 * ProxScore ) / 100")
                                                          }
                                                  },
                                              new Dictionary<string, Type>()
                                                  {{"DocumentId", typeof (long)}, {"Score", typeof (double)}},
                                              GetScoreOccurences(
                                                  GetLookup("search", "Occurrence1"),
                                                  GetLookup("engine", "Occurrence1")
                                                  )
                                           ),
                                       GetLookup("DocumentType", "Occurrence2")
                                   ),
                               GetLookup("DocumentName", "Occurrence2")
                           )
                    );
        }

        /// <summary>Return a the large example query as a DAG with a copy operator.</summary>
        public TestNode GetLargeExampleQuery1()
        {
            TestNode copy = GetCopy(GetSampleQuerySubtree());

            return GetOutput(
                GetSort(10, 0, true, new List<SortDefinition>() {new SortDefinition("Score", false)}, copy),
                GetSort(0, 0, true, new List<SortDefinition>() {new SortDefinition("Count", false)},
                        GetGroup(new List<string>() {"DocumentType"}, false,
                                 new Dictionary<string, Aggregator>() {{"Count", new CountAggregator()}},
                                 new Dictionary<string, Type>() {{"Count", typeof (long)}},
                                 copy)));
        }
    }
}