﻿using System;
using NUnit.Framework;
using Microsoft.Ceres.Optimizer.Engine;
using System.Diagnostics;
using Microsoft.Ceres.Optimizer.Engine.Orders;

namespace Microsoft.Ceres.Optimizer.Tests
{
    /// <summary>
    /// Optimizer speed tests used for benchmarking the optimizer.
    /// </summary>
    [TestFixture]
    [Explicit("LongRunning")]
    public class OptimizerSpeedTest : OptimizationTestBase
    {
        #region Setup/Teardown

        [SetUp]
        public override void SetUp()
        {
            // Be sure to turn off all debug output to not interfere with the timings.
            base.SetUp();
            InitializeIndexStats(cardinalities);
            queryOptimization.PrintStats = false;
            queryOptimization.Visualizer.Visualize = false;
            OrderManager.LIMIT1 = true;
        }

        #endregion

        /// <summary>
        /// Benchmarking test for the optimizer. It executes optimizations
        /// of four types of tasks:
        /// 0: Bushy, chain, 1: Left-deep, chain, 2: Bushy, star, 3: left-deep, star
        /// 
        /// For each tasks, it optimizes an increasing number of joins until the time
        /// spent optimtizing passes a set limit. Each optimization run is repeated RUNS number of times.
        /// Finally, the average runtime for each task and join count is printed.
        /// </summary>
        [Test]
        public void SpeedTest()
        {
            // Time limit for a single run.
            TimeSpan LIMIT = TimeSpan.FromSeconds(60);
            // Number of runs for each task and join count.
            const int RUNS = 10;

            int[,] runs = new int[4, 15];
            TimeSpan[,] timings = new TimeSpan[4, 15];
            queryOptimization.PrintStats = false;

            // Tasks: 0: Bushy, chain, 1: Left-deep, chain, 2: Bushy, star, 3: left-deep, star
            for (int task = 0; task < 4; task++)
            {
                // Increasing number of joins...
                for (int joins = 0; joins < selectivities.Length; joins++)
                {
                    runs[task, joins] = 0;
                    timings[task, joins] = TimeSpan.Zero;
                    // RUNS tasks...
                    for (int run = 0; run < RUNS; run++)
                    {
                        // Construct the query.
                        TestNode leftChild = GetLookup("word", "Mock0");
                        for (int i = 0; i < joins; i++)
                        {
                            TestNode rightChild = GetLookup("word", "Mock" + (i + 1));
                            if (task <= 2)
                            {
                                leftChild = GetJoin("DocumentId" + i, "DocumentId" + (i + 1), selectivities[i],
                                                    leftChild, rightChild);
                            }
                            else
                            {
                                leftChild = GetJoin("DocumentId0", "DocumentId" + (i + 1), selectivities[i],
                                                    leftChild, rightChild);
                            }
                        }
                        OperatorNode query = GetOutput(leftChild);
                        if (task % 2 == 0)
                        {
                            queryOptimization.JoinEnumeration = JoinEnumeration.Bushy;
                        }
                        else
                        {
                            queryOptimization.JoinEnumeration = JoinEnumeration.LeftDeep;
                        }

                        // Time the optimization
                        Stopwatch watch = Stopwatch.StartNew();
                        queryOptimization.Optimize(query);
                        watch.Stop();

                        Console.WriteLine("TASK {0}, JOINS: {1}, RUN: {2}, TIME: {3}", task, joins, run, watch.Elapsed);
                        
                        // Store results
                        if (run != 0)
                        {
                            runs[task, joins]++;
                            timings[task, joins] += watch.Elapsed;
                        }

                        // If we passed the time limit for one task, go to the next task.
                        if (watch.Elapsed > LIMIT)
                        {
                            if (run == 0)
                            {
                                runs[task, joins]++;
                                timings[task, joins] += watch.Elapsed;
                            }
                            goto nextTask;
                        }
                    }
                }
            nextTask:
                int dummy;
            }

            // Print timings summary
            Console.WriteLine("Timings summary:");
            Console.WriteLine("Relations;Bushy-chain;Leftdeep-chain;Bushy-star;Leftdeep-star");
            for (int joins = 0; joins < selectivities.Length; joins++)
            {
                Console.Write("{0};", joins + 1);
                for (int task = 0; task < 4; task++)
                {
                    // Compute the average run time.
                    Console.Write("{0}", timings[task, joins].TotalMilliseconds / runs[task, joins]);
                    if (task < 3)
                    {
                        Console.Write(";");
                    }
                }
                Console.WriteLine();
            }
        }


    }
}
