﻿#region

using Microsoft.Ceres.Optimizer.Engine.Orders;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests.Orders
{
    /// <summary>
    /// Tests for OrderDescription.
    /// </summary>
    [TestFixture]
    public class OrderDescriptionTests
    {
        private const bool ASC = true, DESC = false;

        private OrderDescription getTestData()
        {
            return new OrderDescription(
                new OrderDescriptionItem("A.a", ASC),
                new OrderDescriptionItem("A.b", DESC),
                new OrderDescriptionItem("B.b", ASC),
                new OrderDescriptionItem("A.c", DESC));
        }

        [Test]
        public void TestAppend()
        {
            OrderDescription desc = getTestData();
            OrderDescription desc2 = desc.ConstAppend(new OrderDescriptionItem("B.a", ASC));
            Assert.AreEqual(4, desc.Size);
            Assert.AreEqual(5, desc2.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc.ToString());
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC,B.a ASC]", desc2.ToString());

            desc.Append(new OrderDescriptionItem("B.a", ASC));
            Assert.AreEqual(5, desc.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC,B.a ASC]", desc.ToString());
        }

        [Test]
        public void TestClone()
        {
            OrderDescription desc = getTestData();
            OrderDescription desc2 = (OrderDescription) desc.Clone();

            Assert.AreEqual(4, desc2.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc2.ToString());

            desc2.LimitTo(3);
            desc2[0] = new OrderDescriptionItem("B.b", DESC);
            Assert.AreEqual(4, desc.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc.ToString());
        }

        [Test]
        public void TestEquals()
        {
            OrderDescription desc1 = new OrderDescription(
                new OrderDescriptionItem("A.a", ASC),
                new OrderDescriptionItem("A.b", DESC));
            OrderDescription desc2 = new OrderDescription(
                new OrderDescriptionItem("A.a", ASC),
                new OrderDescriptionItem("A.b", DESC));
            OrderDescription desc3 = new OrderDescription(
                new OrderDescriptionItem("A.a", ASC),
                new OrderDescriptionItem("A.c", DESC));
            OrderDescription desc4 = new OrderDescription(
                new OrderDescriptionItem("A.a", ASC),
                new OrderDescriptionItem("A.b", ASC));

            Assert.IsTrue(desc1.Equals(desc1));
            Assert.IsTrue(desc2.Equals(desc2));
            Assert.IsTrue(desc3.Equals(desc3));
            Assert.IsTrue(desc4.Equals(desc4));
            Assert.IsTrue(desc1.Equals(desc2));
            Assert.IsFalse(desc1.Equals(desc3));
            Assert.IsFalse(desc1.Equals(desc4));
            Assert.IsFalse(desc2.Equals(desc3));
            Assert.IsFalse(desc2.Equals(desc4));
            Assert.IsFalse(desc3.Equals(desc4));
        }

        [Test]
        public void TestInsert()
        {
            OrderDescription desc = getTestData();
            OrderDescription desc2 = desc.Insert(2, new OrderDescriptionItem("B.a", ASC));
            Assert.AreEqual(4, desc.Size);
            Assert.AreEqual(5, desc2.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc.ToString());
            Assert.AreEqual("[A.a ASC,A.b DESC,B.a ASC,B.b ASC,A.c DESC]", desc2.ToString());
        }

        [Test]
        public void TestLimitTo()
        {
            OrderDescription desc = getTestData();
            desc.LimitTo(5);
            Assert.AreEqual(4, desc.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc.ToString());
            desc.LimitTo(3);
            Assert.AreEqual(3, desc.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC]", desc.ToString());
        }

        [Test]
        public void TestOperators()
        {
            OrderDescription desc1 = getTestData();
            OrderDescription desc2 = getTestData();
            OrderDescription desc3 = getTestData();
            desc3[2] = new OrderDescriptionItem("B.b", DESC);
            OrderDescription desc4 = desc2.ConstAppend(new OrderDescriptionItem("B.a", ASC));

            Assert.IsTrue(desc1 == desc2);
            Assert.IsFalse(desc1 == desc3);
            Assert.IsFalse(desc1 == desc4);
            Assert.IsTrue(desc1 <= desc2);
            Assert.IsFalse(desc1 < desc2);
            Assert.IsTrue(desc2 >= desc1);
            Assert.IsFalse(desc2 > desc1);
            Assert.IsFalse(desc1 < desc3);
            Assert.IsFalse(desc1 <= desc3);
            Assert.IsFalse(desc1 > desc3);
            Assert.IsFalse(desc1 >= desc3);
            Assert.IsTrue(desc1 <= desc4);
            Assert.IsTrue(desc1 < desc4);
            Assert.IsFalse(desc1 >= desc4);
            Assert.IsFalse(desc1 > desc4);
            Assert.IsFalse(desc3 <= desc4);
            Assert.IsFalse(desc3 < desc4);
        }

        [Test]
        public void TestRemoveAt()
        {
            OrderDescription desc = getTestData();
            OrderDescription desc2 = desc.RemoveAt(2);
            Assert.AreEqual(4, desc.Size);
            Assert.AreEqual(3, desc2.Size);
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc.ToString());
            Assert.AreEqual("[A.a ASC,A.b DESC,A.c DESC]", desc2.ToString());
        }

        [Test]
        public void TestReplace()
        {
            OrderDescription desc = getTestData(); // [{A.a} ASC,{A.b} DESC,{B.b} ASC,{A.c} DESC]
            OrderDescription desc2 = desc.Replace(2, "B.a");
            OrderDescription desc3 = desc.Replace(3, "B.a");
            OrderDescription desc4 = desc.Replace(2, "A.b");
            OrderDescription desc5 = desc.Replace(0, "B.b");
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc.ToString());
            Assert.AreEqual("[A.a ASC,A.b DESC,B.a ASC,A.c DESC]", desc2.ToString());
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,B.a DESC]", desc3.ToString());
            Assert.AreEqual("[A.a ASC,A.b DESC,A.c DESC]", desc4.ToString());
            Assert.AreEqual("[B.b ASC,A.b DESC,A.c DESC]", desc5.ToString());
        }

        [Test]
        public void TestToString()
        {
            OrderDescription desc = new OrderDescription();
            Assert.AreEqual("[]", desc.ToString());
            desc = getTestData();
            Assert.AreEqual("[A.a ASC,A.b DESC,B.b ASC,A.c DESC]", desc.ToString());
        }
    }
}