﻿#region

using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests.Orders
{
    /// <summary>
    /// Tests for OrderManager.
    /// </summary>
    [TestFixture]
    public class OrderManagerTests
    {
        #region Setup/Teardown

        [SetUp]
        public void SetUp()
        {
            orderManager = new OrderManager();
            orderManager.Visualizer = new OperatorGraphVisualizer();
            orderManager.Visualizer.InvokeDot = false;
        }

        #endregion

        private OrderManager orderManager;
        
        /// <summary>
        /// An uncovered flaw in the orderings framework introducing unecessary sorts.
        /// We will not hit this limitation, so we will probably not fix it.
        /// </summary>
        [Test]
        [Ignore("Bug in orderings we may not fix.")]
        public void CustomTest2()
        {
            // Person(PersonId ASC, Name, Zip, CompanyId)
            // ZipCode(Zip ASC, City)
            // Company(CompanyId ASC, Name)
            //
            // SELECT * FROM Person p
            // JOIN ZipCode z ON p.Zip = z.Zip
            // JOIN Company c ON p.CompanyId = c.CompanyId
            // WHERE c.Name = 'Microsoft'
            // ORDER BY c.Name, z.Zip, z.City

            // All tables clustered on primary key
            OrderingState clusteredPersonId = orderManager.GetOrderingState(
                new OrderDescription("Person.PersonId ASC"), true);
            OrderingState clusteredZip = orderManager.GetOrderingState(new OrderDescription("ZipCode.Zip ASC"), true);
            OrderingState clusteredCompanyId =
                orderManager.GetOrderingState(new OrderDescription("Company.CompanyId ASC"), true);

            // Tested for during joins
            OrderProxy joinPersonZip = orderManager.GetOrder(new OrderDescription("Person.Zip ASC"), false);
            OrderProxy joinZipCodeZip = orderManager.GetOrder(new OrderDescription("ZipCode.Zip ASC"), false);
            OrderProxy joinPersonCompanyId = orderManager.GetOrder(new OrderDescription("Person.CompanyId ASC"), false);
            OrderProxy joinCompanyCompanyId = orderManager.GetOrder(new OrderDescription("Company.CompanyId ASC"), false);

            // Final ORDER BY clause
            OrderProxy personIdzipCity =
                orderManager.GetOrder(new OrderDescription("Person.PersonId ASC, ZipCode.City ASC"), false);

            // Given from primary keys
            Dependency zip_Name = orderManager.GetDependency("ZipCode.Zip->ZipCode.City");
            Dependency companyId_Name = orderManager.GetDependency("Company.CompanyId->Company.Name");
            Dependency personId_NameZipCompanyId =
                orderManager.GetDependency("Person.PersonId->Person.Name,Person.Zip,Person.CompanyId");

            // Join keys
            Dependency person_ZipCode = orderManager.GetDependency("Person.Zip=ZipCode.Zip");
            Dependency person_Company = orderManager.GetDependency("Person.CompanyId=Company.CompanyId");

            // Constant
            Dependency companyName = orderManager.GetDependency("->Company.Name");

            orderManager.PrepareOrders();

            // Base tables
            OrderingState p = clusteredPersonId.Apply(personId_NameZipCompanyId);
            OrderingState z = clusteredZip.Apply(zip_Name);
            OrderingState c = clusteredCompanyId.Apply(companyId_Name);
            // Selection
            c = c.Apply(companyName);
            // Join p X z
            p = p.Apply(person_ZipCode).Apply(z);
            // Join p X c
            p = p.Apply(person_Company).Apply(c);

            Assert.IsTrue(personIdzipCity.Order <= p.Order);
        }

        /// <summary>
        /// Example one from Neumann's disseration.
        /// </summary>
        [Test]
        public void SmokeTestNeumann()
        {
            OrderProxy b = orderManager.GetOrder(new OrderDescription("b ASC"), true);
            OrderProxy ab = orderManager.GetOrder(new OrderDescription("a ASC, b ASC"), true);
            OrderProxy abc = orderManager.GetOrder(new OrderDescription("a ASC, b ASC, c ASC"), false);

            GroupingProxy g_b = orderManager.GetGrouping(new NestableHashSet<string>() {"b"}, true);
            GroupingProxy g_bc = orderManager.GetGrouping(new NestableHashSet<string>() {"b", "c"}, false);

            Dependency b_c = orderManager.GetDependency("b->c");
            Dependency b_d = orderManager.GetDependency("b->d");

            orderManager.PrepareOrders();

            Assert.IsTrue(b.Order <= b.Order);
            Assert.IsTrue(ab.Order <= ab.Order);
            Assert.IsTrue(abc.Order <= abc.Order);
            Assert.IsFalse(b.Order <= ab.Order);
            Assert.IsFalse(b.Order <= abc.Order);
            Assert.IsTrue(ab.Order <= abc.Order);
            Assert.IsFalse(abc.Order <= ab.Order);

            Assert.IsFalse(g_b.Order <= abc.Order);
            Assert.IsFalse(g_bc.Order <= abc.Order);
            Assert.IsTrue(g_b.Order <= b.Order);

            Order applied_abc = ab.Order.Apply(b_c);
            Assert.AreSame(abc.Order, applied_abc);

            Order applied_bc = b.Order.Apply(b_c);
            Assert.IsTrue(g_b.Order <= applied_bc);
            Assert.IsTrue(g_bc.Order <= applied_bc);
        }

        /// <summary>
        /// Example two from Neumann's disseration.
        /// </summary>
        [Test]
        public void SmokeTestNeumann2()
        {
            string[] props = new[]
                                 {
                                     "o_year", "o_partkey", "o_orderkey", "o_custkey",
                                     "p_partkey",
                                     "l_partkey", "l_suppkey", "l_orderkey",
                                     "c_custkey", "c_nationkey",
                                     "n1.n_nationkey", "n2.n_nationkey",
                                     "r_regionkey",
                                     "s_suppkey", "s_nationkey",
                                     // Only props, not orderings:
                                     "p_type", "r_name", "n1.n_regionkey", "r_regionkey"
                                 };

            for (int i = 0; i < props.Length - 4; i++)
                orderManager.GetOrder(new OrderDescription(props[i] + " ASC"), true);

            orderManager.GetDependency("p_partkey=l_partkey");
            orderManager.GetDependency("->p_type");
            orderManager.GetDependency("o_custkey=c_custkey");
            orderManager.GetDependency("->r_name");
            orderManager.GetDependency("c_nationkey=n1.n_nationkey");
            orderManager.GetDependency("s_nationkey=n2.n_nationkey");
            orderManager.GetDependency("l_orderkey=o_orderkey");
            orderManager.GetDependency("s_suppkey=l_suppkey");
            orderManager.GetDependency("n1.n_regionkey=r_regionkey");

            orderManager.PrepareOrders();
        }

        /// <summary>
        /// SELECT * FROM A JOIN B ON A.x = B.x ORDER BY A.x, B.y
        /// </summary>
        [Test]
        public void TestHomogenization()
        {
            OrderProxy requested = orderManager.GetOrder(new OrderDescription("A.x ASC,B.y ASC"), false);
            OrderProxy joinLeft = orderManager.GetOrder(new OrderDescription("A.x ASC"), true);
            OrderProxy joinRight = orderManager.GetOrder(new OrderDescription("B.x ASC"), true);

            OrderingState left = orderManager.GetOrderingState(new OrderDescription("A.x ASC"), true);
            OrderingState right = orderManager.NullOrderingState;

            Dependency joinPredicate = orderManager.GetDependency("A.x=B.x");
            Dependency bPK = orderManager.GetDependency("B.x->B.y");

            orderManager.PrepareOrders();

            right = right.Apply(bPK);
            Assert.IsTrue(joinLeft.Order <= left.Order);
            Assert.IsFalse(joinRight.Order <= right.Order);
            OrderingState join = left.Apply(joinPredicate);
            join = join.Apply(right);
            Assert.IsTrue(requested.Order <= join.Order);
        }

        /// <summary>
        /// SELECT * FROM A JOIN B ON A.x = B.x ORDER BY A.x, B.y
        /// </summary>
        [Test]
        public void TestSeq()
        {
            OrderProxy requested = orderManager.GetOrder(new OrderDescription("A.x ASC,B.x ASC,B.y ASC"), true);
            OrderingState aClustered = orderManager.GetOrderingState(new OrderDescription("A.x ASC"), true);

            Dependency bPK = orderManager.GetDependency("B.x->B.y");
            Dependency joinPredicate = orderManager.GetDependency("A.x=B.x");

            orderManager.PrepareOrders();

            OrderingState order = aClustered;
            order = order.Apply(bPK);
            order = order.Apply(joinPredicate);

            Assert.IsTrue(requested.Order <= order.Order);
        }

        /// <summary>
        /// Tests if constant expressions are treated correctly.
        /// Given the following schema:
        ///     Person(PersonId ASC PK, Name, Phone)
        /// We pretend to optimize the query
        ///     SELECT * FROM Person
        ///     WHERE Name = 'Nissen'
        ///     ORDER BY Name, PersonId, Phone
        /// 
        /// The result of this should be that the order component determines that no
        /// sort operator is necessary. This comes from the fact that Name is constant
        /// and will allways be sorted. We also have PersonId -> Phone, since PersonId
        /// is the PK, so an ordering or (PersonId) is also an ordering of (PersonId, Phone)
        /// 
        /// The test simulates the optimizer using the ordering component, finally verifying
        /// that it tells us that we do not need to sort the result.
        /// </summary>
        [Test]
        public void TestSimpleConstant()
        {
            // Clustered by PK
            OrderingState personId = orderManager.GetOrderingState(new OrderDescription("Person.PersonId ASC"), true);
            // Requested ordering
            OrderProxy requested =
                orderManager.GetOrder(new OrderDescription("Person.Name ASC,Person.PersonId ASC,Person.Phone ASC"),
                                      false);
            // Given by PK
            Dependency personPK = orderManager.GetDependency("Person.PersonId->Person.Name,Person.Phone");
            // Given by WHERE
            Dependency constName = orderManager.GetDependency("->Person.Name");

            // Now let the order manager work...
            orderManager.PrepareOrders();

            // We scan the index...
            OrderingState order = personId;
            // We apply the fact that we have a PK
            order = order.Apply(personPK);
            // We apply the WHERE clause
            order = order.Apply(constName);

            // Verify that the requested ordering is within the set of resulting
            // logical orderings.
            Assert.IsTrue(requested.Order <= order.Order);

            // Also verify that it works if we do it in another sequence
            order = personId;
            order = order.Apply(constName);
            order = order.Apply(personPK);
            Assert.IsTrue(requested.Order <= order.Order);
        }
    }
}