﻿#region

using System.Collections.Generic;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Preprocessors;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests.Preprocessors
{
    [TestFixture]
    public class CopyRemoverTests : OptimizationTestBase
    {
        private readonly TransformationProcessor<PreprocessorAttribute> processor;
        private readonly OperatorGraphVisualizer visualizer;

        public CopyRemoverTests()
        {
            processor = new TransformationProcessor<PreprocessorAttribute>(new QueryOptimization());
            processor.TransformationChain = new List<ITransformationRule>() {new CopyRemover()};

            visualizer = new OperatorGraphVisualizer()
                             {InvokeDot = false, PreserveDotFiles = true, EnumerateFiles = true};
        }

        public void Transform(TestNode node)
        {
            visualizer.GenerateGraph(node, "CopyRemover Input", "CopyRemover_input");
            processor.Transform(node);
            visualizer.GenerateGraph(node, "CopyRemover Transformed", "CopyRemover_transformed");
            visualizer.BumpCounter();
        }

        public static void AssertMatches(TestNode outputOperator, AbstractNodeMatcher expectedPattern)
        {
            Assert.That(!Match.NodeOfType("CopyOperator").GetSearcher().Search(outputOperator),
                        "CopyOperator found in graph!");
            Assert.That(expectedPattern.GetSearcher().Search(outputOperator), "Outputs didn't match expected pattern");
        }

        [Test]
        public void TestCopyRemoval()
        {
            TestNode copy = GetCopy(GetLookup("foo", "Occurrence1"));
            TestNode outputOperator = GetOutput(
                GetSort(10, 0, true, new List<SortDefinition> {new SortDefinition("DocumentId", true)}, copy),
                GetSort(10, 0, true, new List<SortDefinition> {new SortDefinition("DocumentId", false)}, copy)
                );

            Transform(outputOperator);

            AbstractNodeMatcher expectedPattern = Match.NodeOfType("OutputOperator").WithAllChildren(
                Match.NodeOfType("SortOperator").WithChild(
                    Match.NodeOfType("LookupOperator")
                    )
                );

            AssertMatches(outputOperator, expectedPattern);
        }

        [Test]
        public void TestCopyToOutput()
        {
            TestNode copy = GetCopy(GetLookup("foo", "Occurrence1"));
            TestNode outputOperator = GetOutput(
                GetSort(10, 0, true, new List<SortDefinition> {new SortDefinition("DocumentId", true)}, copy),
                copy
                );

            Transform(outputOperator);

            AbstractNodeMatcher expectedPattern = Match.NodeOfType("OutputOperator").WithAllChildren(
                Match.NodeOfType("SortOperator").WithChild(Match.NodeOfType("LookupOperator")) |
                Match.NodeOfType("LookupOperator")
                );

            AssertMatches(outputOperator, expectedPattern);
        }

        [Test]
        public void TestMultipleCopyOperators()
        {
            TestNode copy1 = GetCopy(GetLookup("foo", "Occurrence1"));
            TestNode copy2 =
                GetCopy(GetSort(10, 0, true, new List<SortDefinition> {new SortDefinition("DocumentId", true)}, copy1));

            TestNode outputOperator = GetOutput(
                copy1, copy2,
                GetSelection("DocumentId = 42", copy2)
                );

            Transform(outputOperator);

            AbstractNodeMatcher lookupMatcher = Match.NodeOfType("LookupOperator");
            AbstractNodeMatcher sortMatcher = Match.NodeOfType("SortOperator").WithChild(lookupMatcher);
            AbstractNodeMatcher selectionMatcher = Match.NodeOfType("SelectOperator").WithChild(sortMatcher);

            AbstractNodeMatcher expectedPattern = Match.NodeOfType("OutputOperator").WithAllChildren(
                lookupMatcher | sortMatcher | selectionMatcher
                );

            AssertMatches(outputOperator, expectedPattern);
        }

        [Test]
        public void TestOnLargeExampleQuery1()
        {
            TestNode outputOperator = GetLargeExampleQuery1();

            Transform(outputOperator);

            Assert.That(!Match.NodeOfType("CopyOperator").GetSearcher().Search(outputOperator));
        }
    }
}