﻿#region

using System;
using System.Collections.Generic;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.GraphMatching;
using Microsoft.Ceres.Optimizer.Preprocessors;
using Microsoft.Ceres.Optimizer.Rules;
using Microsoft.Ceres.Optimizer.Tests.GraphMatching;
using NMock2;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests.Preprocessors
{
    [TestFixture]
    public class PreprocessorTests
    {
        #region Setup/Teardown

        [SetUp]
        public void SetUp()
        {
            mocks = new Mockery();
            preprocessor = new TransformationProcessor<PreprocessorAttribute>(new QueryOptimization());
        }

        #endregion

        private Mockery mocks;
        private TransformationProcessor<PreprocessorAttribute> preprocessor;

        private class StubPreProcessor : AbstractTransformationRule
        {
            public override AbstractNodeMatcher Pattern
            {
                get { throw new NotImplementedException(); }
            }

            public override bool Iterative
            {
                get { throw new NotImplementedException(); }
            }

            public override bool Fire(GraphSearcher graphSearcher)
            {
                throw new NotImplementedException();
            }
        }

        private class StubFoodProcessor : StubPreProcessor
        {
        }

        private class StubDrinkProcessor : StubPreProcessor
        {
            public override IEnumerable<Type> DependsOn
            {
                get { return new List<Type>() {typeof (StubFoodProcessor)}; }
            }
        }

        private class StubDesertProcessor : StubPreProcessor
        {
            public override IEnumerable<Type> DependsOn
            {
                get { return new List<Type>() {typeof (StubDrinkProcessor)}; }
            }
        }

        private class CycleProcessor : StubPreProcessor
        {
            public override IEnumerable<Type> DependsOn
            {
                get { yield return typeof (OtherCycleProcessor); }
            }
        }

        private class OtherCycleProcessor : StubPreProcessor
        {
            public override IEnumerable<Type> DependsOn
            {
                get { yield return typeof (CycleProcessor); }
            }
        }


        [Test]
        public void TestCycleDetection()
        {
            preprocessor.TransformationChain = new List<ITransformationRule>()
                                                   {new CycleProcessor(), new OtherCycleProcessor()};
            try
            {
                preprocessor.ResolveDependencies();
                Assert.Fail("Should throw a cycle exception!");
            }
            catch (InvalidOperationException e)
            {
                Assert.That(e.Message.Contains("cycle"), "Wrong exception thrown: " + e);
            }
        }

        [Test]
        public void TestDependencyResolution()
        {
            ITransformationRule foodProcessor = new StubFoodProcessor();
            ITransformationRule drinkProcessor = new StubDrinkProcessor();
            ITransformationRule desertProcessor = new StubDesertProcessor();

            preprocessor.TransformationChain = new List<ITransformationRule>()
                                                   {drinkProcessor, desertProcessor, foodProcessor};

            preprocessor.ResolveDependencies();

            Assert.AreEqual(new List<ITransformationRule>() {foodProcessor, drinkProcessor, desertProcessor},
                            preprocessor.TransformationChain);
        }

        [Test]
        public void TestJoinSplitter()
        {
            MultiJoinSplitter transformer = new MultiJoinSplitter();
            preprocessor.TransformationChain = new List<ITransformationRule>() {transformer};

            Node query = new Node("Query");
            OperatorNode join = new OperatorNode() {OperatorType = typeof (JoinRule)};
            query.AddChild(join);
            join.AddChild(new Node("A"));
            join.AddChild(new Node("B"));
            join.AddChild(new Node("C"));
            join.AddChild(new Node("D"));

            OperatorNode second_join = new OperatorNode {OperatorType = typeof (JoinRule)};
            join.AddChild(second_join);
            second_join.AddChild(new Node("E"));
            second_join.AddChild(new Node("F"));
            second_join.AddChild(new Node("G"));

            GraphSearcher multiJoinSearcher = transformer.Pattern.GetSearcher();

            // We have joins with three or more children.
            multiJoinSearcher.AssertMatches(query);
            Assert.AreEqual(new List<Node>() {join, second_join}, multiJoinSearcher.MatchedNodes);

            // Now apply the preprocessor to the query graph.
            preprocessor.Transform(query);

            // Now all joins are supposed to be binary.
            multiJoinSearcher.AssertNotMatches(query);
        }
    }
}