﻿#region

using System;
using System.Collections.Generic;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Aggregation;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Aggregation.BasicAggregators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.Optimizer.Engine.Orders;
using NUnit.Framework;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests
{
    /// <summary>
    /// Contains smoke tests for the entire optimizer. These
    /// are tests that tries to optimize a complete query, some of them
    /// also asserting that the resulting optimized query is as expected.
    /// All of them verifies that no exceptions are thrown.
    /// </summary>
    [TestFixture]
    public class QueryOptimizationSmokeTests : OptimizationTestBase
    {
        #region Setup/Teardown

        [SetUp]
        public override void SetUp()
        {
            base.SetUp();
            InitializeIndexStats(cardinalities);
        }

        #endregion

        /// <summary>Test optimizing chain joins, both busy and left-deep for the specified number of joins.</summary>
        public void TestChainJoins(int joinCount)
        {
            // Enable the order length limit as no orders will be longer than 1.
            // This is due to a bug in the OrderManager we haven't fixed yet.
            OrderManager.LIMIT1 = true; 
 
            queryOptimization.JoinEnumeration = JoinEnumeration.Bushy;
            queryOptimization.Optimize(getChainJoins(joinCount));

            queryOptimization.JoinEnumeration = JoinEnumeration.LeftDeep;
            queryOptimization.Optimize(getChainJoins(joinCount));
        }

        /// <summary>Test optimizing left deep chain joins.</summary>
        [Test]
        public void TestLeftDeepChainJoins()
        {
            OrderManager.LIMIT1 = true;
            queryOptimization.JoinEnumeration = JoinEnumeration.LeftDeep;
            queryOptimization.Optimize(getChainJoins(5));

        }

        /// <summary>Test optimizing left deep star joins.</summary>
        [Test]
        public void TestLeftDeepStarJoins()
        {
            OrderManager.LIMIT1 = true;
            queryOptimization.JoinEnumeration = JoinEnumeration.LeftDeep;
            queryOptimization.Optimize(getStarJoins(5));
        }

        /// <summary>Test optimizing bushy chain joins.</summary>
        [Test]
        public void TestBushyChainJoins()
        {
            OrderManager.LIMIT1 = true;
            queryOptimization.JoinEnumeration = JoinEnumeration.Bushy;
            queryOptimization.Optimize(getChainJoins(5));
        }

        /// <summary>Test optimizing bushy star joins.</summary>
        [Test]
        public void TestBushyStarJoins()
        {
            OrderManager.LIMIT1 = true;
            queryOptimization.JoinEnumeration = JoinEnumeration.Bushy;
            queryOptimization.Optimize(getStarJoins(5));
        }

        /// <summary>Returns a query with the specified number of chain joins and a selection.</summary>
        private OperatorNode getChainJoins(int joinCount)
        {
            TestNode leftChild = GetLookup("word", "Mock0");
            for (int i = 0; i < joinCount; i++)
            {
                TestNode rightChild = GetLookup("word", "Mock" + (i + 1));
                leftChild = GetJoin("DocumentId" + i, "DocumentId" + (i + 1), selectivities[i],
                                    leftChild, rightChild);
            }
            leftChild = GetSelection("DocumentId1>DocumentId2", 0.01, 0.01, leftChild);
            OperatorNode query = GetOutput(leftChild);
            return query;
        }

        /// <summary>Returns a query with the specified number of star joins and a selection.</summary>
        private OperatorNode getStarJoins(int joinCount)
        {
            TestNode leftChild = GetLookup("word", "Mock0");
            for (int i = 0; i < joinCount; i++)
            {
                TestNode rightChild = GetLookup("word", "Mock" + (i + 1));
                leftChild = GetJoin("DocumentId0", "DocumentId" + (i + 1), selectivities[i],
                                    leftChild, rightChild);
            }
            leftChild = GetSelection("DocumentId1>DocumentId2", 0.01, 0.01, leftChild);
            OperatorNode query = GetOutput(leftChild);
            return query;
        }

        /// <summary>Tests merging multiple map operators into one.</summary>
        [Test]
        public void TestMapMerge()
        {
            OperatorNode query =
                GetOutput(
                    GetMap(false, new Dictionary<IdentifierProperty, ExpressionProperty>() {
                        { new IdentifierProperty("RenameOfDocId"), new ExpressionProperty("DocumentId") },
                        { new IdentifierProperty("SecondRenameOfPosition"), new ExpressionProperty("RenameOfPosition") }},
                        new Dictionary<string, Type>() { { "RenameOfDocId", typeof(long) }, { "SecondRenameOfPosition", typeof(long) } },
                        GetMap(true, new Dictionary<IdentifierProperty, ExpressionProperty>() {
                        { new IdentifierProperty("RenameOfPosition"), new ExpressionProperty("Position") } },
                        new Dictionary<string, Type>() { { "RenameOfPosition", typeof(long) } },
                            GetLookup("olstad", "Occurrence1"))));
            queryOptimization.Optimize(query);
        }

        /// <summary>Test used to generate the multi-query example in the report.</summary>
        [Test]
        public void MultiQueryExample()
        {
            double[] cardinalities = new double[] {1000, 2000, 2000, 1000};
            double[] selectivities = new[] {0.005, 0.0001, 0.0001};
            InitializeIndexStats(cardinalities);

            OperatorNode query = GetOutput(
                GetMergeJoin("DocumentId3", "Input1.DocumentId2", 0.0001,
                             GetLookup("word", "Mock3"),
                             GetMergeJoin("DocumentId0", "DocumentId1", 0.0001,
                                          GetLookup("word", "Mock0"),
                                          GetMergeJoin("DocumentId1", "DocumentId2", 0.005,
                                                       GetLookup("word", "Mock1"),
                                                       GetLookup("word", "Mock2")))),
                GetMergeJoin("DocumentId3", "DocumentId2", 0.0001,
                             GetLookup("word", "Mock3"),
                             GetSelection("DocumentId0 < 42", 0.9, 1,
                                          GetMergeJoin("DocumentId1", "DocumentId2", 0.0001,
                                                       GetMergeJoin("DocumentId0", "DocumentId1", 0.005,
                                                                    GetLookup("word", "Mock0"),
                                                                    GetLookup("word", "Mock1")),
                                                       GetLookup("word", "Mock2"))))
                );

            queryOptimization.Optimize(query);
        }

        /// <summary>Test optimizing simple query with one join and one selection.</summary>
        [Test]
        public void OptimizeSimpleQuery()
        {
            TestNode query = GetOutput(
                GetSelection("DocumentId = 50", 0.001, 10.00001,
                             GetJoin("DocumentId", "DocumentId", 0.006,
                                     GetLookup("search", "Occurrence1"),
                                     GetLookup("engine", "Occurrence1"))));
            queryOptimization.Optimize(query);
        }

        /// <summary>Test optimizing simple left deep query.</summary>
        [Test]
        public void OptimizeLeftDeep()
        {
            TestNode query = GetOutput(
                        GetJoin("DocumentId", "DocumentId", 0.01,
                            GetLookup("olstad", "Occurrence1"),
                            GetJoin("DocumentId", "DocumentId", 0.0006,
                                     GetLookup("search", "Occurrence1"),
                                     GetLookup("engine", "Occurrence1"))));
            queryOptimization.JoinEnumeration = JoinEnumeration.LeftDeep;
            queryOptimization.Optimize(query);
        }

        /// <summary>Query for the "To sort or not to sort" example optimization.
        /// Demonstrates removal of redundant sort operators and assers appropriate prescence of sort
        /// operators.</summary>
        [Test]
        public void OptimizeSimpleQueryWithOrder()
        {
            TestNode query = GetOutput(
                GetSelection("DocumentId0 < 50", 0.001, 10.00001,
                             GetSort(0, 0, true, new List<SortDefinition>() {new SortDefinition("DocumentId0", true)},
                                     GetJoin("DocumentId0", "DocumentId1", 0.006,
                                             GetLookup("word", "Mock0"),
                                             GetLookup("word", "Mock1")))));
            OperatorNode result = queryOptimization.Optimize(query);
            foreach (OperatorNode node in result.IterateDepthFirst(true))
                if (node.OperatorType == typeof (SortOperator))
                    Assert.Fail("The above query does not need a sort operator, as the output are clustered.");


            // Remove the sort on DocumentId for the Mock0 index.
            metadata.SetIndexOrder("Mock0", new OrderDescription());

            query = GetOutput(
                GetSelection("DocumentId0 < 50", 0.001, 10.00001,
                             GetSort(0, 0, true, new List<SortDefinition>() {new SortDefinition("DocumentId0", true)},
                                     GetJoin("DocumentId0", "DocumentId1", 0.006,
                                             GetLookup("word", "Mock0"),
                                             GetLookup("word", "Mock1")))));
            result = queryOptimization.Optimize(query);
            bool foundSort = false;
            foreach (OperatorNode node in result.IterateDepthFirst(true))
            {
                if (node.OperatorType == typeof (SortOperator))
                {
                    foundSort = true;
                    break;
                }
            }
            Assert.IsTrue(foundSort, "The above query needs a sort operator to satisfy the ordering requirement.");

            query = GetOutput(
                GetSelection("DocumentId1 = 50", 0.001, 10.00001,
                             GetSort(0, 0, true, new List<SortDefinition>() {new SortDefinition("DocumentId0", true)},
                                     GetMergeJoin(new List<string>() {"DocumentId0", "DocumentId1"}, 0.006,
                                                  GetSort(0, 0, true,
                                                          new List<SortDefinition>()
                                                              {new SortDefinition("DocumentId0", true)},
                                                          GetLookup("word", "Mock0")),
                                                  GetSort(0, 0, true,
                                                          new List<SortDefinition>()
                                                              {new SortDefinition("DocumentId0", true)},
                                                          GetLookup("word", "Mock1"))))));
            result = queryOptimization.Optimize(query);
            foreach (OperatorNode node in result.IterateDepthFirst(true))
                if (node.OperatorType == typeof (SortOperator))
                    Assert.Fail("The above query does not need a sort operator, since DocumentId == 50 (constant).");
        }

        /// <summary>Test used to generate the example used in the chapter about orderings and groupings.</summary>
        [Test]
        public void PrettyOrderingsTest()
        {
            metadata.SetIndexSchema("OrderTest", null,
                                    new RecordDescriptor(new Field(new NodeAttribute(null, "DocTypeId"), typeof (long)),
                                                         new Field(new NodeAttribute(null, "Ext"), typeof (string)),
                                                         new Field(new NodeAttribute(null, "DocIc"), typeof (long)),
                                                         new Field(new NodeAttribute(null, "DocName"), typeof (string))));
            metadata.SetIndexState("OrderTest", null, new BasicPlanSetState(100, 100));
            metadata.SetIndexOrder("OrderTest", new OrderDescription(new OrderDescriptionItem("DocTypeId", true)));

            TestNode query = GetOutput(
                GetSort(0, 0, true, new List<SortDefinition>()
                    {
                        new SortDefinition("DocTypeId", true)
                    },
                    GetLookup("word", "OrderTest")),
                GetSort(0, 0, true, new List<SortDefinition>()
                    {
                        new SortDefinition("SecondCount", false), 
                        new SortDefinition("DocTypeId", true), 
                        new SortDefinition("DocTypeDesc", true)
                    },
                    GetGroup(new List<string>() { "DocTypeId", "DocTypeDesc" }, false,
                        new Dictionary<string, Aggregator>() { { "SecondCount", new SumAggregator() } },
                        new Dictionary<string, Type>() { { "SecondCount", typeof(long) } },
                        GetMap(true, new Dictionary<IdentifierProperty,ExpressionProperty>()
                            { 
                                { new IdentifierProperty("DocTypeDesc"), new ExpressionProperty("DocTypeId + \" \"") }
                            },
                            new Dictionary<string,Type>() { { "DocTypeDesc", typeof(string) } },
                            GetGroup(new List<string>() { "DocTypeId" }, false,
                                new Dictionary<string, Aggregator>() { { "FirstCount", new CountAggregator() } },
                                new Dictionary<string,Type>() { { "FirstCount", typeof(long) } },
                                GetLookup("word", "OrderTest"))))));
            OperatorNode result = queryOptimization.Optimize(query);
        }

        /// <summary>
        /// Join example from Moerkoette's book, page 38 (PDF page 55).
        /// Also asserts the shape of the optimized query and its costs.
        /// </summary>
        [Test]
        public void SimpleJoins()
        {
            double[] cardinalities = new double[] {1000, 2000, 2000, 1000};
            double[] selectivities = new[] {0.005, 0.0001, 0.0001};
            InitializeIndexStats(cardinalities);

            // Now test bushy enumeration. This should yield the plan (0 X 1) X (2 X 3)
            // with cost 1678.
            TestNode[] expectedLookups = new[]
                                             {
                                                 GetLookup("word", "Mock0"),
                                                 GetLookup("word", "Mock1"),
                                                 GetLookup("word", "Mock2"),
                                                 GetLookup("word", "Mock3")
                                             };
            TestNode[] expectedJoins = new TestNode[3];
            expectedJoins[0] = GetMergeJoin(null, null, expectedLookups[1], expectedLookups[0]);
            expectedJoins[1] = GetMergeJoin(null, null, expectedLookups[2], expectedLookups[3]);
            expectedJoins[2] = GetMergeJoin(null, null, expectedJoins[0], expectedJoins[1]);
            TestNode expected = GetMap(false, null, null, expectedJoins[2]);
            expected.Properties = null;
            expected = GetOutput(expected);
            expectedJoins[2].CustomValidator = (n => Assert.AreEqual(new BasicCost(1678), n.Plan.Costs));

            queryOptimization.JoinEnumeration = JoinEnumeration.Bushy;
            OperatorNode result = queryOptimization.Optimize(GetSimpleJoinsInput(cardinalities, selectivities));
            expected.AreEqualRecursively(result);

            // Now test left-deep enumeration. This should yield the plan ((0 X 1) X 2) X 3
            // with cost 1696.
            expectedLookups = new[]
                                  {
                                      GetLookup("word", "Mock0"),
                                      GetLookup("word", "Mock1"),
                                      GetLookup("word", "Mock2"),
                                      GetLookup("word", "Mock3")
                                  };
            expectedJoins[0] = GetMergeJoin(null, null, expectedLookups[1], expectedLookups[0]);
            expectedJoins[1] = GetMergeJoin(null, null, expectedJoins[0], expectedLookups[2]);
            expectedJoins[2] = GetMergeJoin(null, null, expectedJoins[1], expectedLookups[3]);
            expected = GetMap(false, null, null, expectedJoins[2]);
            expected.Properties = null;
            expected = GetOutput(expected);
            expectedJoins[2].CustomValidator = (n => Assert.AreEqual(new BasicCost(1696), n.Plan.Costs));

            queryOptimization.JoinEnumeration = JoinEnumeration.LeftDeep;
            result = queryOptimization.Optimize(GetSimpleJoinsInput(cardinalities, selectivities));
            expected.AreEqualRecursively(result);
        }

        /// <summary>Test optimizing very simple query with only one lookup
        /// and verify that it has the correct cost.</summary>
        [Test]
        public void SimpleTest()
        {
            TestNode query = GetOutput(GetLookup("olstad", "Occurrence1"));
            OperatorNode result = queryOptimization.Optimize(query);
            Assert.AreEqual(new BasicCost(214), result["Cost"]);
        }

        /// <summary>Test optimizing a query where the only sharing possible is for base plans.
        /// Documents our improved implementation over Neumann.
        /// Asserts the optimized query's cost.</summary>
        [Test]
        public void TestBasePlanSharing()
        {
            TestNode query = GetOutput(
                GetJoin("DocumentId", "DocumentId", 1.0/8000,
                        GetLookup("olstad", "Occurrence1"),
                        GetLookup("DocumentType", "Occurrence2")
                    ),
                GetJoin("DocumentId", "DocumentId", 1.0/8000,
                        GetLookup("olstad", "Occurrence1"),
                        GetLookup("DocumentName", "Occurrence2")
                    )
                );
            OperatorNode result = queryOptimization.Optimize(query);
            // Assert the cost, as this tells us it was probably correctly optimized.
            Assert.AreEqual(new BasicCost(1179), result["Cost"]);
        }

        /// <summary>Tests adding a sort operator on top of a query with joins, scoreoccurrence and lookups.</summary>
        [Test]
        public void TestOneTree()
        {
            TestNode query = GetOutput(
                GetSort(0, 0, true, new List<SortDefinition>() {new SortDefinition("Score", false)},
                        GetSampleQuerySubtree()
                    )
                );
            queryOptimization.Optimize(query);
        }

        /// <summary>Same as the previous test, only without the sort.</summary>
        [Test]
        public void TestOneTreeWithoutSort()
        {
            TestNode query = GetOutput(GetSampleQuerySubtree());
            queryOptimization.Optimize(query);
        }

        /// <summary>Test optimizing a simple query which has sharing opportunities.
        /// Asserts the optimized query's cost.</summary>
        [Test]
        public void TestTwoQueriesWithSharing()
        {
            TestNode query = GetOutput(
                GetGroup(new List<string>() {"Input1.DocumentType"}, false,
                         new Dictionary<string, Aggregator>() {{"Count", new CountAggregator()}},
                         new Dictionary<string, Type>() {{"Count", typeof (long)}},
                         GetJoin("DocumentId", "DocumentId", 1.0/8000,
                                 GetLookup("olstad", "Occurrence1"),
                                 GetLookup("DocumentType", "Occurrence2")
                             )),
                GetJoin("DocumentId", "DocumentId", 1.0/8000,
                        GetJoin("DocumentId", "DocumentId", 1.0/8000,
                                GetLookup("olstad", "Occurrence1"),
                                GetLookup("DocumentType", "Occurrence2")
                            ),
                        GetLookup("DocumentName", "Occurrence2")
                    )
                );
            OperatorNode result = queryOptimization.Optimize(query);
            // Assert the cost, as this tells us it was probably correctly optimized.
            Assert.AreEqual(new BasicCost(1204), result["Cost"]);
        }

        /// <summary>Test optimizing two queries with sharing that is prone to a equivalence class
        /// representative selection problem. Used to document and verify removal of a previous bug.
        /// Asserts the optimized query's cost.</summary>
        [Test]
        public void TestTwoQueriesWithSharingEQClassProblem()
        {
            TestNode query = GetOutput(
                GetGroup(new List<string>() {"BAR", "FOO",}, false,
                         new Dictionary<string, Aggregator>() {{"Count", new CountAggregator()}},
                         new Dictionary<string, Type>() {{"Count", typeof (long)}},
                         GetMap(false, new Dictionary<IdentifierProperty, ExpressionProperty>()
                                           {
                                               {
                                                   new IdentifierProperty("FOO"),
                                                   new ExpressionProperty("Input1.DocumentType")
                                                   },
                                               {
                                                   new IdentifierProperty("BAR"),
                                                   new ExpressionProperty("Input1.DocumentType * 2")
                                                   }
                                           },
                                new Dictionary<string, Type>() {{"FOO", typeof (string)}, {"BAR", typeof (long)}},
                                GetMergeJoin(new List<string>() {"DocumentId", "DocumentId"}, 1.0/8000,
                                             GetScoreOccurences(
                                                 GetLookup("search", "Occurrence1"),
                                                 GetLookup("olstad", "Occurrence1")
                                                 ),
                                             GetLookup("DocumentType", "Occurrence2")
                                    ))
                    ),
                GetMap(false, new Dictionary<IdentifierProperty, ExpressionProperty>()
                                  {{new IdentifierProperty("BING"), new ExpressionProperty("Input1.DocumentType")}},
                       new Dictionary<string, Type>() {{"BING", typeof (string)}},
                       GetMergeJoin(new List<string>() {"DocumentId", "DocumentId"}, 1.0/8000,
                                    GetMergeJoin(new List<string>() {"DocumentId", "DocumentId"}, 1.0/8000,
                                                 GetScoreOccurences(
                                                     GetLookup("search", "Occurrence1"),
                                                     GetLookup("olstad", "Occurrence1")
                                                     ),
                                                 GetLookup("DocumentName", "Occurrence2")
                                        ),
                                    GetLookup("DocumentType", "Occurrence2")
                           )
                    )
                );
            OperatorNode result = queryOptimization.Optimize(query);

            // Assert the cost, as this tells us it was probably correctly optimized.
            Assert.AreEqual(new BasicCost(4612.5), result["Cost"]);
        }

        /// <summary>Test optimizing the two trees version with no sharing of the large example query from the report.</summary>
        [Test]
        public void TestTwoTreesQuery()
        {
            OrderManager.LIMIT1 = true;
            TestNode query = getTwoTreesQuery();
            queryOptimization.Optimize(query);
        }

        private static TestNode getTwoTreesQuery()
        {
            TestNode query = GetOutput(
                GetSort(0, 0, true, new List<SortDefinition>() { new SortDefinition("Score", false) },
                        GetSampleQuerySubtree()
                    ),
                GetSort(0, 0, true, new List<SortDefinition>() { new SortDefinition("Count", false) },
                        GetGroup(new List<string>() { "DocumentType" }, false,
                                 new Dictionary<string, Aggregator>() { { "Count", new CountAggregator() } },
                                 new Dictionary<string, Type>() { { "Count", typeof(long) } },
                                 GetSampleQuerySubtree()
                            )
                    )
                );
            return query;
        }

        /// <summary>Extracts and print costs for a the large example query used in the report.</summary>
        [Test]
        public void TestCostsForSampleQuery()
        {
            double[] cardinalities = new double[] { 82, 328, 1313, 5267, 21699, 87238, 385961, 2022792 };
            queryOptimization.PrintStats = false;

            for (int i = 0; i < cardinalities.Length; i++)
            {
                metadata.SetIndexState("Occurrence1", "search", new BasicPlanSetState(cardinalities[i], 100));
                metadata.SetIndexState("Occurrence1", "engine", new BasicPlanSetState(cardinalities[i], 100));

                queryOptimization.EnableDAG = false;
                TestNode query = getTwoTreesQuery();
                OperatorNode result = queryOptimization.Optimize(query);
                Console.WriteLine("{0}, NO DAG: {1}", i, result["Cost"]);

                queryOptimization.EnableDAG = true;
                query = getTwoTreesQuery();
                result = queryOptimization.Optimize(query);
                Console.WriteLine("{0}, DAG: {1}", i, result["Cost"]);
            }
        }
    }
}