﻿#region

using System;
using System.Reflection;

#endregion

namespace Microsoft.Ceres.Optimizer.Tests
{
    /// <summary>
    /// Helper class containing code to access private and internal members of classes,
    /// such as fields, properties and methods. All methods are implemented as extensions methods.
    /// </summary>
    internal static class ReflectionHelper
    {
        private const BindingFlags INSTANCE_FLAGS = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic;
        private const BindingFlags STATIC_FLAGS = BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic;

        /// <summary>
        /// Create an instance of the specified type, using the public or non-public constructor
        /// which matches the specified parameteres.
        /// </summary>
        /// <param name="t">Type to instantiate.</param>
        /// <param name="parameters">Any parameters to the constructor.</param>
        /// <returns>The inctance created.</returns>
        public static object GetInstance(this Type t, params object[] parameters)
        {
            return t.InvokeMember(null, BindingFlags.CreateInstance | INSTANCE_FLAGS, null, null, parameters);
        }

        /// <summary>
        /// Gets the value of the specified <b>static</b> field of the specified type.
        /// </summary>
        /// <param name="t">Type to access the field for.</param>
        /// <param name="field">Name of the field, case sensitive.</param>
        /// <returns>Value of the field.</returns>
        public static object GetField(this Type t, string field)
        {
            return t.InvokeMember(field, BindingFlags.GetField | STATIC_FLAGS, null, null, null);
        }

        /// <summary>
        /// Gets the value of the specified <b>instance</b> field on the specified instance.
        /// </summary>
        /// <param name="instance">Instance to access the field for.</param>
        /// <param name="field">Name of the field, case sensitive.</param>
        /// <returns>Value of the field.</returns>
        public static object GetField(this object instance, string field)
        {
            return instance.GetType().InvokeMember(field, BindingFlags.GetField | INSTANCE_FLAGS, null, instance, null);
        }

        /// <summary>
        /// Sets the value of the specified <b>static</b> field of the specified type.
        /// </summary>
        /// <param name="t">Type to set the field for.</param>
        /// <param name="field">Name of the field, case sensitive.</param>
        /// <param name="value">Value to set, must be assignable to the field type.</param>
        public static void SetField(this Type t, string field, object value)
        {
            t.InvokeMember(field, BindingFlags.SetField | STATIC_FLAGS, null, null, new[] {value});
        }

        /// <summary>
        /// Sets the value of the specified <b>instance</b> field on the specified instance.
        /// </summary>
        /// <param name="instance">Instance to set the field for.</param>
        /// <param name="field">Name of the field, case sensitive.</param>
        /// <param name="value">Value to set, must be assignable to the field type.</param>
        public static void SetField(this object instance, string field, object value)
        {
            instance.GetType().InvokeMember(field, BindingFlags.SetField | INSTANCE_FLAGS,
                                            null, instance, new[] {value});
        }

        /// <summary>
        /// Gets the value of the specified <b>static</b> property on the specified type,
        /// using the specified parameters (for indexers).
        /// </summary>
        /// <param name="t">Type to access the property for.</param>
        /// <param name="property">Name of the property, case sensitive.</param>
        /// <param name="parameters">Any parameters to the property (for indexers).</param>
        /// <returns>Value of the property.</returns>
        public static object GetProperty(this Type t, string property, params object[] parameters)
        {
            return t.InvokeMember(property, BindingFlags.GetProperty | STATIC_FLAGS, null, null, parameters);
        }

        /// <summary>
        /// Gets the value of the specified <b>instance</b> property on the specified instance,
        /// using the specified parameters (for indexers).
        /// </summary>
        /// <param name="instance">Instance to access the property for.</param>
        /// <param name="property">Name of the property, case sensitive.</param>
        /// <param name="parameters">Any parameters to the property (for indexers).</param>
        /// <returns>Value of the property.</returns>
        public static object GetProperty(this object instance, string property, params object[] parameters)
        {
            return instance.GetType().InvokeMember(property, BindingFlags.GetProperty | INSTANCE_FLAGS, null, instance,
                                                   parameters);
        }

        /// <summary>
        /// Sets the value of the specified <b>static</b> property on the specified type,
        /// using the specified parameters (for indexers).
        /// <param name="t">Type to set the property for.</param>
        /// <param name="property">Name of the property, case sensitive.</param>
        /// <param name="value">Value to set, must be assignable to the property type.</param>
        /// <param name="parameters">Any parameters to the property (for indexers).</param>
        /// </summary>
        public static void SetProperty(this Type t, string property, object value, params object[] parameters)
        {
            object[] temp = new object[parameters.Length + 1];
            parameters.CopyTo(temp, 0);
            temp[temp.Length - 1] = value;
            t.InvokeMember(property, BindingFlags.SetProperty | STATIC_FLAGS, null, null, temp);
        }

        /// <summary>
        /// Set the value of the specified <b>instance</b> property on the specified instance,
        /// using the specified parameters (for indexers).
        /// <param name="instance">Instance to set the property for.</param>
        /// <param name="property">Name of the property, case sensitive.</param>
        /// <param name="value">Value to set, must be assignable to the property type.</param>
        /// <param name="parameters">Any parameters to the property (for indexers).</param>
        /// </summary>
        public static void SetProperty(this object instance, string property, object value, params object[] parameters)
        {
            object[] temp = new object[parameters.Length + 1];
            parameters.CopyTo(temp, 0);
            temp[temp.Length - 1] = value;
            instance.GetType().InvokeMember(property, BindingFlags.SetProperty | INSTANCE_FLAGS, null, instance, temp);
        }

        /// <summary>
        ///	Invokes the specified <b>static</b> method on the specified type, matching
        /// the specified parameter types.
        /// </summary>
        /// <param name="t">Type to call the method on.</param>
        /// <param name="property">Name of the method, case sensitive.</param>
        /// <param name="parameters">Any parameters to the method.</param>
        /// <returns>The return value of the called method.</returns>
        public static object InvokeMethod(this Type t, string method, params object[] parameters)
        {
            return t.InvokeMember(method, BindingFlags.InvokeMethod | STATIC_FLAGS, null, null, parameters);
        }

        /// <summary>
        ///	Invokes the specified <b>instance</b> method on the specified type, matching
        /// the specified parameter types.
        /// </summary>
        /// <param name="instance">Instance to call the method on.</param>
        /// <param name="property">Name of the method, case sensitive.</param>
        /// <param name="parameters">Any parameters to the method.</param>
        /// <returns>The return value of the called method.</returns>
        public static object InvokeMethod(this object instance, string method, params object[] parameters)
        {
            return instance.GetType().InvokeMember(method, BindingFlags.InvokeMethod | INSTANCE_FLAGS, null, instance,
                                                   parameters);
        }
    }
}