﻿#region

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using Microsoft.Ceres.Evaluation.DataModel;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Infrastructure;
using Microsoft.Ceres.Evaluation.Operators.Graphs;
using Microsoft.Ceres.Evaluation.Operators.Parsing;
using Microsoft.Ceres.Evaluation.Operators.PlugIns;
using Microsoft.Ceres.Optimizer.Engine;
using Microsoft.Ceres.SearchCore.Services.QueryEngine;
using Microsoft.Ceres.SearchCore.Services.SystemCatalog;

#endregion

namespace Microsoft.Ceres.Optimizer.Transformer
{
    public class OptimizerFacade
    {
        private const BindingFlags INSTANCE_FLAGS = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic;
        private const BindingFlags STATIC_FLAGS = BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic;
        private readonly OperatorGraphVisualizer visualizer = new OperatorGraphVisualizer();
        private NamedPlugInFactory<OperatorBase> operatorFactory;
        public QueryOptimization Optimizer { get; set; }
        public IOperatorDecorator OperatorDecorator { get; set; }
        public ISystemCatalog Catalog { get; set; }

        /// <summary>
        /// Operator Factory used to create MARS operators.
        /// </summary>
        public NamedPlugInFactory<OperatorBase> OperatorFactory
        {
            get
            {
                // Workaround the fact that the factory field is private.
                if (operatorFactory == null)
                    operatorFactory =
                        (NamedPlugInFactory<OperatorBase>)
                        OperatorDecorator.GetType().InvokeMember("factory", BindingFlags.GetField | INSTANCE_FLAGS, null,
                                                                 OperatorDecorator, null);
                return operatorFactory;
            }
        }

        /// <summary>
        /// Makes an OperatorNode of a MARS Operator.
        /// </summary>
        public static OperatorNode MakeNodeForOperator(OperatorBase physicalOperatorNode)
        {
            OperatorNode node = new OperatorNode
                                    {
                                        Name = physicalOperatorNode.Name,
                                        OperatorType = physicalOperatorNode.GetType()
                                    };

            // Copy over all properties of the operator.
            foreach (PropertyInfo property in physicalOperatorNode.GetType().GetProperties())
            {
                PropertyAttribute attribute =
                    (PropertyAttribute) Attribute.GetCustomAttribute(property, typeof (PropertyAttribute));
                if (attribute == null)
                    continue;

                // No point in keeping attributes we can't get or set.
                if (! (property.CanRead && property.CanWrite))
                    continue;

                node[property.Name] = property.GetValue(physicalOperatorNode, null);
            }
            return node;
        }

        /// <summary>
        /// Makes OperatorNodes for an entire MARS query graph.
        /// </summary>
        public static OperatorNode MakeNodeGraph(OutputOperator root)
        {
            Dictionary<OperatorBase, OperatorNode> nodeForOperator = new Dictionary<OperatorBase, OperatorNode>();
            // Make nodes for every operator. Make depended-on nodes before the dependees.
            foreach (OperatorBase operatorBase in root.IterateTopologicalOrdering())
            {
                OperatorNode node = MakeNodeForOperator(operatorBase);
                nodeForOperator[operatorBase] = node;

                // Set up relationships. Since we're iterating the topological ordering,
                // we can assume that all children have been made already.
                foreach (OperatorBase child in operatorBase.GetChildren())
                    node.AddChild(nodeForOperator[child]);

                if (operatorBase.OutputCount > 0)
                    node.OutputTypeDescriptor = operatorBase.Outputs[0].RecordSetType;
            }
            return nodeForOperator[root];
        }

        /// <summary>Verifies that we can optimize the query (is has only one root -- a QueryOperator).</summary>
        public static bool QueryIsOptimizable(QueryContext qc)
        {
            if (qc.Query == null || qc.Query.RootsCount != 1)
            {
                Debug.WriteLine("Query is null or has more than 1 root.");
                return false;
            }

            if (!(qc.Query.Roots.ElementAt(0) is OutputOperator))
            {
                Debug.WriteLine("Root is not an Output operator");
                return false;
            }
            return true;
        }

        /// <summary>
        /// Optimizes the given MARS query by transforming it to an OperatorNode graph,
        /// invoking the optimizer and transforming the optimized query back.
        /// </summary>
        public IRecord Optimize(IRecord record)
        {
            QueryContext qc = (QueryContext) record["querycontext"].Value;
            OutputOperator rootRecord = (OutputOperator) qc.Query.Roots.ElementAt(0);

            // First, check the sanity of the input. If it's not what we expect, 
            //  simply return the unmodified input.
            if (!QueryIsOptimizable(qc))
                return record;

            try
            {
                // Transform into Optimizer's Node-structure.
                OperatorNode optimizableGraph = MakeNodeGraph(rootRecord);
                visualizer.GenerateGraph(optimizableGraph, "Raw input query", "input");

                // Optimize
                // If the label of the output operator is NO_OPTIMIZATION, do not optimize.
                if (rootRecord.Label != "NO_OPTIMIZATION")
                {
                    qc.Query = MakeMarsQueryGraph(Optimizer.Optimize(optimizableGraph));
                }
            }
            catch (OptimizerException e)
            {
                // Log or something? We don't want to disrupt query processing due to
                //   faults in the optimizer, though.
                Debug.WriteLine(e);
            }
            return record;
        }

        /// <summary>
        /// Makes a MARS operator graph from the optimizer's OperatorNode-graph structure.
        /// </summary>
        private OperatorGraph MakeMarsQueryGraph(OperatorNode optimizableGraph)
        {
            Dictionary<OperatorNode, OperatorBase> operatorForNode = new Dictionary<OperatorNode, OperatorBase>();

            OutputOperator rootOperator = (OutputOperator) OperatorFactory.Resolve("Output");
            operatorForNode[optimizableGraph] = rootOperator;

            OperatorFlow query = new OperatorFlow(rootOperator);

            // We skip the root, as we've already made that above.
            IEnumerable<OperatorNode> inversedTopologicalOrdering =
                optimizableGraph.GetTopologicalOrdering().Cast<OperatorNode>().Reverse().Skip(1);

            // Since the dependencies are created by adding parents, we go over the inversed t. ordering.
            foreach (OperatorNode node in inversedTopologicalOrdering)
            {
                OperatorBase op = MakeMarsNode(node);
                operatorForNode[node] = op;
                foreach (OperatorNode parent in node.Parents)
                    query.AddPredecessor(operatorForNode[parent], op);
            }

            return query;
        }

        /// <summary>
        /// Makes a MARS operator for a single OperatorNode.
        /// </summary>
        private OperatorBase MakeMarsNode(OperatorNode node)
        {
            // We cannot just directly instantiate the operators, we need to use the factory.
            OperatorBase op = OperatorFactory.Resolve(node.OperatorType.Name.Replace("Operator", ""));
            if (op == null)
                throw new OptimizerException("Could not resolve operator type {0}.",
                                             node.OperatorType.Name.Replace("Operator", ""));

            // With the operator instantiated, we need to copy back the properties for it.
            Type opType = op.GetType();
            foreach (KeyValuePair<string, object> property in node.Properties)
            {
                PropertyInfo propInfo = opType.GetProperty(property.Key);

                // We need to use reflection to be able to set the properties.
                if (propInfo != null && propInfo.CanWrite)
                    opType.InvokeMember(property.Key, BindingFlags.SetProperty | INSTANCE_FLAGS, null, op,
                                        new[] {property.Value});
            }
            return op;
        }
    }
}